#ifndef CGUL_BDF_CXX_H
#define CGUL_BDF_CXX_H

/**
 * C++ wrapper for <code>cgul_bdf</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_bdf</code>
 */

#include "cgul_bdf.h"
#include "cgul_bdf_glyph_cxx.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_bdf</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_bdf</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_bdf</code>
 */
class cgul_bdf_cxx {

public:

    /**
     * Construct a new <code>bdf</code> object.  To load a font into
     * the newly created object, call <code>load()</code> or
     * <code>load_explicit()</code>.  If an error occurs, an exception
     * is thrown.
     */
    cgul_bdf_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_bdf__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method frees all internally allocated resources for the
     * object.
     */
    virtual
    ~cgul_bdf_cxx()
    {
        if (obj) {
            cgul_bdf__delete(obj);
        }
    }

    /**
     * This method splits <code>font_list</code> into individual fonts and
     * tries to load and merge each font in turn by repeatedly calling
     * <code>load_explicit()</code>.  The end result is that
     * when the client calls <code>get_glyph()</code>, the fonts
     * will be searched in order and the first matching glyph will be
     * returned. <p>
     *
     * The <code>font_list</code> can be just a single font or can have
     * multiple fonts separated with the "|" character.  A trailing ".bdf"
     * extension in the names of the fonts is optional: <p>
     *
     * <pre>
     *     "foo-14|bar-14|baz-14"
     * </pre>
     *
     * When calling <code>load_explicit()</code>, this method
     * first tries to load the font along with its BDF index file by
     * appending an "i" character to the BDF file name.  Only if this does
     * not succeed will the method try to load the BDF file without an
     * index.  If different behavior is needed, the client should call
     * <code>load_explicit()</code> directly. <p>
     *
     * @param[in] font_list font list
     */
    void
    load(const char* font_list)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__load(&local, obj, font_list);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This methods loads a BDF font from <code>bdf_file_name</code>.  If
     * a font is already loaded, the new font will be merged into the
     * existing font.  To manually unload the existing font first, call
     * <code>unload()</code>. <p>
     *
     * If an error occurs, an exception is thrown, and the
     * <code>cgul_bdf_cxx</code> object will be in an undefined state.
     * In this case, the client should delete the object and start
     * over after correcting the problem. <p>
     *
     * The <code>index_file_name</code> parameter is optional.  It
     * specifies the BDF index file (which should have been generated
     * earlier using <code>write_index()</code>).  The BDF index
     * file contains information on where to find each glyph in the
     * BDF file.  All fonts must be loaded with an index file in order
     * for cache expiration to be enabled; otherwise, there would be
     * no good way to reload any arbitrary glyph that expires.  (See
     * <code>set_cache_size()</code>). <p>
     *
     * If <code>index_file_name</code> is <code>NULL</code> the entire
     * font will be loaded from file including all of the glyphs.  If
     * <code>index_file_name</code> is not <code>NULL</code>, only the
     * meta information for the font will be loaded.  The glyphs for
     * the font will be loaded on demand by <code>get_glyph()</code>
     * using the index information contained in
     * <code>index_file_name</code>. <p>
     *
     * Whether to use an index file or not depends on how the object
     * will be used.  If the object will be used to lookup glyph
     * information in large fonts for characters in random strings, using
     * an index file will probably give the best performance; however, if
     * the object needs to process each glyph once like when writing
     * a new BDF file or when writing a BDF index file, loading the BDF
     * without an index file will give the best performance by a wide
     * margin. <p>
     *
     * The advantages of using an index file is that your program will
     * start much faster and will only load the glyphs that are used
     * thereby saving time and memory.  The disadvantages are that the
     * index file can be large and can potentially get out of sync with
     * the BDF file.  Also, the first time you try to access a glyph, it
     * has to be loaded from disk which is a relatively slow operation,
     * but subsequent accesses will retrieve the glyph from the cache. <p>
     *
     * If the DEFAULT_CHAR property is present in the BDF font, the glyph
     * for that character will be used when the user requests a glyph
     * which is not present in the font.  If the DEFAULT_CHAR property is
     * missing, this method will generate a glyph that looks like an empty
     * box to use instead. <p>
     *
     * @param[in] bdf_file_name bdf file name
     * @param[in] index_file_name index file name
     */
    virtual void
    load_explicit(const char* bdf_file_name,
                  const char* index_file_name = NULL)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__load_explicit(&local, obj, bdf_file_name, index_file_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method unloads the font and frees internally allocated resources.
     */
    virtual void
    unload()
    {
        cgul_exception_t local = NULL;
        cgul_bdf__unload(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the current limit on the size of the internal glyph cache.
     * By convention, a cache size of <code>0</code> means the cache
     * can grow without limit.
     *
     * @return cache size
     */
    virtual unsigned long int
    get_cache_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_cache_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method sets the maximum size of the internal cache that stores
     * glyphs.  The cache size is only relevant if index files are
     * specified for each BDF when <code>load()</code> or
     * <code>load_explicit()</code> is called.  If index files
     * are specified, it means the glyphs can be reloaded from file later
     * so it is safe to remove them from the internal cache. <p>
     *
     * By default, <code>cache_size</code> is set to <code>0</code> which,
     * by convention, means glyphs never expire.  If the caller sets
     * <code>cache_size</code> to a positive value, the least-recently
     * used (LRU) glyphs will be expired first. <p>
     *
     * If glyph expiration is enabled, the client must be careful not to
     * cache glyphs itself because they could expire at any moment.
     * Instead, the client must make sure <code>cache_size</code> is large
     * enough for the task at hand and must call
     * <code>get_glyph()</code> to get the glyph each time it is
     * used so that <code>bdf</code> can keep its LRU list up to date so
     * that glyphs can be expired in the correct order. <p>
     *
     * Calling this method has the potential of cause glyphs to expire
     * immediately. <p>
     *
     * @param[in] cache_size cache size
     */
    virtual void
    set_cache_size(unsigned long int cache_size)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__set_cache_size(&local, obj, cache_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the width and height in pixels of the smallest
     * rectangle that can contain the image of the text in the UTF-8
     * string <code>utf8</code>.  The width and height are returned in
     * <code>width</code> and <code>height</code> provided they are not
     * set to <code>NULL</code>. <p>
     *
     * No attempt is made by this method to trim white space.  Thus, if
     * white space characters appear at the end of a line, their width
     * will be added to the width that is returned.  Likewise, if there
     * are multiple trailing line breaks, the extra ones will add to the
     * height that is returned.  If this is not the behavior you want, you
     * need to trim <code>utf8</code> before passing it into this
     * method. <p>
     *
     * It is also important to understand that the <code>cgul_bdf_cxx</code>
     * class does not draw text.  Instead, it loads fonts and makes the
     * information for each glyph available so that drawing libraries you
     * write can draw text.  Providing your drawing library spaces each
     * character using the value returned by
     * <code>cgul_bdf_glyph_cxx::get_dwidth_x()</code> and providing your
     * drawing library spaces each line using the value returned by
     * <code>cgul_bdf_cxx::get_height()</code>, this method will return
     * accurate values for <code>width</code> and <code>height</code>. <p>
     *
     * It is also important to understand that the width for each line of
     * text is not simply the sum of the "dwidth_x" properties of each
     * character.  It is slightly more complicated than this because some
     * glyphs have pixels that intrude on the character cell to the left
     * or right.  If this happens for the first or last character, these
     * extra pixels will be added to the width of those lines. <p>
     *
     * This method treats the "\n" character as the only character that
     * causes a line break.  If you want to treat "\r\n" or "\r" as line
     * breaks, the correct solution is to use <code>cgul_crlf_file</code>
     * to read in your text because it handles DOS, Mac, and Unix text
     * files allowing you to properly set the standard C new line
     * character for these different formats. <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf8</code> has an invalid encoding or because
     * <code>utf8</code> has a character for which there is no glyph in
     * the <code>bdf</code> font and <code>substitute</code> is not set.
     * Errors can also occur if an index file is being used to load glyphs
     * on demand, and an error occurs during the loading process. <p>
     *
     * @param[in] utf8 UTF-8 string
     * @param[in] utf8_length number of bytes (not characters) in
     * <code>utf8</code>
     * @param[in] substitute whether to substitute missing glyphs with
     * the default
     * @param[out] width width of <code>utf8</code> in pixels
     * @param[out] height height of <code>utf8</code> in pixels
     */
    virtual void
    get_utf8_extents(const char* utf8,
                     size_t utf8_length,
                     int substitute,
                     unsigned int* width,
                     unsigned int* height) const
    {
        cgul_exception_t local = NULL;
        cgul_bdf__get_utf8_extents(&local, obj,
                                   utf8, utf8_length,
                                   substitute,
                                   width, height);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the width and height in pixels of the smallest
     * rectangle that can contain the image of the text in the UTF-32
     * string <code>utf32</code>.  The width and height are returned in
     * <code>width</code> and <code>height</code> provided they are not
     * set to <code>NULL</code>. <p>
     *
     * No attempt is made by this method to trim white space.  Thus, if
     * white space characters appear at the end of a line, their width
     * will be added to the width that is returned.  Likewise, if there
     * are multiple trailing line breaks, the extra ones will add to the
     * height that is returned.  If this is not the behavior you want, you
     * need to trim <code>utf8</code> before passing it into this
     * method. <p>
     *
     * It is also important to understand that the <code>cgul_bdf_cxx</code>
     * class does not draw text.  Instead, it loads fonts and makes the
     * information for each glyph available so that drawing libraries you
     * write can draw text.  Providing your drawing library spaces each
     * character using the value returned by
     * <code>cgul_bdf_glyph_cxx::get_dwidth_x()</code> and providing your
     * drawing library spaces each line using the value returned by
     * <code>cgul_bdf_cxx::get_height()</code>, this method will return
     * accurate values for <code>width</code> and <code>height</code>. <p>
     *
     * It is also important to understand that the width for each line of
     * text is not simply the sum of the "dwidth_x" properties of each
     * character.  It is slightly more complicated than this because some
     * glyphs have pixels that intrude on the character cell to the left
     * or right.  If this happens for the first or last character, these
     * extra pixels will be added to the width of those lines. <p>
     *
     * This method treats the "\n" character as the only character that
     * causes a line break.  If you want to treat "\r\n" or "\r" as line
     * breaks, the correct solution is to use <code>cgul_crlf_file</code>
     * to read in your text because it handles DOS, Mac, and Unix text
     * files allowing you to properly set the standard C new line
     * character for these different formats. <p>
     *
     * If an error occurs, an exception is thrown.  Errors can occur
     * because <code>utf8</code> has an invalid encoding or because
     * <code>utf8</code> has a character for which there is no glyph in
     * the <code>bdf</code> font and <code>substitute</code> is not set.
     * Errors can also occur if an index file is being used to load glyphs
     * on demand, and an error occurs during the loading process. <p>
     *
     * @param[in] utf32 UTF-32 string
     * @param[in] utf32_length number of characters (not bytes) in
     * <code>utf32</code>
     * @param[in] substitute whether to substitute missing glyphs with
     * the default
     * @param[out] width width of <code>utf32</code> in pixels
     * @param[out] height height of <code>utf32</code> in pixels
     */
    virtual void
    get_utf32_extents(const cgul_wchar_t* utf32,
                      size_t utf32_length,
                      int substitute,
                      unsigned int* width,
                      unsigned int* height) const
    {
        cgul_exception_t local = NULL;
        cgul_bdf__get_utf32_extents(&local, obj,
                                    utf32, utf32_length,
                                    substitute,
                                    width, height);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the glyph for a particular code point.  The code point is
     * called <code>encoding</code> to match its BDF name.  If
     * <code>substitute</code> is set, the default character glyph is
     * returned when no glyph is present for <code>encoding</code>;
     * otherwise, <code>NULL</code> is returned when no glyph is
     * present. <p>
     *
     * The returned glyph is still owned by the object so the client
     * must not attempt to delete it.  The client is free however to make
     * a copy using <code>cgul_bdf_glyph_cxx::copy()</code>. <p>
     *
     * Errors can happen if an index file is being used to load glyphs
     * on demand, and an error occurs during the loading process.  If
     * an error occurs, an exception is thrown. <p>
     *
     * @param[in] encoding code point for the glyph that is returned
     * @param[in] substitute whether to substitute missing glyphs with
     * the default
     * @return glyph
     */
    virtual cgul_bdf_glyph_cxx*
    get_glyph(cgul_wchar_t encoding,
              int substitute) const
    {
        cgul_bdf_glyph_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_bdf_glyph_cxx*)cgul_bdf__get_glyph(&local,
                                                      obj,
                                                      encoding,
                                                      substitute);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Shift each glyph in the object the same minimal number of
     * pixels to the right so that no glyph has a negative x-offset.
     * For efficiency, this method requires that all BDF files be
     * loaded without specifying an index file.
     */
    virtual void
    shift_right()
    {
        cgul_exception_t local = NULL;
        cgul_bdf__shift_right(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Force the x-offset of each glyph in the object to zero but add
     * padding to the left side of each glyph to compensate so that
     * the glyph will still be drawn in the same location.  The caller
     * is responsible for making sure <code>shift_right()</code> is
     * called before calling this method.  For efficiency, this method
     * requires that all BDF files be loaded without specifying an
     * index file.
     */
    virtual void
    pad_left()
    {
        cgul_exception_t local = NULL;
        cgul_bdf__pad_left(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method creates or truncates the file
     * <code>bdf_file_name</code> and writes the entire font in BDF
     * format to that file.  For efficiency, this method requires that
     * all BDF files be loaded without specifying an index file. <p>
     *
     * @param[in] bdf_file_name bdf file name
     */
    virtual void
    write(const char* bdf_file_name)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__write(&local, obj, bdf_file_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method writes the entire font in BDF format to the file
     * <code>f</code>.  For efficiency, this method requires that all
     * BDF files be loaded without specifying an index file. <p>
     *
     * @param[in] f file
     */
    virtual void
    write_file(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__write_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method creates or truncates the file
     * <code>index_file_name</code> and writes the entire BDF index file
     * for the BDF file that is currently loaded.  For this method to
     * produce meaningful results only one BDF file can be loaded, and it
     * must not already be loaded using an index file.  You can do this by
     * calling <code>load_explicit()</code> with a
     * <code>NULL</code> value for <code>index_file_name</code>.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] index_file_name file name
     */
    virtual void
    write_index(const char* index_file_name)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__write_index(&local, obj, index_file_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method writes the entire BDF index file to <code>f</code> for
     * the BDF file that is currently loaded.  <code>f</code> must be
     * opened in binary mode.  For this method to produce meaningful
     * results only one BDF file can be loaded, and it must not already be
     * loaded using an index file.  You can do this by calling
     * <code>load_explicit()</code> with a <code>NULL</code>
     * value for <code>index_file_name</code>.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] f file
     */
    virtual void
    write_index_file(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_bdf__write_index_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the ascent of the font.
     *
     * @return ascent
     */
    virtual unsigned int
    get_ascent() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_ascent(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the descent of the font.
     *
     * @return descent
     */
    virtual unsigned int
    get_descent() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_descent(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the height of the font.  This is simply the sum
     * of the ascent and the descent.  The height of the font is typically
     * the value used as the spacing between lines.
     *
     * @return height
     */
    virtual unsigned int
    get_height() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_height(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the list of fonts that have been loaded and merged into the
     * object.  The font list has the same format as the font list
     * passed into <code>load()</code>.  If no font is currently
     * loaded, an empty string is returned.  The caller must not
     * attempt to free the pointer returned by this method.
     *
     * @return font list
     */
    virtual const char*
    get_font_list() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_font_list(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the name of the font as given on the FONT line.  If no font is
     * currently loaded, <code>NULL</code> is returned.  The caller must
     * not attempt to free the pointer returned by this method.
     *
     * @return BDF font name
     */
    virtual const char*
    get_font_name() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_font_name(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the size of the font in points.
     *
     * @return font size
     */
    virtual unsigned int
    get_font_size() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_font_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the x-resolution of the font in dots per inch.
     *
     * @return x-resolution of the font in dots per inch
     */
    virtual unsigned int
    get_dpi_x() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_dpi_x(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the y-resolution of the font in dots per inch.
     *
     * @return y-resolution of the font in dots per inch
     */
    virtual unsigned int
    get_dpi_y() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_dpi_y(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the width of the font's bounding box.
     *
     * @return width of the font's bounding box
     */
    virtual unsigned int
    get_bb_width() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_bb_width(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the height of the font's bounding box.
     *
     * @return height of the font's bounding box
     */
    virtual unsigned int
    get_bb_height() const
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_bb_height(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the x-offset of the lower-left corner of the font's bounding box.
     *
     * @return x-offset of the lower-left corner of the font's bounding box
     */
    virtual int
    get_bb_x_offset() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_bb_x_offset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the y-offset of the lower-left corner of the font's bounding box.
     *
     * @return y-offset of the lower-left corner of the font's bounding box
     */
    virtual int
    get_bb_y_offset() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_bb_y_offset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the number of characters in the font.  This function throws
     * an exception if more than one BDF file is loaded unless all the
     * BDF files are loaded without index files.  The reason for this
     * is that it is hard to know how many characters are in a merged
     * BDF font when index files are being used because the actual
     * merging of glyphs is delayed until <code>get_glyph()</code> is
     * called, and even then, the merging only occurs for the
     * requested encoding.
     *
     * @return number of characters in the font
     */
    virtual unsigned long int
    get_char_count() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_bdf__get_char_count(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_bdf</code> object.
     *
     * @return underlying object
     */
    virtual cgul_bdf_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_bdf</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_bdf__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_bdf_t
    take_obj()
    {
        cgul_bdf_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_bdf_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_bdf__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_bdf_cxx(const cgul_bdf_cxx&);
    cgul_bdf_cxx& operator=(const cgul_bdf_cxx&);

    /* Underlying <code>cgul_bdf</code> object. */
    cgul_bdf_t obj;

};

#endif  /*  CGUL_BDF_CXX_H  */
