#ifndef CGUL_BIG_INTEGER_H
#define CGUL_BIG_INTEGER_H

/**
 * Basic arbitrary precision arithmetic.
 *
 * @file
 * @author Paul Serice
 * @brief arbitrary precision arithmetic
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/*
 * Define cgul_bcd_t before including cgul_big_integer.h to resolve
 * circular dependencies.
 */

/** Opaque pointer to a <code>cgul_big_integer</code> instance. */
typedef struct cgul_big_integer* cgul_big_integer_t;

CGUL_END_C

#include "cgul_bcd.h"

CGUL_BEGIN_C

/**
 * Create a new <code>cgul_big_integer</code> instance initialized to
 * positive 0.  The caller is responsible for calling
 * <code>cgul_big_integer__delete()</code> on the value returned.  If
 * memory allocation fails, <code>NULL</code> is returned, and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT cgul_big_integer_t
cgul_big_integer__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_big_integer</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_big_integer__delete()</code> on the value returned.  If
 * memory allocation fails, <code>NULL</code> is returned and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT cgul_big_integer_t
cgul_big_integer__new_from_big_integer(cgul_exception_t* cex,
                                       cgul_big_integer_t value);

/**
 * Create a new <code>cgul_big_integer</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_big_integer__delete()</code> on the value returned.  If
 * memory allocation fails, <code>NULL</code> is returned and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT cgul_big_integer_t
cgul_big_integer__new_from_long(cgul_exception_t* cex,
                                long value);

/**
 * Create a new <code>cgul_big_integer</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_big_integer__delete()</code> on the value returned.  If
 * memory allocation fails, <code>NULL</code> is returned and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT cgul_big_integer_t
cgul_big_integer__new_from_unsigned_long(cgul_exception_t* cex,
                                         unsigned long value);

/**
 * Create a new <code>cgul_big_integer</code> instance initialized to
 * <code>value</code>.  The caller is responsible for calling
 * <code>cgul_big_integer__delete()</code> on the value returned.  If
 * memory allocation fails, <code>NULL</code> is returned and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT cgul_big_integer_t
cgul_big_integer__new_from_size_type(cgul_exception_t* cex,
                                     size_t value);

/**
 * Create a new <code>cgul_big_integer</code> instance initialized
 * from the string <code>s</code>.  If <code>s</code> starts with "0x"
 * or "0X", it will be treated as hex.  If it starts with "0b" or
 * "0B", it will be treated as binary.  If it starts with "0" followed
 * by any valid octal digit, it will be treated as octal.  All other
 * strings will be treated as decimal.  The caller is responsible for
 * calling <code>cgul_big_integer__delete()</code> on the value
 * returned.  If an error occurs interpreting the string "s", then
 * NULL is returned.  If memory allocation fails, <code>NULL</code> is
 * returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s initial value
 * @return new <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT cgul_big_integer_t
cgul_big_integer__new_from_string(cgul_exception_t* cex,
                                  const char* s);

/**
 * Delete the <code>cgul_big_integer</code> object in
 * <code>big_integer</code>.  The client must not use
 * <code>big_integer</code> after calling this method.
 *
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT void
cgul_big_integer__delete(cgul_big_integer_t big_integer);

/**
 * Return 1 if <code>big_integer</code> is positive.  Return 0 if it
 * is not.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return whether <code>big_integer</code> is positive
 */
extern CGUL_EXPORT int
cgul_big_integer__is_positive(cgul_exception_t* cex,
                              cgul_big_integer_t big_integer);

/**
 * Return 1 if <code>big_integer</code> is negative.  Return 0 if it
 * is not.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return whether <code>big_integer</code> is negative
 */
extern CGUL_EXPORT int
cgul_big_integer__is_negative(cgul_exception_t* cex,
                              cgul_big_integer_t big_integer);

/**
 * Return 1 if <code>big_integer</code> is zero.  Return 0 if it is
 * not.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return whether <code>big_integer</code> is zero
 */
extern CGUL_EXPORT int
cgul_big_integer__is_zero(cgul_exception_t* cex,
                          cgul_big_integer_t big_integer);

/**
 * Set the value of <code>big_integer</code> to zero.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT void
cgul_big_integer__set_to_zero(cgul_exception_t* cex,
                              cgul_big_integer_t big_integer);

/**
 * Return 1 if <code>big_integer</code> is one.  Return 0 if it is
 * not.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return whether <code>big_integer</code> is one
 */
extern CGUL_EXPORT int
cgul_big_integer__is_one(cgul_exception_t* cex,
                         cgul_big_integer_t big_integer);

/**
 * Set the value of <code>big_integer</code> to one.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT void
cgul_big_integer__set_to_one(cgul_exception_t* cex,
                             cgul_big_integer_t big_integer);

/**
 * Return whether <code>big_integer</code> is even.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return whether <code>big_integer</code> is even
 */
extern CGUL_EXPORT int
cgul_big_integer__is_even(cgul_exception_t* cex,
                          cgul_big_integer_t big_integer);

/**
 * Return whether <code>big_integer</code> is odd.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return whether <code>big_integer</code> is odd
 */
extern CGUL_EXPORT int
cgul_big_integer__is_odd(cgul_exception_t* cex,
                         cgul_big_integer_t big_integer);

/**
 * Assign <code>rhs</code> to <code>lhs</code>.  It is safe for
 * <code>rhs</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__assign(cgul_exception_t* cex,
                         cgul_big_integer_t lhs,
                         cgul_big_integer_t rhs);

/**
 * Assign long in <code>rhs</code> to <code>lhs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__assign_from_long(cgul_exception_t* cex,
                                   cgul_big_integer_t lhs,
                                   long rhs);

/**
 * Assign unsigned long in <code>rhs</code> to <code>lhs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__assign_from_unsigned_long(cgul_exception_t* cex,
                                            cgul_big_integer_t lhs,
                                            unsigned long rhs);

/**
 * Assign size type in <code>rhs</code> to <code>lhs</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__assign_from_size_type(cgul_exception_t* cex,
                                        cgul_big_integer_t lhs,
                                        size_t rhs);

/**
 * Assign the string in <code>rhs</code> to <code>lhs</code>.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @return whether <code>rhs</code> parsed
 */
extern CGUL_EXPORT void
cgul_big_integer__assign_from_string(cgul_exception_t* cex,
                                     cgul_big_integer_t lhs,
                                     const char* rhs);

/**
 * Assign the negative of <code>rhs</code> to <code>lhs</code>.  It is
 * safe and efficient for <code>rhs</code> and <code>lhs</code> to
 * point to the same underlying <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__negate(cgul_exception_t* cex,
                         cgul_big_integer_t lhs,
                         cgul_big_integer_t rhs);

/**
 * Assign the absolute value of <code>rhs</code> to <code>lhs</code>.
 * It is safe and efficient for <code>rhs</code> and <code>lhs</code>
 * to point to the same underlying <code>cgul_big_integer</code>
 * object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__abs(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs);

/**
 * This method uses Newton's Method to calculate the scaled
 * multiplicative inverse of <code>rhs</code> and saves it in
 * <code>lhs</code>.  It is permissible for <code>lhs</code> and
 * <code>rhs</code> to point to the same object. <p>
 *
 * The <code>big_integer</code> class does not handle fractional parts
 * of numbers; instead, <code>scale</code> is used to shift
 * <code>scale</code> bits from the fractional side of the radix point
 * to the whole-number side.  The remaining bits in the fractional
 * part are truncated: <p>
 *
 * <pre>
 *            1        scale
 *     lhs = ---  *  2
 *           rhs
 * </pre>
 *
 * Typically, this method works in conjunction with
 * <code>cgul_big_integer__divide_newton()</code> where
 * <code>big_integer</code> corresponds to the divisor.  In this case
 * <code>scale</code> should be at least as large as the
 * following: <p>
 *
 * <pre>
 *     scale = cgul_big_integer__get_inversion_scale(cex, dividend, divisor);
 * </pre>
 *
 * If the same divisor will be used repeatedly, it is usually better
 * to calculate the inverse once, and then perform division by
 * repeatedly multiplying by the inverse using
 * <code>cgul_big_integer__divide_newton()</code>. <p.
 *
 * For example, many pseudo-random number generators (PRNGs)
 * repeatedly divide by the same modulus.  These PRNGs will usually
 * see a significant increase in performance if they multiply by the
 * inverse rather than calculating the division from scratch each
 * time. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @param[in] scale scale
 * @see cgul_big_integer__get_inversion_scale()
 * @see cgul_big_integer__divide_newton()
 */
extern CGUL_EXPORT void
cgul_big_integer__invert(cgul_exception_t* cex,
                         cgul_big_integer_t lhs,
                         cgul_big_integer_t rhs,
                         size_t scale);

/**
 * This method calculates the scale (as a power of 2) needed for
 * inverting <code>divisor</code> using
 * <code>cgul_big_integer__invert()</code> such that it can accurately
 * divide <code>dividend</code> when both are passed into
 * <code>cgul_big_integer__divide_newton()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] dividend dividend
 * @param[in] divisor divisor
 * @return scale
 * @see cgul_big_integer__invert()
 * @see cgul_big_integer__divide_newton()
 */
extern CGUL_EXPORT size_t
cgul_big_integer__get_inversion_scale(cgul_exception_t* cex,
                                      cgul_big_integer_t dividend,
                                      cgul_big_integer_t divisor);

/**
 * Add <code>rhs1</code> to <code>rhs2</code> and store the result in
 * <code>lhs</code>.  It is safe for <code>rhs1</code> or
 * <code>rhs2</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__add(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2);

/**
 * Subtract <code>rhs2</code> from <code>rhs1</code> and store the
 * result in <code>lhs</code>.  It is safe for <code>rhs1</code> or
 * <code>rhs2</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__subtract(cgul_exception_t* cex,
                           cgul_big_integer_t lhs,
                           cgul_big_integer_t rhs1,
                           cgul_big_integer_t rhs2);

/**
 * Multiply <code>rhs1</code> with <code>rhs2</code> and store the
 * result in <code>lhs</code>.  This method picks a suitable
 * multiplication algorithm given the size of the factors
 * <code>rhs1</code> and <code>rhs2</code>.  If in doubt, this is the
 * method to use for multiplication. <p>
 *
 * It is safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object.  This method throws an
 * exception if memory cannot be allocated. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__multiply(cgul_exception_t* cex,
                           cgul_big_integer_t lhs,
                           cgul_big_integer_t rhs1,
                           cgul_big_integer_t rhs2);

/**
 * This method implements multiplication using the "long
 * multiplication" algorithm.  It multiplies <code>rhs1</code> with
 * <code>rhs2</code> and stores the result in <code>lhs</code>.  If in
 * doubt, use <code>cgul_big_integer__multiply()</code> instead. <p>
 *
 * It is safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object.  This method throws an
 * exception if memory cannot be allocated. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__multiply_long(cgul_exception_t* cex,
                                cgul_big_integer_t lhs,
                                cgul_big_integer_t rhs1,
                                cgul_big_integer_t rhs2);

/**
 * This method implements multiplication using the Karatsuba
 * algorithm.  It multiplies <code>rhs1</code> with <code>rhs2</code>
 * and stores the result in <code>lhs</code>.  If in doubt, use
 * <code>cgul_big_integer__multiply()</code> instead. <p>
 *
 * It is safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object.  This method throws an
 * exception if memory cannot be allocated. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__multiply_karatsuba(cgul_exception_t* cex,
                                     cgul_big_integer_t lhs,
                                     cgul_big_integer_t rhs1,
                                     cgul_big_integer_t rhs2);

/**
 * Divide <code>dividend</code> by <code>divisor</code> and store the
 * result in <code>quotient</code> and <code>remainder</code>.  This
 * method picks a suitable division algorithm given the size of
 * <code>dividend</code> and <code>divisor</code>.  If in doubt, this
 * is the method to use for division. <p>
 *
 * It is safe for <code>dividend</code> or <code>divisor</code> and
 * <code>quotient</code> or <code>remainder</code> to point to the
 * same underlying <code>cgul_big_integer</code> object.  This method
 * throws an exception if division by zero is attempted or if memory
 * cannot be allocated. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] quotient quotient
 * @param[out] remainder remainder
 * @param[in] dividend dividend
 * @param[in] divisor divisor
 */
extern CGUL_EXPORT void
cgul_big_integer__divide(cgul_exception_t* cex,
                         cgul_big_integer_t quotient,
                         cgul_big_integer_t remainder,
                         cgul_big_integer_t dividend,
                         cgul_big_integer_t divisor);

/**
 * This method implements division using the "long division"
 * algorithm.  It divides <code>dividend</code> by
 * <code>divisor</code> and stores the result in <code>quotient</code>
 * and <code>remainder</code>.  If in doubt, use
 * <code>cgul_big_integer__divide()</code> instead. <p>
 *
 * It is safe for <code>dividend</code> or <code>divisor</code> and
 * <code>quotient</code> or <code>remainder</code> to point to the
 * same underlying <code>cgul_big_integer</code> object.  This method
 * throws an exception if division by zero is attempted or if memory
 * cannot be allocated.
 *
 * @param[in,out] cex c-style exception
 * @param[out] quotient quotient
 * @param[out] remainder remainder
 * @param[in] dividend dividend
 * @param[in] divisor divisor
 */
extern CGUL_EXPORT void
cgul_big_integer__divide_long(cgul_exception_t* cex,
                              cgul_big_integer_t quotient,
                              cgul_big_integer_t remainder,
                              cgul_big_integer_t dividend,
                              cgul_big_integer_t divisor);

/**
 * This method implements division by multiplying
 * <code>dividend</code> by the pre-calculated, scaled
 * <code>reciprocal</code> of the <code>divisor</code>.  If in doubt,
 * use <code>cgul_big_integer__divide()</code> instead. <p>
 *
 * The <code>divisor</code> must be passed in in addition to
 * <code>reciprocal</code> because the <code>reciprocal</code> is an
 * approximation with some round-off error.  The <code>divisor</code>
 * is used to make minor adjustments to the result of the
 * multiplication so that <code>quotient</code> and
 * <code>remainder</code> are exact. <p>
 *
 * This method is especially useful when repeated divisions need to be
 * performed using the same divisor because the reciprocal of the
 * divisor needs to be calculated just once.  Each division can then
 * be performed using multiplication which is faster. <p>
 *
 * It is safe for <code>dividend</code> or <code>divisor</code> and
 * <code>quotient</code> or <code>remainder</code> to point to the
 * same underlying <code>cgul_big_integer</code> object.  This method
 * throws an exception if division by zero is attempted or if memory
 * cannot be allocated. <p>
 *
 * The basic set up follows: <p>
 *
 * <pre>
 *      scale = cgul_big_integer__get_inversion_scale(cex, dividend, divisor);
 *      cgul_big_integer__invert(cex, reciprocal, divisor, scale);
 *      cgul_big_integer__divide_newton(cex,
 *                                      quotient,
 *                                      remainder,
 *                                      dividend,
 *                                      divisor,
 *                                      reciprocal,
 *                                      scale);
 * </pre>
 *
 * Often, the size of the dividend will be reduced after each
 * iteration.  In this case it is usually possible to almost halve the
 * total run time by keeping the size of the reciprocal in sync with
 * the size of the divisor as shown in the following code: <p>
 *
 * <pre>
 *     reciprocal_size = cgul_big_integer__get_size(cex, reciprocal);
 *     dividend_size = cgul_big_integer__get_size(cex, dividend);
 *     if (reciprocal_size > dividend_size + 17) {
 *         cgul_big_integer__shift_right(cex,
 *                                       reciprocal,
 *                                       reciprocal,
 *                                       64);
 *         scale -= 64;
 *     }
 * </pre>
 *
 * This method derives its name from the fact that Newton's Method is
 * used to calculate the inverse of the divisor. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] quotient quotient
 * @param[out] remainder remainder
 * @param[in] dividend dividend
 * @param[in] divisor divisor
 * @param[in] reciprocal reciprocal
 * @param[in] scale scale applied to <code>reciprocal</code>
 * @see cgul_big_integer__get_inversion_scale()
 * @see cgul_big_integer__invert()
 */
extern CGUL_EXPORT void
cgul_big_integer__divide_newton(cgul_exception_t* cex,
                                cgul_big_integer_t quotient,
                                cgul_big_integer_t remainder,
                                cgul_big_integer_t dividend,
                                cgul_big_integer_t divisor,
                                cgul_big_integer_t reciprocal,
                                size_t scale);

/**
 * This method calculates <code>base</code> raised to the power of
 * <code>exponent</code> and saves the result in <code>lhs</code>.
 * If <code>exponent</code> is negative, an exception is thrown.  If
 * the program runs out of memory, an exception is thrown. <p>
 *
 * This method is not as effecient as
 * <code>cgul_big_integer__power_modulo()</code> when the client is
 * only interested in the modulo and not the power. <p>
 *
 * @note For this implementation, pow(0,0) = 1.  This is in agreement
 * with some, but others leave it undefined. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] base base
 * @param[in] exponent exponent
 * @see cgul_big_integer__power_modulo()
 */
extern CGUL_EXPORT void
cgul_big_integer__power(cgul_exception_t* cex,
                        cgul_big_integer_t lhs,
                        cgul_big_integer_t base,
                        cgul_big_integer_t exponent);

/**
 * This method calculates the modulo of <code>base</code> to the power
 * of <code>exponent</code> using <code>modulus</code> as the
 * modulus.  If <code>exponent</code> is negative, an exception is
 * thrown.  If the program runs out of memory, an exception is
 * thrown. <p>
 *
 * This method is much more effecient than
 * <code>cgul_big_integer__power()</code> when the client is only
 * interested in the modulo and not the power because the modulo
 * operation is used at each stage of the calculation to keep the size
 * of the numbers within reason. <p>
 *
 * @note For this implementation, pow(0,0) = 1.  This is in agreement
 * with some, but others leave it undefined. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] base base
 * @param[in] exponent exponent
 * @param[in] modulus modulus
 * @see cgul_big_integer__power()
 */
extern CGUL_EXPORT void
cgul_big_integer__power_modulo(cgul_exception_t* cex,
                               cgul_big_integer_t lhs,
                               cgul_big_integer_t base,
                               cgul_big_integer_t exponent,
                               cgul_big_integer_t modulus);

/**
 * This method uses the Miller-Rabin Primality Test to determine
 * whether <code>big_integer</code> is probably prime.  The error per
 * iteration for Miller-Rabin is bound by 1/4.  This method performs
 * <code>count</code> iterations.  The following table shows the
 * theoretical upper bound for the error associated with different
 * values of <code>count</code>: <p>
 *
 * <pre>
 *     count | error
 *     ------------------------
 *        10 | 9.54e-7
 *        25 | 8.88e-16
 *        50 | 7.89e-31
 *       100 | 6.22e-61
 *       500 | 9.33e-302
 *      1000 | 8.71e-603
 * </pre>
 *
 * If an exception is not thrown, this method returns <code>1</code>
 * if <code>big_integer</code> is probably prime and <code>0</code> if
 * it is definitely composite. <p>
 *
 * This method throws an exception if <code>big_integer</code> is
 * negative.  It also throws an exception if the program runs out of
 * memory.  If an exception is thrown, <code>0</code> is returned. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] count number of times test is independently executed
 * @return whether <code>big_integer</code> is probably prime
 */
extern CGUL_EXPORT int
cgul_big_integer__is_probably_prime(cgul_exception_t* cex,
                                    cgul_big_integer_t big_integer,
                                    unsigned int count);

/**
 * Shift <code>rhs</code> left by <code>bit_count</code> and put the
 * result in <code>lhs</code>.  It is safe for <code>rhs</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object. <p>
 *
 * Note that this method is a base-2 shift that shifts individual
 * bits.  Contrast this method with
 * <code>cgul_bcd__shift_left()</code> which shifts individual base-10
 * digits instead. <p>
 *
 * This method can be used as a fast way to multiply by a power of 2. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @param[in] bit_count how far to shift <code>rhs</code>
 */
extern CGUL_EXPORT void
cgul_big_integer__shift_left(cgul_exception_t* cex,
                             cgul_big_integer_t lhs,
                             cgul_big_integer_t rhs,
                             size_t bit_count);

/**
 * Shift <code>rhs</code> right by <code>bit_count</code> and put the
 * result in <code>lhs</code>.  It is safe for <code>rhs</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object. <p>
 *
 * Note that this method is a base-2 shift that shifts individual
 * bits.  Contrast this method with
 * <code>cgul_bcd__shift_right()</code> which shifts individual
 * base-10 digits instead. <p>
 *
 * This method can be used as a fast way to divide by a power of 2. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs right-hand side
 * @param[in] bit_count how far to shift <code>rhs</code>
 */
extern CGUL_EXPORT void
cgul_big_integer__shift_right(cgul_exception_t* cex,
                              cgul_big_integer_t lhs,
                              cgul_big_integer_t rhs,
                              size_t bit_count);

/**
 * This method effectively assigns the value of <code>rhs1</code> to
 * <code>lhs</code> and then performs a bit-wise AND of
 * <code>lhs</code> and <code>rhs2</code>.  It is important to
 * understand that this means that <code>lhs</code> will have the same
 * sign as <code>rhs1</code> when this method returns. <p>
 *
 * It is safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__and(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2);

/**
 * This method effectively assigns the value of <code>rhs1</code> to
 * <code>lhs</code> and then performs a bit-wise OR of
 * <code>lhs</code> and <code>rhs2</code>.  It is important to
 * understand that this means that <code>lhs</code> will have the same
 * sign as <code>rhs1</code> when this method returns. <p>
 *
 * It is safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__or(cgul_exception_t* cex,
                     cgul_big_integer_t lhs,
                     cgul_big_integer_t rhs1,
                     cgul_big_integer_t rhs2);

/**
 * This method effectively assigns the value of <code>rhs1</code> to
 * <code>lhs</code> and then performs a bit-wise XOR of
 * <code>lhs</code> and <code>rhs2</code>.  It is important to
 * understand that this means that <code>lhs</code> will have the same
 * sign as <code>rhs1</code> when this method returns. <p>
 *
 * It is safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__xor(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2);

/**
 * Return less than zero if <code>lhs</code> is less than
 * <code>rhs</code>, return zero if <code>lhs</code> equals
 * <code>rhs</code>, and return greater than zero if <code>lhs</code>
 * is greater than <code>rhs</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 * @return comparison of <code>lhs</code> and <code>rhs</code>
 */
extern CGUL_EXPORT int
cgul_big_integer__compare(cgul_exception_t* cex,
                          cgul_big_integer_t lhs,
                          cgul_big_integer_t rhs);

/**
 * Compare just the magnitudes of <code>lhs</code> and
 * <code>rhs</code>.  This function returns less than zero if lhs is
 * less than rhs, zero if lhs equals rhs, and greater than zero if the
 * magnitude of <code>lhs</code> is greater than the magnitude of
 * <code>rhs</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 * @return comparison of the magnitudes of <code>lhs</code> and <code>rhs</code>
 */
extern CGUL_EXPORT int
cgul_big_integer__compare_magnitudes(cgul_exception_t* cex,
                                     cgul_big_integer_t lhs,
                                     cgul_big_integer_t rhs);

/**
 * This method finds the greatest common divisor (GCD) of
 * <code>rhs1</code> and <code>rhs2</code> and saves the result in
 * <code>lhs</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__gcd(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2);

/**
 * This method finds the least common multiple (LCM) of
 * <code>rhs1</code> and <code>rhs2</code> and saves the result
 * in <code>lhs</code>.  The value returned is often used as the
 * lowest common denominator (LCD).  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
extern CGUL_EXPORT void
cgul_big_integer__lcm(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2);

/**
 * Swap the internal data quickly without copying.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer_1 first <code>cgul_big_integer</code> instance
 * @param[in] big_integer_2 second <code>cgul_big_integer</code> instance
 */
extern CGUL_EXPORT void
cgul_big_integer__swap(cgul_exception_t* cex,
                       cgul_big_integer_t big_integer_1,
                       cgul_big_integer_t big_integer_2);

/**
 * Get the internal array of bytes that comprise the value.  The main
 * purpose of this method is to allow the client to access individual
 * bytes after a PRNG with a good bit distribution (like
 * Blum-Blum-Shub) returns a pseudo-random number.  The client can use
 * <code>cgul_big_integer__get_size()</code> to determine the number
 * of bytes in the array. <p>
 *
 * The client must not free the pointer that is returned because it is
 * still owned by <code>v</code>.  The client also must not alter the
 * values in the array.  The pointer returned is only valid until the
 * next math operation. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return internal array
 */
extern CGUL_EXPORT const unsigned char*
cgul_big_integer__get_array(cgul_exception_t* cex,
                            cgul_big_integer_t big_integer);

/**
 * Return the number of bytes in the internal array.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return size of the internal array
 */
extern CGUL_EXPORT size_t
cgul_big_integer__get_size(cgul_exception_t* cex,
                           cgul_big_integer_t big_integer);

/**
 * Convert the value of <code>big_integer</code> to a long.  This
 * function returns 1 if the conversion can be accomplished without
 * losing any significant bits; otherwise, it returns 0.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[out] value converted value
 * @return whether a conversion occurred
 */
extern CGUL_EXPORT int
cgul_big_integer__as_long(cgul_exception_t* cex,
                          cgul_big_integer_t big_integer,
                          long int* value);

/**
 * Convert the value of <code>big_integer</code> to an unsigned long.
 * This function returns 1 if the conversion can be accomplished
 * without losing any significant bits; otherwise, it returns 0.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[out] value converted value
 * @return whether a conversion occurred
 */
extern CGUL_EXPORT int
cgul_big_integer__as_unsigned_long(cgul_exception_t* cex,
                                   cgul_big_integer_t big_integer,
                                   unsigned long int* value);

/**
 * This method converts <code>big_integer</code> to a
 * <code>cgul_bcd</code> object and saves the conversion to
 * <code>bcd</code> which the caller has already instantiated.  If an
 * error occurs, an exception is thrown, and <code>bcd</code> is not
 * altered.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[out] bcd <code>cgul_bcd</code> instance
 */
extern CGUL_EXPORT void
cgul_big_integer__as_bcd(cgul_exception_t* cex,
                         cgul_big_integer_t big_integer,
                         cgul_bcd_t bcd);

/**
 * Return a <code>char*</code> object that holds the string
 * representation of <code>big_integer</code> as a radix-2 number.
 * The caller is responsible for calling <code>free()</code> on the
 * pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return string representation of <code>big_integer</code> as binary
 */
extern CGUL_EXPORT char*
cgul_big_integer__as_binary(cgul_exception_t* cex,
                            cgul_big_integer_t big_integer);

/**
 * Append the binary representation of <code>big_integer</code>
 * to <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_big_integer__append_as_binary(cgul_exception_t* cex,
                                   cgul_big_integer_t big_integer,
                                   cgul_string_t s);

/**
 * Return a <code>char*</code> that holds the string representation of
 * <code>big_integer</code> as a radix-8 number.  The caller is
 * responsible for calling <code>free()</code> on the pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return string representation of <code>big_integer</code> as octal
 */
extern CGUL_EXPORT char*
cgul_big_integer__as_octal(cgul_exception_t* cex,
                           cgul_big_integer_t big_integer);

/**
 * Append the octal representation of <code>big_integer</code> to
 * <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_big_integer__append_as_octal(cgul_exception_t* cex,
                                  cgul_big_integer_t big_integer,
                                  cgul_string_t s);

/**
 * Return a <code>char*</code> object that holds the string
 * representation of <code>big_integer</code> as a radix-10 number.
 * The caller is responsible for calling <code>free()</code> on the
 * pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return string representation of <code>big_integer</code> as decimal
 */
extern CGUL_EXPORT char*
cgul_big_integer__as_decimal(cgul_exception_t* cex,
                             cgul_big_integer_t big_integer);

/**
 * Append the decimal representation of <code>big_integer</code>
 * to <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_big_integer__append_as_decimal(cgul_exception_t* cex,
                                    cgul_big_integer_t big_integer,
                                    cgul_string_t s);

/**
 * Return a <code>char*</code> object that holds the string
 * representation of <code>big_integer</code> as a radix-16 number.
 * The caller is responsible for calling <code>free()</code> on the
 * pointer returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @return string representation of <code>big_integer</code> as hexadecimal
 */
extern CGUL_EXPORT char*
cgul_big_integer__as_hexadecimal(cgul_exception_t* cex,
                                 cgul_big_integer_t big_integer);

/**
 * Append the hexadecimal representation of <code>big_integer</code>
 * to <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] s where to append the string representation
 */
extern CGUL_EXPORT void
cgul_big_integer__append_as_hexadecimal(cgul_exception_t* cex,
                                        cgul_big_integer_t big_integer,
                                        cgul_string_t s);

/**
 * Print the string representation of <code>big_integer</code> as a radix-2
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_big_integer__print_as_binary(cgul_exception_t* cex,
                                  cgul_big_integer_t big_integer,
                                  FILE* f);

/**
 * Print the string representation of <code>big_integer</code> as a radix-8
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_big_integer__print_as_octal(cgul_exception_t* cex,
                                 cgul_big_integer_t big_integer,
                                 FILE* f);

/**
 * Print the string representation of <code>big_integer</code> as a radix-10
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_big_integer__print_as_decimal(cgul_exception_t* cex,
                                   cgul_big_integer_t big_integer,
                                   FILE* f);

/**
 * Print the string representation of <code>big_integer</code> as a radix-16
 * number on <code>f</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] f output file
 */
extern CGUL_EXPORT void
cgul_big_integer__print_as_hexadecimal(cgul_exception_t* cex,
                                       cgul_big_integer_t big_integer,
                                       FILE* f);

CGUL_END_C

#endif  /*  CGUL_BIG_INTEGER_H  */
