#ifndef CGUL_CACHE_CXX_H
#define CGUL_CACHE_CXX_H

/**
 * C++ wrapper for <code>cgul_cache</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_cache</code>
 */

#include "cgul_cache.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_cache</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_cache</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_cache</code>
 */
class cgul_cache_cxx {

public:

    /*
     * The next three typedef statements define the "cacheable" interface.
     * For any class of objects the client wants to cache, the "cacheable"
     * interface must be implemented.
     *
     * Arguably, these typedef should be bundled together in an
     * abstract base class or something that is the functional
     * equivalent of Java's "interface" keyword, but it doesn't really
     * buy anything except added complexity.
     */

    /**
     * This function pointer typedef is needed so that the
     * <code>cgul_cache_cxx</code> knows how to create new default instances
     * of the objects you want to cache.
     */
    typedef cgul_cache__constructor_t constructor_t;

    /**
     * This function pointer typedef is needed so that the
     * <code>cgul_cache_cxx</code> knows how to delete instances of the
     * objects you want to cache.  The client's implementation of this
     * function should not throw an exception.
     */
    typedef cgul_cache__destructor_t destructor_t;

    /**
     * This function pointer typedef is needed so that the
     * <code>cgul_cache_cxx</code> knows how take a fully constructed object
     * and return it to its initial default state.  This callback is
     * invoked when objects are put back on the cache to minimize the
     * memory footprint of objects on the cache.  If the restructor is
     * set to <code>NULL</code>, the client is responsible for resetting
     * the object before putting it back on the cgul_cache.
     */
    typedef cgul_cache__restructor_t restructor_t;

    /**
     * Static method used to return the <code>cgul_cache</code> instance
     * that is an adapter for <code>free()</code>.  Because the object
     * returned is statically allocated, the client should not attempt to
     * call <code>cgul_cache__delete()</code> on the pointer returned.  If
     * the client forgets, <code>cgul_cache__delete()</code> will ignore
     * the request. <p>
     *
     * The purpose of this static method is to allow clients to use
     * methods (like <code>cgul_rbtree_cxx::remove_range()</code>)
     * that accept <code>cgul_cache_cxx</code> objects for cleaning up
     * keys or values even though the objects are allocated directly
     * using one of the <code>malloc()</code> family of function calls
     * (instead of through the <code>cgul_cache</code> interface). <p>
     *
     * The <code>cgul_cache</code> object returned only implements the
     * <code>destructor_t</code> part of the <code>cgul_cache</code>
     * interface.  Thus, any attempt to use the object returned to
     * construct a new object will result in <code>NULL</code> being
     * returned. <p>
     *
     * Because portable C++ has limitations with respect to
     * initializing static members, this static method returns an
     * instance of <code>cgul_cache</code> instead of an instance of
     * <code>cgul_cache_cxx</code>.  Thus, the user is required to
     * wrap the return value as follows: <p>
     *
     * <pre>
     *     cgul_cache_cxx freer(cgul_cache_cxx::get_freer());
     *     trie.clear(&freer, NULL);
     * </pre>
     *
     * @return <code>cgul_cache</code> adopter for <code>free()</code>
     */
    static cgul_cache_t
    get_freer()
    {
        return cgul_cache__get_freer();
    }

    /**
     * Create a new <code>cgul_cache_cxx</code> instance.  If
     * <code>restructor</code> is set to <code>NULL</code>, the client
     * is responsible for resetting the object before putting it back
     * on the <code>cgul_cache_cxx</code>.  If memory cannot be
     * allocated, an exception is thrown. <p>
     *
     * By default, the maximum cache size is set to 0.  This means no
     * objects are cached by default.  When you want to enable caching,
     * you can change the maximum cache size by calling
     * <code>set_size()</code>.
     *
     * @param[in] constructor constructor for the cached objects
     * @param[in] destructor destructor for the cached objects
     * @param[in] restructor restructor for the cached objects
     */
    cgul_cache_cxx(constructor_t constructor,
                   destructor_t destructor,
                   restructor_t restructor)
    {
        cgul_exception_t local = NULL;
        obj = cgul_cache__new(&local, constructor, destructor, restructor);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_cache_cxx</code> object by wrapping an
     * existing <code>cgul_cache</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_cache_cxx(cgul_cache_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method first calls the destructor for each object that is
     * still cached.  It then frees all of its internally allocated
     * resources.
     */
    virtual
    ~cgul_cache_cxx()
    {
        if (obj) {
            cgul_cache__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Get the maximum number of objects the cache can hold when the
     * reserve is zero.  To get the total number of elements the cache can
     * hold, you must add the reserve to the size.
     *
     * @return maximum size of the cache
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_cache__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the maximum number of objects the cache can hold when the
     * reserve is zero. <p>
     *
     * Calling this method with <code>size</code> less than the reserve
     * does not free any of the reserved objects; however, it may free
     * objects that have recently been "moved" from the reserve to the
     * main cache.  (See <code>set_reserve()</code>.) <p>
     *
     * You can release all the cached objects and disable caching of
     * objects by setting the cache reserve to 0 and the cache size to 0
     * (in that order).  This is the default. <p>
     *
     * If an error occurs (while enlarging the cache), an exception is
     * thrown.
     *
     * @param[in] size maximum size of the cache
     */
    virtual void
    set_size(unsigned long int size)
    {
        cgul_exception_t local = NULL;
        cgul_cache__set_size(&local, obj, size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the reserve limit.  This is the number of objects that are
     * guaranteed to be held in reserve by the cache.  These objects will
     * not be returned by the cache until the reserve is reduced.  The
     * reserve is independent of the size of the cache.
     *
     * @return reserve
     */
    virtual unsigned long int
    get_reserve() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_cache__get_reserve(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the reserve limit.  This method causes the cache to
     * reserve objects.  The reserved objects are held indefinitely until
     * the reserve is reduced.  The reserve is independent of the size of
     * the cache. <p>
     *
     * It is important to understand that if these reserved objects are
     * not already being held by the cache, the cache will instantiate
     * them.  This allows the client to immediately instantiate objects
     * without having to worry about out-of-memory exceptions when the
     * objects are used later. <p>
     *
     * Note that if you want to get a pointer to one of the reserved
     * object, you first call <code>set_reserve()</code> to
     * reduce the number of the reserved objects.  This simply "moves" the
     * now unreserved objects to the main cache of objects.  You then
     * immediately call <code>get()</code> to get an object
     * from the main cache. <p>
     *
     * It is done like this so that you can code strictly in terms of
     * <code>get()</code>.  You can then reserve an object
     * now and trigger when it is used later just by calling this method.
     * The code that uses <code>get()</code> does not need to
     * know anything about reserved objects.  It will just naturally use
     * them. <p>
     *
     * Objects that are "moved" from the reserve to the main cache are not
     * guaranteed to be available if <code>set_size()</code>
     * has since been called.
     *
     * @param[in] reserve reserve
     */
    virtual void
    set_reserve(unsigned long int reserve)
    {
        cgul_exception_t local = NULL;
        cgul_cache__set_reserve(&local, obj, reserve);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Try to return a re-initialized object from the cache; if the
     * cache is empty, a newly instantiated object will be returned
     * instead.  If an error occurs, an exception is thrown.
     *
     * @return cached or new object
     */
    virtual void*
    get()
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_cache__get(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Try to put <code>o</code> on the cache.  If the cache is full,
     * <code>o</code> will be delete instead.
     *
     * @param[in] o objct to cache or delete
     */
    virtual void
    put(void* o)
    {
        cgul_exception_t local = NULL;
        cgul_cache__put(&local, obj, o);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_cache</code> object.
     *
     * @return underlying object
     */
    virtual cgul_cache_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_cache</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_cache__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_cache_t
    take_obj()
    {
        cgul_cache_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_cache_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_cache__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_cache_cxx(const cgul_cache_cxx&);
    cgul_cache_cxx& operator=(const cgul_cache_cxx&);

    /* Underlying <code>cgul_cache</code> object. */
    cgul_cache_t obj;
};

#endif  /*  CGUL_CACHE_CXX_H  */
