#ifndef CGUL_CHAR_CXX_H
#define CGUL_CHAR_CXX_H

/**
 * C++ wrapper for <code>cgul_char</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_char</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_char.h"

/**
 * This class provides the C++ bindings for C <code>cgul_char</code>
 * external functions.  The main purpose of this class is to convert
 * the C-style function calls and exception handling in
 * <code>cgul_char</code> into C++-style function calls and exception
 * handling.
 *
 * @brief C++ bindings for <code>cgul_char</code>
 */
class cgul_char_cxx {

public:

    /**
     * Perform a case-significant, 32-bit, string comparison of
     * <code>s1</code> and <code>s2</code>.  Return -1, 0, or 1 if
     * <code>s1</code> is less than, equal to, or greater than
     * <code>s2</code> respectively. <p>
     *
     * For simplicity, this function compares strings stricly by the
     * ordinal value of each character which should be sufficient, for
     * example, when comparing strings when inserting them as keys into a
     * <code>cgul_rbtree_cxx</code>; however, the comparison results are not
     * likely to be the same as what <code>strcoll()</code> would return
     * if used on the UTF-8 version of the strings. <p>
     *
     * @param[in] s1 left-hand side
     * @param[in] s2 right-hand side
     * @return -1, 0, or 1 if <code>s1</code> is less than, equal to, or
     * greater than <code>s2</code> respectively
     * @see cgul_string_cxx::compare()
     */
    static int
    strcmp(const char* s1,
           const char* s2)
    {
        return cgul_char__strcmp(s1, s2);
    }

    /**
     * Return whether the character <code>c</code> is considered to be
     * white-space.  To avoid locale dependencies, white-space is defined
     * as any character from the following list: ' ', '\\t', '\\n', '\\r',
     * '\\f', '\\v'
     *
     * @param[in] c character
     * @return whether the character is white-space
     */
    static int
    isspace(char c)
    {
        int rv;
        cgul_exception_t local = NULL;
        rv = cgul_char__isspace(&local, c);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};

#endif  /*  CGUL_CHAR_CXX_H  */
