#ifndef CGUL_CRLF_H
#define CGUL_CRLF_H

/**
 * Convert blocks of DOS, Mac, and Unix text to lines of arbitrary
 * length. <p>
 *
 * This class operates on in-memory buffers.  Typically, you will want
 * to use <code>cgul_crlf_file</code> instead of this class. <p>
 *
 * For efficiency, this class overwrites parts of the blocks passed
 * into <code>cgul_crlf__convert()</code> with NUL characters.  Thus,
 * if the block is important, the client should pass in a copy of the
 * block instead of the original. <p>
 *
 * Furthermore, because this class inserts NUL characters into the
 * buffer it converts, the buffer must be writable.  Because many
 * operating systems make statically-allocated or stack-allocated
 * variables read-only, this can lead to segfaults if the buffer is
 * not dynamically allocated. <p>
 *
 * @file
 * @author Paul Serice
 * @brief convert blocks of text to lines
 * @see cgul_crlf_file
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_crlf</code> instance. */
typedef struct cgul_crlf* cgul_crlf_t;

/**
 * Create a new <code>cgul_crlf</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_crlf__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_crlf</code> instance
 */
extern CGUL_EXPORT cgul_crlf_t
cgul_crlf__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  Do not try to
 * access <code>crlf</code> after calling this method.
 *
 * @param[in] crlf <code>cgul_crlf</code> instance
 */
extern CGUL_EXPORT void
cgul_crlf__delete(cgul_crlf_t crlf);

/**
 * This method is used to reset the <code>crlf</code> object so that
 * it can process a new stream of text or process the same stream of
 * text after seeking to a different location in the stream. <p>
 *
 * The client must inform this class of the new offset into the
 * underlying file so that subsequent calls to
 * <code>cgul_crlf__get_line_offset()</code> can return correct
 * values.  The value of <code>offset</code> should be zero-based.
 * Thus, to start processing at the beginning of a new file call this
 * method with <code>offset</code> set to <code>0</code>. <p>
 *
 * Calling this method resets the line count to zero.  This can be
 * used by the client to implement a line counter that does not
 * overflow. <p>
 *
 * Calling this method does not reset whether a leading UTF-8
 * byte-order mark (BOM) is stripped. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @param[in] offset zero-based offset
 */
extern CGUL_EXPORT void
cgul_crlf__reset(cgul_exception_t* cex,
                 cgul_crlf_t crlf,
                 unsigned long offset);

/**
 * This method returns whether the leading UTF-8 byte-order mark (BOM)
 * should be stripped from the first line if it is present.
 *
 * @param[in] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @return whether to strip a leading UTF-8 byte-order mark
 */
extern CGUL_EXPORT int
cgul_crlf__get_strip_utf8_bom(cgul_exception_t* cex,
                              cgul_crlf_t crlf);

/**
 * By default, this class detects the leading UTF-8 byte-order mark
 * (BOM) and strips it from the first line returned by
 * <code>cgul_crlf__get_line()</code> if it is present.  It then
 * clears its internal flag so that BOMs internal to the text file
 * will be returned.  This is generally what you want because the
 * leading BOM is not significant but the internal BOMs are. <p>
 *
 * You can alter the way this class handles the leading BOM by calling
 * this method with <code>strip_utf8_bom</code> set to <code>0</code>.
 * This will cause the leading BOM to be returned as part of the first
 * line.  This can be useful, for example, if you just want to convert
 * the text file and are not interested in its contents. <p>
 *
 * The value of <code>strip_utf8_bom</code> is remembered across calls
 * to <code>cgul_crlf__reset()</code>. <p>
 *
 * It should be noted that most operating systems do not save UTF-8
 * text files with a leading BOM because UTF-8 is a character stream
 * and, as such, does not have byte-order problems; however, Microsoft
 * Windows adds the BOM to its UTF-8 text files presumably to help
 * distinguish UTF-8 text files from text files with different
 * encodings. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @param[in] strip_utf8_bom whether to strip leading UTF-8 byte-order mark
 */
extern CGUL_EXPORT void
cgul_crlf__set_strip_utf8_bom(cgul_exception_t* cex,
                              cgul_crlf_t crlf,
                              int strip_utf8_bom);

/**
 * The caller feeds this method a block of text in <code>buf</code> of
 * size <code>bsize</code>.  The blocks you feed this method can end
 * anywhere; they do not have to end exactly on a line boundary.  This
 * method knows how to splice together partial lines from the last
 * call to form arbitrarily long lines using any of the common EOL
 * markers: "\n", "\r", or "\r\n" <p>
 *
 * After each call to this method, you MUST call
 * <code>cgul_crlf__get_line()</code> iteratively until it returns
 * <code>NULL</code> before feeding this method another block. <p>
 *
 * Do not alter <code>buf</code> until after you have exhausted
 * <code>cgul_crlf__get_line()</code>.  This prevents
 * <code>cgul_crlf__convert()</code> from having to make a copy of
 * each block because this method often (but not always) inserts
 * <code>NUL</code> characters directly into <code>buf</code> to
 * produce the lines returned by
 * <code>cgul_crlf__get_line()</code>. <p>
 *
 * After feeding the last block to this method and exhausting
 * <code>cgul_crlf__get_line()</code>, you should call
 * <code>cgul_crlf__get_remainder()</code> to fetch what remained if
 * the last line had no trailing EOL marker. <p>
 *
 * This method dynamically allocates space to hold the lines that are
 * split across calls to this method.  If an error occurs, an
 * exception is thrown, and the <code>crlf</code> object will be in an
 * undefined state. <p>
 *
 * WARNING: Because this class embeds <code>NUL</code> characters
 * directly into <code>buf</code>, it goes without saying that
 * <code>buf</code> must be writable.  What might not be obvious is
 * that this means <code>buf</code> probably should not be statically
 * allocated or allocated on the stack because many operating systems
 * have security mechanisms to prevent unexpected writes to these
 * variables. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @param[in] buf buffer
 * @param[in] bsize buffer size
 */
extern CGUL_EXPORT void
cgul_crlf__convert(cgul_exception_t* cex,
                   cgul_crlf_t crlf,
                   char* buf,
                   unsigned long int bsize);

/**
 * After seeding this object by calling
 * <code>cgul_crlf__convert()</code>, you call this method to fetch
 * the next line.  If a line is ready, this method returns it.  If no
 * line is ready, this method returns <code>NULL</code>.  The caller
 * should not try to call <code>free()</code> on the line returned
 * because it is really just a pointer back to the contents of the
 * buffer passed into <code>cgul_crlf__convert()</code>. <p>
 *
 * If this method does not return <code>NULL</code>, you should keep
 * calling it until it does.  Once it returns <code>NULL</code>, you
 * can either refill this object by calling
 * <code>cgul_crlf__convert()</code> with the next block or call
 * <code>cgul_crlf__get_remainder()</code> to finish. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @return next line of text
 */
extern CGUL_EXPORT const char*
cgul_crlf__get_line(cgul_exception_t* cex,
                    cgul_crlf_t crlf);

/**
 * This method returns the total number of lines returned by
 * <code>cgul_crlf__get_line()</code> and
 * <code>cgul_crlf__get_remainder()</code>.  The line count is
 * one-based.  No attempt is made to prevent the return value from
 * overflowing.  So, the caller is responsible for verifying the
 * return value. <p>
 *
 * Calls to <code>cgul_crlf__reset()</code> reset the line count to
 * zero.  This can be used by the client to implement a line counter
 * that does not overflow. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @return line count
 */
extern CGUL_EXPORT unsigned long
cgul_crlf__get_line_count(cgul_exception_t* cex,
                          cgul_crlf_t crlf);

/**
 * This method returns the offset of the last line returned by
 * <code>cgul_crlf__get_line()</code> or
 * <code>cgul_crlf__get_remainder()</code>.  The offset is zero-based.
 * If you are feeding a binary stream into
 * <code>cgul_crlf__convert()</code> and if the stream is also a
 * random access stream, you can use the return value to directly seek
 * to the line as follows:
 *
 * <pre>
 *     fseek(f, offset, SEEK_SET);
 * </pre>
 *
 * Because the prototype for <code>fseek()</code> requires a
 * <code>long</code> for the offset parameter, no attempt is made to
 * prevent the return value from overflowing.  So, the caller is
 * responsible for verifying the return value. <p>
 *
 * Note that the offset returned is basically the number of bytes from
 * the start of the file to the current line.  This is not necessarily
 * the same as the number of characters which depends on how the file
 * is encoded. <p>
 *
 * To get the offset of the remainder, just call
 * <code>cgul_crlf__get_remainder()</code> before calling this
 * method. <p>
 *
 * This method throws an exception if, after converting a new block,
 * it is called before <code>cgul_crlf__get_line()</code> is
 * called. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @return zero-based offset into file for the current line
 */
extern CGUL_EXPORT unsigned long
cgul_crlf__get_line_offset(cgul_exception_t* cex,
                           cgul_crlf_t crlf);

/**
 * This is the last method you should call, and it should only be
 * called once.  It should be called only after all the blocks have
 * been fed to <code>cgul_crlf__convert()</code> and only after
 * <code>cgul_crlf__get_line()</code> has been exhausted.  At this
 * point, all that is left is the remainder. <p>
 *
 * This method returns <code>NULL</code> if a remainder does not exist
 * or if this method has already been called.  The only time a
 * remainder exists is if the last line in the file is missing the
 * final EOL marker. <p>
 *
 * After calling this function, use
 * <code>cgul_crlf__get_line_offset()</code> to get the offset of the
 * remainder. <p>
 *
 * The caller should not try to call <code>free()</code> on the
 * pointer returned because it points to an internal string that will
 * be freed when <code>cgul_crlf__delete()</code> is called. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @return text remaining after the final EOL
 */
extern CGUL_EXPORT const char*
cgul_crlf__get_remainder(cgul_exception_t* cex,
                         cgul_crlf_t crlf);

/**
 * This method copies <code>fin</code> to <code>fout</code> stripping
 * the original EOL markers and replacing them with <code>eol</code>.
 * <code>fin</code> and <code>fout</code> must have been opened in
 * binary mode.  This method internally uses a <code>cgul_crlf</code>
 * object to perform the conversion.  If an error occurs, an exception
 * is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fin input file
 * @param[out] fout output file
 * @param[in] eol new EOL marker
 */
extern CGUL_EXPORT void
cgul_crlf__convert_file(cgul_exception_t* cex,
                        FILE* fin,
                        FILE* fout,
                        const char* eol);

CGUL_END_C

#endif  /*  CGUL_CRLF_H  */
