#ifndef CGUL_CRLF_CXX_H
#define CGUL_CRLF_CXX_H

/**
 * C++ wrapper for <code>cgul_crlf</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_crlf</code>
 */

#include "cgul_crlf.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_crlf</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_crlf</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_crlf</code>
 */
class cgul_crlf_cxx {

public:

    /**
     * Default Constructor.  If memory cannot be allocated, an
     * exception is thrown.
     */
    cgul_crlf_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_crlf__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_crlf_cxx()
    {
        if (obj) {
            cgul_crlf__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method is used to reset the object so that it can process
     * a new stream of text or process the same stream of text after
     * seeking to a different location in the stream. <p>
     *
     * The client must inform this class of the new offset into the
     * underlying file so that subsequent calls to
     * <code>get_line_offset()</code> can return correct values.  The
     * value of <code>offset</code> should be zero-based.  Thus, to
     * start processing at the beginning of a new file call this
     * method with <code>offset</code> set to <code>0</code>. <p>
     *
     * Calling this method resets the line count to zero.  This can be
     * used by the client to implement a line counter that does not
     * overflow. <p>
     *
     * Calling this method does not reset whether a leading UTF-8
     * byte-order mark (BOM) is stripped. <p>
     *
     * @param[in] offset zero-based offset
     */
    virtual void
    reset(unsigned long offset = 0)
    {
        cgul_exception_t local = NULL;
        cgul_crlf__reset(&local, obj, offset);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns whether the leading UTF-8 byte-order mark (BOM)
     * should be stripped from the first line if it is present.
     *
     * @return whether to strip a leading UTF-8 byte-order mark
     */
    virtual int
    get_strip_utf8_bom() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf__get_strip_utf8_bom(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * By default, this class detects the leading UTF-8 byte-order
     * mark (BOM) and strips it from the first line returned by
     * <code>get_line()</code> if it is present.  It then clears its
     * internal flag so that BOMs internal to the text file will be
     * returned.  This is generally what you want because the leading
     * BOM is not significant but the internal BOMs are. <p>
     *
     * You can alter the way this class handles the leading BOM by calling
     * this method with <code>strip_utf8_bom</code> set to <code>0</code>.
     * This will cause the leading BOM to be returned as part of the first
     * line.  This can be useful, for example, if you just want to convert
     * the text file and are not interested in its contents. <p>
     *
     * The value of <code>strip_utf8_bom</code> is remembered across calls
     * to <code>reset()</code>. <p>
     *
     * It should be noted that most operating systems do not save UTF-8
     * text files with a leading BOM because UTF-8 is a character stream
     * and, as such, does not have byte-order problems; however, Microsoft
     * Windows adds the BOM to its UTF-8 text files presumably to help
     * distinguish UTF-8 text files from text files with different
     * encodings. <p>
     *
     * @param[in] strip_utf8_bom whether to strip leading UTF-8 byte-order mark
     */
    virtual void
    set_strip_utf8_bom(int strip_utf8_bom)
    {
        cgul_exception_t local = NULL;
        cgul_crlf__set_strip_utf8_bom(&local, obj, strip_utf8_bom);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * The caller feeds this method a block of text in <code>buf</code> of
     * size <code>bsize</code>.  The blocks you feed this method can end
     * anywhere; they do not have to end exactly on a line boundary.  This
     * method knows how to splice together partial lines from the last
     * call to form arbitrarily long lines using any of the common EOL
     * markers: "\n", "\r", or "\r\n" <p>
     *
     * After each call to this method, you MUST call
     * <code>get_line()</code> iteratively until it returns
     * <code>NULL</code> before feeding this method another block. <p>
     *
     * Do not alter <code>buf</code> until after you have exhausted
     * <code>get_line()</code>.  This prevents <code>convert()</code>
     * from having to make a copy of each block because this method
     * often (but not always) inserts <code>NUL</code> characters
     * directly into <code>buf</code> to produce the lines returned by
     * <code>get_line()</code>. <p>
     *
     * After feeding the last block to this method and exhausting
     * <code>get_line()</code>, you should call
     * <code>get_remainder()</code> to fetch what remained if the last
     * line had no trailing EOL marker. <p>
     *
     * This method dynamically allocates space to hold the lines that
     * are split across calls to this method.  If an error occurs, an
     * exception is thrown, and the object will be in an undefined
     * state. <p>
     *
     * WARNING: Because this method embeds <code>NUL</code> characters
     * directly into <code>buf</code> to produce the lines returned by
     * <code>get_line()</code>, it goes without saying that
     * <code>buf</code> must be writable.  What might not be obvious
     * is that this means <code>buf</code> probably should not be
     * allocated on the stack because many operating systems have
     * security mechanisms to prevent unexpected writes to stack
     * variables. <p>
     *
     * @param[in] buf buffer
     * @param[in] bsize buffer size
     */
    virtual void
    convert(char* buf,
            unsigned long int bsize)
    {
        cgul_exception_t local = NULL;
        cgul_crlf__convert(&local, obj, buf, bsize);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * After seeding this object by calling <code>convert()</code>,
     * you call this method to fetch the next line.  If a line is
     * ready, this method returns it.  If no line is ready, this
     * method returns <code>NULL</code>.  The caller should not try to
     * call <code>free()</code> on the line returned because it is
     * really just a pointer back to the contents of the buffer passed
     * into <code>convert()</code>. <p>
     *
     * If this method does not return <code>NULL</code>, you should
     * keep calling it until it does.  Once it returns
     * <code>NULL</code>, you can either refill this object by calling
     * <code>convert()</code> with the next block or call
     * <code>get_remainder()</code> to finish. <p>
     *
     * @return next line of text
     */
    virtual const char*
    get_line() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_crlf__get_line(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the total number of lines returned by
     * <code>get_line()</code> and <code>get_remainder()</code>.  The
     * line count is one-based.  No attempt is made to prevent the
     * return value from overflowing.  So, the caller is responsible
     * for verifying the return value. <p>
     *
     * Calls to <code>reset()</code> reset the line count to zero.
     * This can be used by the client to implement a line counter that
     * does not overflow. <p>
     *
     * @return line count
     */
    virtual unsigned long
    get_line_count() const
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf__get_line_count(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the offset of the last line returned by
     * <code>get_line()</code> or <code>get_remainder()</code>.  The
     * offset is zero-based.  If you are feeding a binary stream into
     * <code>convert()</code> and if the stream is also a random
     * access stream, you can use the return value to directly seek to
     * the line as follows:
     *
     * <pre>
     *     fseek(f, offset, SEEK_SET);
     * </pre>
     *
     * Because the prototype for <code>fseek()</code> requires a
     * <code>long</code> for the offset parameter, no attempt is made to
     * prevent the return value from overflowing.  So, the caller is
     * responsible for verifying the return value. <p>
     *
     * Note that the offset returned is basically the number of bytes from
     * the start of the file to the current line.  This is not necessarily
     * the same as the number of characters which depends on how the file
     * is encoded. <p>
     *
     * To get the offset of the remainder, just call
     * <code>get_remainder()</code> before calling this method. <p>
     *
     * This method throws an exception if, after converting a new
     * block, it is called before <code>get_line()</code> is
     * called. <p>
     *
     * @return zero-based offset into file for the current line
     */
    virtual unsigned long
    get_line_offset() const
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_crlf__get_line_offset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This is the last method you should call, and it should only be
     * called once.  It should be called only after all the blocks
     * have been feed to <code>convert()</code> and only after
     * <code>get_line()</code> has been exhausted.  At this point, all
     * that is left is the remainder. <p>
     *
     * This method returns <code>NULL</code> if a remainder does not
     * exist.  The only time a remainder exists is if the last line in the
     * file is missing the final EOL marker. <p>
     *
     * After calling this function, use <code>get_line_offset()</code>
     * to get the offset of the remainder. <p>
     *
     * The caller should not try to call <code>free()</code> on the
     * pointer returned because it points to an internal string that
     * will be freed when <code>delete()</code> is called. <p>
     *
     * @return text remaining after the final EOL
     */
    virtual const char*
    get_remainder()
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_crlf__get_remainder(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method copies <code>fin</code> to <code>fout</code> stripping
     * the original EOL markers and replacing them with <code>eol</code>.
     * <code>fin</code> and <code>fout</code> must have been opened in
     * binary mode.  This method internally uses a <code>cgul_crlf</code>
     * object to perform the conversion.  If an error occurs, an exception
     * is thrown. <p>
     *
     * Note that this method use C-style files because portably
     * dealing with C++ files is very difficult. <p>
     *
     * @param[in] fin input file
     * @param[out] fout output file
     * @param[in] eol new EOL marker
     */
    virtual void
    convert_file(FILE* fin,
                 FILE* fout,
                 const char* eol)
    {
        cgul_exception_t local = NULL;
        cgul_crlf__convert_file(&local, fin, fout, eol);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_crlf</code> object.
     *
     * @return underlying object
     */
    virtual cgul_crlf_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_crlf</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_crlf__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_crlf_t
    take_obj()
    {
        cgul_crlf_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_crlf_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_crlf__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_crlf_cxx(const cgul_crlf_cxx&);
    cgul_crlf_cxx& operator=(const cgul_crlf_cxx&);

    /* Underlying <code>cgul_crlf</code> object. */
    cgul_crlf_t obj;

};

#endif  /*  CGUL_CRLF_CXX_H  */
