#ifndef CGUL_CSV_H
#define CGUL_CSV_H

/**
 * This class is a parser for Comma-Separated Values (CSV) files or
 * Tab-Separated Values (TSV) files.  This class works automatically
 * with DOS, Mac, and Unix text files.  See RFC 4180. <p>
 *
 * @file
 * @author Paul Serice
 * @brief parser for comma-separated values (CSV) files
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_crlf_file.h"
#include "cgul_rbtree.h"
#include "cgul_string.h"
#include "cgul_vector.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_csv</code> instance. */
typedef struct cgul_csv* cgul_csv_t;

/**
 * This method creates a new <code>cgul_csv</code> instance.  After
 * calling this method, an input source should be opened before
 * calling any other method.  Use
 * <code>cgul_csv__set_separator()</code> to change the default
 * separator which is a comma.  The client is responsible for calling
 * <code>cgul_csv__delete()</code> on the value returned.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT cgul_csv_t
cgul_csv__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_csv</code> object and call
 * <code>cgul_csv__open_fname()</code> passing it <code>fname</code>.
 * This method opens <code>fname</code> so that input can be read from
 * it.  For regular files, <code>is_block_buffered</code> should be
 * set to true; for character-oriented or line-oriented files,
 * <code>is_block_buffered</code> should be set to false.  Use
 * <code>cgul_csv__set_separator()</code> to change the default
 * separator which is a comma.  The client is responsible for freeing
 * the object by calling <code>cgul_csv__delete()</code>.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name of stanza file
 * @param[in] is_block_buffered whether <code>fname</code> is block buffered
 * @return new <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT cgul_csv_t
cgul_csv__new_from_fname(cgul_exception_t* cex,
                         const char* fname,
                         int is_block_buffered);

/**
 * Create a new <code>cgul_csv</code> object and call
 * <code>cgul_csv__open_file()</code> passing it <code>f</code>.  This
 * method uses <code>f</code> when reading input.  This class does not
 * take ownership of <code>f</code>.  Thus, the client is still
 * responsible for calling <code>fclose()</code> on <code>f</code>.
 * For regular files, <code>is_block_buffered</code> should be set to
 * true; for character-oriented or line-oriented files,
 * <code>is_block_buffered</code> should be set to false.  Use
 * <code>cgul_csv__set_separator()</code> to change the default
 * separator which is a comma.  The client is responsible for freeing
 * the object by calling <code>cgul_csv__delete()</code>.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[in] is_block_buffered whether <code>f</code> is block buffered
 * @return new <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT cgul_csv_t
cgul_csv__new_from_file(cgul_exception_t* cex,
                        FILE* f,
                        int is_block_buffered);

/**
 * Create a new <code>cgul_csv</code> object and call
 * <code>cgul_csv__open_memory()</code> passing it <code>buf</code>
 * and <code>buf_size</code>.  This method uses <code>buf</code> when
 * reading input.  This class does not take ownership of
 * <code>buf</code>.  Thus, the client is still responsible for
 * calling <code>free()</code> on <code>buf</code> if necessary.  Use
 * <code>cgul_csv__set_separator()</code> to change the default
 * separator which is a comma.  The client is responsible for freeing
 * the object by calling <code>cgul_csv__delete()</code>.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] buf memory buffer
 * @param[in] buf_size size of <code>buf</code> in bytes
 * @return new <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT cgul_csv_t
cgul_csv__new_from_memory(cgul_exception_t* cex,
                          const char* buf,
                          size_t buf_size);

/**
 * Create a new <code>cgul_csv</code> object and call
 * <code>cgul_csv__open_crlf_file()</code> passing it
 * <code>crlf_file</code>.  This makes it possible to read from any
 * file type supported by <code>cgul_crlf_file</code> like memory for
 * example.  Use <code>cgul_csv__set_separator()</code> to change the
 * default separator which is a comma.  The client is responsible for
 * freeing the object returned by calling
 * <code>cgul_csv__delete()</code>.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * The <code>crlf_file</code> object is used as the interface for
 * reading lines of text from a data source.  It should already be
 * initialized before calling this method.  This class does not take
 * ownership of <code>crlf_file</code> so the client is still
 * responsible for calling <code>cgul_crlf_file__delete()</code> on
 * it. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file interface used for reading lines of text
 * @return new <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT cgul_csv_t
cgul_csv__new_from_crlf_file(cgul_exception_t* cex,
                             cgul_crlf_file_t crlf_file);

/**
 * This method deletes <code>csv</code> releasing all internally
 * allocated resources.  The client must not use <code>csv</code>
 * after calling this method.
 *
 * @param[in] csv <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT void
cgul_csv__delete(cgul_csv_t csv);

/**
 * Open the file with name <code>fname</code> so that
 * <code>cgul_csv__load_next()</code> can be called.  If an input
 * source is already open, it will be closed before this method
 * attempts to open the new source.  The new file will be closed when
 * this instance is deleted.  For regular files,
 * <code>is_block_buffered</code> should be set to true; for
 * character-oriented or line-oriented files,
 * <code>is_block_buffered</code> should be set to false.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] fname file name
 * @param[in] is_block_buffered whether <code>f</code> is block buffered
 */
extern CGUL_EXPORT void
cgul_csv__open_fname(cgul_exception_t* cex,
                     cgul_csv_t csv,
                     const char* fname,
                     int is_block_buffered);

/**
 * Open <code>f</code> so that <code>cgul_csv__load_next()</code>
 * can be called.  If an input source is already open, it will be
 * closed before this method attempts to open the new source.  This
 * class does not take ownership of <code>f</code>.  Thus, the client
 * is still responsible for calling <code>fclose()</code> on
 * <code>f</code>.  For regular files, <code>is_block_buffered</code>
 * should be set to true; for character-oriented or line-oriented
 * files, <code>is_block_buffered</code> should be set to false.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] f file
 * @param[in] is_block_buffered whether <code>f</code> is block buffered
 */
extern CGUL_EXPORT void
cgul_csv__open_file(cgul_exception_t* cex,
                    cgul_csv_t csv,
                    FILE* f,
                    int is_block_buffered);

/**
 * Open <code>buf</code> so that <code>cgul_csv__load_next()</code>
 * can be called.  If an input source is already open, it will be
 * closed before this method attempts to open the new source.  This
 * class does not take ownership of <code>buf</code>.  Thus, the
 * client is still responsible for calling <code>free()</code> on
 * <code>buf</code> if necessary.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] buf memory buffer
 * @param[in] buf_size size of <code>buf</code> in bytes
 */
extern CGUL_EXPORT void
cgul_csv__open_memory(cgul_exception_t* cex,
                      cgul_csv_t csv,
                      const char* buf,
                      size_t buf_size);

/**
 * Open <code>crlf_file</code> so that CSV records can be read from
 * it.  If an error occurs, an exception is thrown. <p>
 *
 * The <code>crlf_file</code> object is used as the interface for
 * reading lines of text from a data source.  It should already be
 * initialized before calling this method.  This class does not take
 * ownership of <code>crlf_file</code> so the client is still
 * responsible for calling <code>cgul_crlf_file__delete()</code> on
 * it. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] crlf_file interface used for reading lines of text
 */
extern CGUL_EXPORT void
cgul_csv__open_crlf_file(cgul_exception_t* cex,
                         cgul_csv_t csv,
                         cgul_crlf_file_t crlf_file);

/**
 * This method closes <code>csv</code>.  After calling this method, an
 * input source should be opened before calling any other method.
 *
 * @param[in] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 */
extern CGUL_EXPORT void
cgul_csv__close(cgul_exception_t* cex,
                cgul_csv_t csv);

/**
 * Get the character that is currently being used as the separator.
 * By default, the separator is a comma.
 *
 * @param[in] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @return separator
 */
extern CGUL_EXPORT char
cgul_csv__get_separator(cgul_exception_t* cex,
                        cgul_csv_t csv);

/**
 * Set the character that will be used as the separator.  By default,
 * the separator is a comma.
 *
 * @param[in] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] separator separator
 */
extern CGUL_EXPORT void
cgul_csv__set_separator(cgul_exception_t* cex,
                        cgul_csv_t csv,
                        char separator);

/**
 * This method loads the next csv record.  The client can use
 * <code>cgul_csv__lookup()</code> or
 * <code>cgul_csv__take_value()</code> to lookup the value for a
 * particular field or <code>cgul_csv__get_vector()</code> to iterate
 * over all the fields. <p>
 *
 * The client can also use <code>cgul_csv__lookup_by_header()</code>
 * or <code>cgul_csv__take_value_by_header()</code> to lookup the
 * value associated with a particular header. <p>
 *
 * On success, 1 is returned.  On EOF, 0 is returned.  If an error
 * occurs, 0 is returned, and an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @return whether EOF was not reached
 */
extern CGUL_EXPORT int
cgul_csv__load_next(cgul_exception_t* cex,
                    cgul_csv_t csv);

/**
 * This method returns the number of fields that are present in the
 * current record.  This method throws an exception only if
 * an input source has not been opened yet.
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @return number of fields in the current record
 */
extern CGUL_EXPORT unsigned long int
cgul_csv__get_number_of_fields(cgul_exception_t* cex,
                               cgul_csv_t csv);

/**
 * This method returns the value in the <code>field</code> for the
 * current record.  The pointer returned is owned by this class and
 * will be invalidated the next time
 * <code>cgul_csv__load_next()</code> is called.  If
 * <code>field</code> is out of bounds, <code>NULL</code> is returned,
 * and an exception is thrown. <p>
 *
 * If the client is going to make a copy of the value returned, it
 * should probably use <code>cgul_csv__take_value()</code> instead
 * which functions the same as this method except ownership is
 * transfered to the client thus avoiding a copy. <p>
 *
 * If you need more functionality than what is provided by this
 * method, you can use <code>cgul_csv__get_vector()</code> and
 * iterate over each field instead. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] field field
 * @return value present in the field
 */
extern CGUL_EXPORT const char*
cgul_csv__lookup(cgul_exception_t* cex,
                 cgul_csv_t csv,
                 unsigned long int field);

/**
 * This method returns the value in the field associated with
 * <code>header</code> for the current record.  The pointer returned
 * is owned by this class and will be invalidated the next time
 * <code>cgul_csv__load_next()</code> is called.  If
 * <code>header</code> does not exist, <code>NULL</code> is
 * returned. <p>
 *
 * The basic idea is that many CSV files use the first row to store
 * the headers for each column.  The first time
 * <code>cgul_csv__load_next()</code> is called, it remembers the
 * position of each header allowing this method to retrieve subsequent
 * values by using the string for the header instead of the integer
 * for the position. <p>
 *
 * If two columns have the same header, the value in the first column
 * will be returned. <p>
 *
 * If the client is going to make a copy of the value returned, it
 * should probably use <code>cgul_csv__take_value_by_header()</code>
 * instead which functions the same as this method except ownership is
 * transfered to the client thus avoiding a copy. <p>
 *
 * If headers are manually added to the CSV file, remember that space
 * is significant.  So for this method to work, usually the headers in
 * the CSV file will only be separated by commas (or tabs), not
 * spaces. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] header header
 * @return value present in the field associated with <code>header</code>
 */
extern CGUL_EXPORT const char*
cgul_csv__lookup_by_header(cgul_exception_t* cex,
                           cgul_csv_t csv,
                           const char* header);

/**
 * This method returns the value in the <code>field</code> for the
 * current record.  Ownership of the pointer returned is transfered to
 * the client.  Thus, the client must arrange for <code>free()</code>
 * to be called on the pointer.  If <code>field</code> is out of
 * bounds, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * Note that once the value for a particular field has been taken,
 * subsequent lookups on that field return <code>NULL</code> for the
 * value.  In order to do otherwise would require making a copy of the
 * value which would defeat the purpose of this method. <p>
 *
 * If the client only needs to inspect the value, it should probably
 * use <code>cgul_csv__lookup()</code> instead allowing the
 * <code>cgul_csv</code> instance to retain ownership thereby
 * guaranteeing that the pointer will be freed. <p>
 *
 * If you need more functionality than what is provided by this
 * method, you can use <code>cgul_csv__get_vector()</code> and
 * iterate over each field instead. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] field field
 * @return value present in the field
 */
extern CGUL_EXPORT char*
cgul_csv__take_value(cgul_exception_t* cex,
                     cgul_csv_t csv,
                     unsigned long int field);

/**
 * This method returns the value in the field associated with
 * <code>header</code> for the current record.  Ownership of the
 * pointer returned is transfered to the client.  Thus, the client
 * must arrange for <code>free()</code> to be called on the pointer.
 * If <code>header</code> does not exist, <code>NULL</code> is
 * returned. <p>
 *
 * Note that once the value for a particular header has been taken,
 * subsequent lookups on that field return <code>NULL</code> for the
 * value.  In order to do otherwise would require making a copy of the
 * value which would defeat the purpose of this method. <p>
 *
 * If the client only needs to inspect the value, it should probably
 * use <code>cgul_csv__lookup_by_header()</code> instead allowing the
 * <code>cgul_csv</code> instance to retain ownership thereby
 * guaranteeing that the pointer will be freed. <p>
 *
 * If headers are manually added to the CSV file, remember that space
 * is significant.  So for this method to work, usually the headers in
 * the CSV file will only be separated by commas (or tabs), not
 * spaces. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @param[in] header header
 * @return value present in the field associated with <code>header</code>
 */
extern CGUL_EXPORT char*
cgul_csv__take_value_by_header(cgul_exception_t* cex,
                               cgul_csv_t csv,
                               const char* header);

/**
 * This method returns the internal <code>cgul_vector</code> that is
 * used to hold all the fields of the current record.  The client only
 * needs to get the vector once.  It will not be invalidated until
 * <code>cgul_csv__delete()</code> is called. <p>
 *
 * The fields in the vector are <code>char*</code> types. <p>
 *
 * The client is free to take ownership of the strings for any field
 * provided the client notifies this class that ownership has been
 * transfered by setting the corresponding field in the vector to
 * <code>NULL</code>.  The client must call <code>free()</code> on any
 * pointer taken. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @return vector of fields
 */
extern CGUL_EXPORT cgul_vector_t
cgul_csv__get_vector(cgul_exception_t* cex,
                     cgul_csv_t csv);

/**
 * Get the headers for the CSV file.  This method returns a pointer to
 * the internal <code>cgul_rbtree</code> that maps each header to its
 * zero-based index for its column. This method can be called at any
 * time, but the tree will not be initialized until
 * <code>cgul_csv__load_next()</code> has been called the first
 * time. <p>
 *
 * To iterate over the headers in the order they appear in the CSV
 * file, use <code>cgul_rbtree__get_oldest()</code> and
 * <code>cgul_rbtree_node__get_younger()</code>.  The key for each
 * <code>cgul_rbtree_node</code> is the header for each column. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 * @return tree that maps each header to the zero-based index for its column
 */
extern CGUL_EXPORT cgul_rbtree_t
cgul_csv__get_headers(cgul_exception_t* cex,
                      cgul_csv_t csv);

/**
 * This static method is used to quote <code>bare_field</code> when
 * writing a CSV file.  It appends the result to the
 * <code>quoted_field</code> string.  This method only properly quotes
 * <code>bare_field</code>.  It does not append the commas (or tabs)
 * that need to appear between fields.  If an error occurs, an
 * exception is thrown. <p>
 *
 * To write a CSV file, simply use <code>fopen()</code> to open a text
 * file for writing.  Use this method to quote each field before
 * writing it to the CSV file, and do not forget to manual insert
 * commas (or tabs) between fields as needed. <p>
 *
 * The tests/main_csv_sort.c program demonstrates the process.  It
 * reorders a CSV file by sorting the columns alphabetically based on
 * the headers for each column. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bare_field bare field
 * @param[out] quoted_field quoted field
 */
extern CGUL_EXPORT void
cgul_csv__quote(cgul_exception_t* cex,
                const char* bare_field,
                cgul_string_t quoted_field);

CGUL_END_C

#endif  /*  CGUL_CSV_H  */
