#ifndef CGUL_CSV_CXX_H
#define CGUL_CSV_CXX_H

/**
 * C++ wrapper for <code>cgul_csv</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_csv</code>
 */

#include "cgul_crlf_file_cxx.h"
#include "cgul_csv.h"
#include "cgul_string_cxx.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_csv</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_csv</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_csv</code>
 */
class cgul_csv_cxx {

public:

    /**
     * This method creates a new <code>cgul_csv_cxx</code> instance.
     * After calling this method, an input source should be opened
     * before calling any other method.  Use
     * <code>set_separator()</code> to change the default separator
     * which is a comma.  If an error occurs, an exception is thrown.
     */
    cgul_csv_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_csv__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_csv_cxx</code> object and call
     * <code>open_fname()</code> passing it <code>fname</code>.  This
     * method opens <code>fname</code> so that input can be read from
     * it.  For regular files, <code>is_block_buffered</code> should
     * be set to true; for character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  Use
     * <code>set_separator()</code> to change the default separator
     * which is a comma.  If an error occurs, an exception is thrown.
     *
     * @param[in] fname file name of stanza file
     * @param[in] is_block_buffered whether <code>fname</code> is block buffered
     */
    cgul_csv_cxx(const char* fname,
                 bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        obj = cgul_csv__new_from_fname(&local,
                                       fname,
                                       is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_csv_cxx</code> object and call
     * <code>open_file()</code> passing it <code>f</code>.  This
     * method uses <code>f</code> when reading input.  This class does
     * not take ownership of <code>f</code>.  Thus, the client is
     * still responsible for calling <code>fclose()</code> on
     * <code>f</code>.  For regular files,
     * <code>is_block_buffered</code> should be set to true; for
     * character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  Use
     * <code>set_separator()</code> to change the default separator
     * which is a comma.  If an error occurs, an exception is thrown.
     *
     * @param[in] f file
     * @param[in] is_block_buffered whether <code>f</code> is block buffered
     */
    cgul_csv_cxx(FILE* f,
                 bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        obj = cgul_csv__new_from_file(&local,
                                      f,
                                      is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_csv_cxx</code> object and call
     * <code>open_memory()</code> passing it <code>buf</code> and
     * <code>buf_size</code>.  This method uses <code>buf</code> when
     * reading input.  This class does not take ownership of
     * <code>buf</code>.  Thus, the client is still responsible for
     * calling <code>free()</code> on <code>buf</code> if necessary.
     * Use <code>set_separator()</code> to change the default
     * separator which is a comma.  The client is responsible for
     * freeing the object by calling <code>delete</code> if necessary.
     * If an error occurs, an exception is thrown.
     *
     * @param[in] buf memory buffer
     * @param[in] buf_size size of <code>buf</code> in bytes
     */
    cgul_csv_cxx(const char* buf,
                 size_t buf_size)
    {
        cgul_exception_t local = NULL;
        obj = cgul_csv__new_from_memory(&local, buf, buf_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_csv_cxx</code> object and call
     * <code>open_crlf_file()</code> passing it
     * <code>crlf_file</code>.  This makes it possible to read from
     * any file type supported by <code>cgul_crlf_file_cxx</code> like
     * memory for example.  Use <code>set_separator()</code> to change
     * the default separator which is a comma.  If an error occurs, an
     * exception is thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    cgul_csv_cxx(cgul_crlf_file_cxx& crlf_file)
    {
        cgul_exception_t local = NULL;
        obj = cgul_csv__new_from_crlf_file(&local, crlf_file.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_csv_cxx</code> object and call
     * <code>open_crlf_file()</code> passing it
     * <code>crlf_file</code>.  This makes it possible to read from
     * any file type supported by <code>cgul_crlf_file_cxx</code> like
     * memory for example.  Use <code>set_separator()</code> to change
     * the default separator which is a comma.  If an error occurs, an
     * exception is thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    cgul_csv_cxx(cgul_crlf_file_cxx* crlf_file)
    {
        assert(crlf_file);
        cgul_exception_t local = NULL;
        obj = cgul_csv__new_from_crlf_file(&local, crlf_file->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_csv_cxx</code> object by wrapping an
     * existing <code>cgul_csv</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_csv_cxx(cgul_csv_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method deletes objects of this class releasing all
     * internally allocated resources.
     */
    virtual
    ~cgul_csv_cxx()
    {
        if (obj) {
            cgul_csv__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Open the file with name <code>fname</code> so that
     * <code>load_next()</code> can be called.  If an input source is
     * already open, it will be closed before this method attempts to
     * open the new source.  The new file will be closed when this
     * instance is deleted.  For regular files,
     * <code>is_block_buffered</code> should be set to true; for
     * character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] fname file name
     * @param[in] is_block_buffered whether <code>f</code> is block buffered
     */
    virtual void
    open_fname(const char* fname,
               bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        cgul_csv__open_fname(&local,
                             obj,
                             fname,
                             is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>f</code> so that <code>load_next()</code> can be
     * called.  If an input source is already open, it will be closed
     * before this method attempts to open the new source.  This class
     * does not take ownership of <code>f</code>.  Thus, the client is
     * still responsible for calling <code>fclose()</code> on
     * <code>f</code>.  For regular files,
     * <code>is_block_buffered</code> should be set to true; for
     * character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] f file
     * @param[in] is_block_buffered whether <code>f</code> is block buffered
     */
    virtual void
    open_file(FILE* f,
              bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        cgul_csv__open_file(&local,
                            obj,
                            f,
                            is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>buf</code> so that <code>load_next()</code> can be
     * called.  If an input source is already open, it will be closed
     * before this method attempts to open the new source.  This class
     * does not take ownership of <code>buf</code>.  Thus, the client
     * is still responsible for calling <code>free()</code> on
     * <code>buf</code> if necessary.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] buf memory buffer
     * @param[in] buf_size size of <code>buf</code> in bytes
     */
    virtual void
    open_memory(const char* buf,
                size_t buf_size)
    {
        cgul_exception_t local = NULL;
        cgul_csv__open_memory(&local, obj, buf, buf_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>crlf_file</code> so that input can be read from it.
     * This makes it possible to read from any file type supported by
     * <code>cgul_crlf_file_cxx</code> like memory for example.  If an
     * input source is already open, it will be closed before this
     * method attempts to open the new source.  If an error occurs, an
     * exception is thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    virtual void
    open_crlf_file(cgul_crlf_file_cxx& crlf_file)
    {
        cgul_exception_t local = NULL;
        cgul_csv__open_crlf_file(&local, obj, crlf_file.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>crlf_file</code> so that input can be read from it.
     * This makes it possible to read from any file type supported by
     * <code>cgul_crlf_file_cxx</code> like memory for example.  If an
     * input source is already open, it will be closed before this
     * method attempts to open the new source.  If an error occurs, an
     * exception is thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    virtual void
    open_crlf_file(cgul_crlf_file_cxx* crlf_file)
    {
        assert(crlf_file);
        cgul_exception_t local = NULL;
        cgul_csv__open_crlf_file(&local, obj, crlf_file->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method closes the object.  After calling this method, an
     * input source should be opened before calling any other method.
     */
    virtual void
    close()
    {
        cgul_exception_t local = NULL;
        cgul_csv__close(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the character that is currently being used as the separator.
     * By default, the separator is a comma.
     *
     * @return separator
     */
    virtual char
    get_separator() const
    {
        char rv = '\0';
        cgul_exception_t local = NULL;
        rv = cgul_csv__get_separator(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the character that will be used as the separator.  By default,
     * the separator is a comma.
     *
     * @param[in] separator separator
     */
    virtual void
    set_separator(char separator)
    {
        cgul_exception_t local = NULL;
        cgul_csv__set_separator(&local, obj, separator);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method loads the next csv record.  The client can use
     * <code>lookup()</code> or <code>take_value()</code> to lookup
     * the value for a particular field or <code>get_vector()</code>
     * to iterate over all the fields. <p>
     *
     * The client can also use <code>lookup_by_header()</code> or
     * <code>take_value_by_header()</code> to lookup the value
     * associated with a particular header. <p>
     *
     * On success, 1 is returned.  On EOF, 0 is returned.  If an error
     * occurs, an exception is thrown. <p>
     *
     * @return whether EOF was not reached
     */
    virtual int
    load_next()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_csv__load_next(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the number of fields that are present in
     * the current record.  This method throws an exception only if an
     * input source has not been opened yet.
     *
     * @return number of fields in the current record
     */
    virtual unsigned long int
    get_number_of_fields() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_csv__get_number_of_fields(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the value in the <code>field</code> for the
     * current record.  The pointer returned is owned by this class
     * and will be invalidated the next time <code>load_next()</code>
     * is called.  If <code>field</code> is out of bounds, an
     * exception is thrown. <p>
     *
     * If the caller is going to make a copy of the value returned, it
     * should probably use <code>take_value()</code> instead
     * which functions the same as this method except ownership is
     * transfered to the caller thus avoiding a copy. <p>
     *
     * If you need more functionality than what is provided by this
     * method, you can use <code>get_vector()</code> and
     * iterate over each field instead. <p>
     *
     * @param[in] field field
     * @return value present in the field
     */
    virtual const char*
    lookup(unsigned long int field)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_csv__lookup(&local, obj, field);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the value in the field associated with
     * <code>header</code> for the current record.  The pointer returned
     * is owned by this class and will be invalidated the next time
     * <code>load_next()</code> is called.  If
     * <code>header</code> does not exist, <code>NULL</code> is
     * returned. <p>
     *
     * The basic idea is that many CSV files use the first row to store
     * the headers for each column.  The first time
     * <code>load_next()</code> is called, it remembers the
     * position of each header allowing this method to retrieve subsequent
     * values by using the string for the header instead of the integer
     * for the position. <p>
     *
     * If two columns have the same header, the value in the first column
     * will be returned. <p>
     *
     * If the caller is going to make a copy of the value returned, it
     * should probably use <code>take_value_by_header()</code>
     * instead which functions the same as this method except ownership is
     * transfered to the caller thus avoiding a copy. <p>
     *
     * If headers are manually added to the CSV file, remember that space
     * is significant.  So for this method to work, usually the headers
     * in the CSV file will only be separated by commas (or tabs), not
     * spaces. <p>
     *
     * This method throws an exception only if an input source has not
     * been opened yet. <p>
     *
     * @param[in] header header
     * @return value present in the field associated with <code>header</code>
     */
    virtual const char*
    lookup_by_header(const char* header)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_csv__lookup_by_header(&local, obj, header);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * This method returns the value in the <code>field</code> for the
     * current record.  Ownership of the pointer returned is
     * transfered to the caller.  Thus, the caller must arrange for
     * <code>free()</code> to be called on the pointer.  If
     * <code>field</code> is out of bounds, an exception is
     * thrown. <p>
     *
     * Note that once the value for a particular field has been taken,
     * subsequent lookups on that field return <code>NULL</code> for
     * the value.  In order to do otherwise would require making a
     * copy of the value which would defeat the purpose of this
     * method. <p>
     *
     * If the caller only needs to inspect the value, it should
     * probably use <code>lookup()</code> instead allowing the the
     * object to retain ownership thereby guaranteeing that the
     * pointer will be freed. <p>
     *
     * If you need more functionality than what is provided by this
     * method, you can use <code>get_vector()</code> and iterate over
     * each field instead. <p>
     *
     * @param[in] field field
     * @return value present in the field
     */
    virtual char*
    take_value(unsigned long int field)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_csv__take_value(&local, obj, field);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the value in the field associated with
     * <code>header</code> for the current record.  Ownership of the
     * pointer returned is transfered to the caller.  Thus, the caller
     * must arrange for <code>free()</code> to be called on the pointer.
     * If <code>header</code> does not exist, <code>NULL</code> is
     * returned. <p>
     *
     * Note that once the value for a particular header has been taken,
     * subsequent lookups on that field return <code>NULL</code> for the
     * value.  In order to do otherwise would require making a copy of the
     * value which would defeat the purpose of this method. <p>
     *
     * If the caller only needs to inspect the value, it should probably
     * use <code>lookup_by_header()</code> instead allowing this
     * instance to retain ownership thereby
     * guaranteeing that the pointer will be freed. <p>
     *
     * If headers are manually added to the CSV file, remember that space
     * is significant.  So for this method to work, usually the headers
     * in the CSV file will only be separated by commas (or tabs), not
     * spaces. <p>
     *
     * This method throws an exception only if an input source has not
     * been opened yet. <p>
     *
     * @param[in] header header
     * @return value present in the field associated with <code>header</code>
     */
    virtual char*
    take_value_by_header(const char* header)
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_csv__take_value_by_header(&local, obj, header);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

#if 0
    /**
     * This method returns the internal <code>cgul_vector</code> that is
     * used to hold all the fields of the current record.  The client only
     * needs to get the vector once.  It will not be invalidated until
     * <code>delete</code> is called. <p>
     *
     * The fields in the vector are <code>char*</code> types. <p>
     *
     * The client is free to take ownership of the strings for any field
     * provided the client notifies this class that ownership has been
     * transfered by setting the corresponding field in the vector to
     * <code>NULL</code>.  The client must call <code>free()</code> on any
     * pointer taken.
     *
     * This method throws an exception only if an input source has not
     * been opened yet. <p>
     *
     * @return vector of fields
     */
    virtual cgul_vector_cxx_t
    get_vector() const
    {
        /* NOTE: cgul_vector_cxx was not implemented at the time. */
    }
#endif

    /**
     * This static method is used to quote <code>bare_field</code>
     * when writing a CSV file.  It appends the result to the
     * <code>quoted_field</code> string.  This method only properly
     * quotes <code>bare_field</code>.  It does not append the commas
     * (or tabs) that need to appear between fields.  If an error
     * occurs, an exception is thrown. <p>
     *
     * To write a CSV file, simply use <code>fopen()</code> to open a
     * text file for writing.  Use this method to quote each field
     * before writing it to the CSV file, and do not forget to manual
     * insert commas (or tabs) between fields as needed. <p>
     *
     * The tests/main_csv_sort.c program demonstrates the process.  It
     * reorders a CSV file by sorting the columns alphabetically based on
     * the headers for each column. <p>
     *
     * @param[in] bare_field bare field
     * @param[out] quoted_field quoted field
     */
    static void
    quote(const char* bare_field,
          cgul_string_cxx& quoted_field)
    {
        cgul_exception_t local = NULL;
        cgul_string_t quoted_field_obj = quoted_field.get_obj();
        cgul_csv__quote(&local, bare_field, quoted_field_obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_csv</code> object.
     *
     * @return underlying object
     */
    virtual cgul_csv_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_csv</code> object.  This means
     * the underlying object will not be deleted when the wrapper goes
     * out of scope.  Also, because you have taken the underlying
     * object, no other methods should be called on this wrapper's
     * instance.  Lastly, after taking the underlying object, it is
     * the caller's responsibility to delete the underlying object by
     * calling <code>cgul_csv__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_csv_t
    take_obj()
    {
        cgul_csv_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_csv_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_csv__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_csv_cxx(const cgul_csv_cxx&);
    cgul_csv_cxx& operator=(const cgul_csv_cxx&);

    /** Underlying <code>cgul_csv</code> object. */
    cgul_csv_t obj;

};


#endif  /*  CGUL_CSV_CXX_H  */
