#ifndef CGUL_HASH_H
#define CGUL_HASH_H

/**
 * Hash implementation.
 *
 * @file
 * @author Paul Serice
 * @brief hash implementation
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_hash_node.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_hash</code> instance. */
typedef struct cgul_hash* cgul_hash_t;

/**
 * This typedef defines the interface for hash functions that can be
 * used with custom <code>cgul_hash</code> objects.  The hash function
 * returns the hash of <code>key</code> which should be in the range
 * of [0 - UINT_MAX].  The <code>cgul_hash</code> object will then use
 * the hash value to map to the correct bucket. <p>
 *
 * It may not be necessary to write your own hash function as
 * <code>cgul_hash__hash_block()</code> is a good choice, but you will
 * still need to write an adaptor because
 * <code>cgul_hash__hash_block()</code> needs to know the size of the
 * key which is not something intrinsically stored in the hash
 * itself. <p>
 *
 * The <code>data</code> pointer is just a copy of the same pointer
 * that was registered when the custom <code>cgul_hash</code> object
 * was instantiated. <p>
 *
 * @param[in] key key
 * @param[in] data client data
 * @return hash of <code>key</code>
 */
typedef unsigned int (*cgul_hash__hash_t)(void* key,
                                          void* data);

/**
 * This typedef defines the interface for comparison functions that
 * can be used with custom <code>cgul_hash</code> objects.  After the
 * hash function selects a bucket, the comparison function is used to
 * compare keys within the same bucket in order to determine if a key
 * exists in the hash. <p>
 *
 * When you define your comparison function, you should compare
 * <code>key1</code> to <code>key2</code> and return less than zero,
 * zero, or greater than zero if <code>key1</code> is less than, equal
 * to, or greater than <code>key2</code>.  The reason the comparison
 * function needs to be complete is so
 * <code>cgul_hash__get_sorted_keys()</code> can work.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return result of comparison
 */
typedef int (*cgul_hash__compare_t)(const void* key1,
                                    const void* key2);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_hash__foldl_keys()</code>
 *     <code>cgul_hash__foldr_keys()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_hash__fold_key_t)(cgul_exception_t* cex,
                                     const void* key,
                                     void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_hash__foldl_values()</code>
 *     <code>cgul_hash__foldr_values()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_hash__fold_value_t)(cgul_exception_t* cex,
                                       void* value,
                                       void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_hash__foldl_pairs()</code>
 *     <code>cgul_hash__foldr_pairs()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_hash__fold_pair_t)(cgul_exception_t* cex,
                                      const void* key,
                                      void* value,
                                      void* data);

/**
 * This typedef is the interface for the callback function used by the
 * following methods: <p>
 *
 * <pre>
 *     <code>cgul_hash__traverse()</code>
 *     <code>cgul_hash__traverse_range()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] n node
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_hash__traverse_t)(cgul_exception_t* cex,
                                     cgul_hash_t h,
                                     cgul_hash_node_t n,
                                     void* data);

/**
 * This function computes the hash for <code>block</code> which can be
 * any arbitrary data extending for <code>block_size</code> bytes. <p>
 *
 * With the help of an adaptor, this function can be used as the hash
 * function for custom <code>cgul_hash</code> objects.  An adaptor is
 * needed because the block size is not a value that is intrinsically
 * stored in <code>cgul_hash</code> objects. <p>
 *
 * @param[in] block block
 * @param[in] block_size block size
 * @return hash of <code>block</code>
 * @see cgul_rolling_hash__hash()
 */
extern CGUL_EXPORT unsigned int
cgul_hash__hash_block(void* block,
                      unsigned int block_size);

/**
 * Create a new default <code>cgul_hash</code> object that is already
 * configured to use C-style strings as keys.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_hash__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_hash</code> instance
 * @see cgul_hash__set_load_factor()
 */
extern CGUL_EXPORT cgul_hash_t
cgul_hash__new(cgul_exception_t* cex);

/**
 * Create a new custom <code>cgul_hash</code> object that uses
 * <code>hf</code> as the hash function and <code>cf</code> as the
 * comparison function.  With the help of an adaptor, the
 * <code>cgul_hash__hash_block()</code> function is a good choice for
 * <code>hf</code>.  The caller is responsible for freeing the object
 * by calling <code>cgul_hash__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] hf hash function
 * @param[in] hf_data hash function client data
 * @param[in] cf comparison function
 * @return new <code>cgul_hash</code> instance
 * @see cgul_hash__set_load_factor()
 */
extern CGUL_EXPORT cgul_hash_t
cgul_hash__new_custom(cgul_exception_t* cex,
                      cgul_hash__hash_t hf,
                      void* hf_data,
                      cgul_hash__compare_t cf);

/**
 * The method frees all internally allocated memory.  Do not try to
 * access <code>h</code> after calling this method.
 *
 * @param[in] h <code>cgul_hash</code> instance
 * @see cgul_hash__free_keys()
 * @see cgul_hash__free_values()
 */
extern CGUL_EXPORT void
cgul_hash__delete(cgul_hash_t h);

/**
 * This method calls <code>free()</code> on all the keys in the hash
 * <code>h</code>.  Because this is such a common operation, it is an
 * exception to the rule that cgul containers never free keys.  This
 * method should only ever be called immediately before calling
 * <code>cgul_hash__delete()</code> because it otherwise invalidates
 * the hash.  Because this function is basically an extension of
 * <code>cgul_hash__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] h <code>cgul_hash</code> instance
 */
extern CGUL_EXPORT void
cgul_hash__free_keys(cgul_hash_t h);

/**
 * This method calls <code>free()</code> on all the values in the hash
 * <code>h</code>.  Because this is such a common operation, it is an
 * exception to the rule that cgul containers never free values.  This
 * method should only ever be called immediately before calling
 * <code>cgul_hash__delete()</code> because it otherwise invalidates
 * the hash.  Because this function is basically an extension of
 * <code>cgul_hash__delete()</code>, it does not take a
 * <code>cgul_exception</code> object.
 *
 * @param[in] h <code>cgul_hash</code> instance
 */
extern CGUL_EXPORT void
cgul_hash__free_values(cgul_hash_t h);

/**
 * Get the load factor.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return load factor
 * @see cgul_hash__set_load_factor()
 */
extern CGUL_EXPORT float
cgul_hash__get_load_factor(cgul_exception_t* cex,
                           cgul_hash_t h);

/**
 * The <code>load_factor</code> is the main parameter that governs the
 * speed/space tradeoff.  It is the ratio of number of keys to
 * buckets. <p>
 *
 * This is an important value because each key hashes to a particular
 * bucket.  Then the hash has to perform a linear search of the keys
 * in the bucket.  Because the load factor limits the average number
 * of keys per bucket, it controls how deep the linear search is
 * allowed to get (on average). <p>
 *
 * The default load factor is 0.75. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] load_factor load factor
 */
extern CGUL_EXPORT void
cgul_hash__set_load_factor(cgul_exception_t* cex,
                           cgul_hash_t h,
                           float load_factor);

/**
 * Swap the internal data for <code>h1</code> and <code>h2</code>.
 * This is much faster in general than trying to do the same thing by
 * iterating over the hash elements.
 *
 * @param[in] cex c-style exception
 * @param[in] h1 <code>cgul_hash</code> instance
 * @param[in] h2 second <code>cgul_hash</code> instance
 */
extern CGUL_EXPORT void
cgul_hash__swap(cgul_exception_t* cex,
                cgul_hash_t h1,
                cgul_hash_t h2);

/**
 * Return true if the hash <code>h</code> is empty.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return whether the hash is empty
 */
extern CGUL_EXPORT int
cgul_hash__is_empty(cgul_exception_t* cex,
                    cgul_hash_t h);

/**
 * Return an array of keys in insertion order.  A <code>NULL</code>
 * pointer is appended to the array that is returned and can be used
 * as a sentinel. <p>
 *
 * If <code>key_count</code> is not <code>NULL</code>,
 * <code>*key_count</code> is set to the number of keys returned in
 * the array. <p>
 *
 * The caller is responsible for calling <code>free()</code> on the
 * array that is returned but must not attempt to free the keys
 * themselves. <p>
 *
 * The main reason for calling this method is because you want to sort
 * the keys in the hash.  If you need the keys sorted using the
 * comparison function, as a convenience, you can call
 * <code>cgul_hash__get_sorted_keys()</code> instead.  If you need to
 * sort by some other criteria, you can use
 * <code>cgul_merge_sort_pointers()</code> directly on the array
 * returned by this method. <p>
 *
 * It is wasteful to use this method if all you want to do is iterate
 * over the keys in insertion order.  Use
 * <code>cgul_hash__get_oldest()</code> and
 * <code>cgul_hash_node__get_younger()</code> to do this instead. <p>
 *
 * If possible, <code>cgul_rbtree</code> should be used instead of
 * <code>cgul_hash</code> if the caller needs frequent access to the
 * keys in sorted order. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[out] key_count number of keys returned in the array
 * @return array of keys in chronological order
 */
extern CGUL_EXPORT const void**
cgul_hash__get_keys(cgul_exception_t* cex,
                    cgul_hash_t h,
                    size_t* key_count);

/**
 * Return an array of keys sorted using the comparison function.  A
 * <code>NULL</code> pointer is appended to the array that is returned
 * and can be used as a sentinel. <p>
 *
 * If <code>key_count</code> is not <code>NULL</code>,
 * <code>*key_count</code> is set to the number of keys returned in
 * the array.  <p>
 *
 * The caller is responsible for calling <code>free()</code> on the
 * array that is returned but must not attempt to free the keys
 * themselves. <p>
 *
 * If possible, <code>cgul_rbtree</code> should be used instead of
 * <code>cgul_hash</code> if the caller needs frequent access to the
 * keys in sorted order. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[out] key_count number of keys returned in the array
 * @return sorted array of keys
 */
extern CGUL_EXPORT const void**
cgul_hash__get_sorted_keys(cgul_exception_t* cex,
                           cgul_hash_t h,
                           size_t* key_count);

/**
 * Return the front node or <code>NULL</code> if the hash is empty.
 * The "front" node is just a synonym for the "oldest" node.  You can
 * use the return value to iterate over all the nodes in the hash by
 * calling <code>cgul_hash_node__get_next()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return front node
 * @see cgul_hash__get_oldest()
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash__get_front(cgul_exception_t* cex,
                     cgul_hash_t h);

/**
 * Return the back node or <code>NULL</code> if the hash is empty.
 * The "back" node is just a synonym for the "youngest" node.  You can
 * use the return value to iterate over all the nodes in the hash by
 * calling <code>cgul_hash_node__get_prev()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return back node
 * @see cgul_hash__get_youngest()
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash__get_back(cgul_exception_t* cex,
                    cgul_hash_t h);

/**
 * Get the oldest node.  This operation is O(1).  If the hash is
 * empty, <code>NULL</code> is returned.  This class does not have to
 * search the hash in order to find the oldest node because it
 * efficiently keeps a direct pointer to the oldest node up to
 * date. <p>
 *
 * The following example shows how to iterate over the entire hash in
 * insertion order:
 *
 * <pre>
 *     cgul_hash_node_t n = cgul_hash__get_oldest(cex, h);
 *     for ( ; n ; n = cgul_hash_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return oldest node
 * @see cgul_hash_node__get_younger()
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash__get_oldest(cgul_exception_t* cex,
                      cgul_hash_t h);

/**
 * Set the oldest node in the hash to be <code>n</code>.  Calling this
 * method has the potential to confuse iterators and should be handled
 * with roughly the same level of caution as calling
 * <code>cgul_hash__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * oldest node in the hash and you want to force <code>n</code> to be
 * the next node to expire. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] n node
 * @see cgul_hash__set_youngest()
 */
extern CGUL_EXPORT void
cgul_hash__set_oldest(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash_node_t n);

/**
 * Get the youngest node.  This operation is O(1).  If the hash is
 * empty, <code>NULL</code> is returned.  This class does not have to
 * search the hash in order to find the youngest node because it
 * efficiently keeps a direct pointer to the youngest node up to
 * date. <p>
 *
 * The following example shows how to iterate over the entire hash in
 * reverse chronological order:
 *
 * <pre>
 *     cgul_hash_node_t n = cgul_hash__get_youngest(cex, h);
 *     for ( ; n ; n = cgul_hash_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return youngest node
 * @see cgul_hash_node__get_older()
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash__get_youngest(cgul_exception_t* cex,
                        cgul_hash_t h);

/**
 * Set the youngest node in the hash to be <code>n</code>.  Calling
 * this method has the potential to confuse iterators and should be
 * handled with roughly the same level of caution as calling
 * <code>cgul_hash__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * least-recently used (LRU) node.  By calling
 * <code>cgul_hash__set_youngest()</code> each time a node is used,
 * the least-recently used node will be the oldest node in the
 * hash. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] n node
 * @see cgul_hash__set_oldest()
 */
extern CGUL_EXPORT void
cgul_hash__set_youngest(cgul_exception_t* cex,
                        cgul_hash_t h,
                        cgul_hash_node_t n);

/**
 * Insert <code>key</code> into the hash <code>h</code>.  If
 * <code>key</code> already exists, another node with the same key is
 * not created.  The new or pre-existing node is returned in
 * <code>n</code> if <code>n</code> is not <code>NULL</code>.  If an
 * error occurs, <code>0</code> is returned, <code>*n</code> is set to
 * <code>NULL</code> (if <code>n</code> is not <code>NULL</code>), and
 * an exception is thrown. <p>
 *
 * An important thing to notice is that there is no way to pass in the
 * <code>value</code> pointer that makes up the key/value pair.  This
 * is by design.  The reason is that <code>find()</code> and
 * <code>insert()</code> share the same basic code to search to see if
 * the key already exists.  Thus, if you want to see if a key already
 * exists before inserting a new key (which is very common), you could
 * call <code>cgul_hash__find()</code> followed by
 * <code>cgul_hash__insert()</code>, but this inefficiently causes
 * <code>cgul_hash</code> to do the same lookup twice. <p>
 *
 * By not allowing <code>cgul_hash__insert()</code> to set the
 * <code>value</code> pointer, it allows you to easily use this method
 * in those places where you want <code>cgul_hash__find()</code> to
 * fail before you call <code>cgul_hash__insert()</code>; instead, you
 * just call <code>cgul_hash__insert()</code> and check the return
 * value.  Here is an example:
 *
 * <pre>
 *
 *    int insert_rv = 0;
 *    cgul_hash_node_t* n = NULL;
 *
 *    key = cgul_string__get_value(cex, s);
 *    insert_rv = cgul_hash__insert(cex, h, (void*)key, &n);
 *    if (*cex) {
 *        goto out;
 *    }
 *    if (insert_rv) {
 *        cgul_string__take_value(cex, s);
 *        . . .
 *    } else {
 *        . . .
 *    }
 *
 * </pre>
 *
 * As with all the cgul containers, the client is responsible for
 * deallocating both <code>key</code> and <code>value</code>.  If
 * before inserting the new node the hash detects that the load factor
 * has been exceeded, the hash will be resized. <p>
 *
 * @note The <code>cgul_hash</code> does not make a duplicate copy of
 * the key.  Thus, it is the callers responsibility to make sure
 * <code>key</code> continues to exist until it is removed from the
 * hash. <p>
 *
 * @note An insert operation could cause the average search distance
 * to grow too large which triggers a resize operation.  Resizing the
 * hash is relatively expensive, but the amortized cost should be
 * minimal.  It is possible for the program to run out of memory
 * during a resize operation.  In this case, the hash will still be
 * valid, but it will not be resized. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] key key
 * @param[out] n newly inserted node if successful; otherwise, unaltered
 * @return whether the insertion operation succeeded
 */
extern CGUL_EXPORT int
cgul_hash__insert(cgul_exception_t* cex,
                  cgul_hash_t h,
                  const void* key,
                  cgul_hash_node_t* n);

/**
 * Return the hash node associated with <code>key</code>.  You can use
 * <code>cgul_hash_node__get_value()</code> and
 * <code>cgul_hash_node__set_value()</code> to get and set the value
 * associated with the node.  If no node is found, <code>NULL</code>
 * is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] key key
 * @return node associated with <code>key</code>
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash__find(cgul_exception_t* cex,
                cgul_hash_t h,
                const void* key);

/**
 * Remove the node associated with the key <code>key_in</code> from
 * the hash <code>h</code>.  If <code>key_out</code> is not
 * <code>NULL</code>, the key that is currently stored in the node to
 * be removed will be returned.  If <code>value_out</code> is not
 * <code>NULL</code>, the value stored in the node to be removed will
 * be returned.  The return value is <code>1</code> if the node for
 * <code>key_in</code> was removed; otherwise, it is <code>0</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] key_in key to find
 * @param[out] key_out actual key stored in node
 * @param[out] value_out actual value stored in node
 * @return whether the node associated with <code>key</code> was removed
 */
extern CGUL_EXPORT int
cgul_hash__remove(cgul_exception_t* cex,
                  cgul_hash_t h,
                  const void* key_in,
                  void** key_out,
                  void** value_out);

/**
 * Remove the node <code>n</code> from the hash <code>h</code>.
 * Calling this method has the potential to invalidate iterators.  If
 * you need to remove nodes while simultaneously iterating over the
 * tree, you must call <code>cgul_hash__get_next()</code> (or the
 * equivalent) before removing the current node because, once the node
 * is removed, you cannot use it safely.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] n node
 */
extern CGUL_EXPORT void
cgul_hash__remove_node(cgul_exception_t* cex,
                       cgul_hash_t h,
                       cgul_hash_node_t n);

/**
 * This method removes the front node from the hash <code>h</code>.
 * The "front" node is just a synonym for the "oldest" node.  The key
 * and value pointers stored in the node that is to be removed will be
 * returned in <code>key_out</code> and <code>value_out</code> if you
 * pass in pointers that are not <code>NULL</code>.  If <code>h</code>
 * is not empty, the front node will be removed, and <code>1</code>
 * will be returned; otherwise, <code>0</code> will be returned.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_hash__remove_front(cgul_exception_t* cex,
                        cgul_hash_t h,
                        void** key_out,
                        void** value_out);

/**
 * This method removes the back node from the hash <code>h</code>.
 * The "back" node is just a synonym for the "youngest" node.  The key
 * and value pointers stored in the node that is to be removed will be
 * returned in <code>key_out</code> and <code>value_out</code> if you
 * pass in pointers that are not <code>NULL</code>.  If <code>h</code>
 * is not empty, the back node will be removed, and <code>1</code>
 * will be returned; otherwise, <code>0</code> will be returned.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_hash__remove_back(cgul_exception_t* cex,
                       cgul_hash_t h,
                       void** key_out,
                       void** value_out);

/**
 * This method removes the oldest node from the hash <code>h</code>.
 * The key and value pointers stored in the node that is to be removed
 * will be returned in <code>key_out</code> and <code>value_out</code>
 * if you pass in pointers that are not <code>NULL</code>.  If
 * <code>h</code> is not empty, the oldest node will be removed and
 * <code>1</code> will be returned; otherwise, <code>0</code> will be
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_hash__remove_oldest(cgul_exception_t* cex,
                         cgul_hash_t h,
                         void** key_out,
                         void** value_out);

/**
 * This method removes the youngest node from the hash <code>h</code>.
 * The key and value pointers stored in the node that is to be removed
 * will be returned in <code>key_out</code> and <code>value_out</code>
 * if you pass in pointers that are not <code>NULL</code>.  If
 * <code>h</code> is not empty, the youngest node will be removed and
 * <code>1</code> will be returned; otherwise, <code>0</code> will be
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_hash__remove_youngest(cgul_exception_t* cex,
                           cgul_hash_t h,
                           void** key_out,
                           void** value_out);

/**
 * Remove all elements from the hash <code>h</code>.  You should be
 * careful to deallocate the key/value pairs (if necessary) before
 * calling this method. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 */
extern CGUL_EXPORT void
cgul_hash__clear(cgul_exception_t* cex,
                 cgul_hash_t h);

/**
 * Return the total number of elements stored in the hash.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return total number of elements stored in the hash
 */
extern CGUL_EXPORT size_t
cgul_hash__get_size(cgul_exception_t* cex,
                    cgul_hash_t h);

/**
 * Get the total number of collisions.  Note that the total number of
 * collisions can be more than the size of the hash because the total
 * number of collisions includes all the extra collisions that occur
 * when the hash is resized.
 *
 * @param[in] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 */
extern CGUL_EXPORT size_t
cgul_hash__get_total_collisions(cgul_exception_t* cex,
                                cgul_hash_t h);

/**
 * Return a string with statistics for the <code>cgul_hash</code>.
 * The caller is responsible for freeing the string by calling
 * <code>free()</code>.  If an error occurs while generating the
 * stats, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @return statics for <code>h</code>
 */
extern CGUL_EXPORT char*
cgul_hash__get_stats(cgul_exception_t* cex,
                     cgul_hash_t h);

/**
 * This method performs a left fold of the hash <code>h</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key in the hash starting at the front of the hash and
 * iterating forward to the end of the hash. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the client data which is where the result of the
 * fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__foldl_keys(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash__fold_key_t f,
                      void* data);

/**
 * This method performs a right fold of the hash <code>h</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key in the hash starting at the back of the hash and
 * iterating backward to the front of the hash. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__foldr_keys(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash__fold_key_t f,
                      void* data);

/**
 * This method performs a left fold of the hash <code>h</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the hash starting at the front of the hash
 * and iterating forward to the end of the hash. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__foldl_values(cgul_exception_t* cex,
                        cgul_hash_t h,
                        cgul_hash__fold_value_t f,
                        void* data);

/**
 * This method performs a right fold of the hash <code>h</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the hash starting at the back of the hash
 * and iterating backward to the front of the hash. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__foldr_values(cgul_exception_t* cex,
                        cgul_hash_t h,
                        cgul_hash__fold_value_t f,
                        void* data);

/**
 * This method performs a left fold of the hash <code>h</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key/value pair in the hash starting at the front of
 * the hash and iterating forward to the end of the hash. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__foldl_pairs(cgul_exception_t* cex,
                       cgul_hash_t h,
                       cgul_hash__fold_pair_t f,
                       void* data);

/**
 * This method performs a right fold of the hash <code>h</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key/value pair in the hash starting at the back of
 * the hash and iterating backward to the front of the hash. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__foldr_pairs(cgul_exception_t* cex,
                      cgul_hash_t h,
                      cgul_hash__fold_pair_t f,
                      void* data);

/**
 * Traverse all nodes passing each node to the function
 * <code>f</code>. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the hash <code>h</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_hash__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_hash__traverse()</code> or
 * <code>cgul_hash__traverse_range()</code> in order to iterate over
 * the hash elements.  In fact, I would recommend that you use
 * <code>cgul_hash_node__get_next()</code> for most of your iteration
 * needs.  If you need to remove nodes though, you should probably use
 * this method.  If you look at this method's source however, you will
 * see that the <code>cgul_hash_node</code> class provides all the
 * public methods you need to safely remove nodes while you are
 * iterating over the hash, but you do have to be careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] f traversal callback function
 * @param[in] data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_hash__traverse(cgul_exception_t* cex,
                    cgul_hash_t h,
                    cgul_hash__traverse_t f,
                    void* data);

/**
 * Traverse all nodes starting with <code>first</code> (inclusive) and
 * ending with <code>last</code> (inclusive) passing each node to the
 * function <code>f</code>.  If you know the first node, but are not
 * sure of the last node, just use <code>NULL</code> as the last node.
 * This will cause this method to iterate until it reaches the end of
 * the hash.  You can than have <code>f</code> return 0 when it
 * determines that the last node has been reached. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the hash <code>h</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_hash__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * If <code>first</code> is <code>NULL</code>, iteration starts at the
 * beginning of the hash.  If <code>last</code> is <code>NULL</code>,
 * iteration stops at the end of the hash. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_hash__traverse()</code> or
 * <code>cgul_hash__traverse_range()</code> in order to iterate over
 * the hash elements.  In fact, I would recommend that you use
 * <code>cgul_hash_node__get_next()</code> for most of your
 * iteration needs.  If you need to remove nodes though, you should
 * probably use this method.  If you look at this method's source
 * however, you will see that the <code>cgul_hash_node</code> class
 * provides all the public methods you need to safely remove nodes
 * while you are iterating over the hash, but you do have to be
 * careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] h <code>cgul_hash</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] f traversal callback function
 * @param[in] data client data passed to <code>f</code>
 * @see cgul_hash__traverse
 */
extern CGUL_EXPORT void
cgul_hash__traverse_range(cgul_exception_t* cex,
                          cgul_hash_t h,
                          cgul_hash_node_t first,
                          cgul_hash_node_t last,
                          cgul_hash__traverse_t f,
                          void* data);

CGUL_END_C

#endif  /*  CGUL_HASH_H  */
