#ifndef CGUL_HASH_CXX_H
#define CGUL_HASH_CXX_H

/**
 * C++ wrapper for <code>cgul_hash</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_hash</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_hash.h"
#include "cgul_hash_node_cxx.h"
#include "cgul_string_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_hash</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_hash</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_hash</code>
 * @see cgul_hash_t
 */
class cgul_hash_cxx {

public:

    /**
     * This typedef defines the interface for hash functions that can
     * be used with custom <code>cgul_hash_cxx</code> objects.  The
     * hash function returns the hash of <code>key</code> which should
     * be in the range of [0 - UINT_MAX].  The <code>cgul_hash_cxx</code>
     * object will then use the hash value to map to the correct
     * bucket. <p>
     *
     * It may not be necessary to write your own hash function as
     * <code>hash_block()</code> is a good choice, but you will still
     * need to write an adaptor because <code>hash_block()</code>
     * needs to know the size of the key which is not something
     * intrinsically stored in the hash itself. <p>
     *
     * The <code>data</code> pointer is just a copy of the same pointer
     * that was registered when the custom <code>cgul_hash_cxx</code> object
     * was instantiated. <p>
     */
    typedef cgul_hash__hash_t hash_t;

    /**
     * This typedef defines the interface for comparison functions
     * that can be used with custom <code>cgul_hash_cxx</code>
     * objects.  After the hash function selects a bucket, the
     * comparison function is used to compare keys within the same
     * bucket in order to determine if a key exists in the hash. <p>
     *
     * When you define your comparison function, you should compare
     * <code>key1</code> to <code>key2</code> and return less than
     * zero, zero, or greater than zero if <code>key1</code> is less
     * than, equal to, or greater than <code>key2</code>.  The reason
     * the comparison function needs to be complete is so
     * <code>get_sorted_keys()</code> can work.
     */
    typedef cgul_hash__compare_t compare_t;

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_keys()</code>
     *     <code>foldr_keys()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_key_t)(const void* key,
                              void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_values()</code>
     *     <code>foldr_values()</code>
     * </pre>
     *
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_value_t)(void* value,
                                void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_pairs()</code>
     *     <code>foldr_pairs()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_pair_t)(const void* key,
                               void* value,
                               void* data);

    /**
     * This typedef is the interface for the callback function used by
     * the following methods: <p>
     *
     * <pre>
     *     <code>traverse()</code>
     *     <code>traverse_range()</code>
     * </pre>
     *
     * @param[in] h hash
     * @param[in] n node
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*traverse_t)(cgul_hash_cxx* h,
                              cgul_hash_node_cxx* n,
                              void* data);

    /**
     * Create a new default <code>cgul_hash_cxx</code> object that is
     * already configured to use C-style strings as keys.  If memory
     * cannot be allocated, an exception is thrown.
     *
     * @see set_load_factor()
     */
    cgul_hash_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_hash__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new custom <code>cgul_hash_cxx</code> object that uses
     * <code>hf</code> as the hash function and <code>cf</code> as the
     * comparison function.  With the help of an adaptor, the
     * <code>hash_block()</code> function is a good choice for
     * <code>hf</code>.  If memory cannot be allocated, an exception
     * is thrown.
     *
     * @param[in] hf hash function
     * @param[in] hf_data hash function client data
     * @param[in] cf comparison function
     * @see set_load_factor()
     */
    cgul_hash_cxx(hash_t hf,
                  void* hf_data,
                  compare_t cf)
    {
        cgul_exception_t local = NULL;
        obj = cgul_hash__new_custom(&local, hf, hf_data, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * The method frees all internally allocated memory.
     *
     * @see free_keys()
     * @see free_values()
     */
    virtual
    ~cgul_hash_cxx()
    {
        if (obj) {
            cgul_hash__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the keys in the
     * hash.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free keys.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the hash.
     */
    virtual void
    free_keys()
    {
        cgul_hash__free_keys(obj);
    }

    /**
     * This method calls <code>free()</code> on all the values in the
     * hash.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free values.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the hash.
     */
    virtual void
    free_values()
    {
        cgul_hash__free_values(obj);
    }

    /**
     * Get the load factor.
     *
     * @return load factor
     * @see set_load_factor()
     */
    virtual float
    get_load_factor() const
    {
        float rv = 0.0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_load_factor(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * The <code>load_factor</code> is the main parameter that governs
     * the speed/space tradeoff.  It is the ratio of the number of
     * keys to buckets. <p>
     *
     * This is an important value because each key hashes to a particular
     * bucket.  Then the hash has to perform a linear search of the keys
     * in the bucket.  Because the load factor limits the average number
     * of keys per bucket, it controls how deep the linear search is
     * allowed to get (on average). <p>
     *
     * The default load factor is 0.75. <p>
     *
     * @param[in] load_factor load factor
     */
    virtual void
    set_load_factor(float load_factor)
    {
        cgul_exception_t local = NULL;
        cgul_hash__set_load_factor(&local,
                                   obj,
                                   load_factor);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Swap the internal data for <code>this</code> and <code>rhs</code>.
     * This is much faster in general than trying to do the same thing by
     * iterating over the hash elements.
     *
     * @param[in] rhs righ-hand side
     */
    virtual void
    swap(cgul_hash_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_hash__swap(&local, obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return true if the hash is empty.
     *
     * @return whether the hash is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return an array of keys in insertion order.  A <code>NULL</code>
     * pointer is appended to the array that is returned and can be
     * used as a sentinel. <p>
     *
     * If <code>key_count</code> is not <code>NULL</code>,
     * <code>*key_count</code> is set to the number of keys returned in
     * the array. <p>
     *
     * The caller is responsible for calling <code>free()</code> on the
     * array that is returned but must not attempt to free the keys
     * themselves. <p>
     *
     * The main reason for calling this method is because you want to
     * sort the keys in the hash.  If you need the keys sorted using
     * the comparison function, as a convenience, you can call
     * <code>get_sorted_keys()</code> instead.  If you need to sort by
     * some other criteria, you can use
     * <code>cgul_merge_sort_pointers()</code> directly on the array
     * returned by this method. <p>
     *
     * It is wasteful to use this method if all you want to do is
     * iterate over the keys in insertion order.  Use
     * <code>get_oldest()</code> and
     * <code>cgul_hash_node::get_younger()</code> to do this
     * instead. <p>
     *
     * If possible, <code>cgul_rbtree_cxx</code> should be used instead of
     * <code>cgul_hash_cxx</code> if the caller needs frequent access to the
     * keys in sorted order. <p>
     *
     * @param[out] key_count number of keys returned in the array
     * @return array of keys in insertion order
     */
    virtual const void**
    get_keys(size_t* key_count) const
    {
        const void** rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_keys(&local, obj, key_count);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return an array of keys sorted using the comparison function.
     * A <code>NULL</code> pointer is appended to the array that is
     * returned and can be used as a sentinel. <p>
     *
     * If <code>key_count</code> is not <code>NULL</code>,
     * <code>*key_count</code> is set to the number of keys returned in
     * the array.  <p>
     *
     * The caller is responsible for calling <code>free()</code> on the
     * array that is returned but must not attempt to free the keys
     * themselves. <p>
     *
     * If possible, <code>cgul_rbtree_cxx</code> should be used instead of
     * <code>cgul_hash_cxx</code> if the caller needs frequent access to the
     * keys in sorted order. <p>
     *
     * @param[out] key_count number of keys returned in the array
     * @return array of keys sorted using <code>strcmp()</code>
     */
    virtual const void**
    get_sorted_keys(size_t* key_count) const
    {
        const void** rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_sorted_keys(&local, obj, key_count);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the front node or <code>NULL</code> if the hash is
     * empty.  The "front" node is just a synonym for the "oldest"
     * node.  You can use the return value to iterate over all the
     * nodes in the hash by calling
     * <code>cgul_hash_node::get_next()</code>.
     *
     * @return front node
     * @see get_oldest()
     */
    virtual cgul_hash_node_cxx*
    get_front() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_front(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Return the back node or <code>NULL</code> if the hash is empty.
     * The "back" node is just a synonym for the "youngest" node.  You
     * can use the return value to iterate over all the nodes in the
     * hash by calling <code>cgul_hash_node::get_prev()</code>.
     *
     * @return back node
     * @see get_youngest()
     */
    virtual cgul_hash_node_cxx*
    get_back() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Get the oldest node.  This operation is O(1).  If the hash is
     * empty, <code>NULL</code> is returned.  This class does not have to
     * search the hash in order to find the oldest node because it
     * efficiently keeps a direct pointer to the oldest node up to
     * date. <p>
     *
     * The following example shows how to iterate over the entire hash in
     * insertion order:
     *
     * <pre>
     *     cgul_hash_node_cxx* n = h->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return oldest node
     * @see cgul_hash_node::get_younger()
     */
    virtual cgul_hash_node_cxx*
    get_oldest() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_oldest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Set the oldest node in the hash to be <code>n</code>.  Calling this
     * method has the potential to confuse iterators and should be handled
     * with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * oldest node in the hash and you want to force <code>n</code> to be
     * the next node to expire. <p>
     *
     * @param[in] n node
     * @see set_youngest()
     */
    virtual void
    set_oldest(cgul_hash_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_hash__set_oldest(&local, obj, (cgul_hash_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the youngest node.  This operation is O(1).  If the hash is
     * empty, <code>NULL</code> is returned.  This class does not have to
     * search the hash in order to find the youngest node because it
     * efficiently keeps a direct pointer to the youngest node up to
     * date. <p>
     *
     * The following example shows how to iterate over the entire hash in
     * reverse insertion order:
     *
     * <pre>
     *     cgul_hash_node_cxx* n = h->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return youngest node
     * @see cgul_hash_node::get_older()
     */
    virtual cgul_hash_node_cxx*
    get_youngest() const
    {
        cgul_hash_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_youngest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Set the youngest node in the hash to be <code>n</code>.  Calling
     * this method has the potential to confuse iterators and should be
     * handled with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * least-recently used (LRU) node.  By calling
     * <code>set_youngest()</code> each time a node is used,
     * the least-recently used node will be the oldest node in the
     * hash. <p>
     *
     * @param[in] n node
     * @see cgul_hash__set_oldest()
     */
    virtual void
    set_youngest(cgul_hash_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_hash__set_youngest(&local, obj, (cgul_hash_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the new <code>cgul_hash_node_cxx</code> object created by
     * inserting <code>key</code> into the hash.  If
     * <code>key</code> already exists, another node with the same key is
     * not created.  The new or pre-existing node is returned in
     * <code>n</code> if it is not <code>NULL</code>. <p>
     *
     * An important thing to notice is that there is no way to pass in
     * the <code>value</code> pointer that makes up the key/value
     * pair.  This is by design.  The reason is that
     * <code>find()</code> and <code>insert()</code> share the same
     * basic code to search to see if the key already exists.  Thus,
     * if you want to see if a key already exists before inserting a
     * new key (which is very common), you could call
     * <code>find()</code> followed by <code>insert()</code>, but this
     * inefficiently causes hash to do the same lookup twice. <p>
     *
     * By not allowing <code>insert()</code> to set the
     * <code>value</code> pointer, it allows you to easily use this
     * method in those places where you want <code>find()</code> to
     * fail before you call <code>insert()</code>; instead, you just
     * call <code>insert()</code> and check the return value.  Here is
     * an example:
     *
     * <pre>
     *
     *    int insert_rv = 0;
     *    cgul_hash_node_cxx* n = NULL;
     *
     *    key = s->get_value();
     *    insert_rv = h->insert(key, &n);
     *    if (insert_rv) {
     *        s->take_value();
     *        . . .
     *    } else {
     *        . . .
     *    }
     *
     * </pre>
     *
     * As with all the cgul containers, the client is responsible for
     * deallocating both <code>key</code> and <code>value</code>.  If
     * before inserting the new node the hash detects that the load
     * factor has been exceeded, the hash will be resized. <p>
     *
     * @note The hash does not make a duplicate copy of the key.
     * Thus, it is the callers responsibility to make sure
     * <code>key</code> continues to exist until it is removed from
     * the hash. <p>
     *
     * @note An insert operation could cause the average search distance
     * to grow too large which triggers a resize operation.  Resizing the
     * hash is relatively expensive, but the amortized cost should be
     * minimal.  It is possible for the program to run out of memory
     * during a resize operation.  In this case, the hash will still be
     * valid, but it will not be resized. <p>
     *
     * @param[in] key key
     * @param[out] n newly inserted node if successful; otherwise, unaltered
     * @return whether the insertion operation succeeded
     */
    virtual int
    insert(const void* key,
           cgul_hash_node_cxx** n)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__insert(&local, obj, key, (cgul_hash_node_t*)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the hash node associated with <code>key</code>.  You can use
     * <code>cgul_hash_node::get_value()</code> and
     * <code>cgul_hash_node::set_value()</code> to get and set the value
     * associated with the node.  If no node is found, <code>NULL</code>
     * is returned.
     *
     * @param[in] key key
     * @return node associated with <code>key</code>
     */
    virtual cgul_hash_node_cxx*
    find(const void* key)
    {
        cgul_hash_node_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__find(&local, obj, key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_hash_node_cxx*)rv;
    }

    /**
     * Remove the node associated with the key <code>key_in</code>
     * from the hash.  If <code>key_out</code> is not
     * <code>NULL</code>, the key that is currently stored in the node
     * to be removed will be returned.  If <code>value_out</code> is
     * not <code>NULL</code>, the value stored in the node to be
     * removed will be returned.  The return value is <code>1</code>
     * if the node for <code>key_in</code> was removed; otherwise, it
     * is <code>0</code>.
     *
     * @param[in] key_in key to find
     * @param[out] key_out actual key stored in node
     * @param[out] value_out actual value stored in node
     * @return whether the node associated with <code>key</code> was removed
     */
    virtual int
    remove(const void* key_in,
           void** key_out,
           void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__remove(&local, obj, key_in, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove the node <code>n</code> from the hash.  Calling this
     * method has the potential to invalidate iterators.  If you need
     * to remove nodes while simultaneously iterating over the tree,
     * you must call <code>get_next()</code> (or the equivalent)
     * before removing the current node because, once the node is
     * removed, you cannot use it safely.
     *
     * @param[in] n node
     */
    virtual void
    remove_node(cgul_hash_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_hash__remove_node(&local, obj, (cgul_hash_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method removes the front node from the hash.  The "front"
     * node is just a synonym for the "oldest" node.  The key and
     * value pointers stored in the node that is to be removed will be
     * returned in <code>key_out</code> and <code>value_out</code> if
     * you pass in pointers that are not <code>NULL</code>.  If the
     * hash is not empty, the front node will be removed, and
     * <code>1</code> will be returned; otherwise, <code>0</code> will
     * be returned.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_front(void** key_out,
                 void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__remove_front(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes the back node from the hash.  The "back"
     * node is just a synonym for the "youngest" node.  The key and
     * value pointers stored in the node that is to be removed will be
     * returned in <code>key_out</code> and <code>value_out</code> if
     * you pass in pointers that are not <code>NULL</code>.  If the
     * hash is not empty, the back node will be removed, and
     * <code>1</code> will be returned; otherwise, <code>0</code> will
     * be returned.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_back(void** key_out,
                void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__remove_back(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes the oldest node from the hash.  The key and
     * value pointers stored in the node that is to be removed will be
     * returned in <code>key_out</code> and <code>value_out</code> if
     * you pass in pointers that are not <code>NULL</code>.  If the
     * hash is not empty, the oldest node will be removed and
     * <code>1</code> will be returned; otherwise, <code>0</code> will
     * be returned.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_oldest(void** key_out,
                  void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__remove_oldest(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes the youngest node from the hash.  The key and
     * value pointers stored in the node that is to be removed will be
     * returned in <code>key_out</code> and <code>value_out</code> if
     * you pass in pointers that are not <code>NULL</code>.  If the
     * hash is not empty, the youngest node will be removed and
     * <code>1</code> will be returned; otherwise, <code>0</code> will
     * be returned.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_youngest(void** key_out,
                    void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__remove_youngest(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove all elements from the hash.  You should be careful to
     * deallocate the key/value pairs (if necessary) before calling
     * this method. <p>
     */
    virtual void
    clear()
    {
        cgul_exception_t local = NULL;
        cgul_hash__clear(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the total number of elements stored in the hash.
     *
     * @return total number of elements stored in the hash
     */
    virtual size_t
    get_size() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function computes the hash for <code>block</code> which
     * can be any arbitrary data extending for <code>block_size</code>
     * bytes. <p>
     *
     * With the help of an adaptor, this function can be used as the
     * hash function for custom <code>cgul_hash_cxx</code> objects.
     * An adaptor is needed because the block size is not a value that
     * is intrinsically stored in <code>cgul_hash_cxx</code>
     * objects. <p>
     *
     * @param[in] block block
     * @param[in] block_size block size
     * @return hash of <code>block</code>
     */
    virtual unsigned int
    hash_block(void* block,
               size_t block_size)
    {
        return cgul_hash__hash_block(block, block_size);
    }

    /**
     * Return a string with statistics for the hash.
     * The caller is responsible for freeing the string by calling
     * <code>free()</code>.  If an error occurs while generating the
     * stats, an exception is thrown.
     *
     * @return statics for the hash
     */
    virtual char*
    get_stats() const
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_hash__get_stats(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method performs a left fold of the hash with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each key in the hash starting at the front of the hash and
     * iterating forward to the end of the hash. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_keys(fold_key_t f,
               void* data)
    {
        const void* key = NULL;
        cgul_hash_node_t curr = NULL;
        cgul_hash_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__foldl_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_hash__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_hash_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_hash_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the hash with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key in the hash starting at the back of the hash
     * and iterating backward to the front of the hash. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_keys(fold_key_t f,
               void* data)
    {
        const void* key = NULL;
        cgul_hash_node_t curr = NULL;
        cgul_hash_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__foldr_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_hash__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_hash_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_hash_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the hash with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each value in the hash starting at the front of the hash and
     * iterating forward to the end of the hash. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_values(fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_hash_node_t curr = NULL;
        cgul_hash_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__foldl_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_hash__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_hash_node__get_next(&local, curr);
            /* Call back. */
            value = cgul_hash_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the hash with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the hash starting at the back of the
     * hash and iterating backward to the front of the hash. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_values(fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_hash_node_t curr = NULL;
        cgul_hash_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__foldr_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_hash__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_hash_node__get_prev(&local, curr);
            /* Call back. */
            value = cgul_hash_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the hash with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each key/value pair in the hash starting at the front of the
     * hash and iterating forward to the end of the hash. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_pairs(fold_pair_t f,
                void* data)
    {
        const void* key = NULL;
        void* value = NULL;
        cgul_hash_node_t curr = NULL;
        cgul_hash_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__foldl_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_hash__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_hash_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_hash_node__get_key(&local, curr);
            value = cgul_hash_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the hash with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key/value pair in the hash starting at the back
     * of the hash and iterating backward to the front of the
     * hash. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_pairs(fold_pair_t f,
                void* data)
    {
        const void* key = NULL;
        void* value = NULL;
        cgul_hash_node_t curr = NULL;
        cgul_hash_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__foldr_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_hash__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_hash_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_hash_node__get_key(&local, curr);
            value = cgul_hash_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * Traverse all nodes passing each node to the function
     * <code>f</code>. <p>
     *
     * The first parameter passed into <code>f</code> is the hash
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the hash elements.  In fact, I would
     * recommend that you use
     * <code>cgul_hash_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_hash_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the hash, but you do have to be careful.
     *
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse_range()
     */
    virtual void
    traverse(traverse_t f,
             void* data)
    {
        traverse_range(NULL, NULL, f, data);
    }

    /**
     * Traverse all nodes starting with <code>first</code> (inclusive) and
     * ending with <code>last</code> (inclusive) passing each node to the
     * function <code>f</code>.  If you know the first node, but are not
     * sure of the last node, just use <code>NULL</code> as the last node.
     * This will cause this method to iterate until it reaches the end of
     * the hash.  You can than have <code>f</code> return 0 when it
     * determines that the last node has been reached. <p>
     *
     * The first parameter passed into <code>f</code> is the hash
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * If <code>first</code> is <code>NULL</code>, iteration starts at the
     * beginning of the hash.  If <code>last</code> is <code>NULL</code>,
     * iteration stops at the end of the hash. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the hash elements.  In fact, I would
     * recommend that you use
     * <code>cgul_hash_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_hash_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the hash, but you do have to be careful.
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse()
     */
    virtual void
    traverse_range(cgul_hash_node_cxx* first,
                   cgul_hash_node_cxx* last,
                   traverse_t f,
                   void* data)
    {
        cgul_hash_node_cxx* curr = NULL;
        cgul_hash_node_cxx* next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_hash__traverse_range()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */
        
        if (first == NULL) {
            first = (cgul_hash_node_cxx*)cgul_hash__get_front(&local, obj);
        }
        if (last == NULL) {
            last = (cgul_hash_node_cxx*)cgul_hash__get_back(&local, obj);
        }

        /* Iterate over the nodes. */
        for (curr = first ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = (cgul_hash_node_cxx*)
                       cgul_hash_node__get_next(&local, (cgul_hash_node_t)curr);
            /* Call back. */
            if (!(*f)(this, curr, data)) {
                break;
            }
            /* We are done if we just processed the last node. */
            if (curr == last) {
                break;
            }
        }
    }

    /**
     * Get the underlying <code>cgul_hash</code> object.
     *
     * @return underlying object
     */
    virtual cgul_hash_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_hash</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_hash__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_hash_t
    take_obj()
    {
        cgul_hash_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_hash_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_hash__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_hash_cxx(const cgul_hash_cxx&);
    cgul_hash_cxx& operator=(const cgul_hash_cxx&);

    /** Underlying <code>cgul_hash</code> object. */
    cgul_hash_t obj;

};

#endif  /*  CGUL_HASH_CXX_H  */
