#ifndef CGUL_HEAP_PRIMITIVES_CXX_H
#define CGUL_HEAP_PRIMITIVES_CXX_H

/**
 * C++ wrapper for <code>cgul_heap_primitives</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_heap_primitives</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_heap_primitives.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_heap_primitives</code> functions.  The main purpose of
 * this class is to convert the C-style function calls and exception
 * handling in <code>cgul_heap_primitives</code> into C++-style
 * function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_heap_primitives</code>
 */
class cgul_heap_primitives_cxx {

public:

    /**
     * This typedef is the interface the client must define in order for
     * the <code>cgul_heap_primitives_cxx</code> methods to maintain heap
     * order.  When defining the comparison function, it should compare
     * <code>e1</code> to <code>e2</code> and return less than zero, zero,
     * or greater than zero if <code>e1</code> is less than, equal to, or
     * greater than <code>e2</code>.
     *
     * @param[in] e1 first element
     * @param[in] e2 second element
     * @return result of comparison
     */
    typedef cgul_heap_primitives__compare_t compare_t;

    /**
     * This method converts the unordered array <code>v</code> into a
     * heap.  This is done using the bottom-up algorithm which is O(n).
     *
     * @param[in] v array of pointers to elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] cf comparison function
     */
    static void
    heapify(void** v,
            size_t v_count,
            compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_heap_primitives__heapify(&local, v, v_count, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Shift the element at <code>v[index]</code> down until it is greater
     * than or equal to all of its descendants as determined by the
     * comparison function <code>cf</code>.  (This code is for a
     * max-heap.) <p>
     *
     * If <code>floyd</code> is true, Floyd's Method is used as an
     * optimization to reduce the number of callbacks required to the
     * comparison function.  It works by initially assuming the value
     * being shifted down is smaller than both children causing the larger
     * child to be shifted up at each level until a leaf node is reached.
     * Then, <code>shift_up()</code> is called to
     * find the correct location for the value.  This is an optimization
     * when forced to replace the root value with a value from the back of
     * the array which is probably smaller than all the values along the
     * path described.
     *
     * @note Do *not* use Floyd's Method until after the heap has been
     * fully constructed because <code>shift_up()</code>
     * assumes the heap is fully constructed. <p>
     *
     * @param[in] v array of pointers to elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] index index of element to be shifted down
     * @param[in] use_floyds_method use Floyd's Method
     * @param[in] cf comparison function
     */
    static void
    shift_down(void** v,
               size_t v_count,
               size_t index,
               int use_floyds_method,
               compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_heap_primitives__shift_down(&local,
                                        v,
                                        v_count,
                                        index,
                                        use_floyds_method,
                                        cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Shift the element at <code>v[index]</code> up until it is the root
     * element or until it is greater than or equal to all of its
     * descendants but less than its parent as determined by the
     * comparison function <code>cf</code>.  (This code is for a
     * max-heap.)
     *
     * @param[in] v array of pointers to elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] index index of element to be shifted down
     * @param[in] cf comparison function
     */
    static void
    shift_up(void** v,
             size_t v_count,
             size_t index,
             compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_heap_primitives__shift_up(&local,
                                      v,
                                      v_count,
                                      index,
                                      cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

};

#endif  /*  CGUL_HEAP_PRIMITIVES_CXX_H  */
