#ifndef CGUL_RANDOM_XORSHIFT32_CXX_H
#define CGUL_RANDOM_XORSHIFT32_CXX_H

/**
 * C++ wrapper for <code>cgul_random_xorshift32</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_random_xorshift32</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_random_xorshift32.h"

/**
 * This class provides the C++ bindings for
 * <code>cgul_random_xorshift32</code>.  The main purpose of this class is
 * to convert the C-style function calls and exception handling into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_random_xorshift32</code>
 */
class cgul_random_xorshift32_cxx {

public:

    /**
     * Default constructor that creates a new object seeding it with
     * the value returned by <code>time()</code>, i.e., the 32-bit
     * seconds since the epoch.  The caller is responsible for
     * arranging for <code>delete</code> to be called on the object.
     * If memory cannot be allocated, and an exception is thrown.
     */
    cgul_random_xorshift32_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_random_xorshift32__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new object seeding it with <code>seed</code>.  The
     * caller is responsible for arranging for <code>delete</code> to
     * be called on the object.  If memory cannot be allocated, an
     * exception is thrown.
     *
     * @param[in] seed seed value
     */
    cgul_random_xorshift32_cxx(cgul_uint32_t seed)
    {
        cgul_exception_t local = NULL;
        obj = cgul_random_xorshift32__new_from_seed(&local, seed);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new object by seeding it with the
     * <code>seeds</code> array.  This is the correct way to create a
     * new object; however it can be difficult to acquire the
     * requisite entropy to fill the entire array.  The caller is
     * responsible for arranging for <code>delete</code> to be called
     * on the object.  If memory cannot be allocated or if all the
     * seed values are zero, an exception is thrown. <p>
     *
     * The caller is responsible for making sure <code>seeds</code>
     * has the correct number of elements.  The easiest way to do this
     * is by using <code>allocate_seed_array()</code> to allocate the
     * array and get its size. <p>
     *
     * @param[in] seeds array of 32-bit values used to seed the PRNG
     * @see allocate_seed_array()
     */
    cgul_random_xorshift32_cxx(cgul_uint32_t* seeds)
    {
        cgul_exception_t local = NULL;
        obj = cgul_random_xorshift32__new_from_seed_array(&local, seeds);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_random_xorshift32_cxx</code> object by wrapping
     * an existing <code>cgul_random_xorshift32</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_random_xorshift32_cxx(cgul_random_xorshift32_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  Do not
     * attempt to use the object after calling this method.
     */
    virtual
    ~cgul_random_xorshift32_cxx()
    {
        if (obj) {
            cgul_random_xorshift32__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This function allocates and returns an array of 32-bit seed
     * values <code>v</code> where each element of the array is
     * initialized to zero.  The total number of 32-bit seed values in
     * the array is returned in <code>v_count</code>.  The caller
     * should fill the array with true random values and create a new
     * PRNG seeded with the values by calling the correct constructor.
     * The caller is responsible for calling <code>free(*v)</code>.
     * If an error occurs, an exception is thrown.
     *
     * @param[out] v uninitialized array of 32-bit seed values
     * @param[out] v_count count of the number of 32-bit values in <code>v</code>
     */
    static void
    allocate_seed_array(cgul_uint32_t** v,
                        unsigned int* v_count)
    {
        cgul_exception_t local = NULL;
        cgul_random_xorshift32__allocate_seed_array(&local, v, v_count);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the next element in the pseudo-random number sequence.  The
     * random numbers returned are 32 bits wide and should have a period
     * of (2^4096 - 1). <p>
     *
     * @return next element in the random number sequence
     * @see next_in_range()
     * @see next_in_unit()
     */
    virtual cgul_uint32_t
    next()
    {
        cgul_uint32_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_random_xorshift32__next(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * If the value returned by <code>next()</code> is
     * not the exact range that is needed, this method can limit the
     * return value to [0, range_max).  If <code>range_max</code> is zero,
     * an exception is thrown. <p>
     *
     * This function should avoid the very small round off error
     * associated with just scaling the value returned by
     * <code>next_in_unit()</code>; however, this
     * function is much slower. <p>
     *
     * @param[in] range_max exclusive limit on the maximum value returned
     * @return next element in the random number sequence
     */
    virtual cgul_uint32_t
    next_in_range(cgul_uint32_t range_max)
    {
        cgul_uint32_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_random_xorshift32__next_in_range(&local, obj, range_max);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the next element in the pseudo-random number sequence as
     * type <code>double</code> on the range [0 - 1). <p>
     *
     * @return next element in the random number sequence
     */
    virtual double
    next_in_unit()
    {
        double rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_random_xorshift32__next_in_unit(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_random_xorshift32</code> object.
     *
     * @return underlying object
     */
    virtual cgul_random_xorshift32_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_random_xorshift32</code> object.
     * This means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_random_xorshift32__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_random_xorshift32_t
    take_obj()
    {
        cgul_random_xorshift32_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_random_xorshift32_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_random_xorshift32__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_random_xorshift32_cxx(const cgul_random_xorshift32_cxx&);
    cgul_random_xorshift32_cxx& operator=(const cgul_random_xorshift32_cxx&);

    /** Underlying <code>cgul_random_xorshift32</code> object. */
    cgul_random_xorshift32_t obj;

};

#endif  /*  CGUL_RANDOM_XORSHIFT32_CXX_H  */
