#ifndef CGUL_ROLLING_HASH_H
#define CGUL_ROLLING_HASH_H

/**
 * Rolling hash function.  The rolling hash function allows the hash
 * value to change incrementally by quickly removing the oldest byte
 * and adding a new one to take its place.  It is used by
 * <code>cgul_substring()</code> to implement the Rabin-Karp substring
 * search, but it can be used generally.
 *
 * @file
 * @author Paul Serice
 * @brief rolling hash function
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/**
 * This function calculates the hash value for <code>length</code>
 * bytes in the block <code>block</code>.  The main purpose of this
 * function is to generate the initial hash value that can then be
 * passed into <code>cgul_rolling_hash__roll()</code>, but it can also
 * be used as a stand-alone hash function. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] block block of binary data
 * @param[in] block_length length of <code>block</code> in bytes
 * @return hash value
 * @see cgul_rolling_hash__roll()
 * @see cgul_hash__hash_block()
 */
extern CGUL_EXPORT cgul_uint32_t
cgul_rolling_hash__hash(cgul_exception_t* cex,
                        const char* block,
                        size_t block_length);

/**
 * Peform a rolling hash calculation on the hash value <code>hv</code>
 * by surgically removing the part of the hash value associated with
 * the old character <code>c_old</code> that is being removed and
 * adding the part of the hash value associated with the new character
 * <code>c_new</code> that is being added.  The initial value for
 * <code>hv</code> should be obtained by calling
 * <code>cgul_rolling_hash__hash()</code>.  The value for
 * <code>block_length</code> should be same as what originally was
 * passed into <code>cgul_rolling_hash__hash()</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] hv old hash value
 * @param[in] c_old old character
 * @param[in] c_new new character
 * @param[in] block_length length of the block in bytes
 * @return new hash value
 * @see cgul_rolling_hash__hash()
 */
extern CGUL_EXPORT cgul_uint32_t
cgul_rolling_hash__roll(cgul_exception_t* cex,
                        cgul_uint32_t hv,
                        const char c_old,
                        const char c_new,
                        size_t block_length);

CGUL_END_C

#endif  /*  CGUL_ROLLING_HASH_H  */
