#ifndef CGUL_SHUFFLE_H
#define CGUL_SHUFFLE_H

/**
 * Shuffle an array.
 *
 * @file
 * @author Paul Serice
 * @brief Fisher-Yates Shuffle
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/**
 * This type definition defines the interface to a pseudo-random
 * number generator (PRNG) which the shuffle functions use to generate
 * pseudo-random numbers that are used to shuffle an array.  The
 * quality of the shuffle depends entirely upon the quality of the
 * pseudo-random numbers returned. <p>
 *
 * Functions that implement this interface will be passed the same
 * PRNG <code>prng</code> that is passed into the shuffle function.
 * The return value should be a pseudo-random integer in the half-open
 * interval <code>[0, range_max)</code>.  If an error occurs when this
 * function is called, the client should propagate the error via
 * <code>cex</code>. <p>
 *
 * Note that this interface closely resembles the signature of the
 * <code>cgul_random_*__next_in_range()</code> functions.  So closely
 * in fact, that it is usually not necessary to write a special
 * callback to implement this function.  Instead, this interface can
 * be satisfied by casting <code>cgul_random_*__next_in_range()</code>
 * to have <code>cgul_shuffle__next_in_range_t</code> function type
 * and pass in your PRNG object as <code>prng</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] range_max maximum, non-negative range for pseudo-random numbers
 * @return pseudo-random integer in the half-open interval [0, range_max)
 */
typedef cgul_uint32_t
(*cgul_shuffle__next_in_range_t)(cgul_exception_t* cex,
                                 void* prng,
                                 cgul_uint32_t range_max);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> arbitrary elements of
 * size <code>element_size</code> (in bytes).  If <code>v</code> holds
 * integers or pointers, one of the other shuffle functions tailored
 * to the specific type will perform better.  The quality of the
 * shuffle depends entirely on the quality of the random numbers
 * returned by the <code>next_in_range</code> callback which is passed
 * the PRNG object <code>prng</code> unmodified.  Because this
 * function assumes a 32-bit PRNG, <code>v_count</code> can indicate
 * no more than (2^32 - 1) elements.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code> (in bytes)
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle(cgul_exception_t* cex,
             void* v,
             size_t v_count,
             size_t element_size,
             cgul_shuffle__next_in_range_t next_in_range,
             void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> <code>void*</code>
 * elements.  The quality of the shuffle depends entirely on the
 * quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_pointers(cgul_exception_t* cex,
                      void** v,
                      size_t v_count,
                      cgul_shuffle__next_in_range_t next_in_range,
                      void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>signed char</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_schar(cgul_exception_t* cex,
                   signed char* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>unsigned char</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_uchar(cgul_exception_t* cex,
                   unsigned char* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>short</code>.  The quality of the shuffle depends entirely on
 * the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_short(cgul_exception_t* cex,
                   short int* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>unsigned short</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_ushort(cgul_exception_t* cex,
                    unsigned short int* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>int</code>.  The quality of the shuffle depends entirely on
 * the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_int(cgul_exception_t* cex,
                 int* v,
                 size_t v_count,
                 cgul_shuffle__next_in_range_t next_in_range,
                 void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>unsigned int</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_uint(cgul_exception_t* cex,
                  unsigned int* v,
                  size_t v_count,
                  cgul_shuffle__next_in_range_t next_in_range,
                  void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>long</code>.  The quality of the shuffle depends entirely on
 * the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_long(cgul_exception_t* cex,
                  long int* v,
                  size_t v_count,
                  cgul_shuffle__next_in_range_t next_in_range,
                  void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>unsigned long</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_ulong(cgul_exception_t* cex,
                   unsigned long int* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_int8_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_int8(cgul_exception_t* cex,
                  cgul_int8_t* v,
                  size_t v_count,
                  cgul_shuffle__next_in_range_t next_in_range,
                  void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_uint8_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_uint8(cgul_exception_t* cex,
                   cgul_uint8_t* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_int16_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_int16(cgul_exception_t* cex,
                   cgul_int16_t* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_uint16_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_uint16(cgul_exception_t* cex,
                    cgul_uint16_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_int32_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_int32(cgul_exception_t* cex,
                   cgul_int32_t* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);

/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_uint32_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_uint32(cgul_exception_t* cex,
                    cgul_uint32_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng);

#ifdef CGUL_INT64
/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_int64_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_int64(cgul_exception_t* cex,
                   cgul_int64_t* v,
                   size_t v_count,
                   cgul_shuffle__next_in_range_t next_in_range,
                   void* prng);
#endif /* CGUL_INT64 */

#ifdef CGUL_UINT64
/**
 * Perform an in-place Fisher-Yates shuffle on the array
 * <code>v</code> having <code>v_count</code> elements of type
 * <code>cgul_uint64_t</code>.  The quality of the shuffle depends
 * entirely on the quality of the random numbers returned by the
 * <code>next_in_range</code> callback which is passed the PRNG
 * <code>prng</code> object unmodified.  Because this function assumes
 * a 32-bit PRNG, <code>v_count</code> can indicate no more than (2^32
 * - 1) elements.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to shuffle
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] next_in_range PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_shuffle_uint64(cgul_exception_t* cex,
                    cgul_uint64_t* v,
                    size_t v_count,
                    cgul_shuffle__next_in_range_t next_in_range,
                    void* prng);
#endif /* CGUL_UINT64 */

CGUL_END_C

#endif  /*  CGUL_SHUFFLE_H  */
