#ifndef CGUL_STANZA_H
#define CGUL_STANZA_H

/**
 * This class lets you get arbitrary stanzas of key/value pairs from a
 * text file.  This requires the text file to be in the following
 * format:
 *
 * <pre>
 *     #
 *     # Comments are lines that have a '#' character as the first
 *     # character that isn't white space.  All other lines are
 *     # treated as key value pairs.
 *     #
 *
 *     # First stanza.
 *     key1: value1
 *     key2: value2
 *     key3: value3 (1st line)
 *        +: value3 (2nd line)
 *        +: value3 (3rd line)
 *
 *     # Second stanza.
 *     key1: value4
 *     key2: value5
 *     key3: value6
 *
 *     # Third stanza.
 *     key1: value7
 *     key2: value8
 *     key3: value9
 *
 *     # etc.
 * </pre>
 * 
 * This class has two modes of operation.  The first mode is when you
 * have an array of hashes (or red-black trees).  This intentionally
 * matches the functionality of the XDR layer in RPC.  So anything
 * that can be sent between machines using RPC can be sent using
 * <code>cgul_stanza</code>. <p>
 *
 * The basic skeleton for using it in this mode is as follows where
 * <code>cgul_stanza__load_next()</code> loads the next hash:
 *
 * <pre>
 *     stanza = cgul_stanza__new_from_fname(cex, fname);
 *     while (cgul_stanza__load_next(cex, stanza)) {
 *         value1 = cgul_stanza__lookup(cex, stanza, key1);
 *         value2 = cgul_stanza__lookup(cex, stanza, key2);
 *         ...
 *     }
 * </pre>
 * 
 * The second mode is when you don't need the array feature.  It's the
 * mode to use for configuration files that a user is expected to edit.
 * It has a lot of overlap with the first mode, and there isn't much
 * difference for the programmer:
 * 
 * <pre>
 *     stanza = cgul_stanza__new_from_fname(cex, fname);
 *     cgul_stanza__all_one_stanza(cex, stanza, 1);
 *     cgul_stanza__load_next(cex, stanza);
 *     value1 = cgul_stanza__lookup(cex, stanza, key1);
 *     value2 = cgul_stanza__lookup(cex, stanza, key2);
 *     ...
 * </pre>
 * 
 * Because configuration options are usually long-lived, I also added the
 * ability to "take" the internal tree so it can be queried at any time:
 * 
 * <pre>
 *     // Slurp all the options.
 *     stanza = cgul_stanza__new_from_fname(cex, fname);
 *     cgul_stanza__all_one_stanza(cex, stanza, 1);
 *     cgul_stanza__load_next(cex, stanza);
 *     options = cgul_stanza__take_tree(cex, stanza);
 *     cgul_stanza__close(cex, stanza);
 *     ...
 *     // Lookup options at a later time.
 *     value1 = cgul_stanza__lookup_tree(cex, options, key1);
 *     value2 = cgul_stanza__lookup_tree(cex, options, key2);
 *     ...
 * </pre>
 *
 * @note Any horizontal white space before the key is stripped and
 * does not become part of the key. <p>
 *
 * @note Any vertical white space after the key denotes the end of the
 * stanza unless <code>cgul_stanza__all_one_stanza()</code> is used to
 * force this class to treat the entire file as one stanza thereby
 * allowing vertical white space between keys in the same stanza.  The
 * main reason for enabling <code>cgul_stanza__all_one_stanza()</code>
 * is to provide a conventional configuration file with comments above
 * each entry and vertical white space below. <p>
 *
 * @note The first space between the colon and the value is optional.
 * All subsequent space is considered to be part of the value. <p>
 *
 * @note This class can handle multi-line text.  Just make sure the
 * "key" for subsequent lines is a "+" character.  You can use
 * <code>cgul_stanza__print()</code> to make sure multi-line key/value
 * pairs are printed correctly. <p>
 *
 * @note This class is intentionally designed to avoid having escape
 * sequences. <p>
 *
 * @note This class automatically handles DOS, Mac, and Unix text
 * files. <p>
 *
 * @file
 * @author Paul Serice
 * @brief get arbitrary stanzas of key/value pairs from a text file
 */

#include "cgul_common.h"
#include "cgul_crlf_file.h"
#include "cgul_exception.h"
#include "cgul_rbtree.h"
#include "cgul_vector.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_stanza</code> instance. */
typedef struct cgul_stanza* cgul_stanza_t;

/**
 * Create a new <code>cgul_stanza</code> instance.  After calling this
 * method, an input source should be opened before calling any other
 * method.  The client is responsible for freeing the object by
 * calling <code>cgul_stanza__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT cgul_stanza_t
cgul_stanza__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_stanza</code> object and call
 * <code>cgul_stanza__open_fname()</code> passing it
 * <code>fname</code>.  This method opens <code>fname</code> so that
 * input can be read from it.  For regular files,
 * <code>is_block_buffered</code> should be set to true; for
 * character-oriented or line-oriented files,
 * <code>is_block_buffered</code> should be set to false.  The client
 * is responsible for freeing the object by calling
 * <code>cgul_stanza__delete()</code>.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name of stanza file
 * @param[in] is_block_buffered whether <code>fname</code> is block buffered
 * @return new <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT cgul_stanza_t
cgul_stanza__new_from_fname(cgul_exception_t* cex,
                            const char* fname,
                            int is_block_buffered);

/**
 * Create a new <code>cgul_stanza</code> object and call
 * <code>cgul_stanza__open_file()</code> passing it <code>f</code>.
 * This method uses <code>f</code> when reading input.  This class
 * does not take ownership of <code>f</code>.  Thus, the client is
 * still responsible for calling <code>fclose()</code> on
 * <code>f</code>.  For regular files, <code>is_block_buffered</code>
 * should be set to true; for character-oriented or line-oriented
 * files, <code>is_block_buffered</code> should be set to false.  The
 * client is responsible for freeing the object by calling
 * <code>cgul_stanza__delete()</code>.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[in] is_block_buffered whether <code>f</code> is block buffered
 * @return new <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT cgul_stanza_t
cgul_stanza__new_from_file(cgul_exception_t* cex,
                           FILE* f,
                           int is_block_buffered);

/**
 * Create a new <code>cgul_stanza</code> object and call
 * <code>cgul_stanza__open_memory()</code> passing it <code>buf</code>
 * and <code>buf_size</code>.  This method uses <code>buf</code> when
 * reading input.  This class does not take ownership of
 * <code>buf</code>.  Thus, the client is still responsible for
 * calling <code>free()</code> on <code>buf</code> if necessary.  The
 * client is responsible for freeing the object by calling
 * <code>cgul_stanza__delete()</code>.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] buf memory buffer
 * @param[in] buf_size size of <code>buf</code> in bytes
 * @return new <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT cgul_stanza_t
cgul_stanza__new_from_memory(cgul_exception_t* cex,
                             const char* buf,
                             size_t buf_size);

/**
 * Create a new <code>cgul_stanza</code> object and call
 * <code>cgul_stanza__open_crlf_file()</code> passing it
 * <code>crlf_file</code>.  This makes it possible to read from any
 * file type supported by <code>cgul_crlf_file</code> like memory for
 * example.  The client is responsible for freeing the object returned
 * by calling <code>cgul_stanza__delete()</code>.  If an error occurs,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * The <code>crlf_file</code> object is used as the interface for
 * reading lines of text from a data source.  It should already be
 * initialized before calling this method.  This class does not take
 * ownership of <code>crlf_file</code> so the client is still
 * responsible for calling <code>cgul_crlf_file__delete()</code> on
 * it. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file interface used for reading lines of text
 * @return new <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT cgul_stanza_t
cgul_stanza__new_from_crlf_file(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file);

/**
 * This method closes the file opened when
 * <code>cgul_stanza__new_from_fname()</code> or
 * <code>cgul_stanza__open_fname()</code> was called and frees all
 * internally allocated resources.  Do not try to access
 * <code>stanza</code> after calling this method.
 *
 * @param[in] stanza <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT void
cgul_stanza__delete(cgul_stanza_t stanza);

/**
 * Open the file with name <code>fname</code> so that
 * <code>cgul_stanza__load_next()</code> can be called.  If an input
 * source is already open, it will be closed before this method
 * attempts to open the new source.  The new file will be closed when
 * this instance is deleted.  For regular files,
 * <code>is_block_buffered</code> should be set to true; for
 * character-oriented or line-oriented files,
 * <code>is_block_buffered</code> should be set to false.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] fname file name
 * @param[in] is_block_buffered whether <code>f</code> is block buffered
 */
extern CGUL_EXPORT void
cgul_stanza__open_fname(cgul_exception_t* cex,
                        cgul_stanza_t stanza,
                        const char* fname,
                        int is_block_buffered);

/**
 * Open <code>f</code> so that <code>cgul_stanza__load_next()</code>
 * can be called.  If an input source is already open, it will be
 * closed before this method attempts to open the new source.  This
 * class does not take ownership of <code>f</code>.  Thus, the client
 * is still responsible for calling <code>fclose()</code> on
 * <code>f</code>.  For regular files, <code>is_block_buffered</code>
 * should be set to true; for character-oriented or line-oriented
 * files, <code>is_block_buffered</code> should be set to false.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] f file
 * @param[in] is_block_buffered whether <code>f</code> is block buffered
 */
extern CGUL_EXPORT void
cgul_stanza__open_file(cgul_exception_t* cex,
                       cgul_stanza_t stanza,
                       FILE* f,
                       int is_block_buffered);

/**
 * Open <code>buf</code> so that <code>cgul_stanza__load_next()</code>
 * can be called.  If an input source is already open, it will be
 * closed before this method attempts to open the new source.  This
 * class does not take ownership of <code>buf</code>.  Thus, the
 * client is still responsible for calling <code>free()</code> on
 * <code>buf</code> if necessary.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] buf memory buffer
 * @param[in] buf_size size of <code>buf</code> in bytes
 */
extern CGUL_EXPORT void
cgul_stanza__open_memory(cgul_exception_t* cex,
                         cgul_stanza_t stanza,
                         const char* buf,
                         size_t buf_size);

/**
 * Open <code>crlf_file</code> so that input can be read from it.
 * This makes it possible to read from any file type supported by
 * <code>cgul_crlf_file</code> like memory for example.  If an input
 * source is already open, it will be closed before this method
 * attempts to open the new source.  If an error occurs, an exception
 * is thrown. <p>
 *
 * The <code>crlf_file</code> object is used as the interface for
 * reading lines of text from a data source.  It should already be
 * initialized before calling this method.  This class does not take
 * ownership of <code>crlf_file</code> so the client is still
 * responsible for calling <code>cgul_crlf_file__delete()</code> on
 * it. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] crlf_file interface used for reading lines of text
 */
extern CGUL_EXPORT void
cgul_stanza__open_crlf_file(cgul_exception_t* cex,
                            cgul_stanza_t stanza,
                            cgul_crlf_file_t crlf_file);

/**
 * This method closes <code>stanza</code>.  After calling this method,
 * an input source should be opened before calling any other method.
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 */
extern CGUL_EXPORT void
cgul_stanza__close(cgul_exception_t* cex,
                   cgul_stanza_t stanza);

/**
 * Get the name of the currently open file.  The client must not
 * attempt to call <code>free()</code> on the pointer returned.  This
 * method returns <code>NULL</code> and throws an exception if a data
 * source is not currently open.
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return file name
 */
extern CGUL_EXPORT const char*
cgul_stanza__get_fname(cgul_exception_t* cex,
                       cgul_stanza_t stanza);

/**
 * This methods lets the client determine whether to treat the file as
 * being all one stanza.  By default this class treats empty lines as
 * the separator between stanzas.  Basically, this allows the file to
 * be treat as an array of hashes where vertical white space separates
 * each hash, and the client calls
 * <code>cgul_stanza__load_next()</code> to get the next hash. <p>
 *
 * When used in the default mode, vertical white space is not allowed
 * between key/value pairs that are part of the same hash because the
 * vertical white space indicates the end of one hash and the
 * beginning of the next.  This can be limiting when you just need to
 * save configuration values because white space is traditionally
 * allowed to separate key/value pairs under those circumstances. <p>
 *
 * If you call this method with <code>all_one_stanza</code> set, it
 * will treat the entire file as all one stanza.  This allows you to
 * use vertical white space to separate key/value pairs, and you can
 * slurp the entire file with just one call to
 * <code>cgul_stanza__load_next()</code>. <p>
 *
 * If you set <code>all_one_stanza</code>, you will typically also
 * want to call <code>cgul_stanza__take_tree()</code>.  This lets you
 * get the results of reading the file so that you can continue to use
 * it.  You can then call <code>cgul_stanza__delete()</code> in order
 * to close the file immediately. <p>
 *
 * Calls to <code>cgul_stanza__close()</code> do not alter whether the
 * file is treated as all one stanza. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] all_one_stanza whether the file is all one stanza
 */
extern CGUL_EXPORT void
cgul_stanza__all_one_stanza(cgul_exception_t* cex,
                            cgul_stanza_t stanza,
                            int all_one_stanza);

/**
 * This methods lets the client determine whether duplicate keys are
 * allowed.  By default, duplicate keys are not allowed.  This means
 * an exception is thrown whenever a duplicate key is found in the
 * same stanza.  If duplicate keys are allowed, the last key found is
 * used to set the value.  The values associated with the previous
 * duplicate keys are silently ignored.  Calls to
 * <code>cgul_stanza__close()</code> do not whether duplicate keys are
 * allowed. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] allow_duplicate_keys whether to allow duplicate keys
 */
extern CGUL_EXPORT void
cgul_stanza__allow_duplicate_keys(cgul_exception_t* cex,
                                  cgul_stanza_t stanza,
                                  int allow_duplicate_keys);

/**
 * If this method is called with <code>allowed_keys</code> not set to
 * <code>NULL</code>, this class will verify each key it encounters by
 * making sure it exists in <code>allowed_keys</code> before inserting
 * it into the tree.  If the key is not present in
 * <code>allowed_keys</code>, an exception will be thrown. <p>
 *
 * If this method is called with <code>allowed_keys</code> set to
 * <code>NULL</code> the default behavior of allowing all keys is
 * restored. <p>
 *
 * If <code>cgul_stanza__close()</code> is called, the allowed keys
 * are not automatically reset to <code>NULL</code>.  In this case, if
 * you want to restore the default behavior, this method must
 * explicitly be called with <code>allowed_keys</code> set to
 * <code>NULL</code>.  Thus, if the default behavior is not restored,
 * <code>allowed_keys</code> must remain in scope across calls to
 * <code>cgul_stanza__close()</code>. <p>
 *
 * The distinction between the <code>allowed_keys</code> list and the
 * <code>mandatory_keys</code> list is that the
 * <code>allowed_keys</code> list causes an exception to be thrown
 * whenever an unexpected key is added to a stanza.  The
 * <code>mandatory_keys</code> list causes an exception to be thrown
 * whenever an expected key is missing from a stanza. <p>
 *
 * Note that it is perfectly legal, and often desirable, to use the
 * same list for <code>allowed_keys</code> and
 * <code>mandator_keys</code>.  If you do this, you will specify
 * exactly the format of each stanza. <p>
 *
 * Also note that you can specify optional tags by adding the optional
 * tags to the <code>allowed_keys</code> list but leaving them off of
 * the <code>mandatory_keys</code> list. <p>

 * This class uses <code>allowed_keys</code> directly without taking
 * ownership and without making a copy of it.  Thus,
 * <code>allowed_keys</code> must stay in scope while the client
 * iterates over the file.  Because this class does not take ownership
 * of <code>allowed_keys</code>, the client is still responsible for
 * deleting it. <p>
 *
 * The following is an example of how to set up the list of allowed
 * keys: <p>
 *
 * <pre>
 *     allowed_keys = cgul_rbtree__new(cex, (cgul_rbtree__compare_t)strcmp);
 *     cgul_rbtree__insert(cex, allowed_keys, "foo", NULL);
 *     cgul_rbtree__insert(cex, allowed_keys, "bar", NULL);
 *     cgul_rbtree__insert(cex, allowed_keys, "baz", NULL);
 *     if (cex) {
 *         goto out;
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] allowed_keys list of allowed keys
 */
extern CGUL_EXPORT void
cgul_stanza__set_allowed_keys(cgul_exception_t* cex,
                              cgul_stanza_t stanza,
                              cgul_rbtree_t allowed_keys);

/**
 * If this method is called with <code>mandatory_keys</code> not set
 * to <code>NULL</code>, after reading each stanza in full, this class
 * will verify that all mandatory keys are present.  If any one of the
 * mandatory keys is not present in the stanza, an exception will be
 * thrown. <p>
 *
 * If this method is called with <code>mandatory_keys</code> set to
 * <code>NULL</code> the default behavior of not verifying keys is
 * restored. <p>
 *
 * If <code>cgul_stanza__close()</code> is called, the mandatory keys
 * are not automatically reset to <code>NULL</code>.  In this case, if
 * you want to restore the default behavior, this method must
 * explicitly be called with <code>mandatory_keys</code> set to
 * <code>NULL</code>.  Thus, if the default behavior is not restored,
 * <code>mandatory_keys</code> must remain in scope across calls to
 * <code>cgul_stanza__close()</code>. <p>
 *
 * The distinction between the <code>allowed_keys</code> list and the
 * <code>mandatory_keys</code> list is that the
 * <code>allowed_keys</code> list causes an exception to be thrown
 * whenever an unexpected key is added to a stanza.  The
 * <code>mandatory_keys</code> list causes an exception to be thrown
 * whenever an expected key is missing from a stanza. <p>
 *
 * Note that it is perfectly legal, and often desirable, to use the
 * same list for <code>allowed_keys</code> and
 * <code>mandator_keys</code>.  If you do this, you will specify
 * exactly the format of each stanza. <p>
 *
 * Also note that you can specify optional tags by adding the optional
 * tags to the <code>allowed_keys</code> list but leaving them off of
 * the <code>mandatory_keys</code> list. <p>
 *
 * This class uses <code>mandatory_keys</code> directly without taking
 * ownership and without making a copy of it.  Thus,
 * <code>mandatory_keys</code> must stay in scope while the client
 * iterates over the file.  Because this class does not take ownership
 * of <code>mandatory_keys</code>, the client is still responsible for
 * deleting it. <p>
 *
 * The following is an example of how to set up the list of mandatory
 * keys: <p>
 *
 * <pre>
 *     mandatory_keys = cgul_rbtree__new(cex, (cgul_rbtree__compare_t)strcmp);
 *     cgul_rbtree__insert(cex, mandatory_keys, "foo", NULL);
 *     cgul_rbtree__insert(cex, mandatory_keys, "bar", NULL);
 *     cgul_rbtree__insert(cex, mandatory_keys, "baz", NULL);
 *     if (cex) {
 *         goto out;
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] mandatory_keys list of mandatory keys
 */
extern CGUL_EXPORT void
cgul_stanza__set_mandatory_keys(cgul_exception_t* cex,
                                cgul_stanza_t stanza,
                                cgul_rbtree_t mandatory_keys);

/**
 * This method loads the next stanza.  The client can use
 * <code>cgul_stanza__lookup()</code> or
 * <code>cgul_stanza__take_value()</code> to lookup the value given
 * the key or <code>cgul_stanza__get_tree()</code> to iterate over all
 * key/value pairs.  An input source must be opened before attempting
 * to call this method.  On success, 1 is returned.  On EOF, 0 is
 * returned.  If an error occurs, 0 is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return whether EOF was not reached
 */
extern CGUL_EXPORT int
cgul_stanza__load_next(cgul_exception_t* cex,
                       cgul_stanza_t stanza);

/**
 * This method returns the value associated with the key
 * <code>key</code>.  The pointer returned is owned by this class and
 * will be invalidated the next time
 * <code>cgul_stanza__load_next()</code> is called.  If the key is
 * not present in this stanza, <code>NULL</code> is returned. <p>
 *
 * If the caller is going to make a copy of the value returned, it
 * should probably use <code>cgul_stanza__take_value()</code> instead
 * which functions the same as this method except ownership is
 * transfered to the caller thus avoiding a copy. <p>
 *
 * If you need more functionality than what is provided by this
 * method, you can use <code>cgul_stanza__get_tree()</code> and
 * iterate over each key/value pair instead. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] key key
 * @return value associated with key
 */
extern CGUL_EXPORT const char*
cgul_stanza__lookup(cgul_exception_t* cex,
                    cgul_stanza_t stanza,
                    const char* key);

/**
 * This method returns the value associated with the key
 * <code>key</code>.  Ownership of the pointer returned is transfered
 * to the caller.  Thus, the caller must arrange for
 * <code>free()</code> to be called on the pointer.  If the key is not
 * present in this stanza, <code>NULL</code> is returned.  <p>
 *
 * Note that once the value for a particular key has been taken,
 * subsequent lookups on that key return <code>NULL</code> for the
 * value.  In order to do otherwise would require making a copy of the
 * value which would defeat the purpose of this method. <p>
 *
 * If the caller only needs to inspect the value, it should probably
 * use <code>cgul_stanza__lookup()</code> instead allowing the
 * <code>cgul_stanza</code> instance to retain ownership thereby
 * guaranteeing that the pointer will be freed. <p>
 *
 * If you need more functionality than what is provided by this
 * method, you can use <code>cgul_stanza__get_tree()</code> and
 * iterate over each key/value pair instead. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @param[in] key key
 * @return value associated with key
 */
extern CGUL_EXPORT char*
cgul_stanza__take_value(cgul_exception_t* cex,
                        cgul_stanza_t stanza,
                        const char* key);

/**
 * This method returns the internal <code>cgul_rbtree</code> that is
 * used to hold all the key/value pairs of the current stanza.  The
 * client typically only needs to get the tree once.  It will not be
 * invalidated until <code>cgul_stanza__take_tree()</code> or
 * <code>cgul_stanza__delete()</code> is called. <p>
 *
 * Both the keys and the values in the tree are <code>char*</code>
 * types. <p>
 *
 * The client must not alter the keys in the tree that is returned,
 * but the client is free to take ownership of the values provided the
 * client notifies this class that ownership has been transfered by
 * setting the corresponding value in the tree to <code>NULL</code>.
 * The client must call <code>free()</code> on any pointer taken. <p>
 *
 * Note that you can use <code>cgul_stanza__lookup_tree()</code> to
 * perform looks directly using the value returned. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return tree of key/value pairs
 */
extern CGUL_EXPORT cgul_rbtree_t
cgul_stanza__get_tree(cgul_exception_t* cex,
                      cgul_stanza_t stanza);

/**
 * This method returns the internal <code>cgul_rbtree</code> that is
 * used to hold all the key/value pairs of the current stanza. <p>
 *
 * The client is now responsible for making sure <b>both</b>
 * <code>cgul_stanza__clear_tree()</code> and
 * <code>cgul_rbtree__delete()</code> are called on the value
 * returned. <p>
 *
 * When the current tree is taken, this class creates a new, empty
 * tree to take its place.  If an error occurs while attempting to
 * create the new, empty tree, an exception is thrown, and
 * <code>NULL</code> is returned.  In this case, ownership of the
 * original tree stays with this class. <p>
 *
 * Note that you can use <code>cgul_stanza__lookup_tree()</code> to
 * perform looks directly using the value returned. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return tree of key/value pairs
 */
extern CGUL_EXPORT cgul_rbtree_t
cgul_stanza__take_tree(cgul_exception_t* cex,
                       cgul_stanza_t stanza);

/**
 * This method clears the <code>tree</code> by removing and freeing
 * all key/value pairs.  It does not free the tree itself.  You still
 * must call <code>cgul_rbtree__delete()</code> to do that.  This
 * method is really only relevant if
 * <code>cgul_stanza__take_tree()</code> has been called.
 *
 * @param[in] cex c-style exception
 * @param[in] tree tree to empty
 */
void
cgul_stanza__clear_tree(cgul_exception_t* cex,
                        cgul_rbtree_t tree);

/**
 * This method performs the same lookup as
 * <code>cgul_stanza__lookup()</code> except it uses the
 * <code>cgul_rbtree</code> returned by either
 * <code>cgul_stanza__get_tree()</code()</code> or
 * <code>cgul_stanza__take_tree()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] tree <code>cgul_rbtree</code> instance
 * @param[in] key key
 * @return value associated with key
 */
extern CGUL_EXPORT const char*
cgul_stanza__lookup_tree(cgul_exception_t* cex,
                         cgul_rbtree_t tree,
                         const char* key);

/**
 * This method returns a <code>cgul_vector</code> that holds all the
 * keys in this stanza in the exact order in which they appear in the
 * file.  If <code>cgul_stanza__allow_duplicate_keys()</code> has been
 * called to allow duplicate keys, only the first occurrence of a key
 * is stored in the vector. <p>
 *
 * The vector that is returned is owned by this class and must not be
 * deleted by the caller.  Also, the <code>const char*</code> elements
 * of the vector are also owned by this class and must not be freed by
 * the caller. <p>
 *
 * Furthermore, the <code>const char*</code> elements of the vector
 * actually point to the keys that are owned by the main tree
 * structure that is returned by <code>cgul_stanza__get_tree()</code>
 * or <code>cgul_stanza__take_tree()</code>.  Thus, if
 * <code>cgul_stanza__take_tree()</code> is called, the caller is
 * responsible for not calling <code>cgul_stanza__get_keys()</code> if
 * the caller invalidates the keys in the tree. <p>
 *
 * It is possible to call this method just once to get the vector that
 * holds the keys.  Then each time
 * <code>cgul_stanza__load_next()</code> is called, the vector will
 * automatically be updated to reflect the keys in the new stanza. <p>
 *
 * This means the caller should use
 * <code>cgul_stanza__copy_keys()</code> if the caller needs
 * persistent copies of the keys for the current stanza. <p>
 *
 * It is also worth noting that it is probably more efficient in most
 * cases to use <code>cgul_stanza__get_tree()</code> instead of using
 * this method.  The main reason this method exists is for those times
 * when the caller needs to know the exact order in which the keys
 * appear in the file. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return vector of keys
 * @see cgul_stanza__copy_keys()
 */
extern CGUL_EXPORT cgul_vector_t
cgul_stanza__get_keys(cgul_exception_t* cex,
                      cgul_stanza_t stanza);

/**
 * This method returns a <code>cgul_vector</code> that holds all the
 * keys in this stanza in the exact order in which they appear in the
 * file.  If <code>cgul_stanza__allow_duplicate_keys()</code> has been
 * called to allow duplicate keys, only the first occurrence of a key
 * is stored in the vector. <p>
 *
 * Each element of the vector that is returned is a
 * <code>char*</code>.  The caller is responsible for calling both
 * <code>free()</code> on each elemnt of the vector and
 * <code>cgul_vector__delete()</code> on the vector itself. <p>
 *
 * If an error occurs while creating the copy, an exception is thrown,
 * and <code>NULL</code> is returned. <p>
 *
 * If possible <code>cgul_stanza__get_keys()</code> should be used
 * instead of this method to avoid the overhead of having to copy and
 * free the keys. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return vector of keys
 * @see cgul_stanza__get_keys()
 */
extern CGUL_EXPORT cgul_vector_t
cgul_stanza__copy_keys(cgul_exception_t* cex,
                       cgul_stanza_t stanza);

/**
 * Get the line number where the current stanza starts.  A return
 * value of <code>0</code> means <code>cgul_stanza__load_next()</code>
 * has not been called yet. <p>
 *
 * This method throws an exception only if an input source has not
 * been opened yet. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] stanza <code>cgul_stanza</code> instance
 * @return line number
 */
extern CGUL_EXPORT unsigned long int
cgul_stanza__get_line_number(cgul_exception_t* cex,
                             cgul_stanza_t stanza);

/**
 * This static method returns <code>1</code> if <code>line</code> is
 * empty or comprised of all white space; otherwise, it return
 * <code>0</code>.  This method is convenient when parsing a
 * <code>cgul_stanza</code> file outside of this class. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] line line
 * @param[in] line_length line length
 * @return whether the line is blank
 */
extern CGUL_EXPORT int
cgul_stanza__is_blank_line(cgul_exception_t* cex,
                           const char* line,
                           size_t line_length);

/**
 * This static method returns <code>1</code> if <code>line</code> is a
 * comment line; otherwise, it returns <code>0</code>.  A comment line
 * is one that has a '#' character as the first character that isn't
 * white space.  This method is convenient when parsing a
 * <code>cgul_stanza</code> file outside of this class.
 *
 * @param[in] cex c-style exception
 * @param[in] line line
 * @param[in] line_length line length
 * @return whether the line is a comment
 */
extern CGUL_EXPORT int
cgul_stanza__is_comment_line(cgul_exception_t* cex,
                             const char* line,
                             size_t line_length);

/**
 * This static method splits <code>line</code> into a key/value pair
 * returning the key in <code>*key</code> and the value in
 * <code>*value</code> if <code>key</code> and <code>value</code> are
 * not <code>NULL</code> respectively. <p>
 *
 * The caller is responsible for making sure <code>free()</code> is
 * called on both <code>key</code> and <code>value</code> which are
 * both newly allocated.  If an error occurs, an exception is thrown,
 * and <code>*key</code> and <code>*value</code> are both set to
 * <code>NULL</code>. <p>
 *
 * This method assumes that <code>line</code> is in the following
 * format:
 *
 * <pre>
 *     key: value
 * </pre>
 *
 * If <code>fname</code> is not <code>NULL</code> it will be used
 * when reporting errors.  If <code>fname</code> is
 * <code>NULL</code>, "-" will be used instead.
 * <code>line_count</code> is also only used when reporting
 * errors. <p>
 *
 * This method is convenient when parsing a <code>cgul_stanza</code>
 * file outside of this class. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] line line
 * @param[in] line_length line length
 * @param[in] line_count line in file (for error reporting)
 * @param[in] fname file name (for error reporting)
 * @param[out] key key
 * @param[out] value value
 */
extern CGUL_EXPORT void
cgul_stanza__split_line(cgul_exception_t* cex,
                        const char* line,
                        size_t line_length,
                        const char* fname,
                        unsigned long int line_count,
                        char** key,
                        char** value);

/**
 * This static method properly prints the key/value pair to the file
 * <code>f</code> even if <code>value</code> is a multi-line string.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @param[in] key key
 * @param[in] value value
 */
extern CGUL_EXPORT void
cgul_stanza__print(cgul_exception_t* cex,
                   FILE* f,
                   const char* key,
                   const char* value);

CGUL_END_C

#endif  /*  CGUL_STANZA_H  */
