#ifndef CGUL_STANZA_CXX_H
#define CGUL_STANZA_CXX_H

/**
 * C++ wrapper for <code>cgul_stanza</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_stanza</code>
 */

#include "cgul_crlf_file_cxx.h"
#include "cgul_exception_cxx.h"
#include "cgul_rbtree_cxx.h"
#include "cgul_stanza.h"
#include "cgul_vector_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_stanza</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_stanza</code> into C++-style function calls and
 * exception handling.
 *
 * This class lets you get arbitrary stanzas of key/value pairs from a
 * text file.  This requires the text file to be in the following
 * format:
 *
 * <pre>
 *     #
 *     # Comments are lines that have a '#' character as the first
 *     # character that isn't white space.  All other lines are
 *     # treated as key value pairs.
 *     #
 *
 *     # First stanza.
 *     key1: value1
 *     key2: value2
 *     key3: value3 (1st line)
 *        +: value3 (2nd line)
 *        +: value3 (3rd line)
 *
 *     # Second stanza.
 *     key1: value4
 *     key2: value5
 *     key3: value6
 *
 *     # Third stanza.
 *     key1: value7
 *     key2: value8
 *     key3: value9
 *
 *     # etc.
 * </pre>
 * 
 * This class has two modes of operation.  The first mode is when you
 * have an array of hashes (or red-black trees).  This intentionally
 * matches the functionality of the XDR layer in RPC.  So anything
 * that can be sent between machines using RPC can be sent using
 * <code>cgul_stanza_cxx</code>. <p>
 *
 * The basic skeleton for using it in this mode is as follows where
 * <code>load_next()</code> loads the next hash:
 *
 * <pre>
 *     stanza = new cgul_stanza_cxx(fname);
 *     while (stanza->load_next()) {
 *         value1 = stanza->lookup(key1);
 *         value2 = stanza->lookup(key2);
 *         ...
 *     }
 * </pre>
 * 
 * The second mode is when you don't need the array feature.  It's the
 * mode to use for configuration files that a user is expected to edit.
 * It has a lot of overlap with the first mode, and there isn't much
 * difference for the programmer:
 * 
 * <pre>
 *     stanza = new cgul_stanza_cxx(fname);
 *     stanza->all_one_stanza(1);
 *     stanza->load_next();
 *     value1 = stanza->lookup(key1);
 *     value2 = stanza->lookup(key2);
 *     ...
 * </pre>
 * 
 * Because configuration options are usually long-lived, I also added the
 * ability to "take" the internal tree so it can be queried at any time:
 * 
 * <pre>
 *     // Slurp all the options.
 *     stanza = new cgul_stanza_cxx(fname);
 *     stanza->all_one_stanza(1);
 *     stanza->load_next();
 *     options = stanza->take_tree();
 *     stanza->close();
 *     ...
 *     // Lookup options at a later time.
 *     value1 = cgul_stanza::lookup_tree(options, key1);
 *     value2 = cgul_stanza::lookup_tree(options, key2);
 *     ...
 * </pre>
 *
 * @note Any horizontal white space before the key is stripped and
 * does not become part of the key. <p>
 *
 * @note Any vertical white space after the key denotes the end of the
 * stanza unless <code>all_one_stanza()</code> is used to force this
 * class to treat the entire file as one stanza thereby allowing
 * vertical white space between keys in the same stanza.  The main
 * reason for enabling <code>all_one_stanza()</code> is to provide a
 * conventional configuration file with comments above each entry and
 * vertical white space below. <p>
 *
 * @note The first space between the colon and the value is optional.
 * All subsequent space is considered to be part of the value. <p>
 *
 * @note This class can handle multi-line text.  Just make sure the
 * "key" for subsequent lines is a "+" character.  You can use
 * <code>print()</code> to make sure multi-line key/value pairs are
 * printed correctly. <p>
 *
 * @note This class is intentionally designed to avoid having escape
 * sequences. <p>
 *
 * @note This class automatically handles DOS, Mac, and Unix text
 * files. <p>
 *
 * @brief C++ bindings for <code>cgul_stanza</code>
 */
class cgul_stanza_cxx {

public:

    /**
     * This method creates a new <code>cgul_csv_cxx</code> instance.
     * After calling this method, an input source should be opened
     * before calling any other method.  The client is responsible for
     * freeing the object by making sure the destructor gets called.
     * If memory cannot be allocated, an exception is thrown.
     */
    cgul_stanza_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_stanza__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stanza_cxx</code> object and call
     * <code>open_fname()</code> passing it <code>fname</code>.  This
     * method opens <code>fname</code> so that input can be read from
     * it.  For regular files, <code>is_block_buffered</code> should
     * be set to true; for character-oriented or line-oriented files,
     * <code>is_bleock_buffered</code> should be set to false.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] fname file name of stanza file
     * @param[in] is_block_buffered whether <code>fname</code> is block buffered
     */
    cgul_stanza_cxx(const char* fname,
                    bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        obj = cgul_stanza__new_from_fname(&local,
                                          fname,
                                          is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stanza_cxx</code> object and call
     * <code>open_file()</code> passing it <code>f</code>.  This
     * method uses <code>f</code> when reading input.  This class does
     * not take ownership of <code>f</code>.  Thus, the client is
     * still responsible for calling <code>fclose()</code> on
     * <code>f</code>.  For regular files,
     * <code>is_block_buffered</code> should be set to true; for
     * character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] f file
     * @param[in] is_block_buffered whether <code>f</code> is block buffered
     */
    cgul_stanza_cxx(FILE* f,
                    bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        obj = cgul_stanza__new_from_file(&local,
                                         f,
                                         is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stanza_cxx</code> object and call
     * <code>open_memory()</code> passing it <code>buf</code> and
     * <code>buf_size</code>.  This method uses <code>buf</code> when
     * reading input.  This class does not take ownership of
     * <code>buf</code>.  Thus, the client is still responsible for
     * calling <code>free()</code> on <code>buf</code> if necessary.
     * The client is responsible for freeing the object by calling
     * <code>delete</code> if necessary.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] buf memory buffer
     * @param[in] buf_size size of <code>buf</code> in bytes
     */
    cgul_stanza_cxx(const char* buf,
                    size_t buf_size)
    {
        cgul_exception_t local = NULL;
        obj = cgul_stanza__new_from_memory(&local, buf, buf_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stanza_cxx</code> object and call
     * <code>open_crlf_file()</code> passing it
     * <code>crlf_file</code>.  This makes it possible to read from
     * any file type supported by <code>cgul_crlf_file_cxx</code> like
     * memory for example.  If an error occurs, an exception is
     * thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    cgul_stanza_cxx(cgul_crlf_file_cxx& crlf_file)
    {
        cgul_exception_t local = NULL;
        obj = cgul_stanza__new_from_crlf_file(&local, crlf_file.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stanza_cxx</code> object and call
     * <code>open_crlf_file()</code> passing it
     * <code>crlf_file</code>.  This makes it possible to read from
     * any file type supported by <code>cgul_crlf_file_cxx</code> like
     * memory for example.  If an error occurs, an exception is
     * thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    cgul_stanza_cxx(cgul_crlf_file_cxx* crlf_file)
    {
        assert(crlf_file);
        cgul_exception_t local = NULL;
        obj = cgul_stanza__new_from_crlf_file(&local, crlf_file->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_stanza_cxx</code> object by wrapping
     * an existing <code>cgul_stanza</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_stanza_cxx(cgul_stanza_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method destruct the object by closing the file opened by
     * the constructor and free all internally allocated resources.
     */
    virtual
    ~cgul_stanza_cxx()
    {
        if (obj) {
            cgul_stanza__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method closes the object.  After calling this method, an
     * input source should be opened before calling any other
     * method.  */
    virtual void
    close()
    {
        cgul_exception_t local = NULL;
        cgul_stanza__close(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open the file with name <code>fname</code> so that
     * <code>load_next()</code> can be called.  If an input source is
     * already open, it will be closed before this method attempts to
     * open the new source.  The new file will be closed when this
     * instance is deleted.  For regular files,
     * <code>is_block_buffered</code> should be set to true; for
     * character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] fname file name
     * @param[in] is_block_buffered whether <code>f</code> is block buffered
     */
    virtual void
    open_fname(const char* fname,
               bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__open_fname(&local,
                                obj,
                                fname,
                                is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>f</code> so that <code>load_next()</code> can be
     * called.  If an input source is already open, it will be closed
     * before this method attempts to open the new source.  This class
     * does not take ownership of <code>f</code>.  Thus, the client is
     * still responsible for calling <code>fclose()</code> on
     * <code>f</code>.  For regular files,
     * <code>is_block_buffered</code> should be set to true; for
     * character-oriented or line-oriented files,
     * <code>is_block_buffered</code> should be set to false.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] f file
     * @param[in] is_block_buffered whether <code>f</code> is block buffered
     */
    virtual void
    open_file(FILE* f,
              bool is_block_buffered = true)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__open_file(&local,
                               obj,
                               f,
                               is_block_buffered ? 1 : 0);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>buf</code> so that <code>load_next()</code> can be
     * called.  If an input source is already open, it will be closed
     * before this method attempts to open the new source.  This class
     * does not take ownership of <code>buf</code>.  Thus, the client
     * is still responsible for calling <code>free()</code> on
     * <code>buf</code> if necessary.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] buf memory buffer
     * @param[in] buf_size size of <code>buf</code> in bytes
     */
    virtual void
    open_memory(const char* buf,
                size_t buf_size)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__open_memory(&local, obj, buf, buf_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>crlf_file</code> so that input can be read from it.
     * This makes it possible to read from any file type supported by
     * <code>cgul_crlf_file_cxx</code> like memory for example.  If an
     * input source is already open, it will be closed before this
     * method attempts to open the new source.  If an error occurs, an
     * exception is thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    virtual void
    open_crlf_file(cgul_crlf_file_cxx& crlf_file)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__open_crlf_file(&local, obj, crlf_file.get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Open <code>crlf_file</code> so that input can be read from it.
     * This makes it possible to read from any file type supported by
     * <code>cgul_crlf_file_cxx</code> like memory for example.  If an
     * input source is already open, it will be closed before this
     * method attempts to open the new source.  If an error occurs, an
     * exception is thrown. <p>
     *
     * The <code>crlf_file</code> object is used as the interface for
     * reading lines of text from a data source.  It should already be
     * initialized before calling this method.  This class does not
     * take ownership of <code>crlf_file</code> so the client is still
     * responsible for calling <code>delete</code> on it if
     * necessary. <p>
     *
     * @param[in] crlf_file interface used for reading lines of text
     */
    virtual void
    open_crlf_file(cgul_crlf_file_cxx* crlf_file)
    {
        assert(crlf_file);
        cgul_exception_t local = NULL;
        cgul_stanza__open_crlf_file(&local, obj, crlf_file->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the name of the currently open file.  The client must not
     * attempt to call <code>free()</code> on the pointer returned.
     * This method returns <code>NULL</code> and throws an exception
     * if a data source is not currently open.
     *
     * @return file name
     */
    virtual const char*
    get_file_name() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__get_fname(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This methods lets the client determine whether to treat the
     * file as being all one stanza.  By default this class treats
     * empty lines as the separator between stanzas.  Basically, this
     * allows the file to be treat as an array of hashes where
     * vertical white space separates each hash, and the client calls
     * <code>load_next()</code> to get the next hash. <p>
     *
     * When used in the default mode, vertical white space is not
     * allowed between key/value pairs that are part of the same hash
     * because the vertical white space indicates the end of one hash
     * and the beginning of the next.  This can be limiting when you
     * just need to save configuration values because white space is
     * traditionally allowed to separate key/value pairs under those
     * circumstances. <p>
     *
     * If you call this method with <code>all_one_stanza</code> set,
     * it will treat the entire file as all one stanza.  This allows
     * you to use vertical white space to separate key/value pairs,
     * and you can slurp the entire file with just one call to
     * <code>load_next()</code>. <p>
     *
     * If you set <code>all_one_stanza</code>, you will typically also
     * want to call <code>take_tree()</code>.  This lets you get the
     * results of reading the file so that you can continue to use it.
     * You can then delete the object in order to close the file
     * immediately. <p>
     *
     * @param[in] all_one_stanza whether the file is all one stanza
     */
    virtual void
    all_one_stanza(int all_one_stanza)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__all_one_stanza(&local, obj, all_one_stanza);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This methods lets the client determine whether duplicate keys
     * are allowed.  By default, duplicate keys are not allowed.  This
     * means an exception is thrown whenever a duplicate key is found
     * in the same stanza.  If duplicate keys are allowed, the last
     * key found is used to set the value.  The values associated with
     * the previous duplicate keys are silently ignored.
     *
     * @param[in] allow_duplicate_keys whether to allow duplicate keys
     */
    virtual void
    allow_duplicate_keys(int allow_duplicate_keys)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__allow_duplicate_keys(&local, obj, allow_duplicate_keys);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * If this method is called with <code>allowed_keys</code> not set
     * to <code>NULL</code>, this class will verify each key it
     * encounters by making sure it exists in
     * <code>allowed_keys</code> before inserting it into the tree.
     * If the key is not present in <code>allowed_keys</code>, an
     * exception will be thrown. <p>
     *
     * If this method is called with <code>allowed_keys</code> set to
     * <code>NULL</code> the default behavior of allowing all keys is
     * restored. <p>
     *
     * The distinction between the <code>allowed_keys</code> list and
     * the <code>mandatory_keys</code> list is that the
     * <code>allowed_keys</code> list causes an exception to be thrown
     * whenever an unexpected key is added to a stanza.  The
     * <code>mandatory_keys</code> list causes an exception to be
     * thrown whenever an expected key is missing from a stanza. <p>
     *
     * Note that it is perfectly legal, and often desirable, to use
     * the same list for <code>allowed_keys</code> and
     * <code>mandator_keys</code>.  If you do this, you will specify
     * exactly the format of each stanza. <p>
     *
     * Also note that you can specify optional tags by adding the
     * optional tags to the <code>allowed_keys</code> list but leaving
     * them off of the <code>mandatory_keys</code> list. <p>
     *
     * This class uses <code>allowed_keys</code> directly without taking
     * ownership and without making a copy of it.  Thus,
     * <code>allowed_keys</code> must stay in scope while the client
     * iterates over the file.  Because this class does not take ownership
     * of <code>allowed_keys</code>, the client is still responsible for
     * deleting it. <p>
     *
     * The following is an example of how to set up the list of allowed
     * keys: <p>
     *
     * <pre>
     *     cgul_rbtree_cxx allowed_keys((cgul_rbtree__compare_t)strcmp);
     *     allowed_keys.insert("foo", NULL);
     *     allowed_keys.insert("bar", NULL);
     *     allowed_keys.insert("baz", NULL);
     * </pre>
     *
     * @param[in] allowed_keys list of allowed keys
     */
    virtual void
    set_allowed_keys(cgul_rbtree_cxx* allowed_keys)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__set_allowed_keys(&local, obj, allowed_keys->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * If this method is called with <code>mandatory_keys</code> not
     * set to <code>NULL</code>, after reading each stanza in full,
     * this class will verify that all mandatory keys are present.  If
     * any one of the mandatory keys is not present in the stanza, an
     * exception will be thrown. <p>
     *
     * If this method is called with <code>mandatory_keys</code> set to
     * <code>NULL</code> the default behavior of not verifying keys is
     * restored. <p>
     *
     * The distinction between the <code>allowed_keys</code> list and
     * the <code>mandatory_keys</code> list is that the
     * <code>allowed_keys</code> list causes an exception to be thrown
     * whenever an unexpected key is added to a stanza.  The
     * <code>mandatory_keys</code> list causes an exception to be
     * thrown whenever an expected key is missing from a stanza. <p>
     *
     * Note that it is perfectly legal, and often desirable, to use
     * the same list for <code>allowed_keys</code> and
     * <code>mandator_keys</code>.  If you do this, you will specify
     * exactly the format of each stanza. <p>
     *
     * Also note that you can specify optional tags by adding the
     * optional tags to the <code>allowed_keys</code> list but leaving
     * them off of the <code>mandatory_keys</code> list. <p>
     *
     * This class uses <code>mandatory_keys</code> directly without taking
     * ownership and without making a copy of it.  Thus,
     * <code>mandatory_keys</code> must stay in scope while the client
     * iterates over the file.  Because this class does not take ownership
     * of <code>mandatory_keys</code>, the client is still responsible for
     * deleting it. <p>
     *
     * The following is an example of how to set up the list of mandatory
     * keys: <p>
     *
     * <pre>
     *     cgul_rbtree_cxx mandatory_keys((cgul_rbtree__compare_t)strcmp);
     *     mandatory_keys.insert("foo", NULL);
     *     mandatory_keys.insert("bar", NULL);
     *     mandatory_keys.insert("baz", NULL);
     * </pre>
     *
     * @param[in] mandatory_keys list of mandatory keys
     */
    virtual void
    set_mandatory_keys(cgul_rbtree_cxx* mandatory_keys)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__set_mandatory_keys(&local,
                                        obj,
                                        mandatory_keys->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method loads the next stanza.  The client can use
     * <code>lookup()</code> or <code>take_value()</code> to lookup
     * the value given the key or <code>get_tree()</code> to iterate
     * over all key/value pairs.  An input source must be opened
     * before attempting to call this method.  On success,
     * <code>1</code> is returned.  On EOF, <code>0</code> is
     * returned.  If an error occurs, an exception is thrown. <p>
     *
     * @return whether EOF was not reached
     */
    virtual int
    load_next()
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__load_next(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the value associated with the key
     * <code>key</code>.  The pointer returned is owned by this class
     * and will be invalidated the next time <code>load_next()</code>
     * is called.  If the key is not present in this stanza,
     * <code>NULL</code> is returned. <p>
     *
     * If the caller is going to make a copy of the value returned, it
     * should probably use <code>take_value()</code> instead which
     * functions the same as this method except ownership is
     * transfered to the caller thus avoiding a copy. <p>
     *
     * If you need more functionality than what is provided by this
     * method, you can use <code>get_tree()</code> and iterate over
     * each key/value pair instead. <p>
     *
     * This method throws an exception only if an input source has not
     * been opened yet. <p>
     *
     * @param[in] key key
     * @return value associated with key
     */
    virtual const char*
    lookup(const char* key) const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__lookup(&local, obj, key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the value associated with the key
     * <code>key</code>.  Ownership of the pointer returned is
     * transfered to the caller.  Thus, the caller must arrange for
     * <code>free()</code> to be called on the pointer.  If the key is
     * not present in this stanza, <code>NULL</code> is returned. <p>
     *
     * Note that once the value for a particular key has been taken,
     * subsequent lookups on that key return <code>NULL</code> for the
     * value.  In order to do otherwise would require making a copy of the
     * value which would defeat the purpose of this method. <p>
     *
     * If the caller only needs to inspect the value, it should
     * probably use <code>lookup()</code> instead allowing the
     * <code>cgul_stanza_cxx</code> instance to retain ownership
     * thereby guaranteeing that the pointer will be freed. <p>
     *
     * If you need more functionality than what is provided by this
     * method, you can use <code>get_tree()</code> and iterate over
     * each key/value pair instead. <p>
     *
     * This method throws an exception only if an input source has not
     * been opened yet. <p>
     *
     * @param[in] key key
     * @return value associated with key
     */
    virtual char*
    take_value(const char* key) const
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__take_value(&local, obj, key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the internal <code>cgul_rbtree</code> that is
     * used to hold all the key/value pairs of the current stanza. <p>
     *
     * The client is now responsible for making sure <b>both</b>
     * <code>cgul_stanza::clear_tree()</code> and
     * <code>cgul_rbtree::delete()</code> are called on the value
     * returned. <p>
     *
     * When the current tree is taken, this class creates a new, empty
     * tree to take its place.  If an error occurs while attempting to
     * create the new, empty tree, an exception is thrown, and
     * <code>NULL</code> is returned.  In this case, ownership of the
     * original tree stays with this class. <p>
     *
     * Note that you can use <code>cgul_stanza_cxx::lookup_tree()</code>
     * to perform looks directly using the value returned. <p>
     *
     * @return tree of key/value pairs
     */
    virtual cgul_rbtree_cxx*
    take_tree()
    {
        cgul_rbtree_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__take_tree(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return new cgul_rbtree_cxx(rv);
    }

    /**
     * This method clears the <code>tree</code> by removing and
     * freeing all key/value pairs.  It does not free the tree itself.
     * You still must call <code>delete</code> to do that.  This
     * method is really only relevant if <code>take_tree()</code> has
     * been called.  Note that this method is <code>static</code> and
     * should be called as
     * <code>cgul_stanza_cxx::clear_tree()</code>.
     *
     * @param[in] tree tree to empty
     */
    static void
    clear_tree(cgul_rbtree_cxx* tree)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__clear_tree(&local, tree->get_obj());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method performs the same lookup as <code>lookup()</code>
     * except it uses the <code>cgul_rbtree_cxx</code> pointer
     * returned by either <code>get_tree()</code()</code> or
     * <code>take_tree()</code>. <p>
     *
     * Note that this method is <code>static</code> and should be
     * called as <code>cgul_stanza_cxx::lookup_tree()</code>.
     *
     * @param[in] tree <code>cgul_rbtree_cxx</code> instance
     * @param[in] key key
     * @return value associated with key
     */
    static const char*
    lookup_tree(cgul_rbtree_cxx* tree,
                const char* key)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__lookup_tree(&local, tree->get_obj(), key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns a <code>cgul_vector_cxx</code> that holds
     * all the keys in this stanza in the exact order in which they
     * appear in the file.  If <code>allow_duplicate_keys()</code> has
     * been called to allow duplicate keys, only the first occurrence
     * of a key is stored in the vector. <p>
     *
     * Each element of the vector that is returned is a
     * <code>char*</code>.  The caller is responsible for calling
     * <code>free()</code> on each elemnt of the vector and for
     * deleting the vector itself. <p>
     *
     * If an error occurs while creating the copy, an exception is
     * thrown. <p>
     *
     * @return vector of keys
     */
    virtual cgul_vector_cxx*
    copy_keys()
    {
        cgul_vector_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__copy_keys(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return new cgul_vector_cxx(rv);
    }

    /**
     * Get the line number where the current stanza starts.  A return
     * value of <code>0</code> means <code>load_next()</code> has not
     * been called yet. <p>
     *
     * This method throws an exception only if an input source has not
     * been opened yet. <p>
     *
     * @return line number
     */
    virtual unsigned long int
    get_line_number() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__get_line_number(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This static method returns <code>1</code> if <code>line</code> is
     * empty or comprised of all white space; otherwise, it return
     * <code>0</code>.  This method is convenient when parsing a
     * <code>cgul_stanza_cxx</code> file outside of this class.
     *
     * @param[in] line line
     * @param[in] line_length line length
     * @return whether the line is blank
     */
    static int
    is_blank_line(const char* line,
                  size_t line_length)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__is_blank_line(&local, line, line_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This static method returns <code>1</code> if <code>line</code> is a
     * comment line; otherwise, it returns <code>0</code>.  A comment line
     * is one that has a '#' character as the first character that isn't
     * white space.  This method is convenient when parsing a
     * <code>cgul_stanza_cxx</code> file outside of this class.
     *
     * @param[in] line line
     * @param[in] line_length line length
     * @return whether the line is a comment
     */
    static int
    is_comment_line(const char* line,
                    size_t line_length)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_stanza__is_comment_line(&local, line, line_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This static method splits <code>line</code> into a key/value pair
     * returning the key in <code>*key</code> and the value in
     * <code>*value</code> if <code>key</code> and <code>value</code> are
     * not <code>NULL</code> respectively. <p>
     *
     * The caller is responsible for making sure <code>free()</code> is
     * called on both <code>key</code> and <code>value</code> which are
     * both newly allocated.  If an error occurs, an exception is thrown,
     * and <code>*key</code> and <code>*value</code> are both set to
     * <code>NULL</code>. <p>
     *
     * This method assumes that <code>line</code> is in the following
     * format:
     *
     * <pre>
     *     key: value
     * </pre>
     *
     * If <code>fname</code> is not <code>NULL</code> it will be used
     * when reporting errors.  If <code>fname</code> is
     * <code>NULL</code>, "-" will be used instead.
     * <code>line_count</code> is also only used when reporting
     * errors. <p>
     *
     * This method is convenient when parsing a <code>cgul_stanza_cxx</code>
     * file outside of this class. <p>
     *
     * @param[in] line line
     * @param[in] line_length line length
     * @param[in] line_count line in file (for error reporting)
     * @param[in] fname file name (for error reporting)
     * @param[out] key key
     * @param[out] value value
     */
    static void
    split_line(const char* line,
               size_t line_length,
               const char* fname,
               unsigned long int line_count,
               char** key,
               char** value)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__split_line(&local,
                                line,
                                line_length,
                                fname,
                                line_count,
                                key,
                                value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This static method properly prints the key/value pair to the
     * file <code>f</code> even if <code>value</code> is a multi-line
     * string.
     *
     * @param[in] f file
     * @param[in] key key
     * @param[in] value value
     */
    static void
    print(FILE* f,
          const char* key,
          const char* value)
    {
        cgul_exception_t local = NULL;
        cgul_stanza__print(&local, f, key, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_stanza</code> object.
     *
     * @return underlying object
     */
    virtual cgul_stanza_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_stanza</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_stanza__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_stanza_t
    take_obj()
    {
        cgul_stanza_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_stanza_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_stanza__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_stanza_cxx(const cgul_stanza_cxx&);
    cgul_stanza_cxx& operator=(const cgul_stanza_cxx&);

    /** Underlying <code>cgul_stanza</code> object. */
    cgul_stanza_t obj;

};

#endif  /*  CGUL_STANZA_CXX_H  */
