#ifndef CGUL_STRTOK_H
#define CGUL_STRTOK_H

/**
 * String tokenizer designed to replace <code>strtok()</code>.
 *
 * @file
 * @author Paul Serice
 * @brief string tokenizer
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_strtok</code> instance. */
typedef struct cgul_strtok* cgul_strtok_t;

/**
 * Create a new <code>cgul_strtok</code> instance.  The caller is
 * responsible for caling <code>cgul_strtok__delete()</code> on the
 * pointer returned.  If memory allocation fails, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_strtok</code> instance
 */
extern CGUL_EXPORT cgul_strtok_t
cgul_strtok__new(cgul_exception_t* cex);

/**
 * Delete the <code>cgul_strtok</code> object in <code>tok</code>.
 * The client must not use <code>tok</code> after calling this method.
 *
 * @param[in] tok <code>cgul_strtok</code> instance
 */
extern CGUL_EXPORT void
cgul_strtok__delete(cgul_strtok_t tok);

/**
 * Set the value to tokenize.  This method makes an internal copy of
 * <code>value</code> out of which tokens will be returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] tok <code>cgul_strtok</code> instance
 * @param[in] value value to tokenize
 */
extern CGUL_EXPORT void
cgul_strtok__set_value(cgul_exception_t* cex,
                       cgul_strtok_t tok,
                       const char* value);

/**
 * Get the next token that is delimited by the current list of
 * <code>separators</code>.  The list of separators can change each
 * time this method is called.  If all the tokens have been consumed,
 * <code>NULL</code> is returned.  <p>
 *
 * If <code>allow_empty_tokens</code> is true, this method will return
 * empty tokens when two consecutive separators are encountered
 * including the case where a separator immediately precedes the end
 * of string.  If <code>allow_empty_tokens</code> is false,
 * consecutive separators will be telescoped into one, and all tokens
 * returned will be non-empty. <p>
 *
 * The pointer returned is owned by <code>tok</code> and must not be
 * freed by the client. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] tok <code>cgul_strtok</code> instance
 * @param[in] separators list of separators that delimit tokens
 * @param[in] allow_empty_tokens whether to allow returning empty tokens
 * @return next token
 * @see cgul_split()
 */
extern CGUL_EXPORT const char*
cgul_strtok__get_next(cgul_exception_t* cex,
                      cgul_strtok_t tok,
                      const char* separators,
                      int allow_empty_tokens);

/**
 * Get the trailing separator for the current token.  This value can
 * be NUL if the current token terminated because it reached the end
 * of the string.
 *
 * @param[in] cex c-style exception
 * @param[in] tok <code>cgul_strtok</code> instance
 * @return trailing separator for the current token
 */
extern CGUL_EXPORT char
cgul_strtok__get_separator(cgul_exception_t* cex,
                           cgul_strtok_t tok);

/**
 * Get the starting index for the current token.
 *
 * @param[in] cex c-style exception
 * @param[in] tok <code>cgul_strtok</code> instance
 * @return starting index for the current token 
 */
extern CGUL_EXPORT size_t
cgul_strtok__get_start(cgul_exception_t* cex,
                       cgul_strtok_t tok);

/**
 * Get the ending index for the current token.  The value returned is
 * actually one beyond the end of the current token.  So in effect,
 * the value returned is the index of the trailing separator for the
 * current token.
 *
 * @param[in] cex c-style exception
 * @param[in] tok <code>cgul_strtok</code> instance
 * @return ending index for the current token 
 */
extern CGUL_EXPORT size_t
cgul_strtok__get_end(cgul_exception_t* cex,
                     cgul_strtok_t tok);

CGUL_END_C

#endif  /*  CGUL_STRTOK_H  */
