#ifndef CGUL_SUBSTRING_CXX_H
#define CGUL_SUBSTRING_CXX_H

/**
 * C++ wrapper for <code>cgul_substring</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_substring</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_substring.h"

/**
 * This class provides the C++ bindings for <code>cgul_substring</code>.
 * The main purpose of this class is to convert the C-style function
 * calls and exception handling in <code>cgul_substring</code> into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_substring</code>
 */
class cgul_substring_cxx {

public:

    /*************************************************************************
     * EZ API
     *************************************************************************/

    /**
     * This function is the easy way to search because it does not
     * require the user to allocate a <code>cgul_substring</code>
     * instance.  Instead, just pass in the c-style substring
     * <code>ss</code> along with the c-style string <code>s</code> to
     * search.  If the substring is found, a pointer to the first
     * matching character of the substring in <code>s</code> is
     * returned; otherwise, <code>NULL</code> is returned. <p>
     *
     * If you are only going to search <code>s</code> once, in
     * addition to being easier to use, this function will be faster
     * than <code>search()</code> because this function does not copy
     * the substring. <p>
     *
     * @param[in] ss c-style substring
     * @param[in] s c-style string to search
     * @return pointer to first match or <code>NULL</code> if no match
     */
    static const char*
    search_ez(const char* ss,
              const char* s)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_substring__search_ez(&local, ss, strlen(ss), s, strlen(s));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is the easy way to search because it does not
     * require the user to allocate a <code>cgul_substring</code>
     * instance.  Instead, just pass in the binary substring
     * <code>ss</code> and its length <code>ss_length</code> in bytes
     * along with the binary block <code>block</code> to search and
     * its length <code>block_length</code> in bytes.  If the
     * substring is found, a pointer to the first matching character
     * of the substring in <code>block</code> is returned; otherwise,
     * <code>NULL</code> is returned. <p>
     *
     * If you are only going to search <code>block</code> once, in
     * addition to being easier to use, this function will be faster
     * than <code>search()</code> because this function does not copy
     * the substring. <p>
     *
     * @param[in] ss binary substring
     * @param[in] ss_length substring length in bytes
     * @param[in] block block of binary data to search
     * @param[in] block_length block length in bytes
     * @return pointer to first match or <code>NULL</code> if no match
     */
    static const char*
    search_ez(const char* ss,
              size_t ss_length,
              const char* block,
              size_t block_length)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_substring__search_ez(
                 &local, ss, ss_length, block, block_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /*************************************************************************
     * Multiple Search API
     *************************************************************************/

    /**
     * Default constructor that creates a new instance with an
     * undefined substring.  If an error occurs, an exception is
     * thrown.
     */
    cgul_substring_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_substring__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new instance initialized from the C-style substring
     * <code>substring</code>.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] substring c-style substring
     */
    cgul_substring_cxx(const char* substring)
    {
        cgul_exception_t local = NULL;
        obj = cgul_substring__new_from_cstring(&local, substring);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Construct a new instance initialized from the binary substring
     * <code>substring</code> that extends for <code>length</code>
     * bytes.  If an error occurs, an exception is thrown.
     *
     * @param[in] substring binary substring
     * @param[in] length length of <code>substring</code>
     */
    cgul_substring_cxx(const char* substring,
                       size_t length)
    {
        cgul_exception_t local = NULL;
        obj = cgul_substring__new_from_bstring(&local, substring, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy constructor that creates a new instance by copying the
     * value from <code>rhs</code>.  If an error occurs, an exception
     * is thrown.
     *
     * @param[in] rhs right-hand side
     */
    cgul_substring_cxx(const cgul_substring_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_substring__new_from_bstring(&local,
                                               rhs.get_value(),
                                               rhs.get_length());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Destructor.
     */
    virtual
    ~cgul_substring_cxx()
    {
        if (obj) {
            cgul_substring__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Assign the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_substring_cxx&
    operator=(const cgul_substring_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        if (this != &rhs) {
            cgul_substring__set_value(&local,
                                      obj,
                                      rhs.get_value(),
                                      rhs.get_length());
            if (local) {
                throw cgul_exception_cxx(local);
            }
        }
        return *this;
    }

    /**
     * Assign the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_substring_cxx&
    operator=(const char* rhs)
    {
        cgul_exception_t local = NULL;
        cgul_substring__set_value(&local, obj, rhs, strlen(rhs));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Get the value used as the substring in subsequent searches.
     * The value can be set by calling <code>set_value()</code>.  If
     * the value is not set, <code>NULL</code> is returned.
     *
     * @return value used as the substring in subsequent searches
     */
    virtual const char*
    get_value() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_substring__get_value(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the value to use as the substring in subsequent searches
     * from the C-style substring <code>substring</code>.  This
     * function must be called before calling <code>search()</code>.
     * It caches the pre-computations based on the substring so that
     * subsequent searches for the same substring will not have to
     * recompute the same values.  If an error occurs, an exception is
     * thrown. <p>
     *
     * In addition to caching the pre-computed values, this function
     * also caches a copy of <code>substring</code> for use with
     * <code>search()</code>.  Because <code>search_ez()</code> does
     * not copy the substring, it is not always clear whether
     * <code>search()</code> or <code>search_ez()</code> will run
     * faster. <p>
     *
     * @param[in] substring c-style substring
     */
    virtual void
    set_value(const char* substring)
    {
        cgul_exception_t local = NULL;
        cgul_substring__set_value(&local, obj, substring, strlen(substring));
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Set the value to use as the substring in subsequent searches.
     * This function must be called before calling
     * <code>search()</code>.  It caches the pre-computations based on
     * the substring so that subsequent searches for the same
     * substring will not have to recompute the same values.  If an
     * error occurs, an exception is thrown. <p>
     *
     * In addition to caching the pre-computed values, this function
     * also caches a copy of <code>substring</code> for use with
     * <code>search()</code>.  Because <code>search_ez()</code> does
     * not copy the substring, it is not always clear whether
     * <code>search()</code> or <code>search_ez()</code> will run
     * faster. <p>
     *
     * @param[in] substring binary substring
     * @param[in] length length of <code>substring</code>
     */
    virtual void
    set_value(const char* substring,
              size_t length)
    {
        cgul_exception_t local = NULL;
        cgul_substring__set_value(&local, obj, substring, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the length of the substring that will be used in subsequent
     * searches.  The substring can be set by calling
     * <code>set_value()</code>.  If the substring is not set,
     * <code>0</code> is returned.
     *
     * @return length of the substring that will be used in subsequent searches
     */
    virtual size_t
    get_length() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_substring__get_length(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Search the c-style string <code>s</code> for the substring set
     * by the most recent call to <code>set_value()</code>.  If the
     * substring is found, a pointer to the first matching character
     * in <code>block</code> is returned; otherwise, <code>NULL</code>
     * is returned.
     *
     * @param[in] s c-style string
     * @return pointer to first match or <code>NULL</code> if no match
     */
    virtual const char*
    search(const char* s)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_substring__search(&local, obj, s, strlen(s));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Search the first <code>length</code> bytes of block
     * <code>block</code> for the substring set by the most recent
     * call to <code>set_value()</code>.  If the substring is found, a
     * pointer to the first matching character in <code>block</code>
     * is returned; otherwise, <code>NULL</code> is returned.
     *
     * @param[in] block block of binary data to search
     * @param[in] length block length in bytes
     * @return pointer to first match or <code>NULL</code> if no match
     */
    virtual const char*
    search(const char* block,
           size_t length)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_substring__search(&local, obj, block, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

private:

    /** Underlying <code>cgul_substring</code> object. */
    cgul_substring_t obj;

};

#endif  /*  CGUL_SUBSTRING_CXX_H  */
