#ifndef CGUL_TRIE_CXX_H
#define CGUL_TRIE_CXX_H

/**
 * C++ wrapper for <code>cgul_trie</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_trie</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_trie.h"
#include "cgul_trie_node_cxx.h"

/**
 * This class provides the C++ bindings for C <code>cgul_trie</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_trie</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_trie</code>
 * @see cgul_trie_t
 */
class cgul_trie_cxx {

public:

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_keys()</code>
     *     <code>foldr_keys()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_key_t)(const char* key,
                              void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_values()</code>
     *     <code>foldr_values()</code>
     * </pre>
     *
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_value_t)(void* value,
                                void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_pairs()</code>
     *     <code>foldr_pairs()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_pair_t)(const char* key,
                               void* value,
                               void* data);

    /**
     * This typedef is the interface for the callback function used by
     * the following methods: <p>
     *
     * <pre>
     *     <code>traverse()</code>
     *     <code>traverse_range()</code>
     * </pre>
     *
     * @param[in] trie trie
     * @param[in] node node
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*traverse_t)(cgul_trie_cxx* trie,
                              cgul_trie_node_cxx* node,
                              void* data);

    /**
     * Create a new <code>cgul_trie_cxx</code> object.  The caller is
     * responsible for freeing the object by arranging for
     * <code>delete</code> to be called.  If memory cannot be
     * allocated, an exception is thrown.
     *
     * Be default <code>with_indexing</code> is false which means the
     * ability to index into the trie as though it were an array will
     * not be enabled.  If <code>with_indexing</code> is set to true,
     * each node will be larger containing three extra <code>unsigned
     * long</code> values and will run about 15% slower. <p>
     *
     * This class does not take ownership of the inserted key/value pairs.
     * Instead, the client is responsible for freeing the key/value pairs
     * only after their nodes have been permanently removed from the
     * <code>cgul_trie_cxx</code>.  The <code>traverse()</code>
     * method can be used to remove each node safely before calling
     * <code>delete</code> on the object created. <p>
     *
     * @param[in] with_indexing whether to enable indexing
     */
    cgul_trie_cxx(int with_indexing = 0)
    {
        cgul_exception_t local = NULL;
        if (with_indexing) {
            obj = cgul_trie__new_with_indexing(&local);
        } else {
            obj = cgul_trie__new(&local);
        }
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_trie_cxx</code> object by wrapping
     * an existing <code>cgul_trie</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_trie_cxx(cgul_trie_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  This does
     * not include the key/value pairs stored in the trie.  The client
     * is responsible for freeing the key/value pairs when the client
     * thinks it is convenient and safe to do so; however, the client
     * must understand that freeing a key before removing its node
     * invalidates the data structure. <p>
     *
     * As a convenience, you may want to call <code>clear()</code>
     * before calling this method in order to properly put the keys
     * and values back on their respective <code>cgul_cache_cxx</code>
     * objects (if <code>cgul_cache_cxx</code> objects are being used)
     * before you delete the trie.  If <code>cgul_cache_cxx</code>
     * objects are not being used, the client needs to arrange some
     * other mechanism to free the keys or values. <p>
     *
     * @see free_keys()
     * @see free_values()
     */
    virtual
    ~cgul_trie_cxx()
    {
        if (obj) {
            cgul_trie__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the keys in the trie
     * trie.  Because this is such a common operation, it is
     * an exception to the rule that cgul containers never free keys.
     * This method should only ever be called immediately before calling
     * <code>delete</code> because it otherwise invalidates the trie.
     */
    virtual void
    free_keys()
    {
        cgul_trie__free_keys(obj);
    }

    /**
     * This method calls <code>free()</code> on all the values in the trie.
     * Because this is such a common operation, it is
     * an exception to the rule that cgul containers never free values.
     * This method should only ever be called immediately before calling
     * <code>delete</code> because it otherwise invalidates the trie.
     */
    virtual void
    free_values()
    {
        cgul_trie__free_values(obj);
    }

    /**
     * Get the size of the cache of nodes. <p>
     *
     * @return size of the cache of nodes
     */
    virtual unsigned long int
    get_cache_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__get_cache_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the size of the cache of nodes. <p>
     *
     * For efficiency, the <code>cgul_trie_cxx</code> object can keep
     * a cache of nodes that can be reused.  In many situations, this
     * can greatly reduce the number of calls to <code>malloc()</code>
     * which can greatly improve performance and reduce memory
     * fragmentation. <p>
     *
     * You can release all the cached nodes and disable caching of nodes
     * by setting the cache reserve to 0 and the cache size to 0 (in that
     * order).  This is the default. <p>
     *
     * If an error occurs (while enlarging the cache), an exception is
     * thrown. <p>
     *
     * @param[in] size of the cache of nodes
     */
    virtual void
    set_cache_size(unsigned long int size)
    {
        cgul_exception_t local = NULL;
        cgul_trie__set_cache_size(&local, obj, size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the reserve limit of the cache of nodes. <p>
     *
     * @return reserve limit of the cache of nodes
     */
    virtual unsigned long int
    get_cache_reserve() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__get_cache_reserve(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the reserve limit of the cache of nodes and guarantee that at
     * least that many nodes are allocated and waiting in reserve. <p>
     *
     * When you use this method to increase the reserve, you are
     * guaranteed that at least <code>reserve</code> count of nodes will
     * be allocated and reserved for future use. <p>
     *
     * This is useful because <code>insert()</code> can
     * throw an exception if it cannot allocate a node to hold the
     * key/value pair.  Thus, by allocating and reserving a node now, we
     * know that <code>insert()</code> will be able to run
     * without error later when the node is "unreserved".  Normally, this
     * is not an issue because you can almost always just insert at the
     * moment when it is requested, but if the <code>cgul_trie_cxx</code>
     * is busy at that moment (perhaps it is iterating over its nodes),
     * you'll have to queue insert (and remove) requests for later
     * execution. <p>
     *
     * To "unreserve" a node so that it can be used, you simply call this
     * method again with a value for <code>reserve</code> that is less
     * than the current reserve limit. <p>
     *
     * If an error occurs while allocating the reserved nodes, an
     * exception is thrown. <p>
     *
     * WARNING: If you reserve nodes without "unreserving" them, you will
     * introduce a memory leak that is difficult to track down because
     * this class is careful to free all of the cached nodes (including
     * the reserved nodes) when it is deleted, but the reserved nodes will
     * just sit around in memory until that time.  Thus, after reserving
     * nodes by calling this method, you must remember to timely set the
     * reserve level back to zero in order to make those nodes
     * available. <p>
     *
     * @param[in] reserve of the cache of nodes
     * @see cgul_cache_cxx::set_reserve()
     */
    virtual void
    set_cache_reserve(unsigned long int reserve)
    {
        cgul_exception_t local = NULL;
        cgul_trie__set_cache_reserve(&local, obj, reserve);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return 1 if the trie is empty; otherwise, return 0.
     *
     * @return whether the trie is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Balance the trie.  If an error occurs, an exception is
     * thrown. <p>
     *
     * Because the key/value pairs are reinserted randomly into the trie,
     * the resulting older/younger linked list (which tracks insertion
     * order) will generally not match the original insertion order.
     * However, if <code>maintain_relative_ages</code> is true, this
     * method will take the extra, relatively expensive step of adjusting
     * the older/younger list to match the original order. <p>
     *
     * Ternary Search Tries have worst-case search behavior of O(n).  This
     * problem occurs when the keys are inserted in sort order.  This
     * method can be used to reinsert all the keys at random which will
     * likely come very close to minimizing the depth of the trie which
     * will result in search behavior of O(log n).  This is an expensive
     * operation which should be called rarely over the lifetime of the
     * trie. <p>
     *
     * One plausible scenario where this method could be called is if all
     * the keys are read at once from a sorted list.  After inserting all
     * the keys, this method could be called to balance the trie.
     * (Another solution to this problem would be to insert all the keys
     * through <code>insert_from_arrays()</code>.) <p>
     *
     * @param[in] maintain_relative_ages whether to maintain relative node ages
     * @see insert_from_arrays()
     */
    virtual void
    balance(int maintain_relative_ages)
    {
        cgul_exception_t local = NULL;
        cgul_trie__balance(&local, obj, maintain_relative_ages);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Insert <code>key</code> into the trie <code>trie</code>.  If
     * <code>key</code> already exists, another node with the same key is
     * not created.  The new or pre-existing node is returned in
     * <code>n</code> if <code>n</code> is not <code>NULL</code>. <p>
     *
     * If a matching node is not found, <code>1</code> is returned because
     * a new node can and will be inserted.  If a matching node is found,
     * <code>0</code> is returned because the insertion failed because the
     * key already exists.<p>
     *
     * If an error occurs, 0 is returned, <code>*n</code> is set to
     * <code>NULL</code> (if <code>n</code> is not <code>NULL</code>), and
     * an exception is thrown. <p>
     *
     * It is important to understand that the <code>cgul_trie</code>
     * class (like all containers in the cgul library) does not take
     * ownership of <code>key</code>.  It also does not attempt to make a
     * copy of the thing pointed to by <code>key</code>.  This gives the
     * user complete control over the lifetime of the key and gives a real
     * performance boost in many cases; however, this means that
     * <code>key</code> must not be invalidated while it is still being
     * used by this class. <p>
     *
     * This method operates as it does so that you only have to search the
     * trie once when doing an insert.  Logically, you would usually like
     * to do a find to see if the trie holds your key.  If it does, you
     * use the value associated with that key in some way.  If it does
     * not, you then want to insert a new key/value pair into the trie.
     * The problem with divorcing the find from the insert is that it then
     * requires you to search the trie twice.  Once to see if the key
     * exists and once to do the insertion. <p>
     *
     * To prevent having to search the trie twice, you can just call this
     * method to handle both inserting and updating nodes.  It will search
     * the trie once, and it will always return a pointer to the correct
     * node.  You can then use <code>cgul_trie_node_cxx::get_value()</code>
     * or <code>cgul_trie_node_cxx::set_value()</code> to get or set the
     * value indexed by "key". <p>
     *
     * TIP: You might find it convenient to use <code>cgul_string_cxx</code>
     * to build up a string that you use as a key.  You can then use
     * <code>cgul_string_cxx::get_value()</code> [yes, get_value()] and pass
     * the string to <code>insert()</code>.  If the insert
     * succeeds, you can inform the <code>cgul_string_cxx</code> object that
     * it is no longer the owner of the underlying C-style string by
     * calling <code>cgul_string_cxx::take_value()</code> [yes, take_value()].
     * An example is as follows: <p>
     *
     * <pre>
     *
     *     cgul_string_cxx word;
     *     cgul_trie_node_cxx* n = NULL;
     *
     *     if (trie->insert(word.get_value(), &n)) {
     *         word.take_value();
     *         . . .
     *     } else {
     *         . . .
     *     }
     *
     * </pre>
     *
     * NOTE: Inserting a key/value pair does not cause this class to take
     * ownership of <code>key</code> or <code>value</code>.  The client
     * must make sure the values pointed to by <code>key</code> and
     * <code>value</code> are not invalidated during the life of the newly
     * created node.  If the client needs to delete the <code>key</code>
     * or the <code>value</code> or both, the client needs to manually
     * arrange for this. <p>
     *
     * @param[in] key key
     * @param[out] n node that already existed or was newly created
     * @return 1 if a new node was inserted; 0 if a node with the same key
     * already exists
     * @see balance()
     * @see insert_from_arrays()
     */
    virtual int
    insert(const char* key,
           cgul_trie_node_cxx** n)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__insert(&local, obj, key, (cgul_trie_node_t*)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert into the trie the key/value pairs formed
     * by interleaving keys from the <code>keys</code> array with values
     * from the <code>values</code> array.  If <code>values</code> is
     * <code>NULL</code>, all values will be set to <code>NULL</code>.
     * The number of new nodes insert is returned.  If an error occurs, an
     * exception is thrown. <p>
     *
     * This method copies the <code>keys</code> and <code>values</code>
     * arrays and then performs the same shuffle on each copy before
     * inserting the shuffled key/value pairs.  It shuffles the key/value
     * pairs because inserting sorted keys into a Ternary Search Trie
     * leads to worst-case search behavior of O(N).  By shuffling the keys
     * first, the trie will very likely be balanced leading to O(log N)
     * search behavior. <p>
     *
     * Because the key/value pairs are inserted randomly into the trie,
     * the resulting older/younger linked list (which tracks insertion
     * order) will generally not match the order of the key/value pairs
     * from the <code>keys</code> and <code>values</code> arrays.
     * However, if <code>maintain_relative_ages</code> is true, this
     * method will take the extra, relatively expensive step of adjusting
     * the older/younger list to match the original order of the key/value
     * pairs. <p>
     *
     * @param[in] keys keys
     * @param[in] values values
     * @param[in] kv_size number of key/value pairs (i.e., size of the arrays)
     * @param[in] maintain_relative_ages whether to maintain relative node ages
     * @return number of new nodes inserted
     * @see balance()
     */
    virtual unsigned long int
    insert_from_arrays(const char** keys,
                       void** values,
                       unsigned long int kv_size,
                       int maintain_relative_ages)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__insert_from_arrays(&local,
                                           obj,
                                           keys,
                                           values,
                                           kv_size,
                                           maintain_relative_ages);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node indexed by <code>key</code>.  If the key does not
     * exist, <code>NULL</code> is returned; otherwise, the node
     * associated with the key is returned.
     *
     * @param[in] key key
     * @return node associated with the key or <code>NULL</code>
     */
    virtual cgul_trie_node_cxx*
    find(const char* key)
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__find(&local, obj, key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node indexed by position.  This lets you use a
     * <code>cgul_trie_cxx</code> instances like you would an array.  Even in
     * the best-case scenario where indexing into <code>cgul_trie_cxx</code>
     * is a logarithmic function, it will still be slower than indexing
     * into an array, but you get all the advantages of a trie including
     * relatively fast insertion and deletion.  So, if you ever feel the
     * need to shift the elements of an array, this method might be for
     * you. <p>
     *
     * If <code>index</code> is out of range, <code>NULL</code> is
     * returned; otherwise, the node returned is the same node that would
     * be returned if you had an array <code>trie</code> that held all the
     * nodes in the trie in sorted order, and you asked for
     * <code>trie[index]</code>. <p>
     *
     * This method throws an exception if the trie was not created
     * with indexing enabled. <p>
     *
     * @param[in] index index into sorted keys of node to return
     * @return node associated with <code>index</code>
     * @see find_rank()
     */
    virtual cgul_trie_node_cxx*
    find_at(unsigned long int index)
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__find_at(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether the rank of the key <code>key</code> could be
     * determined.  If the rank could be determined it is returned in
     * <code>*rank</code> (if <code>rank</code> is not
     * <code>NULL</code>). <p>
     *
     * The rank is the number of keys less than <code>key</code>.  It can
     * be used as the index parameter to <code>find_at()</code>
     * in order to lookup <code>key</code> by index. <p>
     *
     * This method throws an exception if <code>trie</code> was not
     * created with indexing enabled. <p>
     *
     * @param[in] key key
     * @param[out] rank rank of key
     * @return whether the key exists in the trie
     * @see find_at()
     */
    virtual int
    find_rank(const char* key,
              unsigned long int* rank)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_rank(&local, obj, key, rank);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node having the largest key less than or equal to the
     * search key <code>search_key</code> and return it.  If such a key
     * does not exist, <code>NULL</code> is returned.
     *
     * @param[in] search_key search key
     * @return floor node or <code>NULL</code>
     */
    virtual cgul_trie_node_cxx*
    find_floor(const char* search_key)
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__find_floor(&local,
                                                        obj,
                                                        search_key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node having the smallest key greater than or equal to the
     * search key <code>search_key</code> and return it.  If such a key
     * does not exist, <code>NULL</code> is returned.
     *
     * @param[in] search_key search key
     * @return ceiling node or <code>NULL</code>
     */
    virtual cgul_trie_node_cxx*
    find_ceiling(const char* search_key)
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__find_ceiling(&local,
                                                          obj,
                                                          search_key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * When dealing with prefixes that generate a large number of matches,
     * it may be more responsive from the user's perspective to iterate
     * over the results.  To that end, this method returns the closed
     * range <code>[begin_node, end_node]</code> for the nodes in the trie
     * associated with keys that start with the prefix
     * <code>prefix</code>.  If any nodes exist in the range, this method
     * returns <code>1</code>; otherwise, this methods returns
     * <code>0</code>.  If an error occurs, an exception is thrown. <p>
     *
     * @param[in] prefix prefix
     * @param[out] begin_node beginning node (inclusive) or <code>NULL</code>
     * @param[out] end_node ending node (inclusive) or <code>NULL</code>
     * @return whether any nodes were found
     * @see find_half_opened_range_by_prefix()
     */
    virtual int
    find_closed_range_by_prefix(const char* prefix,
                                cgul_trie_node_cxx** begin_node,
                                cgul_trie_node_cxx** end_node)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_closed_range_by_prefix(
                 &local,
                 obj,
                 prefix,
                 (cgul_trie_node_t*)begin_node,
                 (cgul_trie_node_t*)end_node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * When dealing with prefixes that generate a large number of matches,
     * it may be more responsive from the user's perspective to iterate
     * over the results.  To that end, this method returns the half-opened
     * range <code>[begin_node, end_node]</code> for the nodes in the trie
     * associated with keys that start with the prefix
     * <code>prefix</code>.  If any nodes exist in the range, this method
     * returns <code>1</code>; otherwise, this methods returns
     * <code>0</code>.  If an error occurs, an exception is thrown. <p>
     *
     * @param[in] prefix prefix
     * @param[out] begin_node beginning node (inclusive) or <code>NULL</code>
     * @param[out] end_node ending node (inclusive) or <code>NULL</code>
     * @return whether any nodes were found
     * @see find_closed_range_by_prefix()
     */
    virtual int
    find_half_opened_range_by_prefix(const char* prefix,
                                     cgul_trie_node_cxx** begin_node,
                                     cgul_trie_node_cxx** end_node)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_half_opened_range_by_prefix(
                 &local,
                 obj,
                 prefix,
                 (cgul_trie_node_t*)begin_node,
                 (cgul_trie_node_t*)end_node);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * All keys in the trie that start with the prefix
     * <code>prefix</code> are returned in the arrays <code>keys</code>
     * (if not <code>NULL</code>) and the number of keys is returned in
     * <code>keys_size</code> (if not <code>NULL</code>).  If any keys are
     * found, this method returns <code>1</code>; otherwise, this methods
     * returns <code>0</code>.  If an error occurs, an exception is
     * thrown. <p>
     *
     * The client is responsible for calling <code>free()</code> on the
     * <code>keys</code> array but must not modify or delete any of the
     * individual keys. <p>
     *
     * The following example code shows how to print all the keys in the
     * trie that match a particular prefix and how to clean up afterward: <p>
     *
     * <pre>
     *
     *      void
     *      print_keys_by_prefix(cgul_trie_cxx& trie,
     *                           const char* prefix)
     *      {
     *          unsigned long int i = 0;
     *          const char** keys = NULL;
     *          unsigned long int keys_size = 0;
     *      
     *          // Iterate over all the keys that match the prefix.
     *          if (trie.find_keys_by_prefix(prefix, &keys, &keys_size)) {
     *              for (i = 0 ; i < keys_size ; ++i) {
     *                  printf("%s -> %s\n", prefix, keys[i]);
     *              }
     *          }
     *      
     *          // Clean up.
     *          if (keys) {
     *              free(keys);
     *          }
     *      
     *      }
     *
     * </pre>
     *
     * @param[in] prefix prefix
     * @param[out] keys keys found by prefix
     * @param[out] keys_size number of keys found
     * @return whether any keys were found
     * @see find_nodes_by_prefix()
     * @see find_closed_range_by_prefix()
     * @see find_half_opened_range_by_prefix()
     */
    virtual int
    find_keys_by_prefix(const char* prefix,
                        const char*** keys,
                        unsigned long int* keys_size)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_keys_by_prefix(&local,
                                            obj,
                                            prefix,
                                            keys,
                                            keys_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * All values in the trie associated with keys that
     * start with the prefix <code>prefix</code> are returned in the
     * arrays <code>values</code> (if not <code>NULL</code>) and the
     * number of values is returned in <code>values_size</code> (if not
     * <code>NULL</code>).  If any values are found, this method returns
     * <code>1</code>; otherwise, this methods returns <code>0</code>.  If
     * an error occurs, an exception is thrown. <p>
     *
     * The client is responsible for calling <code>free()</code> on the
     * <code>values</code> array.  The client is free to modify but not
     * delete any of the individual values unless the value is also used
     * as a key for the trie.  If a value needs to be delete the client
     * should do so only after using
     * <code>cgul_trie_node_cxx::set_value()</code> to clear the value. <p>
     *
     * The following example code shows how to print all the values in the
     * trie associated with keys that match a particular prefix and how to
     * clean up afterward: <p>
     *
     * <pre>
     *
     *      void
     *      print_values_by_prefix(cgul_trie_cxx& trie,
     *                             const char* prefix)
     *      {
     *          unsigned long int i = 0;
     *          void** values = NULL;
     *          unsigned long int values_size = 0;
     *      
     *          // Iterate over all the values that match the prefix.
     *          if (trie.find_values_by_prefix(prefix, &values, &values_size)) {
     *              for (i = 0 ; i < values_size ; ++i) {
     *                  printf("%s -> %p\n", prefix, values[i]);
     *              }
     *          }
     *      
     *          // Clean up.
     *          if (values) {
     *              free(values);
     *          }
     *      
     *      }
     *
     * </pre>
     *
     * @param[in] prefix prefix
     * @param[out] values values found by prefix
     * @param[out] values_size number of values found
     * @return whether any values were found
     * @see find_nodes_by_prefix()
     * @see find_closed_range_by_prefix()
     * @see find_half_opened_range_by_prefix()
     */
    virtual int
    find_values_by_prefix(const char* prefix,
                          void*** values,
                          unsigned long int* values_size)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_values_by_prefix(&local,
                                              obj,
                                              prefix,
                                              values,
                                              values_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * All nodes in the trie associated with keys that
     * start with the prefix <code>prefix</code> are returned in the
     * arrays <code>nodes</code> (if not <code>NULL</code>) and the number
     * of nodes is returned in <code>nodes_size</code> (if not
     * <code>NULL</code>).  If any nodes are found, this method returns
     * <code>1</code>; otherwise, this methods returns <code>0</code>.  If
     * an error occurs, an exception is thrown. <p>
     *
     * The client is responsible for calling <code>free()</code> on the
     * <code>nodes</code> array.  The client must not delete any of the
     * nodes nor delete or modify any of the keys stored in the nodes, but
     * the client is free to modify but not delete any of the individual
     * values stored in the nodes unless the value is also used as a key
     * for the trie.  If a value needs to be delete the client should do
     * so only after using <code>cgul_trie_node_cxx::set_value()</code> to
     * clear the value. <p>
     *
     * The following example code shows how to print all the key/value
     * pairs in the trie where the keys match a particular prefix and how
     * to clean up afterward: <p>
     *
     * <pre>
     *
     *      void
     *      print_nodes_by_prefix(cgul_trie_cxx& trie,
     *                            const char* prefix)
     *      {
     *          unsigned long int i = 0;
     *          const char* key = NULL;
     *          void* value = NULL;
     *          cgul_trie_node_cxx** nodes = NULL;
     *          unsigned long int nodes_size = 0;
     *      
     *          // Iterate over all the nodes that match the prefix.
     *          if (trie.find_nodes_by_prefix(prefix, &nodes, &nodes_size)) {
     *              for (i = 0 ; i < nodes_size ; ++i) {
     *                  key = nodes[i]->get_key();
     *                  value = nodes[i]->get_value();
     *                  printf("%s -> (%s, %p)\n", prefix, key, value);
     *              }
     *          }
     *      
     *          // Clean up.
     *          if (nodes) {
     *              free(nodes);
     *          }
     *      
     *      }
     *
     * </pre>
     *
     * @param[in] prefix prefix
     * @param[out] nodes nodes found by prefix
     * @param[out] nodes_size number of nodes found
     * @return whether any nodes were found
     * @see find_closed_range_by_prefix()
     * @see find_half_opened_range_by_prefix()
     */
    virtual int
    find_nodes_by_prefix(const char* prefix,
                         cgul_trie_node_cxx*** nodes,
                         unsigned long int* nodes_size)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_nodes_by_prefix(&local,
                                             obj,
                                             prefix,
                                             (cgul_trie_node_t**)nodes,
                                             nodes_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the key in the trie that is the longest
     * prefix of <code>complete_string</code>.  To match, the entire key
     * must be a prefix of <code>complete_string</code>.  If no keys are a
     * prefix, <code>NULL</code> is returned. <p>
     *
     * For example, assume "abcx" and "abcxyz123" are the only keys in the
     * trie and this method is called with <code>complete_string</code>
     * equal to "abcxyz", the longest key that is a prefix of the complete
     * string is "abcx".  The result is not "abcxyz123" because the entire
     * key must be a prefix of <code>complete_string</code>. <p>
     *
     * One practical example for the use of this function is if the trie
     * holds IP routing information for subnetworks.  To route a packet to
     * a particular IP address, you will likely want to use the route that
     * is the longest prefix of the IP address.  (See Wikipedia's "Longest
     * Prefix Match" entry.) <p>
     *
     * @param[in] complete_string complete string
     * @return key that is the longest prefix of the complete string
     */
    virtual const char*
    find_longest_prefix_of(const char* complete_string)
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_trie__find_longest_prefix_of(&local, obj, complete_string);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the node holding the first key according to sort order.
     * This operation is *not* O(1) because it has to descend the trie
     * looking for the front node.  If the trie is empty,
     * <code>NULL</code> is returned. <p>
     *
     * The following example shows how to iterate over the entire trie in
     * sort order: <p>
     *
     * <pre>
     *     cgul_trie_node_cxx* n = trie->get_front();
     *     for ( ; n ; n = n->get_next()) {
     *         ...
     *     }
     * </pre>
     *
     * @return front node
     */
    virtual cgul_trie_node_cxx*
    get_front() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__get_front(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the node holding the last key according to sort order.  This
     * operation is *not* O(1) because it has to descend the trie looking
     * for the back node.  If the trie is empty, <code>NULL</code> is
     * returned. <p>
     *
     * The following example shows how to iterate over the entire trie in
     * reverse sort order: <p>
     *
     * <pre>
     *     cgul_trie_node_cxx* n = trie->get_back();
     *     for ( ; n ; n = n->get_prev()) {
     *         ...
     *     }
     * </pre>
     *
     * @return back node
     */
    virtual cgul_trie_node_cxx*
    get_back() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__get_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the oldest node according to chronological order (i.e., the
     * order in which the nodes are inserted).  This operation is O(1).
     * If the trie is empty, <code>NULL</code> is returned.  This class
     * does not have to search the trie in order to find the oldest node
     * because it efficiently keeps a direct pointer to the oldest node up
     * to date. <p>
     *
     * The following example shows how to iterate over the entire trie in
     * chronological order:
     *
     * <pre>
     *     cgul_trie_node_cxx* n = trie->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return oldest node
     * @see cgul_trie_node_cxx::get_younger()
     */
    virtual cgul_trie_node_cxx*
    get_oldest() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__get_oldest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the oldest node in the trie to be <code>n</code>.  Calling this
     * method has the potential to confuse iterators and should be handled
     * with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * oldest node in the trie, and you want to force <code>n</code> to be
     * the next node to expire. <p>
     *
     * @param[in] n node
     * @see set_youngest()
     */
    virtual void
    set_oldest(cgul_trie_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_trie__set_oldest(&local, obj, (cgul_trie_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the youngest node according to chronological order (i.e.,
     * the order in which the nodes are inserted).  This operation is
     * O(1).  If the trie is empty, <code>NULL</code> is returned.  This
     * class does not have to search the trie in order to find the
     * youngest node because it efficiently keeps a direct pointer to the
     * youngest node up to date. <p>
     *
     * The following example shows how to iterate over the entire trie in
     * reverse chronological order:
     *
     * <pre>
     *     cgul_trie_node_cxx* n = trie->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return youngest node
     * @see cgul_trie_node_cxx::get_older()
     */
    virtual cgul_trie_node_cxx*
    get_youngest() const
    {
        cgul_trie_node_cxx* rv = NULL;
        cgul_exception_t local = NULL;
        rv = (cgul_trie_node_cxx*)cgul_trie__get_youngest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the youngest node in the trie to be <code>n</code>.  Calling
     * this method has the potential to confuse iterators and should be
     * handled with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * least-recently used (LRU) node.  By calling
     * <code>set_youngest()</code> each time a node is used,
     * the least-recently used node will be the oldest node in the
     * trie. <p>
     *
     * @param[in] n node
     * @see set_oldest()
     */
    virtual void
    set_youngest(cgul_trie_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_trie__set_youngest(&local, obj, (cgul_trie_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove the node in the trie associated with
     * <code>key_in</code>.  The key and value pointers stored in the node
     * that is to be removed will be returned in <code>key_out</code> and
     * <code>value_out</code> if you pass in pointers that are not
     * <code>NULL</code>.  This method returns <code>1</code> if the node
     * was removed; it returns <code>0</code> otherwise. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the trie.  See <code>remove_node()</code>
     * for details. <p>
     *
     * @param[in] key_in key associated with the node to be removed
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove(const char* key_in,
           char** key_out,
           void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__remove(&local, obj, key_in, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove <code>node</code> from the trie.  The key
     * and value pointers stored in the node that is to be removed will be
     * returned in <code>key_out</code> and <code>value_out</code> if you
     * pass in pointers that are not <code>NULL</code>. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the trie.  The problem is that calling this method
     * invalidates the node making it impossible to call
     * <code>cgul_trie_node_cxx::get_next()</code> afterward.  The solution is
     * simple.  Just call <code>cgul_trie_node_cxx::get_next()</code> before
     * calling this method: <p>
     *
     * <pre>
     *
     *     for (curr = trie->find(key) ; curr ; curr = next) {
     *         next = curr->get_next();
     *         trie->remove_node(curr, &key_out, &value_out);
     *         free(key_out);
     *         free(value_out);
     *     }
     *
     * </pre>
     *
     * Alternatively, you can use <code>remove_range()</code>,
     * <code>traverse()</code>, or <code>traverse_range()</code>. <p>
     *
     * @param[in] n node to be removed
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     */
    virtual void
    remove_node(cgul_trie_node_cxx* n,
                char** key_out,
                void** value_out)
    {
        cgul_exception_t local = NULL;
        cgul_trie__remove_node(&local,
                               obj,
                               (cgul_trie_node_t)n,
                               key_out,
                               value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method lets you remove a node based on its sorted order
     * <code>index</code>.  If the index is out of bounds, 0 is returned;
     * otherwise 1 is returned.  If you are interested in getting your
     * hands on the key and value pointers stored in the node that is to
     * be removed, they will be returned in <code>key_out</code> and
     * <code>value_out</code> if you pass in pointers that are not
     * <code>NULL</code>.  <code>index</code> is zero-based. <p>
     *
     * This method throws an exception if the trie was not
     * created with indexing enabled. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the trie.  See <code>remove_node()</code>
     * for details. <p>
     *
     * @param[in] index index of the sorted key associated with the node
     * to be removed
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_at(unsigned long int index,
              char** key_out,
              void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__remove_at(&local, obj, index, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove the first key/value pair as determined by sort order from
     * the trie.  The key and value pointers stored in
     * the node that is to be removed will be returned in
     * <code>key_out</code> and <code>value_out</code> if you pass in
     * pointers that are not <code>NULL</code>.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_front(char** key_out,
                 void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__remove_front(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove the last key/value pair as determined by sort order from the
     * trie.  The key and value pointers stored in the
     * node that is to be removed will be returned in <code>key_out</code>
     * and <code>value_out</code> if you pass in pointers that are not
     * <code>NULL</code>.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_back(char** key_out,
                void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__remove_back(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes nodes in the range <code>first</code>
     * (inclusive) to <code>last</code> (inclusive).  Strictly as a
     * convenience, this method is an exception to the rule that cgul
     * containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches.  <p>
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @return whether any nodes were removed.
     * @see cgul_cache_cxx::get_freer()
     */
    virtual int
    remove_range(cgul_trie_node_cxx* first,
                 cgul_trie_node_cxx* last,
                 cgul_cache_cxx* keys_cache = NULL,
                 cgul_cache_cxx* values_cache = NULL)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__remove_range(
                 &local,
                 obj,
                 (cgul_trie_node_t)first,
                 (cgul_trie_node_t)last,
                 (keys_cache ? keys_cache->get_obj() : NULL),
                 (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method clears the trie by removing each node individually.
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches. <p>
     *
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @see remove_range
     * @see cgul_cache_cxx::get_freer()
     */
    virtual void
    clear(cgul_cache_cxx* keys_cache,
          cgul_cache_cxx* values_cache)
    {
        cgul_exception_t local = NULL;
        cgul_trie__clear(&local,
                         obj,
                         (keys_cache ? keys_cache->get_obj() : NULL),
                         (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the number of key/value pairs stored in the trie.
     *
     * @return number of key/value pairs stored in the trie
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_trie__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large tries, this should be much faster than trying to do the
     * same thing using removes and inserts.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_trie_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_trie__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method performs a left fold of the trie with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each key in the trie starting at the front of the trie and
     * iterating forward to the end of the trie. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_keys(fold_key_t f,
               void* data)
    {
        const char* key = NULL;
        cgul_trie_node_t curr = NULL;
        cgul_trie_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__foldl_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_trie__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_trie_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_trie_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the trie with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key in the trie starting at the back of the trie
     * and iterating backward to the front of the trie. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_keys(fold_key_t f,
               void* data)
    {
        const char* key = NULL;
        cgul_trie_node_t curr = NULL;
        cgul_trie_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__foldr_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_trie__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_trie_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_trie_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the trie with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each value in the trie starting at the front of the trie and
     * iterating forward to the end of the trie. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_values(fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_trie_node_t curr = NULL;
        cgul_trie_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__foldl_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_trie__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_trie_node__get_next(&local, curr);
            /* Call back. */
            value = cgul_trie_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the trie with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the trie starting at the back of the
     * trie and iterating backward to the front of the trie. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_values(fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_trie_node_t curr = NULL;
        cgul_trie_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__foldr_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_trie__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_trie_node__get_prev(&local, curr);
            /* Call back. */
            value = cgul_trie_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the trie with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each key/value pair in the trie starting at the front of the
     * trie and iterating forward to the end of the trie. <p>
     * 
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_pairs(fold_pair_t f,
                void* data)
    {
        const char* key = NULL;
        void* value = NULL;
        cgul_trie_node_t curr = NULL;
        cgul_trie_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__foldl_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_trie__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_trie_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_trie_node__get_key(&local, curr);
            value = cgul_trie_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the trie with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key/value pair in the trie starting at the back
     * of the trie and iterating backward to the front of the
     * trie. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_pairs(fold_pair_t f,
                void* data)
    {
        const char* key = NULL;
        void* value = NULL;
        cgul_trie_node_t curr = NULL;
        cgul_trie_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__foldr_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_trie__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_trie_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_trie_node__get_key(&local, curr);
            value = cgul_trie_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * Traverse all nodes passing each node to the function
     * <code>f</code>. <p>
     *
     * The first parameter passed into <code>f</code> is the trie
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the trie elements.  In fact, I would
     * recommend that you use
     * <code>cgul_trie_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_trie_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the trie, but you do have to be careful.
     *
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse_range()
     */
    virtual void
    traverse(traverse_t f,
             void* data)
    {
        traverse_range(NULL, NULL, f, data);
    }

    /**
     * Traverse all nodes starting with <code>first</code> (inclusive) and
     * ending with <code>last</code> (inclusive) passing each node to the
     * function <code>f</code>.  If you know the first node, but are not
     * sure of the last node, just use <code>NULL</code> as the last node.
     * This will cause this method to iterate until it reaches the end of
     * the trie.  You can than have <code>f</code> return 0 when it
     * determines that the last node has been reached. <p>
     *
     * The first parameter passed into <code>f</code> is the trie
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * If <code>first</code> is <code>NULL</code>, iteration starts at the
     * beginning of the trie.  If <code>last</code> is <code>NULL</code>,
     * iteration stops at the end of the trie. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the trie elements.  In fact, I would
     * recommend that you use
     * <code>cgul_trie_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_trie_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the trie, but you do have to be careful.
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse()
     */
    virtual void
    traverse_range(cgul_trie_node_cxx* first,
                   cgul_trie_node_cxx* last,
                   traverse_t f,
                   void* data)
    {
        cgul_trie_node_cxx* curr = NULL;
        cgul_trie_node_cxx* next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_trie__traverse_range()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */
        
        if (first == NULL) {
            first = (cgul_trie_node_cxx*)cgul_trie__get_front(&local, obj);
        }
        if (last == NULL) {
            last = (cgul_trie_node_cxx*)cgul_trie__get_back(&local, obj);
        }

        /* Iterate over the nodes. */
        for (curr = first ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = (cgul_trie_node_cxx*)
                       cgul_trie_node__get_next(&local,
                                                (cgul_trie_node_t)curr);
            /* Call back. */
            if (!(*f)(this, curr, data)) {
                break;
            }
            /* We are done if we just processed the last node. */
            if (curr == last) {
                break;
            }
        }
    }

    /**
     * Get the underlying <code>cgul_trie</code> object.
     *
     * @return underlying object
     */
    virtual cgul_trie_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_trie</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_trie__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_trie_t
    take_obj()
    {
        cgul_trie_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_trie_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_trie__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_trie_cxx(const cgul_trie_cxx&);
    cgul_trie_cxx& operator=(const cgul_trie_cxx&);

    /** Underlying <code>cgul_trie</code> object. */
    cgul_trie_t obj;

};

#endif  /*  CGUL_TRIE_CXX_H  */
