#ifndef CGUL_TRIE_NODE_H
#define CGUL_TRIE_NODE_H

/**
 * This class implements a trie node which holds one key/value pair.
 *
 * @file
 * @author Paul Serice
 * @brief trie node
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to <code>cgul_trie_node</code> instance. */
typedef struct cgul_trie_node* cgul_trie_node_t;

/*
 * The constructor and destructor are both private.  Only the
 * <code>cgul_trie</code> class should create are destroy its
 * nodes.
 */

/**
 * Return the key stored in this node.  The key is used to determine
 * the sort order of the node.  If you want to set the value of the
 * key, there is no method to do so because changing the value of a
 * key requires changing the node's position in the
 * <code>cgul_trie</code>.  Thus, if you need to change the value of a
 * key, you should delete the node from the <code>cgul_trie</code> and
 * insert a new node with the new value for the key.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return key
 */
extern CGUL_EXPORT const char*
cgul_trie_node__get_key(cgul_exception_t* cex,
                        cgul_trie_node_t n);

/**
 * Return the value stored in this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return value
 */
extern CGUL_EXPORT void*
cgul_trie_node__get_value(cgul_exception_t* cex,
                          cgul_trie_node_t n);

/**
 * Set the value stored in this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @param[in] value
 */
extern CGUL_EXPORT void
cgul_trie_node__set_value(cgul_exception_t* cex,
                          cgul_trie_node_t n,
                          void* value);

/**
 * Return the previous node.  Together with
 * <code>cgul_trie_node__get_next()</code>, this method lets you
 * traverse the tree in sorted order. If there is no previous node,
 * this method returns <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return previous node
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__get_prev(cgul_exception_t* cex,
                         cgul_trie_node_t n);

/**
 * Return the next node.  Together with
 * <code>cgul_trie_node__get_prev()</code>, this method lets you
 * traverse the tree in sorted order.  If there is no next node, this
 * method returns <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return next node
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__get_next(cgul_exception_t* cex,
                         cgul_trie_node_t n);

/**
 * Return the next older node.  Together with
 * <code>cgul_trie_node__get_younger()</code>, this method lets you
 * traverse the tree in chronological order. If there is no older
 * node, this method returns <code>NULL</code>. <p>
 *
 * <code>cgul_trie_node__get_older()</code> and
 * <code>cgul_trie_node__get_younger()</code> are faster than
 * <code>cgul_trie_node__get_prev()</code> and
 * <code>cgul_trie_node__get_next()</code> because the latter pair
 * of functions has to walk the tree to find the previous and next
 * nodes in sort order, but the former pair has direct pointers to the
 * younger and older nodes. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * reverse chronological order.  Notice that you have to start with
 * the youngest node when calling
 * <code>cgul_trie_node__get_older()</code>:
 *
 * <pre>
 *     cgul_trie_node_t n = cgul_trie__get_youngest(cex, t);
 *     for ( ; n ; n = cgul_trie_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return older node
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__get_older(cgul_exception_t* cex,
                          cgul_trie_node_t n);

/**
 * Return the next younger node.  Together with
 * <code>cgul_trie_node__get_older()</code>, this method lets you
 * traverse the tree in chronological order. If there is no younger
 * node, this method returns <code>NULL</code>. <p>
 *
 * <code>cgul_trie_node__get_older()</code> and
 * <code>cgul_trie_node__get_younger()</code> are faster than
 * <code>cgul_trie_node__get_prev()</code> and
 * <code>cgul_trie_node__get_next()</code> because the latter pair
 * of functions has to walk the tree to find the previous and next
 * nodes in sort order, but the former pair has direct pointers to the
 * younger and older nodes. <p>
 *
 * The following example shows how to iterate over the entire tree in
 * chronological order.  Notice that you have to start with the oldest
 * node when calling <code>cgul_trie_node__get_younger()</code>:
 *
 * <pre>
 *     cgul_trie_node_t n = cgul_trie__get_oldest(cex, t);
 *     for ( ; n ; n = cgul_trie_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return younger node
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__get_younger(cgul_exception_t* cex,
                            cgul_trie_node_t n);

CGUL_END_C

#endif  /*  CGUL_TRIE_NODE_H  */
