#ifndef CGUL_VARSUB_CXX_H
#define CGUL_VARSUB_CXX_H

/**
 * C++ wrapper for <code>cgul_varsub</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_varsub</code>
 */

#include "cgul_convspec_cxx.h"
#include "cgul_exception_cxx.h"
#include "cgul_string_cxx.h"
#include "cgul_varsub.h"

/**
 * This typedef defines the lookup function used by
 * <code>cgul_varsub_cxx()</code> to substitute a variable with its
 * value.  When <code>cgul_varsub_cxx()</code> encounters a variable,
 * it passes the name of the variable to the lookup function in
 * <code>varname</code>.  The lookup function should then append the
 * value associated with <code>varname</code> to the destination
 * string <code>dst</code>. <p>
 *
 * Note that the <code>dst</code> that is passed into the lookup
 * function is the same <code>dst</code> that is passed into
 * <code>cgul_varsub_cxx()</code>.  By altering <code>dst</code>, the
 * lookup function is directly altering the result that will be
 * returned to the caller. <p>
 *
 * The lookup function should return <code>1</code> if the lookup was
 * performed and substituted for <code>varname</code>.  On the other
 * hand, if the lookup failed, <code>0</code> should be returned so
 * <code>cgul_varsub_cxx()</code> will know to insert <code>varname</code>
 * into <code>dst</code> literally. <p>
 *
 * If an error occurs during the execution of the lookup function, it
 * should be handled in the lookup function.  If the lookup function
 * lets a C++ exception escape, it will be quashed by
 * <code>cgul_varsub_cxx__callback_adaptor()</code> to avoid having
 * the C++ exception stomp on the C stack. <p>
 *
 * @param[in] varname variable name
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 */
typedef int
(*cgul_varsub_cxx__lookup_t)(const char* varname,
                             cgul_string_cxx* dst,
                             void* data);


/**
 * This typedef defines the lookup function used by
 * <code>cgul_varsub_formatted_cxx()</code> to substitute the variable
 * <code>varname</code> with its value using an optional
 * printf()-style conversion specification <code>cs</code>.  The
 * lookup function should not try to delete <code>cs</code>. <p>
 *
 * If <code>cgul_varsub_formatted_cxx()</code> does not find a conversion
 * specification or if it detects that there is a problem with the
 * conversion specification, it will set <code>cs</code> to
 * <code>NULL</code>.  Thus, the lookup function must check to make
 * sure <code>cs</code> is not <code>NULL</code> before attempting to
 * dereference it. <p>
 *
 * <code>cgul_varsub_formatted_cxx()</code> considers conversion
 * specifications like "%*s", "%.*s", or "%*.*s" to be invalid because
 * they serve no good purpose in this context, and they have negative
 * security implications.  The lookup function will be called with
 * <code>cs</code> set to <code>NULL</code> if they are detected. <p>
 *
 * Thus to validate <code>cs</code>, the lookup function primarily
 * needs to verify that the type of the conversion specification
 * matches the type of the value that is being appended.  After this
 * is verified, it should then be safe to call
 * <code>cs->serialize()</code> to convert <code>cs</code> to a format
 * string that can then be used with <code>dst->append_sprintf()</code>
 * in order to append <code>varname</code> to <code>dst</code>. <p>
 *
 * Note that the <code>dst</code> that is passed into the lookup
 * function is the same <code>dst</code> that is passed into
 * <code>cgul_varsub_formatted_cxx()</code>.  By altering
 * <code>dst</code>, the lookup function is directly altering the
 * result that will be returned to the caller. <p>
 *
 * The lookup function should return <code>1</code> if the lookup was
 * performed and substituted for <code>varname</code>.  On the other
 * hand, if the lookup failed, <code>0</code> should be returned so
 * <code>cgul_varsub_formatted_cxx()</code> will know to insert
 * <code>varname</code> into <code>dst</code> literally. <p>
 *
 * If an error occurs during the execution of the lookup function, it
 * should be handled in the lookup function.  If the lookup function
 * lets a C++ exception escape, it will be quashed by
 * <code>cgul_varsub_formatted_cxx__callback_adaptor()</code> to avoid
 * having the C++ exception stomp on the C stack. <p>
 *
 * @param[in] varname variable name
 * @param[in] cs conversion specification or <code>NULL</code>
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 */
typedef int
(*cgul_varsub_formatted_cxx__lookup_t)(const char* varname,
                                       cgul_convspec_cxx* cs,
                                       cgul_string_cxx* dst,
                                       void* data);

/**
 * The adoptor used by <code>cgul_varsub_cxx()</code> as the direct
 * callback for the C <code>cgul_varsub()</code> function.  It in turn
 * calls the user's C++ callback.
 *
 * @param[in] cex c-style exception
 * @param[in] varname variable name
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 */
inline int
cgul_varsub_cxx__callback_adaptor(cgul_exception_t* cex,
                                  const char* varname,
                                  cgul_string_t dst,
                                  void* data)
{
    int rv = 0;
    try {
        /* Recover the client data for the adaptor. */
        void** argv = (void**)data;
        cgul_string_cxx* dst2 = (cgul_string_cxx*)argv[0];
        cgul_varsub_cxx__lookup_t f2 = (cgul_varsub_cxx__lookup_t)argv[1];
        void* data2 = argv[2];
        /* Call the client's C++ callback. */
        rv = (*f2)(varname, dst2, data2);
    } catch (...) {
        /* Quash all C++ exceptions to prevent them from stomping on
         * the C stack. */
    }
    return rv;
}

/**
 * The adoptor used by <code>cgul_varsub_formatted_cxx()</code> as the
 * direct callback for the C <code>cgul_varsub_formatted()</code>
 * function.  It in turn calls the user's C++ callback.
 *
 * @param[in,out] cex c-style exception
 * @param[in] varname variable name
 * @param[in] cs conversion specification or <code>NULL</code>
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 */
inline int
cgul_varsub_formatted_cxx__callback_adaptor(cgul_exception_t* cex,
                                            const char* varname,
                                            cgul_convspec_t cs,
                                            cgul_string_t dst,
                                            void* data)
{
    int rv = 0;
    try {
        /* Recover the client data for the adaptor. */
        void** argv = (void**)data;
        cgul_string_cxx* dst2 = (cgul_string_cxx*)argv[0];
        cgul_varsub_formatted_cxx__lookup_t f2 =
            (cgul_varsub_formatted_cxx__lookup_t)argv[1];
        void* data2 = argv[2];
        /* cs could be NULL if the user's format string is bad. */
        if (cs) {
            /* Create a C++-style copy of cs. */
            cgul_convspec_cxx cs2(cs);
            /* Call the client's C++ callback. */
            rv = (*f2)(varname, &cs2, dst2, data2);
        } else {
            /* Call the client's C++ callback. */
            rv = (*f2)(varname, NULL, dst2, data2);
        }
    } catch (...) {
        /* Quash all C++ exceptions to prevent them from stomping on
         * the C stack. */
    }
    return rv;
}

/**
 * This function performs variable substitution on the string
 * <code>src</code> appending the result to <code>dst</code>.  If an
 * error occurs, an exception is thrown.  If an exception is thrown,
 * <code>dst</code> will contain a partial result. <p>
 *
 * The syntax for variable names embedded in <code>src</code> should
 * follow the same syntax that is used by the Bourne shell.  For
 * example if name="Mr. President", "Hello, $name!" or "Hello,
 * ${name}!" will yield "Hello, Mr. President!" after basic variable
 * substitution is performed. <p>
 *
 * Any character can be escaped by prefixing it with the backslash
 * character.  The following conventional escape sequences produce the
 * conventional results: <p>
 *
 * <pre>
 *     "\a", "\b", "\t", "\n", "\v", "\f", "\r", "\\", "\$"
 * </pre>
 *
 * The client is responsible for providing the lookup function
 * <code>f</code> that maps variable names to values.  <code>f</code>
 * is free to treat the variable names creatively.  By placing the
 * variable names in braces, it is possible to embed extra information
 * in the source string <code>src</code> because the name is
 * everything between the braces and is not limited to alphanumeric or
 * underscore characters.  <code>f</code> just needs to be written to
 * handle the extra information.  (See
 * <code>cgul_varsub_formatted_cxx()</code> for an example of how to take
 * advantage of this feature.) <p>
 *
 * When using braces to demarcate variable names, it is worth knowing
 * that braces nest, and escape sequences are enabled.  If you escape
 * a brace, it will be treated like any other character at the current
 * brace level.  An escaped brace will not increase or decrease the
 * current brace level, and it will not match any other brace. <p>
 *
 * <code>data</code> will be passed to <code>f</code> as the client
 * data.  Typically, <code>data</code> will hold the data needed by
 * <code>f</code> to perform the lookup. <p>
 *
 * See tests/main_varsub_cxx.cc for a simple example of how to write a
 * lookup function <code>f</code>. <p>
 *
 * @param[in] src source string
 * @param[out] dst destination string
 * @param[in] f lookup function
 * @param[in] data client data
 * @see cgul_varsub_formatted_cxx()
 * @see cgul_varsub_cxx__lookup_t
 */
inline void
cgul_varsub_cxx(const char* src,
                cgul_string_cxx* dst,
                cgul_varsub_cxx__lookup_t f,
                void* data)
{
    /* We use "argv" to hold the client data for the
     * cgul_varsub_cxx__callback_adaptor().  We use a simple array to
     * avoid having to define a separate structure. */
    void* argv[3] = {(void*)dst, (void*)f, data};
    cgul_exception_t local = NULL;
    cgul_varsub(&local,
                src,
                dst->get_obj(),
                (cgul_varsub__lookup_t)cgul_varsub_cxx__callback_adaptor,
                argv);
    if (local) {
        throw cgul_exception_cxx(local);
    }
}

/**
 * This function performs variable substitution on the string
 * <code>src</code> appending the result to <code>dst</code>.  If an
 * optional printf()-style conversion specification is embedded in the
 * variable name, it will be used to format the variable when
 * appending to <code>dst</code>.  If an error occurs, an exception is
 * thrown.  If an exception is thrown, <code>dst</code> will contain a
 * partial result. <p>
 *
 * As explained in the comments for <code>cgul_varsub()</code>, it
 * understands the Bourne shell convention of placing braces around
 * the name of a variable.  When braces are used, everything between
 * the braces is treat as the literal name of the variable including
 * non-alphanumeric characters.  This makes it possible to embed an
 * optional printf()-style conversion specification in the name of the
 * variable. <p>
 *
 * This underlying <code>cgul_varsub_formatted()</code> function
 * registers a local callback with <code>cgul_varsub()</code> so that
 * <code>cgul_varsub()</code> can parse the source string
 * <code>src</code>.  Each time <code>cgul_varsub()</code> detects a
 * variable name, the local callback is invoked.  The local callback
 * then splits the string into the real variable name and an option
 * printf()-style conversion specification which is (eventually)
 * passed to <code>f</code>. <p>
 *
 * For example, if <code>src</code> is set to "${key:%32s}",
 * <code>cgul_varsub()</code> will call the local callback with the
 * variable name set to "key:%32s".  The local callback will split the
 * string into the real variable name ("key") and its conversion
 * specification ("%32s").  The local callback will then (eventually)
 * cause <code>f</code> to be passed the real variable name and its
 * conversion specification. <p>
 *
 * If the optional conversion specification is missing or invalid,
 * <code>f</code> will be called with the conversion specification set
 * to <code>NULL</code>.  Conversion specifications like "%*s",
 * "%.*s", or "%*.*s" are considered to be invalid because they serve
 * no good purpose in this context, and they have negative security
 * implications.  <p>
 *
 * Any character can be escaped by prefixing it with the backslash
 * character.  The following conventional escape sequences produce the
 * conventional results: <p>
 *
 * <pre>
 *     "\a", "\b", "\t", "\n", "\v", "\f", "\r", "\\", "\$"
 * </pre>
 *
 * <code>data</code> will be passed to <code>f</code> as the client
 * data.  Typically, <code>data</code> will hold the data needed by
 * <code>f</code> to perform the lookup. <p>
 *
 * See tests/main_varsub_formatted_cxx.cc for an example of how to
 * write the lookup function <code>f</code>. <p>
 *
 * @param[in] src source string
 * @param[out] dst destination string
 * @param[in] f lookup function
 * @param[in] data client data
 * @see cgul_varsub_cxx()
 * @see cgul_varsub_formatted_cxx__lookup_t
 */
inline void
cgul_varsub_formatted_cxx(const char* src,
                          cgul_string_cxx* dst,
                          cgul_varsub_formatted_cxx__lookup_t f,
                          void* data)
{
    /* We use "argv" to hold the client data for the
     * cgul_varsub_formatted_cxx__callback_adaptor().  We use a simple
     * array to avoid having to define a separate structure. */
    void* argv[3] = {(void*)dst, (void*)f, data};
    cgul_exception_t local = NULL;
    cgul_varsub_formatted(&local,
                          src,
                          dst->get_obj(),
                          (cgul_varsub_formatted__lookup_t)
                              cgul_varsub_formatted_cxx__callback_adaptor,
                          argv);
    if (local) {
        throw cgul_exception_cxx(local);
    }
}


#endif  /*  CGUL_VARSUB_CXX_H  */
