#ifndef CGUL_VECTOR_H
#define CGUL_VECTOR_H

/**
 * Vector implementation.
 *
 * @file
 * @author Paul Serice
 * @brief vector implementation
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_vector</code> instance. */
typedef struct cgul_vector* cgul_vector_t;

/**
 * This typedef is the interface for the comparison function used by
 * <code>cgul_vector__sort()</code>.  When you define your comparison
 * function, you should compare <code>value1</code> to
 * <code>value2</code> and return less than zero, zero, or greater
 * than zero if <code>value1</code> is less than, equal to, or greater
 * than <code>value2</code>.
 *
 * @param[in] value1 first value
 * @param[in] value2 second value
 * @return result of comparison
 */
typedef int (*cgul_vector__compare_t)(const void* value1,
                                      const void* value2);

/**
 * This type definition defines the interface to a pseudo-random
 * number generator (PRNG) which <code>cgul_vector__shuffle()</code>
 * uses to generate pseudo-random numbers that are used to shuffle the
 * elements of the vector.  The quality of the shuffle depends
 * entirely upon the quality of the pseudo-random numbers
 * returned. <p>
 *
 * Functions that implement this interface will be passed the same
 * PRNG <code>prng</code> that is passed into the shuffle function.
 * The return value should be a pseudo-random integer in the half-open
 * interval <code>[0, range_max)</code>.  If an error occurs when this
 * function is called, the client should propagate the error via
 * <code>cex</code>. <p>
 *
 * Note that this interface closely resembles the signature of the
 * <code>cgul_random_*__next_in_range()</code> functions.  So closely
 * in fact, that it is usually not necessary to write a special
 * callback to implement this function.  Instead, this interface can
 * be satisfied by casting <code>cgul_random_*__next_in_range()</code>
 * to have <code>cgul_vector__next_in_range_t</code> function type and
 * pass in your PRNG object as <code>prng</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] prng PRNG
 * @param[in] range_max maximum, non-negative range for pseudo-random numbers
 * @return pseudo-random integer in the half-open interval [0, range_max)
 */
typedef cgul_uint32_t
(*cgul_vector__next_in_range_t)(cgul_exception_t* cex,
                                void* prng,
                                cgul_uint32_t range_max);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_vector__foldl()</code>
 *     <code>cgul_vector__foldr()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_vector__fold_t)(cgul_exception_t* cex,
                                   void* value,
                                   void* data);

/**
 * Create a new <code>cgul_vector</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_vector__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_vector</code> instance
 */
extern CGUL_EXPORT cgul_vector_t
cgul_vector__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This method
 * does not free the elements stored in the vector.  The caller
 * allocated those elements so the caller is responsible for freeing
 * them.  Do not try to access <code>v</code> after calling this method.
 *
 * @param[in] v <code>cgul_vector</code> instance
 * @see cgul_vector__free_values()
 */
extern CGUL_EXPORT void
cgul_vector__delete(cgul_vector_t v);

/**
 * This method calls <code>free()</code> on all the values in the
 * vector <code>v</code>.  Because this is such a common operation, it
 * is an exception to the rule that cgul containers never free values.
 * This method should only ever be called immediately before calling
 * <code>cgul_vector__delete()</code> because it otherwise invalidates
 * the vector.  Because this function is basically an extension of
 * <code>cgul_vector__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] v <code>cgul_vector</code> instance
 */
extern CGUL_EXPORT void
cgul_vector__free_values(cgul_vector_t v);

/**
 * Return true if the vector <code>v</code> is empty.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return whether <code>v</code> is empty
 */
extern CGUL_EXPORT int
cgul_vector__is_empty(cgul_exception_t* cex,
                      cgul_vector_t v);

/**
 * Assign <code>ptr</code> to <code>v[index]</code>.  If
 * <code>index</code> is out of bounds, an exception is thrown.
 * Indexing is zero-based.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] index index into <code>v</code>
 * @param[in] ptr pointer to assign to <code>v[index]</code>
 */
extern CGUL_EXPORT void
cgul_vector__assign(cgul_exception_t* cex,
                    cgul_vector_t v,
                    size_t index,
                    const void* ptr);

/**
 * Insert <code>ptr</code> at <code>v[index]</code> shifting other
 * pointers to the right as necessary.  If <code>index</code> is out
 * of bounds or if memory cannot be allocated, an exception is thrown.
 * Indexing is zero-based. <p>
 *
 * It is legal to insert at the end of the vector such that
 * <code>index</code> is one beyond the last element.  Typically, you
 * will use <code>cgul_vector__push_back()</code> for inserting at the
 * end however.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] index index into <code>v</code>
 * @param[in] ptr pointer to insert at <code>v[index]</code>
 */
extern CGUL_EXPORT void
cgul_vector__insert(cgul_exception_t* cex,
                    cgul_vector_t v,
                    size_t index,
                    const void* ptr);

/**
 * Erase the element at <code>v[index]</code> shifting other pointers
 * to the left as necessary.  If <code>index</code> is out of bounds,
 * an exception is thrown.  Indexing is zero-based. <p>
 *
 * If it is not necessary to keep the elements in the underlying array
 * in order, it is generally more efficient to call
 * <code>cgul_vector__erase_unordered()</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] index index into <code>v</code>
 */
extern CGUL_EXPORT void
cgul_vector__erase(cgul_exception_t* cex,
                   cgul_vector_t v,
                   size_t index);

/**
 * Erase the element at <code>v[index]</code> replacing it with the
 * element at the back of the underlying array.  If <code>index</code>
 * is out of bounds, an exception is thrown.  Indexing is
 * zero-based. <p>
 *
 * If it is necessary to keep the elements in the underlying array in
 * order, <code>cgul_vector__erase()</code> should be used instead of
 * this method.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] index index into <code>v</code>
 */
extern CGUL_EXPORT void
cgul_vector__erase_unordered(cgul_exception_t* cex,
                             cgul_vector_t v,
                             size_t index);

/**
 * This method returns the element at <code>v[index]</code>.  If
 * <code>index</code> is out of bounds, <code>NULL</code> is returned
 * and an exception is thrown.  Indexing is zero-based.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] index index into <code>v</code>
 * @return element at <code>v[index]</code>
 */
extern CGUL_EXPORT void*
cgul_vector__at(cgul_exception_t* cex,
                cgul_vector_t v,
                size_t index);

/**
 * Append <code>ptr</code> to the end of <code>v</code>.  If memory
 * cannot be allocated, an exception will be thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] ptr pointer to add to the back of <code>v</code>
 */
extern CGUL_EXPORT void
cgul_vector__push_back(cgul_exception_t* cex,
                       cgul_vector_t v,
                       const void* ptr);

/**
 * Returns the pointer held in the last element of <code>v</code>
 * without removing it from <code>v</code>.  If there are no more
 * elements, <code>NULL</code> is returned.  If you want to remove the
 * last element from <code>v</code>, see
 * <code>cgul_vector__pop_back()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return last element of <code>v</code>
 */
extern CGUL_EXPORT void*
cgul_vector__get_back(cgul_exception_t* cex,
                      cgul_vector_t v);

/**
 * Return and remove the pointer held in the last element of
 * <code>v</code>.  If there are no more elements, <code>NULL</code>
 * is returned.  If you want to inspect the last element without
 * removing it, see <code>cgul_vector__get_back()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return last element of <code>v</code>
 */
extern CGUL_EXPORT void*
cgul_vector__pop_back(cgul_exception_t* cex,
                      cgul_vector_t v);

/**
 * Sort the elements of the vector <code>v</code> using the comparison
 * function <code>f</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] f comparison function
 */
extern CGUL_EXPORT void
cgul_vector__sort(cgul_exception_t* cex,
                  cgul_vector_t v,
                  cgul_vector__compare_t f);

/**
 * Shuffle the elements of the vector <code>v</code> using the
 * function <code>f</code> to generate the pseudo-random numbers that
 * will be used to shuffle the vector.  The PRNG <code>prng</code>
 * will be passed through to the PRNG callback function <code>f</code>
 * without modification.  If an error occurs, an exception is
 * thrown. <p>
 *
 * For an example of how to use this method, see the following test
 * program that accompanies the cgul source <p>
 *
 * <pre>
 *     tests/main_vector_shuffle.c
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] f PRNG callback
 * @param[in] prng PRNG
 */
extern CGUL_EXPORT void
cgul_vector__shuffle(cgul_exception_t* cex,
                     cgul_vector_t v,
                     cgul_vector__next_in_range_t f,
                     void* prng);

/**
 * Remove all elements from the vector <code>v</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 */
extern CGUL_EXPORT void
cgul_vector__clear(cgul_exception_t* cex,
                   cgul_vector_t v);

/**
 * Get the internal array.  The client must not free the pointer that
 * is returned because it is still owned by <code>v</code>.  The
 * pointer returned is only valid until the next insert operation.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return array of pointers
 */
extern CGUL_EXPORT void**
cgul_vector__get_array(cgul_exception_t* cex,
                       cgul_vector_t v);

/**
 * Take ownership of the internal array.  The client is responsible
 * for calling <code>free()</code> on the pointer returned. <p>
 *
 * Just prior to taking ownership, the internal state of
 * <code>v</code> is reset.  If an error occurs when attempting to
 * reset <code>v</code>, <code>v</code> is left unaltered,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return array of pointers
 */
extern CGUL_EXPORT void**
cgul_vector__take_array(cgul_exception_t* cex,
                        cgul_vector_t v);

/**
 * Return the size of the vector.  This is a count of the total number
 * of elements stored in the vector.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return size of the vector
 */
extern CGUL_EXPORT size_t
cgul_vector__get_size(cgul_exception_t* cex,
                      cgul_vector_t v);

/**
 * When you are through adding elements, you can trim any extra
 * allocated space by calling this method.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 */
extern CGUL_EXPORT void
cgul_vector__trim(cgul_exception_t* cex,
                  cgul_vector_t v);

/**
 * Get the minimum capacity.  The default value is <code>128</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @return minimum capacity
 */
extern CGUL_EXPORT size_t
cgul_vector__get_minimum_capacity(cgul_exception_t* cex,
                                  cgul_vector_t v);

/**
 * Set the minimum capacity.  The default value is <code>128</code>.
 * Decreasing the capacity will decrease the amount of space used and
 * (typically) also decrease performance.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] minimum_capacity minimum capacity
 */
extern CGUL_EXPORT void
cgul_vector__set_minimum_capacity(cgul_exception_t* cex,
                                  cgul_vector_t v,
                                  size_t minimum_capacity);

/**
 * Swap the underlying data for <code>v1</code> and <code>v2</code>.
 * For large vectors, this should be much faster than trying to do the
 * same thing using <code>cgul_vector__push_back()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v1 <code>cgul_vector</code> instance
 * @param[in] v2 <code>cgul_vector</code> instance
 */
extern CGUL_EXPORT void
cgul_vector__swap(cgul_exception_t* cex,
                  cgul_vector_t v1,
                  cgul_vector_t v2);

/**
 * This method performs a left fold of the vector <code>v</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the vector starting at the front of the
 * vector and iterating forward to the end of the vector. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_vector__foldl(cgul_exception_t* cex,
                   cgul_vector_t v,
                   cgul_vector__fold_t f,
                   void* data);

/**
 * This method performs a right fold of the vector <code>v</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the vector starting at the back of the
 * vector and iterating backward to the front of the vector. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v <code>cgul_vector</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_vector__foldr(cgul_exception_t* cex,
                   cgul_vector_t v,
                   cgul_vector__fold_t f,
                   void* data);


CGUL_END_C

#endif  /*  CGUL_VECTOR_H  */
