#ifndef CGUL_VECTOR_CXX_H
#define CGUL_VECTOR_CXX_H

/**
 * C++ wrapper for <code>cgul_vector</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_vector</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_vector.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_vector</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_vector</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_vector</code>
 * @see cgul_vector_t
 */
class cgul_vector_cxx {

public:

    /**
     * This typedef is the interface for the comparison function used
     * by <code>sort()</code>.
     */
    typedef cgul_vector__compare_t compare_t;

    /**
     * This type definition defines the interface to a pseudo-random
     * number generator (PRNG) which <code>shuffle()</code> uses to
     * generate pseudo-random numbers that are used to shuffle the
     * elements of the vector.  The quality of the shuffle depends
     * entirely upon the quality of the pseudo-random numbers
     * returned. <p>
     *
     * Functions that implement this interface will be passed the same
     * PRNG <code>prng</code> that is passed into the shuffle function.
     * The return value should be a pseudo-random integer in the half-open
     * interval <code>[0, range_max)</code>.  If an error occurs when this
     * function is called, the client should propagate the error via
     * <code>cex</code>. <p>
     *
     * Note that this interface closely resembles the signature of the
     * <code>cgul_random_*__next_in_range()</code> functions.  So closely
     * in fact, that it is usually not necessary to write a special
     * callback to implement this function.  Instead, this interface can
     * be satisfied by casting <code>cgul_random_*__next_in_range()</code>
     * to have <code>next_in_range_t</code> function type and
     * pass in your PRNG object as <code>prng</code>. <p>
     */
    typedef cgul_vector__next_in_range_t next_in_range_t;

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl()</code>
     *     <code>foldr()</code>
     * </pre>
     *
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_t)(void* value,
                          void* data);

    /**
     * Create a new <code>cgul_vector_cxx</code> object.  If memory
     * cannot be allocated, an exception is thrown.
     */
    cgul_vector_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_vector__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_vector_cxx</code> object by wrapping
     * an existing <code>cgul_vector</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_vector_cxx(cgul_vector_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  This method
     * does not free the elements stored in the vector.  The caller
     * allocated those elements so the caller is responsible for freeing
     * them.
     *
     * @see free_values()
     */
    virtual
    ~cgul_vector_cxx()
    {
        if (obj) {
            cgul_vector__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the values in the
     * vector.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free values.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the vector.
     */
    virtual void
    free_values()
    {
        cgul_vector__free_values(obj);
    }


    /**
     * Return true if the vector is empty.
     *
     * @return whether the vector is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_vector__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Assign <code>ptr</code> to the slot specified by
     * <code>index</code>.  If <code>index</code> is out of bounds, an
     * exception is thrown.  Indexing is zero-based.
     *
     * @param[in] index index into the vector
     * @param[in] ptr pointer to assign
     */
    virtual void
    assign(unsigned long int index,
           const void* ptr)
    {
        cgul_exception_t local = NULL;
        cgul_vector__assign(&local, obj, index, ptr);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Insert <code>ptr</code> at the slot specified by
     * <code>index</code> shifting other pointers to the right as
     * necessary.  If <code>index</code> is out of bounds or if memory
     * cannot be allocated, an exception is thrown.  Indexing is
     * zero-based. <p>
     *
     * It is legal to insert at the end of the vector such that
     * <code>index</code> is one beyond the last element.  Typically,
     * you will use <code>push_back()</code> for inserting at the end
     * however.
     *
     * @param[in] index index into the vector
     * @param[in] ptr pointer to insert
     */
    virtual void
    insert(unsigned long int index,
           const void* ptr)
    {
        cgul_exception_t local = NULL;
        cgul_vector__insert(&local, obj, index, ptr);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Erase the element at the slot specified by <code>index</code>
     * shifting other pointers to the left as necessary.  If
     * <code>index</code> is out of bounds, an exception is thrown.
     * Indexing is zero-based. <p>
     *
     * If it is not necessary to keep the elements in the underlying array
     * in order, it is generally more efficient to call
     * <code>erase_unordered()</code>.
     *
     * @param[in] index index into the vector
     */
    virtual void
    erase(unsigned long int index)
    {
        cgul_exception_t local = NULL;
        cgul_vector__erase(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Erase the element at the slot specified by <code>index</code>
     * replacing it with the element at the back of the underlying
     * array.  If <code>index</code> is out of bounds, an exception is
     * thrown.  Indexing is zero-based. <p>
     *
     * If it is necessary to keep the elements in the underlying array
     * in order, <code>erase()</code> should be used instead of this
     * method.
     *
     * @param[in] index index into the vector
     */
    virtual void
    erase_unordered(unsigned long int index)
    {
        cgul_exception_t local = NULL;
        cgul_vector__erase_unordered(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the pointer at the slot specified by
     * <code>index</code>.  If <code>index</code> is out of bounds, an
     * exception is thrown.  Indexing is zero-based. <p>
     *
     * Unlike <code>std::vector::at()</code> this method does not
     * return a reference to the pointer.  It just returns the
     * pointer.  This is much safer because the caller does not have
     * to worry about the reference that is returned being invalidated
     * when the vector is resized because of an insert operation. <p>
     *
     * Furthermore, the underlying implementation of this method is
     * written in C which cannot return references because they are
     * not part of the language.  Thus, to be consistent, this method
     * returns the same value as the underlying C implementation. <p>
     *
     * @param[in] index index into the vector
     * @return element at the slot for <code>index</code>
     */
    virtual void*
    at(unsigned long int index)
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_vector__at(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns a reference to the slot specified by
     * <code>index</code> making it possible to assign directly to the
     * return value.  This behavior is similiar to how
     * <code>std::vector::operator[]()</code> behaves. <p>
     *
     * The reference that is returned is only valid until the next
     * insert operation so, if possible, use <code>at()</code> instead
     * of this method to avoid having your pointers invalidated
     * unexpectedly. <p>
     *
     * As with <code>std::vector::operator[]()</code>, this method
     * does not bounds check <code>index</code>.  Thus, if
     * <code>index</code> is out of bounds, behavior is undefined.  If
     * bounds checking is desired, use <code>at()</code> instead. <p>
     *
     * Indexing is zero-based. <p>
     *
     * @param[in] index index into the vector
     * @return reference to the element at the slot for <code>index</code>
     */
    virtual void*&
    operator[](unsigned long int index)
    {
        void** v = NULL;
        cgul_exception_t local = NULL;
        v = cgul_vector__get_array(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return v[index];
    }

    /**
     * Append <code>ptr</code> to the end of the vector.  If memory
     * cannot be allocated, an exception will be thrown.
     *
     * @param[in] ptr pointer to add to the back of the vector
     */
    virtual void
    push_back(const void* ptr)
    {
        cgul_exception_t local = NULL;
        cgul_vector__push_back(&local, obj, ptr);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Returns the pointer held in the last element of the vector
     * without removing it from the vector.  If there are no more
     * elements, <code>NULL</code> is returned.  If you want to remove
     * the last element from the vector, see <code>pop_back()</code>.
     *
     * @return last element of the vector
     */
    virtual void*
    get_back() const
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_vector__get_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return and remove the pointer held in the last element of
     * the vector.  If there are no more elements, <code>NULL</code>
     * is returned.  If you want to inspect the last element without
     * removing it, see <code>get_back()</code>.
     *
     * @return last element of the vector
     */
    virtual void*
    pop_back()
    {
        void* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_vector__pop_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Sort the elements of the vector using the comparison function
     * <code>f</code>.
     *
     * @param[in] f comparison function
     */
    virtual void
    sort(compare_t f)
    {
        cgul_exception_t local = NULL;
        cgul_vector__sort(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Shuffle the elements of the vector using the
     * function <code>f</code> to generate the pseudo-random numbers that
     * will be used to shuffle the vector.  The PRNG <code>prng</code>
     * will be passed through to the PRNG callback function <code>f</code>
     * without modification.  If an error occurs, an exception is
     * thrown. <p>
     *
     * The following is an an example of how to use this method: <p>
     *
     * <pre>
     *     cgul_vector_cxx v;
     *     cgul_random_lfg32_cxx prng;
     *
     *     v.shuffle((cgul_vector_cxx::next_in_range_t)
     *                   cgul_random_lfg32__next_in_range,
     *               prng.get_obj());
     * </pre>
     *
     * @param[in] f PRNG callback
     * @param[in] prng PRNG
     */
    virtual void
    shuffle(next_in_range_t f,
            void* prng)
    {
        cgul_exception_t local = NULL;
        cgul_vector__shuffle(&local, obj, f, prng);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove all elements from the vector.
     */
    virtual void
    clear()
    {
        cgul_exception_t local = NULL;
        cgul_vector__clear(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the internal array.  The client must not free the pointer that
     * is returned because it is still owned by the vector.  The
     * pointer returned is only valid until the next insert operation.
     *
     * @return array of pointers
     */
    virtual void**
    get_array() const
    {
        void** rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_vector__get_array(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Take ownership of the internal array.  The client is
     * responsible for calling <code>free()</code> on the pointer
     * returned. <p>
     *
     * Just prior to taking ownership, the internal state of the
     * vector is reset.  If an error occurs when attempting to reset
     * the vector, the vector is left unaltered, <code>NULL</code> is
     * returned, and an exception is thrown. <p>
     *
     * @return array of pointers
     */
    virtual void**
    take_array()
    {
        void** rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_vector__take_array(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the size of the vector.  This is a count of the total number
     * of elements stored in the vector.
     *
     * @return size of the vector
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_vector__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * When you are through adding elements, you can trim any extra
     * allocated space by calling this method.
     */
    virtual void
    trim()
    {
        cgul_exception_t local = NULL;
        cgul_vector__trim(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the minimum capacity.  The default value is <code>128</code>.
     *
     * @return minimum capacity
     */
    virtual size_t
    get_minimum_capacity() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_vector__get_minimum_capacity(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the minimum capacity.  The default value is <code>128</code>.
     * Decreasing the capacity will decrease the amount of space used and
     * (typically) also decrease performance.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] minimum_capacity minimum capacity
     */
    virtual void
    set_minimum_capacity(size_t minimum_capacity)
    {
        cgul_exception_t local = NULL;
        cgul_vector__set_minimum_capacity(&local, obj, minimum_capacity);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large vectors, this should be much faster than trying to do
     * the same thing using <code>push_back()</code>.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_vector_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_vector__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method performs a left fold of the vector with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each value in the vector starting at the front of the vector and
     * iterating forward to the end of the vector. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order
     * for iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl(fold_t f,
          void* data)
    {
        unsigned long int i = 0;
        unsigned long int i_max = 0;

        /*
         * NOTE: We do NOT forward the call to cgul_vector__foldl()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over each value. */
        i_max = get_size();
        for (i = 0 ; i < i_max ; ++i) {
            /* Call back. */
            if (!(*f)(at(i), data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the vector with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the vector starting at the back of the
     * vector and iterating backward to the front of the vector. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr(fold_t f,
          void* data)
    {
        unsigned long int i = 0;
        unsigned long int i_max = 0;

        /*
         * NOTE: We do NOT forward the call to cgul_vector__foldl()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over each value. */
        i_max = get_size();
        for (i = i_max ; i > 0 ; --i) {
            /* Call back. */
            if (!(*f)(at(i - 1), data)) {
                break;
            }
        }
    }

    /**
     * Get the underlying <code>cgul_vector</code> object.
     *
     * @return underlying object
     */
    virtual cgul_vector_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_vector</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_vector__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_vector_t
    take_obj()
    {
        cgul_vector_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_vector_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_vector__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_vector_cxx(const cgul_vector_cxx&);
    cgul_vector_cxx& operator=(const cgul_vector_cxx&);

    /** Underlying <code>cgul_vector</code> object. */
    cgul_vector_t obj;

};

#endif  /*  CGUL_VECTOR_CXX_H  */
