#ifndef CGUL_BITWISE_TRIE_NODE_P_H
#define CGUL_BITWISE_TRIE_NODE_P_H

#include "cgul_common.h"
#include "cgul_bitwise_trie_node.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_bitwise_trie_node</code> class.
 */
struct cgul_bitwise_trie_node {

    /**
     * Left node is children[0], and the right node is children[1].
     * It is done like this to simplify traversing the tree because
     * when the least-significant bit is zero we go left and when it
     * is 1 we go right.
     */
    cgul_bitwise_trie_node_t children[2];

    /**
     * Pointer to older node.  Together with <code>younger</code>,
     * <code>older</code> forms a doubly linked list that has nothing
     * to do with sort order; rather, the <code>older</code> node
     * points to the next older node.  If there is no older node,
     * <code>older</code> is <code>NULL</code>.  By calling
     * <code>cgul_bitwise_trie_node__get_older()</code>, the user can
     * iterate over the nodes in reverse chronological order.
     */
    cgul_bitwise_trie_node_t older;

    /**
     * Pointer to younger node.  Together with <code>older</code>,
     * <code>younger</code> forms a doubly linked list that has
     * nothing to do with sort order; rather, the <code>younger</code>
     * node points to the next younger node.  If there is no younger
     * node, <code>younger</code> is <code>NULL</code>.  By calling
     * <code>cgul_bitwise_trie_node__get_younger()</code>, the user
     * can iterate over the nodes in chronological order.
     */
    cgul_bitwise_trie_node_t younger;

    /** Index into the trie where the value is stored. */
    unsigned long int index;

    /** Value returned by a lookup based on the index. */
    const void* value;
};

/*
 * Private Methods
 */

/**
 * Create a new <code>cgul_bitwise_trie_node</code> object.  The new
 * object will be initialized with <code>index</code> set to zero and
 * <code>value</code> set to <code>NULL</code>.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_bitwise_trie_node__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * This class does not take ownership of <code>value</code>.  If the
 * client needs to delete the <code>value</code>, the client should
 * first call <code>cgul_bitwise_trie__remove()</code> on the
 * <code>index</code>. <p>
 *
 * The nodes constructed by this method are suitable for inclusion in
 * <code>cgul_bitwise_trie</code> objects constructed with the
 * <code>cgul_bitwise_trie__new()</code> method. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_bitwise_trie_node</code> instance
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the <code>value</code> pointers stored in the node.  The
 * value for <code>n</code> should have been instantiated by
 * <code>cgul_bitwise_trie_node__new()</code>.  Do not try to access
 * <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 */
extern CGUL_EXPORT void
cgul_bitwise_trie_node__delete(cgul_bitwise_trie_node_t n);

/**
 * This method is shared between
 * <code>cgul_bitwise_trie_node__new()</code> and the
 * <code>restructor</code> for the <code>cgul_cache()</code>.  It
 * initializes new nodes and re-initializes existing nodes.
 *
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance to initialize
 */
extern CGUL_EXPORT void
cgul_bitwise_trie_node__initialize(cgul_bitwise_trie_node_t n);

/**
 * Set the index used to lookup this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @param[in] index index
 */
extern CGUL_EXPORT void
cgul_bitwise_trie_node__set_index(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n,
                                  unsigned long int index);

/**
 * Return the left child node.
 *
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @return the left child
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_left_child(cgul_exception_t* cex,
                                       cgul_bitwise_trie_node_t n);

/**
 * Return the right child node.
 *
 * @param[in] n <code>cgul_bitwise_trie_node</code> instance
 * @return the right child
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_right_child(cgul_exception_t* cex,
                                        cgul_bitwise_trie_node_t n);

CGUL_END_C

#endif  /*  CGUL_BITWISE_TRIE_NODE_P_H  */
