#include "cgul_crlf_p.h"
#include "cgul_libc.h"

#define CGUL_CRLF__CONVERT_FILE_BUFSIZE (16384)

/**
 * This method lazily allocates the <code>crlf->lines</code> and
 * <code>crlf->line_offsets</code> arrays.  Note that the number of
 * possible lines in a buffer has an upper bound equal to the number
 * of characters in the buffer (or "bsize").
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf <code>cgul_crlf</code> instance
 * @param[in] bsize buffer size
 */
static void
cgul_crlf__allocate_arrays(cgul_exception_t* cex,
                           cgul_crlf_t crlf,
                           unsigned long int bsize)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    if (bsize == 0) {
        goto out;
    }

    /* Free both arrays. */
    if (crlf->lines) {
        free(crlf->lines);
        crlf->lines = NULL;
    }
    if (crlf->line_offsets) {
        free(crlf->line_offsets);
        crlf->line_offsets = NULL;
    }
    crlf->array_size = 0;       /* in case we throw an exception */
    crlf->max_line_index = 0;   /* just to be safe */
    crlf->next_line_index = 0;  /* just to be safe */

    /* Allocate both arrays. */
    crlf->lines =
        (char**)cgul_libc__calloc(cex, bsize, sizeof(char*));

    crlf->line_offsets =
        (unsigned long int*)cgul_libc__calloc(cex,
                                              bsize,
                                              sizeof(unsigned long int));

    /* Throw exceptions. */
    if (*cex) {
        goto out;
    }

    /* Remember the new buffer size. */
    crlf->array_size = bsize;

 out:
    return;
}

cgul_crlf_t
cgul_crlf__new(cgul_exception_t* cex)
{
    cgul_crlf_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_crlf_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    rv->strip_utf8_bom = 1;
    rv->test_utf8_bom = 1;
    rv->remainder = cgul_string__new(cex);
    rv->first_line = cgul_string__new(cex);

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_crlf__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_crlf__delete(cgul_crlf_t crlf)
{
    if(crlf) {
        if (crlf->first_line) {
            cgul_string__delete(crlf->first_line);
        }
        if (crlf->remainder) {
            cgul_string__delete(crlf->remainder);
        }
        if (crlf->line_offsets) {
            free(crlf->line_offsets);
        }
        if (crlf->lines) {
            free(crlf->lines);
        }
        free(crlf);
    }
}

void
cgul_crlf__reset(cgul_exception_t* cex,
                 cgul_crlf_t crlf,
                 unsigned long offset)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    /*
     * Reset all of the member variables.
     */

    crlf->skip_leading_new_line = 0;

    /* Keep crlf->strip_utf8_bom between resets. */

    /* If the user repositions to the start of the file, we need to
     * test the UTF-8 byte-order mark again. */
    if (offset == 0) {
        crlf->test_utf8_bom = 1;
    } else {
        crlf->test_utf8_bom = 0;
    }

    crlf->line_count = 0;

    /* Remember the new offset. */
    crlf->block_offset = offset;

    crlf->next_line_index = 0;
    crlf->max_line_index = 0;

    /* Keep crlf->lines buffer between resets. */
    /* Keep crlf->line_offsets between resets. */
    /* Keep crlf->array_size same between resets. */

    cgul_string__set_value(cex, crlf->remainder, "");
    crlf->remainder_offset = 0;
    crlf->has_get_remainder_been_called = 0;
    cgul_string__set_value(cex, crlf->first_line, "");

 out:
    return;
}

int
cgul_crlf__get_strip_utf8_bom(cgul_exception_t* cex,
                              cgul_crlf_t crlf)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    rv = crlf->strip_utf8_bom;

 out:
    return rv;
}

void
cgul_crlf__set_strip_utf8_bom(cgul_exception_t* cex,
                              cgul_crlf_t crlf,
                              int strip_utf8_bom)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    /* Normalize strip_utf8_bom so that we always use 0 or 1 internally
     * regardless of the value passed in by the user. */
    if (strip_utf8_bom) {
        crlf->strip_utf8_bom = 1;
    } else {
        crlf->strip_utf8_bom = 0;
    }

 out:
    return;
}

void
cgul_crlf__convert(cgul_exception_t* cex,
                   cgul_crlf_t crlf,
                   char* buf,
                   unsigned long int bsize)
{
    char c = '\0';
    unsigned long int i = 0;
    unsigned long int abs_offset = 0;
    char* start = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);
    assert(buf);

    /* Make sure there is at least one character in "buf". */
    if( bsize == 0 ) {
        goto out;
    }

    /* Lazily allocate the crlf->lines and crlf->line_offsets arrays. */
    if (crlf->array_size < bsize) {
        cgul_crlf__allocate_arrays(cex, crlf, bsize);
    }

    /* Reset the index information. */
    crlf->max_line_index = 0;
    crlf->next_line_index = 0;

    /* Initialize the starting point by possibly skipping a left over
     * '\n' that was part of an incomplete "\r\n" combination that was
     * started last go around. */
    i = 0;
    start = buf;
    if( crlf->skip_leading_new_line ) {
        /* Because of "bsize != 0" check above, it is safe to
         * dereference "buf[0]" here. */
        if( buf[0] == '\n' ) {
            ++i;
            ++start;
        }
        crlf->skip_leading_new_line = 0;
    }

    for( ; i < bsize ; ++i ) {
        c = buf[i];
        if( (c == '\n') || (c == '\r') ) {
            buf[i] = '\0';
            /* If this is the first line, see if we need to prepend
             * the remainder from the last go around */
            if ((crlf->max_line_index == 0)
                && (cgul_string__get_length(cex, crlf->remainder) > 0))
            {
                /* We cannot just push_back "start".  Instead, we have
                 * to combine what remained from the last go around
                 * with what we currently have. */

                /* Start with the remainder. */
                cgul_string__set_value(
                    cex,
                    crlf->first_line,
                    cgul_string__get_value(cex, crlf->remainder));

                /* Append the rest of the line. */
                cgul_string__append(
                    cex,
                    crlf->first_line,
                    start);

                if (*cex) {
                    goto out;
                }

                /* Sanity check the max_line_index. */
                if (crlf->max_line_index >= crlf->array_size) {
                    *cex = cgul_exception__new(
                           "cgul_crlf__convert: Internal error: "
                           "Attempt to write beyond array.");
                    goto out;
                }

                /* Store the first line of text. */
                crlf->lines[crlf->max_line_index] =
                    cgul_string__get_value(cex, crlf->first_line);

                /* Store the offset for the first line of text. */
                crlf->line_offsets[crlf->max_line_index] =
                    crlf->remainder_offset;

                /* Increment the index. */
                ++crlf->max_line_index;

                /* Clear the remainder. */
                cgul_string__set_value(
                    cex,
                    crlf->remainder,
                    "");

                crlf->remainder_offset = 0;

            } else {

                /* Sanity check the max_line_index. */
                if (crlf->max_line_index >= crlf->array_size) {
                    *cex = cgul_exception__new(
                           "cgul_crlf__convert: Internal error: "
                           "Attempt to write beyond array.");
                    goto out;
                }

                /* Store "start". */
                crlf->lines[crlf->max_line_index] = start;

                /* Store the absolute offset for "start". */
                abs_offset = crlf->block_offset + (unsigned long)(start - buf);
                crlf->line_offsets[crlf->max_line_index] = abs_offset;

                /* Increment the index. */
                ++crlf->max_line_index;

            }

            /* Check for "\r\n" combination. */
            if( c == '\r' ) {
                unsigned long int ip1 = i + 1;  /* "ip1" means "i plus 1" */
                if( ip1 < bsize ) {
                    if( buf[ip1] == '\n' ) {
                        ++i;
                    }
                } else {
                    /* The end of buffer happened right in the middle
                     * of detecting "\r\n". */
                    crlf->skip_leading_new_line = 1;
                }
            }
            start = buf + i + 1;

            if (*cex) {
                goto out;
            }
        }
    }

    /* Determine the remainder. */
    if( start != buf + bsize ) {
        /* Only change the remainder's offset if this is a new remainder. */
        if (cgul_string__get_length(cex, crlf->remainder) == 0) {
            crlf->remainder_offset =
                crlf->block_offset + (unsigned long)(start - buf);
        }
        /* Append the remainder. */
        cgul_string__append_block(cex,
                                  crlf->remainder,
                                  start,
                                  (buf + bsize) - start);
    }

    /* Update the block_offset to point to the start of the next block. */
    crlf->block_offset += bsize;

 out:
    return;
}

const char*
cgul_crlf__get_line(cgul_exception_t* cex,
                    cgul_crlf_t crlf)
{
    const unsigned char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    /* Check if there is another line to return. */
    if (crlf->next_line_index < crlf->max_line_index) {

        /* Get the line to return. */
        rv = (unsigned char*)crlf->lines[crlf->next_line_index];

        /* Strip off the leading UTF-8 byte-order mark when necessary. */
        if (crlf->test_utf8_bom) {
            if (crlf->strip_utf8_bom) {
                if (   (rv[0] == (unsigned char)0xef)
                    && (rv[1] == (unsigned char)0xbb)
                    && (rv[2] == (unsigned char)0xbf))
                {
                    /* Skip the BOM. */
                    rv += 3;
                    /* Update the offset for this line. */
                    crlf->line_offsets[crlf->next_line_index] += 3;
                }
            }
            crlf->test_utf8_bom = 0;
        }

        /* Increment the next_line_index and line_count. */
        ++crlf->next_line_index;
        ++crlf->line_count;
    }

 out:
    return (char*)rv;
}

unsigned long
cgul_crlf__get_line_count(cgul_exception_t* cex,
                          cgul_crlf_t crlf)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    rv = crlf->line_count;

 out:
    return rv;
}

unsigned long
cgul_crlf__get_line_offset(cgul_exception_t* cex,
                           cgul_crlf_t crlf)
{
    unsigned long int rv = 0;
    unsigned long int index = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    /* Check if we need to return the remainder's offset. */
    if (crlf->has_get_remainder_been_called) {
        /* Make sure there is a remainer. */
        if (cgul_string__get_length(cex, crlf->remainder) > 0) {
            rv = crlf->remainder_offset;
            goto out;
        }
        /* Fall through because there is no remainder.  Instead, just
         * return the offset of the last line returned by
         * cgul_crlf__get_line(). */
    }

    /*
     * Return the offset for the last line returned by cgul_crlf__get_line().
     */

    /* Make sure the user has called cgul_crlf__get_line() at least
     * once.  That function increments crlf->next_line_index.  So, if
     * crlf->next_line_index is still zero, we know
     * cgul_crlf__get_line() has not been called. */
    if (crlf->next_line_index == 0) {
        *cex = cgul_exception__new("cgul_crlf__get_line_offset: "
                                   "Must call cgul_crlf__get_line() first.");
        goto out;
    }

    /* Because cgul_crlf__get_line() increments crlf->next_line_index
     * before returning, we need to subtract one in order to use the
     * same index.  By checking if crlf->line_offsets is zero above,
     * we guard against underflow. */
    index = crlf->next_line_index - 1;

    /* Make sure "index" does not point beyond the last line.  This is
     * necessary if this method is called after the user exhausts
     * cgul_crlf__get_line(). */
    if (index >= crlf->max_line_index) {
        index = crlf->max_line_index - 1;
    }

    /* Return the offset. */
    rv = crlf->line_offsets[index];

 out:
    return rv;
}

const char*
cgul_crlf__get_remainder(cgul_exception_t* cex,
                         cgul_crlf_t crlf)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    if (!crlf->has_get_remainder_been_called) {
        if (cgul_string__get_length(cex, crlf->remainder) > 0) {
            rv = cgul_string__get_value(cex, crlf->remainder);
            ++crlf->line_count;
        }
        crlf->has_get_remainder_been_called = 1;
    }

 out:
    return rv;
}

unsigned long
cgul_crlf__get_remainder_offset(cgul_exception_t* cex,
                                cgul_crlf_t crlf)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf);

    rv = crlf->remainder_offset;

 out:
    return rv;
}

void
cgul_crlf__convert_file(cgul_exception_t* cex,
                        FILE* fin,
                        FILE* fout,
                        const char* eol)
{
    unsigned long int rcount = CGUL_CRLF__CONVERT_FILE_BUFSIZE;
    char* buf = NULL;
    const char* line = NULL;
    const char* remainder = NULL;
    cgul_crlf_t crlf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fin);
    assert(fout);
    assert(eol);

    /* I would prefer to not put 16K on the stack. */
    buf = (char*)cgul_libc__malloc(cex, CGUL_CRLF__CONVERT_FILE_BUFSIZE);

    /* Allocate a new crlf object in order to do the conversion. */
    crlf = cgul_crlf__new(cex);

    if (*cex) {
        goto out;
    }

    for (;;) {

        /* Read a block. */
        rcount = (unsigned long int)
                     cgul_libc__fread(cex,
                                      buf,
                                      sizeof(char),
                                      CGUL_CRLF__CONVERT_FILE_BUFSIZE,
                                      fin);
        if (*cex) {
            goto out;
        }
        if (rcount == 0) {
            break;
        }

        /* Convert the block. */
        cgul_crlf__convert(cex, crlf, buf, rcount);
        if (*cex) {
            goto out;
        }

        /* Write the lines from this block appending "eol". */
        for (;;) {
            line = cgul_crlf__get_line(cex, crlf);
            if (line == NULL) {
                break;
            }
            cgul_libc__fputs(cex, line, fout);
            cgul_libc__fputs(cex, eol, fout);
            if (*cex) {
                goto out;
            }
        }
    }

    /* Copy the remainder over without adding an EOL. */
    remainder = cgul_crlf__get_remainder(cex, crlf);
    if (remainder) {
        cgul_libc__fputs(cex, remainder, fout);
    }

 out:

    if (crlf) {
        cgul_crlf__delete(crlf);
    }
    if (buf) {
        free(buf);
    }
}
