#ifndef CGUL_HASH_P_H
#define CGUL_HASH_P_H

#include "cgul_common.h"
#include "cgul_hash.h"
#include "cgul_list.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_hash</code> class.
 */
struct cgul_hash {

    /**
     * When a custom <code>cgul_hash</code> object is instantiated,
     * this is the callback hash function.  This value is set to
     * <code>NULL</code> by default which causes the in-line hash
     * function that works only with C-style strings to be used.
     */
    cgul_hash__hash_t hf;

    /**
     * Client data for <code>hf</code>.
     */
    void* hf_data;

    /**
     * When a custom <code>cgul_hash</code> object is instantiated,
     * this is the callback comparison function.  This value is set to
     * <code>strcmp</code> by default which works only with C-style
     * strings.
     */
    cgul_hash__compare_t cf;

    /**
     * Vector of linked lists that makes up the hash.  For size and
     * space efficiency, <code>cgul_list</code> objects are not used
     * here.  Instead, the <code>cgul_hash_node</code> objects have an
     * embedded linked list.
     */
    cgul_hash_node_t* v;

    /**
     * Current capacity of <code>v.  You must remember to recalculate
     * <code>size_limit</code> whenever you change this value.
     */
    unsigned int v_capacity;

    /**
     * The hash is always a power of two which allows us to use a
     * simple mask to calculate the index into <code>v</code> rather
     * than a modulo operation.
     */
    unsigned int key_mask;

    /**
     * Total number of elements stored in the hash.
     */
    size_t size;

    /**
     *  The oldest node in the hash.  This is the first node that was
     *  inserted.
     */
    cgul_hash_node_t oldest;

    /**
     *  The youngest node in the hash.  This is the last node that was
     *  inserted.
     */
    cgul_hash_node_t youngest;

    /**
     * The <code>load_factor</code> is the main parameter that governs
     * the speed/space tradeoff.  It is the ratio of the average
     * number of keys per bucket that is enforced by the hash.  You
     * must remember to recalculate <code>size_limit</code> whenever
     * you change this value.
     */
    float load_factor;

    /**
     * The size limit of the hash.  When this value is exceeded, the
     * hash will automatically resize.  You must remember to
     * recalculate this value whenever the <code>load_factor</code> or
     * the <code>v_capacity</code> change.
     */
    size_t size_limit;

    /**
     * The total number of collisions.  This can be helpful when
     * trying to determine how effective a particular hash function is
     * for a specific set of data.  Note that the total number of
     * collisions can be more than the number of keys because the
     * total number of collisions includes all the extra collisions
     * that occur when the hash is resized.
     */
    size_t total_collisions;

};

CGUL_END_C

#endif  /*  CGUL_HASH_P_H  */
