#ifndef CGUL_LIST_NODE_P_H
#define CGUL_LIST_NODE_P_H

#include "cgul_common.h"
#include "cgul_list_node.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_list_node</code> class.
 */
struct cgul_list_node {

    /** Next node. */
    cgul_list_node_t next;

    /** Previous node. */
    cgul_list_node_t prev;

    /** Value stored in the node. */
    const void* value;

};


/*
 * Private Methods
 */

/**
 * Create a new <code>cgul_list_node</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_list_delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_list_node</code> instance
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list_node__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  It does not
 * free the <code>value</code> stored in the list.  The client is
 * responsible for freeing the values stored in the list when the
 * client thinks it is convenient and safe to do so.  It also does not
 * manipulate <code>next</code> and <code>prev</code> pointers
 * (assuming the node is in a linked list) for the elements that come
 * before it or after it in the list.  Manipulating <code>next</code>
 * and <code>prev</code> pointers is the job of the linked list.  Do
 * not try to access <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_list_node</code> instance
 */
extern CGUL_EXPORT void
cgul_list_node__delete(cgul_list_node_t n);

/**
 * This method is shared between <code>cgul_list_node__new()</code>
 * and the "restructor" for the <code>cgul_cache()</code>.  It
 * initializes new nodes and re-initializes existing nodes.  Its
 * signature is designed to loosely match
 * <code>cgul_cache_restructor_t</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] obj 
 */
extern CGUL_EXPORT void
cgul_list_node__initialize(cgul_list_node_t n);

/**
 * Set the pointer to the next node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_list_node</code> instance
 * @param[in] next next node
 */
extern CGUL_EXPORT void
cgul_list_node__set_next(cgul_exception_t* cex,
                         cgul_list_node_t n,
                         cgul_list_node_t next);

/**
 * Set the pointer to the previous node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_list_node</code> instance
 * @param[in] prev previous node
 */
extern CGUL_EXPORT void
cgul_list_node__set_prev(cgul_exception_t* cex,
                         cgul_list_node_t n,
                         cgul_list_node_t prev);


CGUL_END_C

#endif  /*  CGUL_LIST_NODE_P_H  */
