#ifndef CGUL_MULTIMAP_NODE_P_H
#define CGUL_MULTIMAP_NODE_P_H

#include "cgul_common.h"
#include "cgul_list.h"
#include "cgul_multimap_p.h"
#include "cgul_multimap_node.h"
#include "cgul_rbtree_node.h"

CGUL_BEGIN_C

/** Data encapsulated by the <code>cgul_multimap_node</code> class. */
struct cgul_multimap_node {

    /**
     * Value.
     */
    const void* value;

    /**
     * Pointer back to the node in the <code>cgul_rbtree</code> that
     * holds the linked list as its value.  This is needed to be able
     * to look up the key and to be able to iterate to the next node
     * when we exhaust all the values for this key.
     */
    cgul_rbtree_node_t tree_node;

    /**
     * Pointer back to the node in the <code>cgul_list</code> that
     * holds this <code>cgul_multimap_node</code>.  This is needed so
     * we can move to the next and previous nodes in the list.  It is
     * also needed so we can efficiently remove nodes from the list
     * without having to do a linear search for them.
     */
    cgul_list_node_t list_node;

    /**
     * Pointer to older node.  Together with <code>younger</code>,
     * <code>older</code> forms a doubly linked list that has nothing
     * to do with sort order; rather, the <code>older</code> node
     * points to the next older node.  If there is no older node,
     * <code>older</code> is <code>NULL</code>.  By calling
     * <code>cgul_multimap_node__get_older()</code>, the user can
     * iterate over the nodes in reverse chronological order.
     */
    cgul_multimap_node_t older;

    /**
     * Pointer to younger node.  Together with <code>older</code>,
     * <code>younger</code> forms a doubly linked list that has
     * nothing to do with sort order; rather, the <code>younger</code>
     * node points to the next younger node.  If there is no younger
     * node, <code>younger</code> is <code>NULL</code>.  By calling
     * <code>cgul_multimap_node__get_younger()</code>, the user can
     * iterate over the nodes in chronological order.
     */
    cgul_multimap_node_t younger;

};

/*
 * Private Methods
 */

/**
 * Create a new <code>cgul_multimap_node</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_multimap_node__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_multimap_node</code> instance
 */
extern CGUL_EXPORT cgul_multimap_node_t
cgul_multimap_node__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the <code>key</code> or <code>value</code> pairs.  Do not
 * try to access <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_multimap_node</code> instance
 */
extern CGUL_EXPORT void
cgul_multimap_node__delete(cgul_multimap_node_t n);

/**
 * This method is shared between <code>cgul_multimap_node__new()</code>
 * and the <code>restructor</code> for the <code>cgul_cache()</code>.
 * It initializes new nodes and re-initializes existing nodes.
 *
 * @param[in] n <code>cgul_multimap_node</code> instance to initialize
 */
extern CGUL_EXPORT void
cgul_multimap_node__initialize(cgul_multimap_node_t n);

CGUL_END_C

#endif  /*  CGUL_MULTIMAP_NODE_P_H  */
