#include "cgul_progress_bar_p.h"
#include "cgul_libc.h"

/**
 * This method converts the <code>tick</code> to its string
 * representation.  The strings returned are all static so the client
 * must not attempt to free them.
 *
 * @param[in] cex c-style exception
 * @param[in] tick tick
 */
static const char*
cgul_progress_bar__simple_tick_to_string(cgul_exception_t* cex,
                                         int tick)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Convert "tick" to its string representation.  Remember that
     * there are 40 ticks in a progress bar, not 100. */
    switch (tick) {
    case 0:
        rv = "0";
        break;
    case 4:
        rv = "10";
        break;
    case 8:
        rv = "20";
        break;
    case 12:
        rv = "30";
        break;
    case 16:
        rv = "40";
        break;
    case 20:
        rv = "50";
        break;
    case 24:
        rv = "60";
        break;
    case 28:
        rv = "70";
        break;
    case 32:
        rv = "80";
        break;
    case 36:
        rv = "90";
        break;
    case 40:
        rv = "100%";
        break;
    default:
        rv = ".";
        break;
    }

 out:
    return rv;
}

/**
 * Draw the <code>CGUL_PROGRESS_BAR__SIMPLE</code> progress bar.  If
 * an error occurs while trying write to <code>bar->f</code>, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] progress raw progress
 * @param[in] distance distance of current progress
 * @param[in] total_distance total distance of range of progress
 */
static void
cgul_progress_bar__draw_simple(cgul_exception_t* cex,
                               cgul_progress_bar_t bar,
                               double progress,
                               double distance,
                               double total_distance)
{
    int i = 0;
    int ticks = 0;
    const char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);
    assert(total_distance != 0.0);

    /* Convert "progress" into ticks.  The equation below basically
     * finds the current percentage for the progress.  Because we have
     * 40 ticks, we then simply scale the fraction to 40.  We throw in
     * the 0.5 to facilitate round off.  Note that we check for
     * division by zero in cgul_progress_bar__update(). */
    ticks = (int)(distance / total_distance * 40.0 + 0.5);
    if (ticks > 40) {
        ticks = 40;
    }

    /* We do not want to display 100% until we are actually finished. */
    if ((ticks == 40) && (progress < bar->max)) {
        ticks = 39;
    }

    /* Draw the extension to the bar. */
    for (i = bar->ticks + 1 ; i <= ticks ; ++i) {

        /* Get the tick as a string. */
        s = cgul_progress_bar__simple_tick_to_string(cex, i);

        /* Write the tick. */
        cgul_libc__fputs(cex, s, bar->f);
        if (*cex) {
            goto out;
        }

        /* Remember the last tick that was drawn.  We do this inside
         * the loop so we will know where we left off if an exception
         * is thrown. */
        bar->ticks = i;

    }

 out:
    return;
}

/**
 * Draw the <code>CGUL_PROGRESS_BAR__COMMON</code> progress bar.  If
 * an error occurs while trying write to <code>bar->f</code>, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] progress raw progress
 * @param[in] distance distance of current progress
 * @param[in] total_distance total distance of range of progress
 * @param[in] before custom text to draw before the progress bar
 * @param[in] after custom text to draw after the progress bar
 */
static void
cgul_progress_bar__draw_common(cgul_exception_t* cex,
                               cgul_progress_bar_t bar,
                               double progress,
                               double distance,
                               double total_distance,
                               const char* before,
                               const char* after)
{
    int i = 0;
    int ticks = 0;
    int percent_as_int = 0;
    double percent_as_double = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);
    assert(total_distance != 0.0);

    /* Find the percent as a double.  We store this separately because
     * we need it in multiple calculations below.  Note that we check
     * for division by zero in cgul_progress_bar__update(). */
    percent_as_double = distance / total_distance;

    /* Find the percent complete as an integer.  We add 0.5 to
     * facilitate round off. */
    percent_as_int = (int)(percent_as_double * 100 + 0.5);

    /* Do not display 100% until we are actually finished. */
    if ((percent_as_int == 100) && (progress < bar->max)) {
        percent_as_int = 99;
    }

    /* Convert "progress" into ticks.  The equation below basically
     * finds the current percentage for the progress.  Because we have
     * 36 ticks, we then simply scale the fraction to 36.  We add 0.5
     * to facilitate round off. */
    ticks = (int)(percent_as_double * 36 + 0.5);
    if (ticks > 36) {
        ticks = 36;
    }

    /* We do not want to display 100% until we are actually finished. */
    if ((ticks == 36) && (progress < bar->max)) {
        ticks = 35;
    }

    /* Start at the beginning of the line. */
    cgul_libc__fputc(cex, '\r', bar->f);
    if (*cex) {
        goto out;
    }

    /* Draw the "before" message if requested by the user. */
    if (before) {
        cgul_libc__fprintf(cex, bar->f, "%s ", before);
        if (*cex) {
            goto out;
        }
    }

    /* Print the leading part that shows the percent and the open
     * bracket.  */
    cgul_libc__fprintf(cex, bar->f, "%3d%% [", percent_as_int);
    if (*cex) {
        goto out;
    }

    /* Draw the bar part. */
    for (i = 0 ; i < ticks ; ++i) {
        cgul_libc__fputc(cex, '#', bar->f);
        if (*cex) {
            goto out;
        }
    }

    /* Print spaces to indicate the remaining work. */
    for ( ; i < 36 ; ++i) {
        cgul_libc__fputc(cex, ' ', bar->f);
        if (*cex) {
            goto out;
        }
    }

    /* Print the close bracket. */
    cgul_libc__fputc(cex, ']', bar->f);
    cgul_libc__fputc(cex, ' ', bar->f);
    if (*cex) {
        goto out;
    }

    /* Draw the "after" message if requested by the user. */
    if (after) {
        cgul_libc__fprintf(cex, bar->f, "%s", after);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

cgul_progress_bar_t
cgul_progress_bar__new(cgul_exception_t* cex)
{
    cgul_progress_bar_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate space for the new object. */
    rv = (cgul_progress_bar_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Initialize. */
    rv->min = 0.0;
    rv->max = 100.0;
    rv->ticks = -1;
    rv->style = CGUL_PROGRESS_BAR__COMMON;
    rv->f = stdout;

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_progress_bar__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_progress_bar__delete(cgul_progress_bar_t bar)
{
    if (bar) {
        free(bar);
    }
}

void
cgul_progress_bar__reset(cgul_exception_t* cex,
                         cgul_progress_bar_t bar)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    bar->ticks = -1;

 out:
    return;
}

cgul_progress_bar__style_t
cgul_progress_bar__get_style(cgul_exception_t* cex,
                             cgul_progress_bar_t bar)
{
    cgul_progress_bar__style_t rv = CGUL_PROGRESS_BAR__INVALID;
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    rv = bar->style;

 out:
    return rv;
}

void
cgul_progress_bar__set_style(cgul_exception_t* cex,
                             cgul_progress_bar_t bar,
                             cgul_progress_bar__style_t style)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    if (style < CGUL_PROGRESS_BAR__INVALID) {
        bar->style = style;
    } else {
        *cex = cgul_exception__new(
                   "cgul_progress_bar__set_style: Invalid style.");
        goto out;
    }

 out:
    return;
}

double
cgul_progress_bar__get_min(cgul_exception_t* cex,
                           cgul_progress_bar_t bar)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    rv = bar->min;

 out:
    return rv;
}

void
cgul_progress_bar__set_min(cgul_exception_t* cex,
                           cgul_progress_bar_t bar,
                           double min)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    bar->min = min;

 out:
    return;
}

double
cgul_progress_bar__get_max(cgul_exception_t* cex,
                           cgul_progress_bar_t bar)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    rv = bar->max;

 out:
    return rv;
}

void
cgul_progress_bar__set_max(cgul_exception_t* cex,
                           cgul_progress_bar_t bar,
                           double max)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    bar->max = max;

 out:
    return;
}

FILE*
cgul_progress_bar__get_file(cgul_exception_t* cex,
                            cgul_progress_bar_t bar)
{
    FILE* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    rv = bar->f;

 out:
    return rv;
}

void
cgul_progress_bar__set_file(cgul_exception_t* cex,
                            cgul_progress_bar_t bar,
                            FILE* f)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);
    assert(f);

    bar->f = f;

 out:
    return;
}

void
cgul_progress_bar__update(cgul_exception_t* cex,
                          cgul_progress_bar_t bar,
                          double progress,
                          const char* before,
                          const char* after)
{
    double distance = 0.0;
    double total_distance = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bar);

    /* Bounds check "progress." */
    if (progress < bar->min) {
        progress = bar->min;
    } else if (progress > bar->max) {
        progress = bar->max;
    }

    /* Determine the distance of progress. */
    distance = progress - bar->min;

    /* Make sure the minimum and maximum values are sane. */
    if (bar->min >= bar->max) {
        *cex = cgul_exception__new(
                   "cgul_progress_bar__update: minimum value cannot be "
                   "greater than or equal to maximum value.");
        goto out;
    }

    /* Determine the total distance for the range. */
    total_distance = bar->max - bar->min;
    if (total_distance == 0.0) {
        *cex = cgul_exception__new(
                   "cgul_progress_bar__update: division by zero");
        goto out;
    }

    /* Draw the bar depending on the style. */
    switch (bar->style) {

    case CGUL_PROGRESS_BAR__SIMPLE:
        cgul_progress_bar__draw_simple(cex,
                                       bar,
                                       progress,
                                       distance,
                                       total_distance);
        break;

    case CGUL_PROGRESS_BAR__COMMON:
        cgul_progress_bar__draw_common(cex,
                                       bar,
                                       progress,
                                       distance,
                                       total_distance,
                                       before,
                                       after);
        break;

    default:
        *cex = cgul_exception__new("cgul_progress_bar__update: "
                                   "Invalid progress bar style: %d",
                                   (int)bar->style);
        goto out;
    }

    /* Make sure the user sees the new output in a timely fashion. */
    if (bar->f) {
        fflush(bar->f);
    }

 out:
    return;
}
