#include "cgul_random_lcg32_p.h"
#include "cgul_libc.h"

/*
 * This class implements two 32-bit LCGs both of which have the
 * following form:
 *
 *     next = (a * curr + c) % m
 *
 * It is well-known that the lower bits of the LCG output are not
 * random having small periods of 2^n where n is the bit number.  For
 * example, the first bit has period 2^1 or 2, i.e., it just toggles.
 *
 * To help mitigate this problem, this program uses two LCGs reversing
 * the bits of the second LCG before combining it with the first LCG.
 * Also, this class skips zero whenever the second LCG produces it.
 * This is so that the periods of the first and second LCGs are
 * relatively prime.  The period for the first LCG will be 2^32.  The
 * period for the second LCG will be (2^32 - 1).  Thus, the period for
 * the combined LCG will be 2^32 * (2^32 - 1) or (2^64 - 2^32) when
 * combined as follows:
 *
 *     Z = (X + Y) mod 2^32
 */

/**
 * The "a" value for the first LCG.
 */
static cgul_uint32_t cgul_random_lcg32__a1 = 177435565U;

/**
 * The "c" value for the first LCG.
 */
static cgul_uint32_t cgul_random_lcg32__c1 = 156788809U;

/**
 * The "a" value for the second LCG.
 */
static cgul_uint32_t cgul_random_lcg32__a2 = 160912581U;

/**
 * The "c" value for the second LCG.
 */
static cgul_uint32_t cgul_random_lcg32__c2 = 92392009U;

/**
 * This is the odd constant in the Weyl Generator often denoted as omega.
 */
static cgul_uint32_t cgul_random_lcg32__omega = 2654435769U;

/**
 * The size of the vector that will hold the PRNs until they are
 * randomly selected.  This must be a power of 2 because the code
 * below just masks off the high-order bits.  Again, we can do this
 * because this LCG's low-order bits are randomized.  If you change
 * this value, you must also change v_mask immediately below.
 */
static size_t cgul_random_lcg32__v_size = 256;

/**
 * The mask used to select a particular index into the vector that
 * holds the PRNs waiting to be randonly selected.  If you change this
 * value, you must change v_size to match.
 */
static size_t cgul_random_lcg32__v_mask = 0xff;

/**
 * Iterate both LCGs and the Weyl Generator composed as part of
 * <code>r</code>.  Reverse the bits in the PRN from the second LCG.
 * Then, combine all the PRNs returning the result in <code>x</code>.
 * We resort to a #define here because the "inline" keyword is not
 * portable.
 *
 * @param[in] r <code>cgul_random_lcg32</code> instance
 * @param[out] x PRN combined from the first and second LCGs
 */
#define CGUL_RANDOM_LCG32__COMBINE(r, x) \
do { \
    cgul_uint32_t tmp2 = 0; \
\
    /* Generate the next PRN for both LCGs.  Remember that it is \
     * necessary to skip a few values for some of the PRNs so that \
     * their periods are relatively prime thereby maximizing the total \
     * period.  The period for lcg1 will be (2^32 - 1), and the period \
     * of lcg2 will be (2^32 - 3). */ \
    do { \
        r->lcg1 = cgul_random_lcg32__a1 * r->lcg1 + cgul_random_lcg32__c1; \
    } while (r->lcg1 == 0x0); \
\
    do { \
        r->lcg2 = cgul_random_lcg32__a2 * r->lcg2 + cgul_random_lcg32__c2; \
    } while ((r->lcg2 == 0x0) \
          || (r->lcg2 == 0x9e82eed9) \
          || (r->lcg2 == 0xa869da3f)); \
\
    /* Reverse the bits of the second PRN. */ \
    tmp2 = r->lcg2; \
    tmp2 = ((tmp2 >>  1) & 0x55555555) | ((tmp2 & 0x55555555) <<  1); \
    tmp2 = ((tmp2 >>  2) & 0x33333333) | ((tmp2 & 0x33333333) <<  2); \
    tmp2 = ((tmp2 >>  4) & 0x0F0F0F0F) | ((tmp2 & 0x0F0F0F0F) <<  4); \
    tmp2 = ((tmp2 >>  8) & 0x00FF00FF) | ((tmp2 & 0x00FF00FF) <<  8); \
    tmp2 = ((tmp2 >> 16))              | ((tmp2)              << 16); \
\
    /* Combine all PRNs. */ \
    x = r->lcg1 + tmp2 + (r->weyl += cgul_random_lcg32__omega); \
\
} while (0)

cgul_random_lcg32_t
cgul_random_lcg32__new(cgul_exception_t* cex)
{
    cgul_uint32_t seed = 0;
    cgul_random_lcg32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a new object seeded with time(). */
    seed = (cgul_uint32_t)time(NULL);
    rv = cgul_random_lcg32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

cgul_random_lcg32_t
cgul_random_lcg32__new_from_seed(cgul_exception_t* cex,
                                 cgul_uint32_t seed)
{
    unsigned int seed_count = 0;
    cgul_uint32_t* seeds = NULL;
    cgul_random_lcg32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the seed array and get its size. */
    cgul_random_lcg32__allocate_seed_array(cex, &seeds, &seed_count);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (seed_count != 2) {
        *cex = cgul_exception__new("cgul_random_lcg32__new_from_seed: "
                                   "Internal Error: seed count is not 2.");
        goto out;
    }

    /* Initialize the seed array. */
    seeds[0] = seed;
    seeds[1] = 2096027753U;

    /* Create a new object seeding the first LCG with the user's seed
     * and the second LCG with a constant. */
    rv = cgul_random_lcg32__new_from_seed_array(cex, seeds);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (seeds) {
        free(seeds);
    }

    return rv;
}


cgul_random_lcg32_t
cgul_random_lcg32__new_from_seed_array(cgul_exception_t* cex,
                                       cgul_uint32_t* seeds)
{
    size_t i = 0;
    cgul_random_lcg32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(seeds);

    /* Allocate the object. */
    rv = (cgul_random_lcg32_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the vector that will hold the PRNs until they are
     * randomly selected. */
    rv->v = (cgul_uint32_t*)cgul_libc__calloc(cex,
                                              cgul_random_lcg32__v_size,
                                              sizeof(*rv->v));
    if (*cex) {
        goto out;
    }

    /* Seed the PRNGs.  Note that the comments for this method
     * recommend that the caller use allocate_seed_array() to allocate
     * "seeds" so it should be the correct size. */
    rv->lcg1 = seeds[0];
    rv->lcg2 = seeds[1];
    rv->weyl = 166947538U;

    /* Fill the vector with PRNs waiting to be randomly selected. */
    for (i = 0 ; i < cgul_random_lcg32__v_size ; ++i) {
        CGUL_RANDOM_LCG32__COMBINE(rv, rv->v[i]);
    }

    /* Establish the next index into v. */
    CGUL_RANDOM_LCG32__COMBINE(rv, rv->v_index);
    rv->v_index &= cgul_random_lcg32__v_mask;

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_random_lcg32__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_random_lcg32__delete(cgul_random_lcg32_t r)
{
    if (r) {

        /* v */
        if (r->v) {
            free(r->v);
        }

        /* Free the object itself. */
        free(r);
    }
}

void
cgul_random_lcg32__allocate_seed_array(cgul_exception_t* cex,
                                       cgul_uint32_t** v,
                                       unsigned int* v_count)
{
    size_t count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count);

    /* Number of elements in the array. */
    count = 2;  /* Number of LCGs*/

    /* Allocate the seed array. */
    if (v) {
        *v = (cgul_uint32_t*)cgul_libc__calloc(cex, count, sizeof(**v));
        if (*cex) {
            goto out;
        }
    }

    /* Return the count. */
    if (v_count) {
        *v_count = count;
    }

 out:
    return;
}

cgul_uint32_t
cgul_random_lcg32__next(cgul_exception_t* cex,
                        cgul_random_lcg32_t r)
{
    cgul_uint32_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);

    /* Randomly select the next PRN from the vector.  Note that
     * v_index was established on the previous iteration. */
    rv = r->v[r->v_index];

    /* Replace the PRN we randomly selected with a new PRN. */
    CGUL_RANDOM_LCG32__COMBINE(r, r->v[r->v_index]);

    /* Establish the next index into v. */
    r->v_index = rv & cgul_random_lcg32__v_mask;

 out:
    return rv;
}

cgul_uint32_t
cgul_random_lcg32__next_in_range(cgul_exception_t* cex,
                                 cgul_random_lcg32_t r,
                                 cgul_uint32_t range_max)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t upper_limit = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (range_max == 0) {
        *cex = cgul_exception__new("cgul_random_lcg32__next_in_range: "
                                   "exclusive range_max is zero");
        goto out;
    }

    /* Use integer arithmetic to find the largest multiple of
     * range_max that is less than the maximum random value that could
     * be produced by the PRNG. */
    upper_limit = 0xffffffff - (0xffffffff % range_max);

    /* Iterate the PRNG until it returns a raw value that is less than
     * the largest multiple of range_max.  In effect, we are just
     * discarding the upper few PRNs so that the mod operation is
     * fair.
     *
     * Note that upper_limit is an _exclusive_ upper bound because if
     * it were included (upper_limit % range) would necessarily be
     * zero (because "upper_limit" is defined to be a multiple of
     * "range") which would cause there to be one more potential zero
     * than any other value that could be returned.  This is why
     * "greater than or equal to" is used rather than just "greater
     * than." */
    do {
        rv = cgul_random_lcg32__next(cex, r);
    } while (rv >= upper_limit);

    /* Return a PRN in the user's range that is bias-free.  Note that
     * this class combines the LCGs and the Weyl Generator in such a
     * way as to remove the non-randomness in the lower bits making it
     * ok to use the mod operator. */
    rv %= range_max;

 out:
    return rv;
}

double
cgul_random_lcg32__next_in_unit(cgul_exception_t* cex,
                                cgul_random_lcg32_t r)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next PRN and divide by 2^32 by multiplying by the
     * precomputed reciprocal of 2^32. */
    rv = cgul_random_lcg32__next(cex, r) * CGUL_INVERSE_2_32;

 out:
    return rv;
}
