#include "cgul_rbtree_p.h"
#include "cgul_rbtree_node_p.h"
#include "cgul_rbtree_node_with_indexing_p.h"
#include "cgul_libc.h"

/************************************************************************
 * General Comments
 ************************************************************************/

/* When working with the code below, you should always keep four
 * different ways a single 2-3-4 node can map to 1, 2, or 3 nodes in
 * the red-black tree.
 *
 * First, you can have a 2-3-4 node that holds a single key/value pair
 * and has 2 children.  This maps to the red-black tree as follows
 * where N is black:
 *
 *              |
 *              |
 *            +---+
 *            | N |
 *            +---+
 *           /     \
 *          /       \
 *         1         2
 *
 * Second and third, you can have a 2-3-4 node that holds two
 * key/value pairs and has 3 children.  This maps to the red-black
 * tree in two slightly different (but equivalent) ways:
 *
 *                  |                         |             
 *                  |                         |             
 *                +----+                    +----+          
 *                | N2 |                    | N1 |          
 *                +----+                    +----+          
 *         (red) /      \       or         /      \ (red)   
 *         +----+        \                /        +----+   
 *         | N1 |         3              1         | N2 |   
 *         +----+                                  +----+   
 *        /      \                                /      \  
 *       /        \                              /        \ 
 *      1          2                            2          3
 *
 * Lastly, you can have a 2-3-4 node that holds 3 key/value pairs and
 * has 4 children.  This maps to the red-black tree as follows where
 * N2 is black and N1 and N3 are red.
 *
 *                               |
 *                               |
 *                             +----+
 *                             | N2 |
 *                             +----+
 *                      (red) /      \ (red)
 *                      +----+        +----+
 *                      | N1 |        | N3 |
 *                      +----+        +----+
 *                     /      \      /      \
 *                    /        \    /        \
 *                   1          2  3          4
 *
 * There is just no getting around the fact that for most things, you
 * have to explicitly code for each of the four diagrams above because
 * they all required different handling.
 */

/************************************************************************
 * Tree Utilities
 ************************************************************************/

/**
 * Rotate the red-black tree right.  This will raise <code>n</code> up
 * one level, and the former parent of <code>n</code> will drop down
 * to become the right child of <code>n</code>.  The former right
 * child of <code>n</code> will drop down to become the left child of
 * the former parent of <code>n</code>.  This operation preserves
 * sorting order; however, the caller has to correct for red-black
 * color.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to rotate
 */
static void
cgul_rbtree__rotate_right_ignoring_color(cgul_exception_t* cex,
                                         cgul_rbtree_t t,
                                         cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    if (n->parent) {
        /* n is not the root node so it makes sense to rotate. */
        cgul_rbtree_node_t c = n->right;
        cgul_rbtree_node_t p = n->parent;
        cgul_rbtree_node_t gp = p->parent;  /* grandparent */
        if (gp) {
            if (gp->left == p) {
                gp->left = n;
            } else {
                assert(gp->right == p);
                gp->right = n;
            }
        }
        n->parent = p->parent;
        n->right = p;
        p->parent = n;
        p->left = c;
        if (c) {
            c->parent = p;
        }

        /* Update the counts if the tree is indexed. */
        if (t->is_indexed) {
            /* Down cast. */
            cgul_rbtree_node_with_indexing_t nwi =
                (cgul_rbtree_node_with_indexing_t)n;
            cgul_rbtree_node_with_indexing_t pwi =
                (cgul_rbtree_node_with_indexing_t)p;
            cgul_rbtree_node_with_indexing_t cwi =
                (cgul_rbtree_node_with_indexing_t)c;
            /* Update. */
            nwi->rcount += pwi->rcount + 1;
            if (cwi) {
                pwi->lcount = cwi->lcount + cwi->rcount + 1;
            } else {
                pwi->lcount = 0;
            }
        }

        /* We have a new root if after the rotate n's parent is NULL. */
        if (n->parent == NULL) {
            t->root = n;
        }
    }

 out:
    return;
}

/**
 * Rotate the red-black tree left.  This will raise <code>n</code> up
 * one level, and the former parent of <code>n</code> will drop down
 * to become the left child of <code>n</code>.  The former left child
 * of <code>n</code> will drop down to become the right child of the
 * former parent of <code>n</code>.  This operation preserves sorting
 * order; however, the caller has to correct for red-black color.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to rotate
 */
static void
cgul_rbtree__rotate_left_ignoring_color(cgul_exception_t* cex,
                                        cgul_rbtree_t t,
                                        cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    if (n->parent) {
        /* n is not the root node so it makes sense to rotate. */
        cgul_rbtree_node_t c = n->left;
        cgul_rbtree_node_t p = n->parent;
        cgul_rbtree_node_t gp = p->parent;  /* grandparent */
        if (gp) {
            if (gp->left == p) {
                gp->left = n;
            } else {
                assert(gp->right == p);
                gp->right = n;
            }
        }
        n->parent = p->parent;
        n->left = p;
        p->parent = n;
        p->right = c;
        if (c) {
            c->parent = p;
        }

        /* Update the counts if the tree is indexed. */
        if (t->is_indexed) {
            /* Down cast. */
            cgul_rbtree_node_with_indexing_t nwi =
                (cgul_rbtree_node_with_indexing_t)n;
            cgul_rbtree_node_with_indexing_t pwi =
                (cgul_rbtree_node_with_indexing_t)p;
            cgul_rbtree_node_with_indexing_t cwi =
                (cgul_rbtree_node_with_indexing_t)c;
            /* Update. */
            nwi->lcount += pwi->lcount + 1;
            if (cwi) {
                pwi->rcount = cwi->lcount + cwi->rcount + 1;
            } else {
                pwi->rcount = 0;
            }
        }

        /* We have a new root if after the rotate n's parent is NULL. */
        if (n->parent == NULL) {
            t->root = n;
        }
    }

 out:
    return;
}

/**
 * A full 2-3-4 node is a node in the corresponding 2-3-4 tree that
 * holds 3 key/value pairs and (necessarily) pointers to 4 children.
 * This type of node is full because there are no more slots for
 * inserting key/value pairs.
 *
 * Because we split every full 2-3-4 node during an insert operation
 * as we traverse downward, you are guaranteed that the parent 2-3-4
 * node has room if you want to insert a new key/value pair.
 * Unfortunately, when you are inserting a new key/value pair, you
 * always insert at the bottom of the tree (of course).  Thus, if we
 * are to pass key/value pairs up the tree (in order to keep it
 * balanced -- which is the whole point), we must pass up key/value
 * pairs that have already been inserted.  So, wait until we have a
 * full node and "split" it.  The "split" operation passes the middle
 * key/value pair up to the parent and splits the remaining two keys
 * into two independent 2-3-4 nodes that get attached to the parent
 * 2-3-4 node on either side of the middle node that was passed up.
 *
 * Most of the time, splitting causes key/value pairs to trickle up
 * the tree; however, if you have to split the root node, you actually
 * end up adding a new level to your tree because the root node has no
 * parent.  The middle key/value pair then becomes the new root node.
 * It is a 2-3-4 node with two children which are the other two
 * key/value pairs each also 2-3-4 nodes with just two children.
 *
 * As usual, mapping what happens in a 2-3-4 tree to what happens in
 * the corresponding red-black tree is complicated, but if you
 * understand the mapping, it makes it much easier to implement.  At
 * this point, I recommend getting out a sheet of paper and drawing
 * all the possible combinations of splitting a 3 key/value node and
 * passing up the middle value to parent nodes that have 1 or 2
 * key/value pairs.  Here's an example:
 *
 * Assume you have the following 2-3-4 tree:
 *
 *               |
 *               |
 *        +-------------+
 *        |  A       B  |
 *        +-------------+
 *       /       |       \
 *      1        2        +-----------+  
 *                        | C   D   E |  
 *                        +-----------+  
 *                       /    |   |    \ 
 *                      3     4   5     6
 *
 *               (Diagram #1)
 *
 * After splitting the C-D-E node in diagram #1, you are left with the
 * following 2-3-4 tree:
 *
 *                   |
 *                   |
 *        +---------------------+
 *        |  A       B       D  |
 *        +---------------------+
 *       /       |       |       \
 *      1        2     +---+      +---+
 *                     | C |      | E |
 *                     +---+      +---+
 *                    /     \    /     \
 *                   3       4  5       6
 *
 *                        (Diagram #2)
 *
 * The trick is mapping the above two diagrams of 2-3-4 trees to the
 * the red-black tree implementation of 2-3-4 trees.  First, you need
 * to map diagram #1 (above) to a red-black tree.  Then you need to
 * map diagram #2 (above) to a red-black tree and see how they differ.
 * You then need to code the differences.
 *
 * You can draw diagram #1 in two ways.  The first red-black
 * representation of diagram #1 has A as the black (top) node of the
 * 2-3-4 node that holds both A and B:
 *
 *        |
 *        |
 *      +---+
 *      | A |
 *      +---+
 *     /    \ (red)
 *    1      +---+
 *           | B |
 *           +---+
 *          /     \
 *         2       \
 *                  \
 *                   +---+
 *                   | D |
 *                   +---+
 *            (red) /     \ (red)
 *             +---+       +---+
 *             | C |       | E |
 *             +---+       +---+
 *            /     \     /     \
 *           3       4   5       6
 *
 *                (Diagram #3)
 *
 * The second representation of diagram #1 has B as the black (top) node:
 *
 *                       |
 *                       |
 *                     +---+
 *                     | B |
 *                     +---+
 *                    /     \ 
 *             (red) /       \
 *              +---+         \
 *              | A |          \
 *              +---+           \
 *             /     \           \
 *            1       2           +---+        
 *                                | D |        
 *                                +---+        
 *                         (red) /     \ (red) 
 *                          +---+       +---+  
 *                          | C |       | E |  
 *                          +---+       +---+  
 *                         /     \     /     \ 
 *                        3       4   5       6
 *
 *                             (Diagram #4)
 *
 * Now, lets see how diagram #2 maps to the red-black tree:
 *
 *                      |
 *                      |
 *                    +---+
 *                    | B |
 *                    +---+
 *            (red) /       \ (red)
 *            +---+           +---+
 *            | A |           | D |
 *            +---+           +---+
 *           /     \         /     \
 *          1       2   +---+       +---+
 *                      | C |       | E |
 *                      +---+       +---+
 *                     /     \     /     \
 *                    3       4   5       6
 *
 *                           (Diagram #5)
 *
 * So, the question is how do you go from diagram #3 to diagram #5 and
 * how do you also go from diagram #4 to diagram #5.  The first thing
 * to notice is that if you rotate B up in diagram #3 and make B black
 * and A red, you have not altered the tree any.  In fact, you have
 * just converted diagram #3 into diagram #4.  Now, diagram #4 is
 * identical to diagram #5 except for the colors which are simple
 * enough to rearrange.
 *
 * Remember, the above example is just one of the many ways things can
 * be arranged before a split.
 *
 * You can pass in any node in the red-black tree that is part of the
 * corresponding 2-3-4 tree node.  This function will then split that
 * node as described above.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to split
*/
static void
cgul_rbtree__split_234(cgul_exception_t* cex,
                       cgul_rbtree_t t,
                       cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t p = NULL;   /* parent */
    cgul_rbtree_node_t gp = NULL;  /* grandparent */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

#ifndef NDEBUG
    {
        /* You only split full nodes. */
        int key_count_234 = 0;
        cgul_rbtree_node__get_234_key_count(cex, n, key_count_234);
        assert(key_count_234 == 3);
    }
#endif

    /* If this is a red node, it is part of a larger node on the
     * corresponding 2-3-4 tree.  Thus, you need to deal with
     * the parent. */
    if (n->red) {
        n = n->parent;
    }

    assert(n);
    assert(!n->red);

    /* The red-black tree parent of n. */
    p = n->parent;

    if (p == NULL) {
        /* We have to split the _root_ node of the 2-3-4 tree.  This
         * is just a simple color change because there is no parent to
         * pass the middle key/value pair up to. */
        n->left->red = 0;
        n->right->red = 0;
        goto out;
    }

    /* First, take care of the easy case.  If n's parent is black
     * (regardless of whether the corresponding 2-3-4 tree's parent
     * node is a 2-child or 3-child node), it means you can split the
     * current 2-3-4 node with a simple color flip. */
    if (!n->parent->red) {
        n->red = 1;
        n->left->red = 0;
        n->right->red = 0;
        goto out;
    }

    /*
     * If you are here, we know n's parent is red.  We also know that
     * we have already split n's parent as we descended through it.
     * Thus we do not have to write any code to handle the case where
     * n's parent is part of a 2-3-4 node that holds 3 key/value
     * pairs.  The work that remains is only to deal with the case
     * where n's parent is part of a 2-3-4 node that holds exactly 2
     * key/value pairs.
     */

    /* At this point, it is safe to assume that n has a grandparent in
     * the red-black tree.  This is true because the above if()
     * statement has take care of all the cases where n is attached to
     * a black node.  Thus, n must be attached to a red node, and all
     * red nodes must be attached to black nodes. */
    gp = p->parent;
    assert(gp);

    /* Next, deal with the case where a single rotation is required to
     * balance the corresponding 2-3-4 tree's parent node to
     * accommodate the new key/value pair we are passing up to it.
     * This happens when the middle key in n is less than both keys or
     * greater than both keys in n's parent. */
    if ((n == p->left) && (p == gp->left)) {
        /* n's key is less than all the keys in the 2-3-4 parent's node. */
        cgul_rbtree__rotate_right_ignoring_color(cex, t, p);
        p->red = 0;
        n->red = 1;
        gp->red = 1;
        n->left->red = 0;
        n->right->red = 0;
        goto out;
    }
    if ((n == p->right) && (p == gp->right)) {
        /* n's key is greater than all the keys in the 2-3-4 parent's node. */
        cgul_rbtree__rotate_left_ignoring_color(cex, t, p);
        p->red = 0;
        n->red = 1;
        gp->red = 1;
        n->left->red = 0;
        n->right->red = 0;
        goto out;
    }

    /* Lastly, deal with the case where a double rotation is required
     * to balance the corresponding 2-3-4 tree's parent node to
     * accommodate the new key/value pair we are passing up to it.
     * This happens when the middle key of the current node should be
     * the middle key of the 2-3-4 parent's node. */
    if ((n == p->left) && (p == gp->right)) {
        /* While we still have pointers to n's previously red children ...*/
        n->left->red = 0;
        n->right->red = 0;
        /* Rotate n up to the top of the red-black implementation of
         * the parents 2-3-4 node. */
        cgul_rbtree__rotate_right_ignoring_color(cex, t, n);
        cgul_rbtree__rotate_left_ignoring_color(cex, t, n);
        n->red = 0;
        p->red = 1;
        gp->red = 1;
        goto out;
    }
    if ((n == p->right) && (p == gp->left)) {
        /* While we still have pointers to n's previously red children ...*/
        n->right->red = 0;
        n->left->red = 0;
        /* Rotate n up to the top of the red-black implementation of
         * the parents 2-3-4 node. */
        cgul_rbtree__rotate_left_ignoring_color(cex, t, n);
        cgul_rbtree__rotate_right_ignoring_color(cex, t, n);
        n->red = 0;
        p->red = 1;
        gp->red = 1;
        goto out;
    }

    /* You should not reach here. */
    assert(0);

 out:
    return;
}

/**
 * This function descends the cgul_rbtree <code>t</code> until it
 * finds the correct leaf node for inserting a new key/value pair with
 * key equal to <code>key</code>.  In order to only call the
 * comparison function once for each node during an insert operation,
 * <code>compare</code> is returned holding the final comparison
 * between the <code>key</code> you pass in and the key for the leaf
 * node that is returned .
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] key key
 * @param[out] compare result of the final comparison between
 * <code>key</code> and the key for the leaf node that is returned
 * @return leaf node where insertion should occur
 */
static cgul_rbtree_node_t
cgul_rbtree__find_insertion_leaf_node(cgul_exception_t* cex,
                                      cgul_rbtree_t t,
                                      const void* key,
                                      int* compare)
{
    int compare_rv = 0;
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(t->root);
    
    /* As with other trees, you can obviously only insert new nodes at
     * the bottom of the tree.  So, walk down to the bottom. */
    rv = t->root;
    for ( ;; ) {

        /* Along the way to the bottom, if you pass through a 2-3-4
         * tree node that is full, split it.  You only need to check
         * the black nodes. */
        if (!rv->red) {
            if (cgul_rbtree_node__get_234_key_count_is_exactly_three(cex,
                                                                      rv))
            {
                cgul_rbtree__split_234(cex, t, rv);
            }
        }

        /* Compare keys. */
        compare_rv = (*t->compare)(key, rv->key);

        /* New key is less than key at current node. */
        if (compare_rv < 0) {
            if (rv->left == NULL) {
                break;
            }
            rv = rv->left;
            continue;
        }

        /* New key is greater than key at current node. */
        if (compare_rv > 0) {
            if (rv->right == NULL) {
                break;
            }
            rv = rv->right;
            continue;
        }

        /* New key is equal to key at current node. */
        break;
    }

    *compare = compare_rv;

 out:
    return rv;
}


/**
 * 3-child nodes in the 2-3-4 tree can be represented in either of two
 * ways in the red-black tree:
 *
 *             |                             |   
 *             |                             |   
 *           +---+                         +---+ 
 *           | B |                         | A | 
 *           +---+                         +---+ 
 *          /     \                       /     \
 *     +---+                                     +---+
 *     | A |                                     | B |
 *     +---+                                     +---+
 *    /     \                                   /     \
 *
 * This routing will toggle between the two representations.  The
 * caller is responsible for making sure n is a three child node.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to realign
 */
static void
cgul_rbtree__realign_234_three_child_node(cgul_exception_t* cex,
                                          cgul_rbtree_t t,
                                          cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

#ifndef NDEBUG
    {
        /* A 2-3-4 node with 3 children has 2 keys. */
        int key_count_234 = 0;
        cgul_rbtree_node__get_234_key_count(cex, n, key_count_234);
        assert(key_count_234 == 2);
    }
#endif

    /* If this is a red node, it is part of a larger node on the
     * corresponding 2-3-4 tree.  We need the black node. */
    if (n->red) {
        n = n->parent;
        assert(n);
    }

    assert(!n->red);

    /* Toggle the alignment. */
    if (n->left && n->left->red) {
        n->red = 1;
        n->left->red = 0;
        cgul_rbtree__rotate_right_ignoring_color(cex, t, n->left);
    } else if (n->right) {
        assert(n->right->red);
        n->red = 1;
        n->right->red = 0;
        cgul_rbtree__rotate_left_ignoring_color(cex, t, n->right);
    }

 out:
    return;
}

/**
 * Remember that a red-black tree is just one way to implement a 2-3-4
 * tree.  So, when you insert your new node, you have to take into
 * account what you are doing to the 2-3-4 tree.
 *
 * The value of <code>n</code> that you pass in should point to one of
 * the red-black nodes that comprise the 2-3-4 leaf node where you
 * need to insert your key/value pair.  Below you will find Diagrams
 * #6, #7, and #8.
 *
 * Diagram #6 is the diagram of how the red-black tree models a
 * 2-child 2-3-4 node.  Inserting into this diagram is a simple matter
 * of adding the child and making the child red.
 *
 * Diagram #7 and diagram #8 are more complicated.  These diagrams
 * show how the red-black tree models 3-child 2-3-4 nodes.  (Because
 * you split on the way down, you do not have to worry about having to
 * insert a new node into a 4-child 2-3-4 node.)
 *
 * If <code>n</code> ends up pointing to the "P" (parent) node and if
 * the new node needs to be inserted in the "1" slot of either diagram
 * #7 or #8, then insertion is just like it is for diagram #6.  You
 * add the child to the "1" slot and change its color to red.  This
 * will produce a 4-child 2-3-4 node.
 *
 * If <code>n</code> ends up pointing to the "C" (child) node in
 * diagrams #7 and #8, you have to be more careful.  The problem is
 * that after the insertion you are going to have a 4-child 2-3-4
 * node, and a 4-child 2-3-4 node is represented in a red-black tree
 * as one black parent with 2 red children; it has no other
 * representation.  If you blindly insert the new node in slots 2 or 3
 * in diagrams #7 and #8 without doing anything else, you end up with
 * "P" having a red grandchild (which throws things out of balance).
 *
 * You can solve this problem by inserting the new node as a child of
 * "C".  If the key for "C" lies between the key for "P" and the key
 * for the new node, you can simply rotate "C" up and make the new
 * node red.  Otherwise, the key for the new node lies between "P" and
 * "C".  In this case, you need to rotate the new node up twice and
 * make "P" red.  (<code>cgul_rbtree_node__new()</code> creates new
 * nodes that are black by default.)
 *
 *
 *                            |
 *                            |
 *                          +---+   
 *                          | C |   
 *                          +---+   
 *                         /     \  
 *                        /       \ 
 *                       /         \
 *                      1           2
 *                                   
 *                       (Diagram #6)
 *
 *
 *           |                                 |
 *           |                                 |
 *         +---+                             +---+     
 *         | P |                             | P |     
 *         +---+                             +---+     
 *        /     \                           /     \    
 *       /       \ (red)             (red) /       \   
 *      /         \                       /         \  
 *     1           +---+             +---+           1 
 *                 | C |             | C |             
 *                 +---+             +---+             
 *                /     \           /     \            
 *               /       \         /       \           
 *              2         3       3         2
 *
 *       (Diagram #7)                    (Diagram #8)
 *
 * You pass in the node <code>n</code> that you want to insert by
 * attaching to parent <code>p</code>.  You also need to pass in the
 * <code>compare</code> value returned by
 * <code>cgul_rbtree__find_insertion_leaf_node()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] p parent node of <code>n</code>
 * @param[in] node leaf node where new child is to be inserted
 * @param[in] comparison result between new key and leaf node's key
 */
static void
cgul_rbtree__insert_new_child_into_leaf_node(cgul_exception_t* cex,
                                             cgul_rbtree_t t,
                                             cgul_rbtree_node_t p,
                                             cgul_rbtree_node_t n,
                                             int compare)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(p);
    assert(n);
    assert(compare != 0);

    /* Handle the easy case where we are either inserting into either
     * slot "1" or slot 2 in diagram #6 or into slot "1" on diagrams
     * #7 or #8. */
    if (!p->red && (compare < 0) && (p->left == NULL)) {
        p->left = n;
        n->parent = p;
        n->red = 1;
        goto out;
    }
    if (!p->red && (compare > 0) && (p->right == NULL)) {
        p->right = n;
        n->parent = p;
        n->red = 1;
        goto out;
    }

    /* p must be red here.  Thus, p must have a parent. */
    assert(p->red);
    assert(p->parent);

    /* Handle the cases where you are inserting into slot 3 on either
     * diagram #7 or #8.  These are the single rotation cases. */
    if ((compare > 0) && (p == p->parent->right)) {
        /* Insert the new node. */
        p->right = n;
        n->parent = p;
        n->red = 1;
        /* Rotate n left. */
        cgul_rbtree__rotate_left_ignoring_color(cex, t, p);
        p->red = 0;
        p->left->red = 1;
        goto out;
    }
    if ((compare < 0) && (p == p->parent->left)) {
        /* Insert the new node. */
        p->left = n;
        n->parent = p;
        n->red = 1;
        /* Rotate n right. */
        cgul_rbtree__rotate_right_ignoring_color(cex, t, p);
        p->red = 0;
        p->right->red = 1;
        goto out;
    }

    /* Handle the cases where you are inserting into slot 2 on either
     * diagram #7 or #8.  These are the double rotation cases. */
    if (compare < 0) {
        /* Insert the new node.  It will be black after the double rotation. */
        p->left = n;
        n->parent = p;
        /* Rotate n right then rotate left. */
        cgul_rbtree__rotate_right_ignoring_color(cex, t, n);
        cgul_rbtree__rotate_left_ignoring_color(cex, t, n);
        n->left->red = 1;
        n->right->red = 1;
        goto out;
    }
    if (compare > 0) {
        /* Insert the new node.  It will be black after the double rotation. */
        p->right = n;
        n->parent = p;
        /* Rotate n right then rotate left. */
        cgul_rbtree__rotate_left_ignoring_color(cex, t, n);
        cgul_rbtree__rotate_right_ignoring_color(cex, t, n);
        n->left->red = 1;
        n->right->red = 1;
        goto out;
    }

    /* You should never reach here. */
    assert(0);

 out:
    if (*cex == NULL) {
        /* Increment the size. */
        ++t->size;
        if (t->size == 0) {
            *cex = cgul_exception__new(
                "cgul_rbtree__insert: overflow detected in size of tree");
        }
    }
}

/**
 * This function will determine if it is possible to rotate a
 * key/value pair from one of n's 2-3-4 siblings up to n's 2-3-4
 * parent forcing the 2-3-4 parent to rotate a key/value pair down to
 * n.  If such a sibling is found, this function will perform the
 * rotation and return 1; otherwise, it returns 0.
 *
 * Before you call this code, you must make sure that n is directly
 * attached to a black node in the red-black tree that represents a
 * 2-child (one key/value pair) 2-3-4 node or that n is directly
 * attached to a red node.  The reason for this is that this function
 * only does a simple binary tree rotation which works to rotate
 * values in the 2-3-4 tree correctly only if n and its sibling with
 * the extra key/value pair are directly connected to the same parent
 * node that you want to rotate down to n's 2-3-4 node.
 *
 * At this point, you may ask yourself how n could ever be connected
 * to a 2-3-4 node that holds exactly one key/value pair.  Shouldn't
 * all of our rotation and unsplitting as we descend prevent n from
 * being connected to a 2-3-4 node that holds exactly one key/value
 * pair?  Usually, yes; however, when the parent is the root node,
 * there is no way to rotate an extra key/value pair into the root
 * node from a sibling because the root node has no sibling because
 * the root node has no parent.  So all the root node can do is
 * unsplit.  Fortunately, this works out ok because the root node
 * unsplits when both of its 2-3-4 children have only one key/value
 * pair.  Thus children of the root node are guaranteed to be able to
 * rotate a value in or to be able to unsplit.  This is enough to keep
 * the tree balanced.
 *
 * At this point, you may ask yourself if unsplitting is sufficient to
 * keep the tree in balance why don't we just unsplit and forget about
 * all this rotation stuff?  The answer is that just unsplitting only
 * works for the root node.  If you think carefully about unsplitting,
 * you will see that if you unsplit every time you will quickly start
 * killing entire subtrees.  This problem just doesn't exist for the
 * root node because when you unsplit the root node, it does not have
 * a parent node to leave a hole in.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node used to find sibling to rotate
 * @return whether a 2-3-4 sibling was rotated
 */
static int
cgul_rbtree__rotate_234_sibling_sharing_same_parent(cgul_exception_t* cex,
                                                    cgul_rbtree_t t,
                                                    cgul_rbtree_node_t n)
{
    int rv = 0;
    int key_count_234 = 0;
    cgul_rbtree_node_t p = NULL;
    cgul_rbtree_node_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    p = n->parent;
    assert(p);

    /* Either this parent is red or it is the root node as explained
     * in the comments above this function. */
    assert(p->red || (p->parent == NULL));

    /* If n is attached as the left child of the parent, try to take a
     * key/value pair from the sibling on the right.  If you work it
     * out, you will see that it does not matter whether the sibling
     * has 2 or 3 key/value pairs. */
    if (n == p->left) {

        /* Get the black node for the 2-3-4 sibling. */
        s = p->right;
        assert(s);
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {

            /* If there is a lesser key/value pair in the same 2-3-4
             * sibling, use it instead.  */
            if (s->left) {
                if (s->left->red) {
                    s = s->left;
                    /* Get s into position. */
                    cgul_rbtree__rotate_right_ignoring_color(cex, t, s);
                }
            }

            /* While we still have access to s->right, set it to black. */
            assert(s->right);
            s->right->red = 0;

            /* Rotate the siblings key/value pair up to the
             * parent and the parent's key/value pair down to
             * n. */
            cgul_rbtree__rotate_left_ignoring_color(cex, t, s);

            /* Correct for color. */
            s->red = p->red;
            p->red = 0;
            n->red = 1;
            rv = 1;
        }
        goto out;
    }

    /* If n is attached as the right child of the red parent, try
     * to take a key/value pair from the sibling on the left.  If
     * you work it out, you will see that it does not matter
     * whether the sibling has 2 or 3 key/value pairs. */
    if (n == p->right) {

        /* Get the black node for the 2-3-4 sibling. */
        s = p->left;
        assert(s);
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {

            /* If there is a greater key/value pair in the same 2-3-4
             * sibling, use it instead.  */
            if (s->right) {
                if (s->right->red) {
                    s = s->right;
                    /* Get s into position. */
                    cgul_rbtree__rotate_left_ignoring_color(cex, t, s);
                }
            }

            /* While we still have access to s->left, set it to black. */
            assert(s->left);
            s->left->red = 0;

            /* Rotate the siblings key/value pair up to the
             * parent and the parent's key/value pair down to
             * n. */
            cgul_rbtree__rotate_right_ignoring_color(cex, t, s);

            /* Correct for color. */
            s->red = p->red;
            p->red = 0;
            n->red = 1;
            rv = 1;
        }
        goto out;
    }

 out:
    return rv;
}

/**
 * This function handles the difficult scenario of rotating a
 * key/value pair from one middle child to another middle child when
 * the parent is a 4-child node.  See Diagram #10 above the definition
 * of cgul_rbtree__rotate_234().  The caller is responsible for making
 * sure that n's parent is a 4-child node.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node used to find sibling to rotate
 * @return whether sibling was rotated
 */
static int
cgul_rbtree__rotate_234_sibling_across_great_divide(cgul_exception_t* cex,
                                                    cgul_rbtree_t t,
                                                    cgul_rbtree_node_t n)
{
    int rv = 0;
    int key_count_234 = 0;
    cgul_rbtree_node_t gp = NULL;  /* grandparent */
    cgul_rbtree_node_t  p = NULL;  /* parent      */
    cgul_rbtree_node_t  s = NULL;  /* sibling     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    p = n->parent;
    assert(p);
    assert(p->red);

    gp = p->parent;
    assert(gp);
    assert(!gp->red);

#ifndef NDEBUG
    cgul_rbtree_node__get_234_key_count(cex, p, key_count_234);
    assert(key_count_234 == 3);
#endif

    /* If n is the left-middle child of its 2-3-4 parent node, try to
     * rotate a key/value pair from its sibling that is the
     * right-middle child of the same 2-3-4 parent.*/
    if ((n == p->right) && (p == gp->left)) {
        s = gp->right->left;
        assert(s);
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {
            if (s->left) {
                if (s->left->red) {
                    s = s->left;
                    /* Get s into position. */
                    cgul_rbtree__rotate_right_ignoring_color(cex, t, s);
                }
                /* While we still have access to s->right, set it to
                 * black. */
                assert(s->right);
                s->right->red = 0;

                /* Do all five rotations. */
                cgul_rbtree__rotate_right_ignoring_color(cex, t, s);
                cgul_rbtree__rotate_left_ignoring_color(cex, t, n);
                cgul_rbtree__rotate_left_ignoring_color(cex, t, s);
                cgul_rbtree__rotate_right_ignoring_color(cex, t, n);
                cgul_rbtree__rotate_right_ignoring_color(cex, t, p);

                s->red = 0;
                gp->red = 1;
                rv = 1;
            }
        }
        goto out;
    }

    /* If n is the right-middle child of its 2-3-4 parent node, try to
     * rotate a key/value pair from its sibling that is the
     * left-middle child of the same 2-3-4 parent.*/
    if ((n == p->left) && (p == gp->right)) {
        s = gp->left->right;
        assert(s);
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {
            if (s->right && s->right->red) {
                s = s->right;
                /* Get s into position. */
                cgul_rbtree__rotate_left_ignoring_color(cex, t, s);
            }
            /* While we still have access to s->left, set it to
             * black. */
            assert(s->left);
            s->left->red = 0;

            /* Do all five rotations. */
            cgul_rbtree__rotate_left_ignoring_color(cex, t, s);
            cgul_rbtree__rotate_right_ignoring_color(cex, t, n);
            cgul_rbtree__rotate_right_ignoring_color(cex, t, s);
            cgul_rbtree__rotate_left_ignoring_color(cex, t, n);
            cgul_rbtree__rotate_left_ignoring_color(cex, t, p);

            s->red = 0;
            gp->red = 1;
            rv = 1;
        }
        goto out;
    }

 out:
    return rv;
}

/**
 * You have to unsplit the root in case someone wants to remove the
 * only key/value pair in the root node.  You have to unsplit the root
 * differently than in the general case.  The caller is responsible
 * for making sure r's children are both holding only one key/value
 * pair or that they don't exist.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] r root node
 * @return whether the root node was unsplit
 */
static int
cgul_rbtree__unsplit_234_for_root(cgul_exception_t* cex,
                                  cgul_rbtree_t t,
                                  cgul_rbtree_node_t r)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(r);
    assert(r->parent == NULL);
    assert(!r->red);

    /* There have to be more than just two nodes left in the tree in
     * order to have enough nodes for an unsplit operation. */
    if (r->left &&  r->right) {

#ifndef NDEBUG
        {
            int key_count_234 = 0;
            cgul_rbtree_node__get_234_key_count(cex,r->left, key_count_234);
            assert(key_count_234 == 1);
            cgul_rbtree_node__get_234_key_count(cex, r->right, key_count_234);
            assert(key_count_234 == 1);
        }
#endif

        r->left->red = 1;
        r->right->red = 1;
        rv = 1;
    }

 out:
    return rv;
}

/**
 * This function implements the general unsplit operation.  This is
 * the operation of last resort when descending a tree with the intent
 * to remove a node.  This function should only be called if
 * <code>cgul_rbtree__rotate_234()</code> was unable to provide
 * <code>n</code> with an extra key/value pair.  (See the comments for
 * the <code>cgul_rbtree__rotate_234()</code> function.)
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to unsplit
 * @return whether the node was unsplit
 */
static int
cgul_rbtree__unsplit_234(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree_node_t n)
{
    int rv = 0;
    cgul_rbtree_node_t p = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    p = n->parent;

    /* The root node gets special treatment. */
    if (p == NULL) {
        rv = cgul_rbtree__unsplit_234_for_root(cex, t, n);

    } else {
        /* In the general case, we just need to color flip. */

        /* If the parent is black, this must be a 3-child node.  We
         * just need to realign the node in order to get a red
         * parent. */
        if (!p->red) {
            cgul_rbtree__realign_234_three_child_node(cex, t, p);
        }

        /* If everything is working right, p should now have two
         * children (of which n is one) that are both only holding one
         * key/value pair. */
        assert(p->red);
        assert(p->left && p->right);

#ifndef NDEBUG
        {
            int key_count_234 = 0;
            cgul_rbtree_node__get_234_key_count(cex, p->left, key_count_234);
            assert(key_count_234 == 1);
            cgul_rbtree_node__get_234_key_count(cex, p->right, key_count_234);
            assert(key_count_234 == 1);
        }
#endif

        /* With a simple color flip, we can unsplit. */
        p->red = 0;
        p->left->red = 1;
        p->right->red = 1;
        rv = 1;
    }

 out:
    return rv;
}

/**
 * This function helps to break down the 2-3-4 rotation of key/value
 * pairs between siblings to a more manageable size.  <code>n</code>
 * is the node that needs another key/value pair rotated to it.  If
 * the "easy" sibling is found AND it has an extra key/value pair, we
 * return 1 and adjust the tree <code>t</code> so that <code>n</code>
 * and its sibling share the same parent.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node that needs to have an extra key/value pair rotated to it
 * @return whether an "easy" 2-3-4 sibling was found
 */
static int
cgul_rbtree__find_easy_234_sibling(cgul_exception_t* cex,
                                   cgul_rbtree_t t,
                                   cgul_rbtree_node_t n)
{
    int rv = 0;
    int key_count_234 = 0;
    cgul_rbtree_node_t p = NULL;
    cgul_rbtree_node_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    p = n->parent;

    /* When the root node is the parent of n and holds only one
     * key/value pair, you are guaranteed to be able to rotate a
     * key/value pair out of n's only sibling.  You will still need to
     * test for when the root node holds 2 or 3 key/value pairs, but
     * for those cases, you can treat the root node like any other
     * node.  (While the root node can be n's parent, the root node
     * itself is never passed into this fucntion.) */
    if (p->parent == NULL) {
        cgul_rbtree_node__get_234_key_count(cex, p, key_count_234);
        if (key_count_234 == 1) {
            /* The tree is already ready for the rotation. */
            assert(p->left);
            assert(p->right);
            rv = 1;
            goto out;
        }
    }

    /* Only if the parent is the root node do we allow it to hold only
     * one key/value pair, and we just tested for that.  For all other
     * nodes, we are able to guarantee that the parent holds two or
     * more key/value pairs. */
#ifndef NDEBUG
    cgul_rbtree_node__get_234_key_count(cex, p, key_count_234);
    assert(key_count_234 > 1);
#endif

    /* The easy sibling to rotate is the one that is already attached
     * to the same parent.  This includes some 2 key/value pair nodes
     * and all 3 key/value pair nodes. */
    if (p->red) {
        s = cgul_rbtree_node__get_sibling(cex, n);
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {
            /* The tree is already ready for rotation. */
            rv = 1;
        }
        goto out;
    }

    /* If you are here, it is because n is attached to the black node
     * of a 2-3-4 parent that has 2 key/value pairs.  We need to find
     * the sibling and see if it is suitable for use in the rotation
     * operation. */
    if (n == p->left) {
        assert(p->right);
        s = p->right->left;
    } else {
        assert(n == p->right);
        assert(p->left);
        s = p->left->right;        
    }
    assert(s);
    /* If the sibling is suitable for rotation, we realign it so that
     * n and the sibling share the same red parent in the red-black
     * tree. */
    cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
    if (key_count_234 > 1) {
        cgul_rbtree__realign_234_three_child_node(cex, t, p);
        rv = 1;
    }
    goto out;


 out:
    return rv;
}

/**
 * This function helps to break down the 2-3-4 rotation of key/vlaue
 * pairs between siblings to a more manageable size.  <code>n</code>
 * is the node that needs another key/value pair rotated to it.  If
 * the "hard" sibling of a 3-child (i.e., 2-key) parent is found AND
 * it has an extra key/value pair, we return 1 and adjust the tree
 * <code>t</code> so that <code>n</code> and its sibling share the
 * same parent.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node that needs to have an extra key/value pair rotated to it
 * @return whether "hard" two-key sibling was found
 */
static int
cgul_rbtree__find_hard_two_key_sibling(cgul_exception_t* cex,
                                       cgul_rbtree_t t,
                                       cgul_rbtree_node_t n)
{
    int rv = 0;
    int key_count_234 = 0;
    cgul_rbtree_node_t gp = NULL;  /* grandparent */
    cgul_rbtree_node_t  p = NULL;  /* parent      */
    cgul_rbtree_node_t  s = NULL;  /* sibling     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    p = n->parent;
    assert(p);

#ifndef NDEBUG
    /* The parent should hold 2 key/value pairs. */
    cgul_rbtree_node__get_234_key_count(cex, p, key_count_234);
    assert(key_count_234 == 2);
#endif

    /* If the parent is black, n is on the edge and only has one
     * sibling in the red-black tree (not the 2-3-4 tree) which we've
     * already checked in cgul_rbtree__find_easy_234_sibling(). */
    if (!p->red) {
        goto out;
    }

    /* The grandparent cannot be NULL at this point because this would
     * mean the root node is the parent, and the root node is black
     * which we just tested. */
    gp = p->parent;
    assert(gp);

    /* If n is the middle child of p, it can use the sibling currently
     * attached to the black node of p. */
    if ((n == p->right) && (p == gp->left)) {
        s = gp->right;
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {
            /* The sibling has an extra key so rotate things so that s
             * and n share the same parent, */
            cgul_rbtree__realign_234_three_child_node(cex, t, p);
            rv = 1;
        }
    } else if ((n == p->left) && (p == gp->right)) {
        s = gp->left;
        cgul_rbtree_node__get_234_key_count(cex, s, key_count_234);
        if (key_count_234 > 1) {
            /* The sibling has an extra key so rotate things so that s
             * and n share the same parent, */
            cgul_rbtree__realign_234_three_child_node(cex, t, p);
            rv = 1;
        }
    }

 out:
    return rv;
}

/**
 * As you walk down the tree intending to remove a node, you have to
 * prepare the tree so that removing a node does not put the tree out
 * of balance.  The thing you are trying to avoid is removing the last
 * key/value pair from _any_ node.  If you can do this, your tree will
 * be no worse than a balanced binary tree except for the red nodes
 * which hold together the 2-3-4 nodes and create an imbalance, but
 * the imbalance is bounded.
 *
 * You can force each node on the way down to have more than one
 * key/value pair by rotating and unsplitting.  This way, when you
 * reach the leaf node you want to remove, you can do so without
 * messing up the red-black tree structure.
 *
 * This is going to get complicated.  We know that "n" is a 2-child
 * (i.e., 1-key/value pair) node.  This helps, but we have to account
 * for all the different ways the parent "p" can be attached to both
 * "n" and its 2-, 3-, or 4-child sibling "s".  You MUST draw out all
 * of these combinations on a sheet of paper if you intend to work on
 * this function.  However, I will draw some of these below for
 * reference.
 *
 * The following drawing is one of many ways to draw "n" attached to a
 * 3-child parent "p" which is attached to a 3-child sibling "s".  The
 * idea is that you can reduce all but one of the many different ways
 * that parents, siblings, and our target node can be combined to look
 * like the following diagram:
 *
 *                                   |
 *                                   |
 *                                 +----+
 *                                 | P2 |
 *                                 +----+
 *                          (red) /      \
 *                          +----+        +----+
 *                          | P1 |        | S2 |
 *                          +----+        +----+
 *                        /        \
 *                      /            \
 *               +----+                +---+
 *               | S1 |                | N |
 *               +----+                +---+
 *              /      \ (red)        /     \
 *                       +----+
 *                       | S1 |
 *                       +----+
 *                      /      \
 *       
 *                       (Diagram #9)
 *
 * In diagram #9 above, if you rotate S1 left then right, you will
 * pass S1 up to the parent and P1 down to N.  You just need to flip
 * the colors of P1 and N.  You should notice something else about
 * diagram #9.  N has two siblings because N is the middle sibling
 * sort-order-wise.  If you just rotate P1 up, you'll get the mirror
 * image of diagram #9 except P2 will be located between N and S2.
 *
 * The most difficult case is when N is one of the middle children of
 * a 4-child parent node and the only sibling that has an extra
 * key/value pair is the other middle child.
 *
 *                                   |
 *                                   |
 *                                 +----+
 *                                 | P2 |
 *                                 +----+
 *                              /         \
 *                           /               \
 *                        /                     \
 *              (red)  /                           \ (red)
 *              +----+                               +----+
 *              | P1 |                               | P3 |
 *              +----+                               +----+
 *             /      \                             /      \
 *            /        \                           /        \
 *                      +---+                +----+
 *                      | N |                | S2 |
 *                      +---+                +----+
 *                     /     \              /      \
 *                                   (red) /        \
 *                                   +----+
 *                                   | S1 |
 *                                   +----+
 *                                  /      \
 *
 *                               (Diagram #10)
 *
 * Diagram #10 is a difficult case because S1 needs to be rotated so
 * that it occupies the space currently occupied by P2 because S1 is
 * in the middle of P1 and P3 sort-order-wise.  While you are rotating
 * S1 up, you need to simultaneously rotate P2 down so that it can
 * join with N.  These rotations do not compliment each other.  To get
 * it right in the worse case, you will have to make six rotations:
 *
 *     Rotation #1: Rotate S1 right
 *     Rotation #2: Rotate S1 right
 *     Rotation #3: Rotate N  left
 *     Rotation #4: Rotate S1 left
 *     Rotation #5: Rotate N  right
 *     Rotation #6: Rotate P1 right
 *
 * Also remember that you need to adjust the color of various nodes
 * after all rotations are complete.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node that needs to have an extra key/value pair rotated to it
 * @return whether n was rotated an extra key/value pair
 */
static int
cgul_rbtree__rotate_234(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree_node_t n)
{
    int rv = 0;
    int key_count_234 = 0;
    cgul_rbtree_node_t  p = NULL;  /* parent      */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);
#ifndef NDEBUG
    cgul_rbtree_node__get_234_key_count(cex, n, key_count_234);
    assert(key_count_234 == 1);
#endif

    p = n->parent;

    /* The root node should never venture down this path as "n".  It
     * makes no sense to try to rotate in a value from a sibling
     * because the root node has no sibling because the root node has
     * no parent. */
    assert(p);

    /* Because each 2-3-4 node holds 1, 2, or 3 children it is
     * possible for a child to have two siblings.  We must be able to
     * rotate key/value pairs from both siblings.  Each rotation is
     * very detailed so we try to prepare for the rotation.  If
     * successful, we then do the rotation. */
    if (cgul_rbtree__find_easy_234_sibling(cex, t, n)) {
        rv = cgul_rbtree__rotate_234_sibling_sharing_same_parent(cex, t, n);
        goto out;
    }

    cgul_rbtree_node__get_234_key_count(cex, p, key_count_234);
    if (key_count_234 == 2) {
        if (cgul_rbtree__find_hard_two_key_sibling(cex, t, n)) {
            rv = cgul_rbtree__rotate_234_sibling_sharing_same_parent(cex, t, n);
        }
        goto out;
    }

    if (key_count_234 == 3) {
        /* This is the case where n and s are both middle children.
         * This makes things much more complicated. */
        rv = cgul_rbtree__rotate_234_sibling_across_great_divide(cex, t, n);
        goto out;
    }

 out:
    return rv;
}

/**
 * Determine whether an unsplit operation is needed for the root node.
 *
 * @param[in] cex c-style exception
 * @param[in] r root node
 * @return whether an unsplit operation is needed for the root node
 */
static int
cgul_rbtree__is_unsplit_needed_for_root_node(cgul_exception_t* cex,
                                             cgul_rbtree_node_t r)
{
    int rv = 0;
    int key_count_234 = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);
    assert(r->parent == NULL);

    if (r->left && r->right) {
        cgul_rbtree_node__get_234_key_count(cex, r->left, key_count_234);
        if (key_count_234 == 1) {
            cgul_rbtree_node__get_234_key_count(cex, r->right, key_count_234);
            if (key_count_234 == 1) {
                rv = 1;
            }
        }
    }

 out:
    return rv;
}

/**
 * This function makes sure each node you pass through when walking
 * down the tree when intending to remove a node has more than one
 * key/value pair.  This is discussed in comments further up the call
 * stack, but this is done first by trying to rotate key/value pairs
 * from sibling nodes into the parent which forces the parent to
 * rotate n an extra key/value pair.  If there is no way to rotate a
 * new key/value pair to <code>n</code>, an unsplit operation should
 * work unless <code>n</code> is both the root node and a leaf node.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to insure has more than one key
 */
static void
cgul_rbtree__insure_node_has_more_than_one_key(cgul_exception_t* cex,
                                               cgul_rbtree_t t,
                                               cgul_rbtree_node_t n)
{
    int unsplit_needed = 0;
    int key_count_234 = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* If this node only has one key, we have work to do. */
    cgul_rbtree_node__get_234_key_count(cex, n, key_count_234);
    if (key_count_234 == 1) {

        /* Unlike the other nodes, the root node can unsplit while
         * holding just one key/value pair.  If any other node did
         * this, it would leave a hole in the tree.  The root does not
         * have to worry about this because it does not have a parent
         * to leave a hole in.  Also, it does not make much sense to
         * try to rotate data from a sibling that the root node
         * doesn't have up to a parent the root node doesn't have.
         * Thus, the root node does not perform the rotate
         * operation in this context. */
        if (n->parent) {
            /* Nodes other than the root node can try to solve their
             * population problems by rotating. */
            if (!cgul_rbtree__rotate_234(cex, t, n)) {
                unsplit_needed = 1;
            }
        } else {
            unsplit_needed =
                cgul_rbtree__is_unsplit_needed_for_root_node(cex, n);
        }

        if (unsplit_needed) {
            if (!cgul_rbtree__unsplit_234(cex, t, n)) {
                /* Something is very wrong if these asserts fail. */
                assert(n == t->root);
                assert(cgul_rbtree_node__is_234_leaf_node(cex, t->root));
            }
        }

    }

 out:
    return;
}

/**
 * This code is called with <code>curr</code> pointing to the node you
 * want to remove.  It is the caller's responsibility to make sure
 * this function is only called if <code>curr</code> is not in a 2-3-4
 * leaf node (by first making sure
 * <code>cgul_rbtree_node__is_234_leaf_node()</code> is false).  The
 * value returned is the value that comes after <code>curr</code> in
 * sort order.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] curr node you want to remove
 * @return next node in sorted order or <code>NULL</code> if
 * <code>curr</code> ends up as a leaf node because of unsplitting and
 * rotating as we descend the tree
 */
static cgul_rbtree_node_t
cgul_rbtree__find_next_with_intent_to_remove(cgul_exception_t* cex,
                                             cgul_rbtree_t t,
                                             cgul_rbtree_node_t curr)
{
    cgul_rbtree_node_t n = NULL;
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(curr);
    assert(curr->right);

    /* The next node is (more or less) the left-most node of curr's
     * right subtree. */
    n = curr->right;

    while (n != NULL) {
        /* Yes, you still have to rotate or unsplit. */
        if (!n->red) {
            cgul_rbtree__insure_node_has_more_than_one_key(cex, t, n);
        }
        /* Continue descending the tree.  You have to be careful here.
         * Certain combinations of unsplitting and rotating (preformed
         * by cgul_rbtree__insure_node_has_more_than_one_key() above)
         * can cause the node you are trying to replace to be rotated
         * into this search path.  Thus, if you just blindly go left,
         * you could end up getting curr's previous value instead of
         * curr's next value! */
        if (n == curr) {
            if (n->right == NULL) {
                /* curr is now a leaf node. */
                break;
            }
            n = n->right;
        } else if (n->left == NULL) {
            /* Found "next". */
            rv = n;
            break;
        } else {
            n = n->left;
        }
    }

    assert(n);

 out:
    return rv;
}

/**
 * Walk down the tree rotating and unsplitting as necessary.  After
 * calling this function, if the node to remove was found,
 * <code>n</code> will point to it; otherwise, <code>n</code> will be
 * NULL and so will <code>next</code>.  If <code>n</code> is not part
 * of a 2-3-4 leaf node, this function will continue walking down the
 * tree rotating and unsplitting as necessary until it finds the node
 * that follows <code>n</code> in sort order.  It will return this
 * node as <code>next</code>, and you can move it from the leaf node
 * to where <code>n</code> was.  This allows you to only remove leaf
 * nodes which should keep the tree in sorted.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] key key to match for the node to remove
 * @param[out] n node that matches key or <code>NULL</code> if no
 * match was found
 * @param[out] next the next node in sorted order or <code>NULL</code>
 * if <code>n</code> is a leaf node
 */
static void
cgul_rbtree__find_node_to_remove(cgul_exception_t* cex,
                                 cgul_rbtree_t t,
                                 const void* key,
                                 cgul_rbtree_node_t* n,
                                 cgul_rbtree_node_t* next)
{
    int compare = 0;
    cgul_rbtree_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);
    assert(next);

    *n = NULL;
    *next = NULL;

    /* Find the node to remove. */
    curr = t->root;
    while (curr != NULL) {
        /* Rotate or unsplit the node if necessary.  You only need to
         * rotate or unsplit when you first enter the 2-3-4 node,
         * i.e., when you arrive at a black node. */
        if (!curr->red) {
            cgul_rbtree__insure_node_has_more_than_one_key(cex, t, curr);
        }
        /* Continue descending the tree. */
        compare = (*t->compare)(key, curr->key);
        if (compare < 0) {
            curr = curr->left;
        } else if (compare > 0) {
            curr = curr->right;
        } else {
            break;
        }
    }

    if (curr) {
        /* We found the red-black tree node to delete. */
        *n = curr;
        if (!cgul_rbtree_node__is_234_leaf_node(cex, curr)) {
            *next = cgul_rbtree__find_next_with_intent_to_remove(cex, t, curr);
        }
    }

 out:
    return;
}

/**
 * Find the node indexed by position.  This method is the basically
 * the same as the public API method
 * <code>cgul_rbtree__find_at()</code> except it has been
 * parameterized to accept the <code>is_for_removal</code> flag so the
 * same code can be shared by <code>cgul_rbtree__remove_at()</code>. <p>
 *
 * If <code>index</code> is out of range, <code>NULL</code> is
 * returned; otherwise, the node returned is the same node that would
 * be returned if you had an array <code>t</code> that held all the
 * nodes in the tree in sorted order, and you asked for
 * <code>t[index]</code>. <p>
 *
 * This method throws an exception if <code>t</code> was not created
 * with <code>cguil_rbtree__new_with_indexing()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] index index into sorted keys of node to return
 * @param[in] is_for_remove whether for <code>cgul_rbtree__remove_at()</code>
 * @return node associated with <code>index</code>
 */
static cgul_rbtree_node_t
cgul_rbtree__find_node_at(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          unsigned long int index,
                          int is_for_removal)
{
    unsigned long int lesser_count = 0;
    cgul_rbtree_node_with_indexing_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Make sure the tree is using nodes from the
     * cgul_rbtree_node_with_indexing class instead of from the base
     * cgul_rbtree_node class. */
    if (!t->is_indexed) {
        *cex = cgul_exception__new(
                  "cgul_rbtree__find_node_at: "
                  "Indexing not supported for this tree.\n"
                  "cgul_rbtree object not created with "
                  "cgul_rbtree__new_with_indexing().");
        goto out;
    }

    if (index < t->size) {

        /*
         * The user passes in a C-style index.  So, if the user passed
         * in 5 for index (for example), he or she is looking for the
         * 6th element in sort order.  We can find the 6th element by
         * finding the node with 5 lesser children.
         *
         * The algorithm is relatively simple.  "lesser_count" holds
         * the number of keys that are less than the key for the
         * current node.
         *
         * If you go right, "lesser_count" is *increased* by 1 for the
         * key in the parent node you just left and by "lcount" for
         * the new right node you are now visiting because all the
         * other lesser nodes are already accounted for in
         * "lesser_count".
         *
         * If you go left, "lesser_count" is *decreased* by 1 for the
         * key in the parent node you just left and by "rcount" for
         * the new left node you are now visiting because these are
         * the only nodes still in "lesser_count" that are greater
         * than the new current node.
         *
         * See cgul_trie__find_at() for a more detailed example.
         */

        rv = (cgul_rbtree_node_with_indexing_t)t->root;
        assert(rv);
        lesser_count = rv->lcount;
        for (;;) {
            /* Rotate or unsplit the node if necessary.  You only need
             * to rotate or unsplit when you first enter the 2-3-4
             * node, i.e., when you arrive at a black node. */
            if (is_for_removal && !rv->base.red) {
                cgul_rbtree__insure_node_has_more_than_one_key(
                    cex, t, (cgul_rbtree_node_t)rv);
            }
            if (lesser_count == index) {
                break;
            }
            if (index < lesser_count) {
                rv = (cgul_rbtree_node_with_indexing_t)rv->base.left;
                assert(rv);
                lesser_count -= rv->rcount + 1;
                continue;
            }
            if (index > lesser_count) {
                rv = (cgul_rbtree_node_with_indexing_t)rv->base.right;
                assert(rv);
                lesser_count += rv->lcount + 1;
                continue;
            }
            /* Not reached. */
            assert(0);
        }
    }

 out:
    return (cgul_rbtree_node_t)rv;
}

/**
 * Remove the leaf node from the tree but do not free the node.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to remove from the tree
 */
static void
cgul_rbtree__remove_leaf_node_without_freeing_node(cgul_exception_t* cex,
                                                   cgul_rbtree_t t,
                                                   cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t p = NULL;
    cgul_rbtree_node_t* pc = NULL;  /* parent's child */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);
    assert(cgul_rbtree_node__is_234_leaf_node(cex, n));

    p = n->parent;

    if (p) {
        if (n == p->left) {
            pc = &p->left;
        } else {
            assert(n == p->right);
            pc = &p->right;
        }
    }

    /* If n is red, it is attached to a black node.  Just remove it.*/
    if (n->red) {
        assert(p);
        *pc = NULL;
        goto out;
    }

    /*
     * n must be black and could have zero, one, or two children.
     */

    /* If we are deleting the root when the root is the only child in
     * the tree, update t->root.  While black nodes that are not root
     * that have both children null do exist in leaf nodes, the way we
     * descend the tree rotating and unsplitting guarantees that only
     * the root node will be the only black node that will ever have
     * both children null when being deleted. */
    if ((n->left == NULL) && (n->right == NULL)) {
        assert(n->parent == NULL);
        t->root = NULL;
        t->front = NULL;
        t->back = NULL;
        t->oldest = NULL;
        t->youngest = NULL;
        goto out;
    }

    /* n has just a left child.  It must be made black and attached to
     * the current parent of n. */
    if (n->left && (n->right == NULL)) {
        if (p) {
            *pc = n->left;
        } else {
            t->root = n->left;
        }
        n->left->parent = p;
        n->left->red = 0;
        goto out;
    }

    /* n has just a right child.  It must be made black and attached to
     * the current parent of n. */
    if (n->right && (n->left == NULL)) {
        if (p) {
            *pc = n->right;
        } else {
            t->root = n->right;
        }
        n->right->parent = p;
        n->right->red = 0;
        goto out;
    }

    /* n has two children.  One of the children must be made black and
     * attached to the current parent of n.  The other child must be
     * attached to the child that is attached to the current parent of
     * n. */
    if (n->right && n->left) {
        assert(n->left->red);
        assert(n->right->red);
        if (p) {
            *pc = n->right;
        } else {
            t->root = n->right;
        }
        n->right->parent = p;
        n->right->left = n->left;
        n->right->red = 0;
        if (t->is_indexed) {
            ((cgul_rbtree_node_with_indexing_t)n->right)->lcount = 1;
        }
        n->left->parent = n->right;
        goto out;
    }

    /* You should not reach here. */
    assert(0);     

 out:
    return;
}

/**
 * This function will properly remove <code>n</code> from the tree
 * <code>t</code> PROVIDED you have called
 * <code>cgul_rbtree__insure_node_has_more_than_one_key()</code> for
 * every black node you descended through on your way to finding
 * <code>n</code>.  ALSO, <code>replacement</code> should be
 * <code>NULL</code> if <code>n</code> is part of a 2-3-4 leaf node;
 * otherwise, <code>replacement</code> should be the next value larger
 * than <code>n</code> in sort order.  After calling this function,
 * you should not dereference <code>n</code> because the entire node
 * will have been deleted.
 *
 * The reason we sometimes need a <code>replacement<code> is that, if
 * you want the tree to stay balanced, you can only remove key/value
 * pairs that are in 2-3-4 leaf nodes.  The solution, therefore, is to
 * find the key/value pair that comes just after <code>n</code> in
 * sort order.  When <code>n</code> is not in a 2-3-4 leaf node, the
 * next node in sort order is guaranteed to be in a 2-3-4 leaf node.
 * We can then remove the <code>replacement</code> node without
 * affecting the tree's balance, and we can simply move it to where
 * <code>n</code> currently sits in the tree without affecting the
 * sort order.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] n node to remove
 * @param[in] replacement leaf node that can assume the place of
 * <code>n</code> in the tree
 */
static void
cgul_rbtree__remove_node_after_proper_descent(cgul_exception_t* cex,
                                              cgul_rbtree_t t,
                                              cgul_rbtree_node_t n,
                                              cgul_rbtree_node_t replacement)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(n);

    /* While n is still valid, update "front".  If t->front has a
     * right child, the right child is the new front; otherwise, the
     * parent is the new front. */
    if (n == t->front) {
        assert(t->front->left == NULL);
        if (t->front->right) {
            t->front = t->front->right;
        } else {
            t->front = n->parent;
        }
    }

    /* While n is still valid, update "back".  If t->back has a left
     * child, the left child is the new back; otherwise, the parent is
     * the new back. */
    if (n == t->back) {
        assert(t->back->right == NULL);
        if (t->back->left) {
            t->back = t->back->left;
        } else {
            t->back = n->parent;
        }
    }

    /* While n is still valid, update "oldest". */
    if (n == t->oldest) {
        t->oldest = t->oldest->younger;
    }

    /* While n is still valid, update "youngest". */
    if (n == t->youngest) {
        t->youngest = t->youngest->older;
    }

    /* While n is still valid, update the "older/younger" linked list
     * that runs through all the nodes. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }


    /* Remove the correct node from the tree. */
    if (replacement == NULL) {
        /* n is a leaf node. */
        cgul_rbtree__remove_leaf_node_without_freeing_node(cex, t, n);
        cgul_cache__put(cex, t->cache, n);

    } else {

        /* n is not a leaf node.  In order to remove it from the tree
         * while simultaneously maintaining sort order and balance, we
         * have to replace it with its "replacement" which we pull off
         * the bottom of the tree. */
        cgul_rbtree__remove_leaf_node_without_freeing_node(cex,
                                                            t,
                                                            replacement);

        /*
         * We now put the replacement nodes into n's spot in the tree.
         * It would be nice if we could just do a shallow copy of the
         * replacement node, and then delete the replacement node.
         * The problem with this approach is that it invalidates any
         * iterator or "older/younger" pointer that points to the
         * replacement node (because it is the replacement node that
         * would be deleted in this scheme).  Instead, by moving the
         * replacement node itself to a new spot in the tree, any
         * iterator that points to the replacement node should remain
         * valid.
         */

        /* Update the rbtree. */
        if (t->root == n) {
            t->root = replacement;
        }
        if (t->front == n) {
            t->front = replacement;
        }
        if (t->back == n) {
            t->back = replacement;
        }

        /* Update the replacement node.  The replacement node is n's
         * "next" node.  This means the left count stays the same, but
         * the right count is now one less; however, this actually
         * means one less right count all the way up the tree from the
         * leaf node where "replacement" was residing.  To take care
         * of this, cgul_rbtree__decrement_counts_along_path() is
         * called before entering this function. */
        replacement->red = n->red;
        replacement->parent = n->parent;
        replacement->left = n->left;
        replacement->right = n->right;
        if (t->is_indexed) {
            /* Down cast. */
            cgul_rbtree_node_with_indexing_t rwi =
                (cgul_rbtree_node_with_indexing_t)replacement;
            cgul_rbtree_node_with_indexing_t nwi =
                (cgul_rbtree_node_with_indexing_t)n;
            /* Adjust. */
            rwi->lcount = nwi->lcount;
            rwi->rcount = nwi->rcount;
        }

        /* Update n's parent. */
        if (n->parent) {
            if (n->parent->left == n) {
                n->parent->left = replacement;
            } else {
                assert(n->parent->right == n);
                n->parent->right = replacement;
            }
        }

        /* Update n's left child. */
        if (n->left) {
            n->left->parent = replacement;
        }

        /* Update n's right child. */
        if (n->right) {
            n->right->parent = replacement;
        }

        /* It is now safe to delete n. */
        cgul_cache__put(cex, t->cache, n);
    }

 out:

    if (*cex == NULL) {
        --t->size;
    }

    return;
}

/**
 * This function increments the lcount and rcount of the nodes along
 * the path from <code>n</code> (exclusive) up to the root node
 * (inclusive) to account for a newly inserted node along that path.
 * You pass in the the newly created node as <code>n</code>.  Any
 * changes that need to occur to the lcount and rcount of nodes below
 * the parent should be taken care of by your insert routine.
 *
 * @param[in] cex c-style exception
 * @param[in] n node used to describe a path from the root node
 */
static void
cgul_rbtree__increment_counts_along_path(cgul_exception_t* cex,
                                         cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    for ( ; n->parent != NULL ; n = n->parent) {
        if (n == n->parent->left) {
            /* t->size will overflow before lcount.  So we don't check here. */
            ++(((cgul_rbtree_node_with_indexing_t)n->parent)->lcount);
        } else {
            /* t->size will overflow before rcount.  So we don't check here. */
            assert(n == n->parent->right);
            ++(((cgul_rbtree_node_with_indexing_t)n->parent)->rcount);
        }
    }

 out:
    return;
}

/**
 * After removing a node "c" from the tree, you call this with n set
 * to c's parent.  It will then decrement the lcount and rcount of all
 * the nodes along the path from n (exclusive) up to the root node
 * (inclusive).  Any changes that need to occur to the lcount and
 * rcount of nodes below the parent should be taken care of by your
 * remove routine.
 *
 * @param[in] cex c-style exception
 * @param[in] n node used to describe a path from the root node
 */
static void
cgul_rbtree__decrement_counts_along_path(cgul_exception_t* cex,
                                         cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    for ( ; n->parent != NULL ; n = n->parent) {
        if (n == n->parent->left) {
            --(((cgul_rbtree_node_with_indexing_t)n->parent)->lcount);
        } else {
            assert(n == n->parent->right);
            --(((cgul_rbtree_node_with_indexing_t)n->parent)->rcount);
        }
    }

out:
    return;
}

/**
 * This method is the callback portion of the
 * <code>cgul_rbtree__clear()</code> convenience method that is
 * designed to remove all the nodes of a <code>cgul_rbtree</code>
 * _and_ to put back all the keys and values to their respective
 * <code>cgul_cache</code> caches.  If the keys cache is
 * <code>NULL</code>, the key will not be put back or otherwise freed.
 * The same applies for the values cache. <p>
 *
 * This method should not be invoked directly.  Instead, it should be
 * invoked as a callback function for
 * <code>cgul_rbtree__traverse()</code>.  When you invoke
 * <code>cgul_rbtree__traverse()</coded>, you should do so in such a
 * way that <code>client_data</code> holds a
 * <code>cgul_rbtree__clear_cb_data_t</code>.  Typically, your code
 * will work with <code>cgul_rbtree__clear()</code> or
 * <code>cgul_rbtree__remove_range()</code> instead of with this
 * method.
 *
 * @param[in] cex c-style exception
 * @param[in] t tree
 * @param[in] n tree node
 * @param[in] client_data client data
 * @return whether to continue traversing the tree
 */
static int
cgul_rbtree__clear_cb(cgul_exception_t* cex,
                      cgul_rbtree_t tree,
                      cgul_rbtree_node_t n,
                      void* client_data)
{
    int rv = 0;
    void* key = NULL;
    void* value = NULL;
    cgul_rbtree__clear_cb_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree);
    assert(n);
    assert(client_data);

    data = (cgul_rbtree__clear_cb_data_t)client_data;

    /* Remove the node. */
    cgul_rbtree__remove_node(cex, tree, n, &key, &value);
    data->rv = 1;

    /* Only after the node is removed is it safe to put back the key. */
    if (data->keys_cache && key) {
        cgul_cache__put(cex, data->keys_cache, key);
    }

    /* The value can be put back at any time. */
    if (data->values_cache && value) {
        cgul_cache__put(cex, data->values_cache, value);  
    }

    rv = 1;

 out:
    return rv;
}

/**
 * This method calls <code>cgul_cache__put()</code> on each node in
 * <code>t</code>.  It does not bother to keep the tree balanced
 * during the deletion because this function should only be called
 * from <code>cgul_rbtree__delete()</code>.  Because this method is
 * only called from the destructor, it is intentional that a
 * <code>cgul_exception</code> instance is not passed to this method.
 *
 * @param[in] t <code>cgul_rbtree</code> instance
 */
static void
cgul_rbtree__free_all_nodes(cgul_rbtree_t t)
{
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t next = NULL;
    cgul_exception_t local = NULL;

    assert(t);

    /* Iterate over all the nodes in the tree using the embedded
     * older/younger linked list.  This makes it possible to iterate
     * even if the user has already called free() on all the keys in
     * the tree.
     *
     * The original code for this method worked by walking the tree
     * without dereferencing any of the keys, but the new code is a
     * little more efficient. */
    for (curr = t->youngest ; curr ; curr = next) {
        /* Get the next node while we still can. */
        next = curr->older;
        /* Put the current node back on its cache. */
        cgul_cache__put(&local, t->cache, curr);
    }

    return;
}

/************************************************************************
 * Public Methods
 ************************************************************************/

cgul_rbtree_t
cgul_rbtree__new(cgul_exception_t* cex,
                 cgul_rbtree__compare_t compare)
{
    cgul_rbtree_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(compare);

    rv = (cgul_rbtree_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    rv->cache = cgul_cache__new(
                    cex,
                    (cgul_cache__constructor_t)&cgul_rbtree_node__new,
                    (cgul_cache__destructor_t)&cgul_rbtree_node__delete,
                    (cgul_cache__restructor_t)&cgul_rbtree_node__initialize);
    if (*cex) {
        goto out;
    }

    rv->compare = compare;

 out:

    /* Clean up any partially instantiated objects. */
    if (*cex && rv) {
        cgul_rbtree__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_rbtree_t
cgul_rbtree__new_with_indexing(cgul_exception_t* cex,
                               cgul_rbtree__compare_t compare)
{
    cgul_rbtree_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(compare);

    rv = (cgul_rbtree_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    rv->cache =
        cgul_cache__new(cex,
                        (cgul_cache__constructor_t)
                            &cgul_rbtree_node_with_indexing__new,
                        (cgul_cache__destructor_t)
                            &cgul_rbtree_node_with_indexing__delete,
                        (cgul_cache__restructor_t)
                            &cgul_rbtree_node_with_indexing__initialize);
    if (*cex) {
        goto out;
    }

    rv->is_indexed = 1;
    rv->compare = compare;

 out:

    /* Clean up any partially instantiated objects. */
    if (*cex && rv) {
        cgul_rbtree__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_rbtree__delete(cgul_rbtree_t t)
{
    if (t) {

        if (t->cache) {
            /* Delete all the nodes.  This might move some of the
             * nodes to the cache.  Thus, this must be done before
             * deleting the cache. */
            cgul_rbtree__free_all_nodes(t);
            cgul_cache__delete(t->cache);
        }

        /* Delete the object itself. */
        free(t);
    }
}

void
cgul_rbtree__free_keys(cgul_rbtree_t t)
{
    cgul_rbtree_node_t curr = NULL;

    if (t) {
        /* Iterate over each node. */
        for (curr = t->oldest ; curr ; curr = curr->younger) {
            /* Free the key. */
            free((void*)curr->key);
        }
    }
}

void
cgul_rbtree__free_values(cgul_rbtree_t t)
{
    cgul_rbtree_node_t curr = NULL;

    if (t) {
        /* Iterate over each node. */
        for (curr = t->oldest ; curr ; curr = curr->younger) {
            /* Free the value. */
            free((void*)curr->value);
        }
    }
}

unsigned long int
cgul_rbtree__get_cache_size(cgul_exception_t* cex,
                            cgul_rbtree_t t)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = cgul_cache__get_size(cex, t->cache);

 out:
    return rv;
}

void
cgul_rbtree__set_cache_size(cgul_exception_t* cex,
                            cgul_rbtree_t t,
                            unsigned long int size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    cgul_cache__set_size(cex, t->cache, size);

 out:
    return;
}

unsigned long int
cgul_rbtree__get_cache_reserve(cgul_exception_t* cex,
                               cgul_rbtree_t t)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = cgul_cache__get_reserve(cex, t->cache);

 out:
    return rv;
}

void
cgul_rbtree__set_cache_reserve(cgul_exception_t* cex,
                               cgul_rbtree_t t,
                               unsigned long int reserve)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    cgul_cache__set_reserve(cex, t->cache, reserve);

 out:
    return;
}

int
cgul_rbtree__is_empty(cgul_exception_t* cex,
                      cgul_rbtree_t t)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = (t->size == 0);

 out:
    return rv;
}

/* As with all other binary trees, 2-3-4 trees insert at the bottom.
 * You have to insert 3 key/value pairs into the 2-3-4 tree's nodes
 * that reside at the bottom of the tree.  The brilliant thing about
 * requiring you to insert 3 key/value pairs into the bottom nodes is
 * that this forces you to sort nodes as they enter the tree.
 *
 * Thus, you NEVER end up with something like the following:
 *
 *       |
 *       |
 *     +---+
 *     | A |
 *     +---+
 *          \
 *           +---+
 *           | B |
 *           +---+
 *                \
 *                 +---+
 *                 | C |
 *                 +---+
 *
 *               (Diagram #11)
 *
 * The reason you never end up with the situation in diagram #11 is
 * that the 2-3-4 node can hold exactly three children which is
 * perfect for a red-black binary tree.  Thus instead of diagram #11
 * you get the following:
 *
 *             |
 *             |
 *           +---+
 *           | B |
 *           +---+
 *    (red) /     \ (red)
 *     +---+       +---+
 *     | A |       | C |
 *     +---+       +---+
 *
 *        (Diagram #12)
 *
 * So, from the very beginning values are balanced when inserted into
 * 2-3-4 trees.
 *
 * Now, think about a perfectly balanced binary tree.  Your root node
 * must always be the middle node for all your keys.  The left child
 * of the root node must be the middle node for all the keys that are
 * less than the root node, and the right child of the root node must
 * be the middle node for all the keys that are greater than the root
 * node.  And so on.
 *
 * So, as you insert more and more key/value pairs on just one side of
 * the tree, you have to have someway of percolating those values up
 * toward the top of tree because those values are always inserted at
 * the bottom of the tree, and one of those values may end up needing
 * to be the root of your tree at some point.
 *
 * In a 2-3-4 tree, the "split" operation is what is used to percolate
 * key/value pairs up the tree.  When you have three values stored
 * (and sorted) in a 2-3-4 node, it causes the middle value to move up
 * the tree.  This is the only way values move up the tree.
 *
 * While descending the tree with intent to insert a new value it is
 * imperative that we create room for the key/value pairs that might
 * need to percolate up the tree.  To do this we split each node that
 * is full (i.e., that is currently holding 3 key/value pairs).  To
 * split a full 2-3-4 node, we replace the full 2-3-4 node with three
 * new 2-3-4 nodes each having 1 key/value pair from the original
 * 2-3-4 node.  We insert the middle key/value pair into the original
 * parent and attach the other 2 key/value pairs to the parent on
 * either side of the middle key that we passed up. */
int
cgul_rbtree__insert(cgul_exception_t* cex,
                    cgul_rbtree_t t,
                    const void* key,
                    cgul_rbtree_node_t* n)
{
    int rv = 0;
    int compare = 0;
    cgul_rbtree_node_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* We have to return NULL in n if we fail to insert or find the node. */
    if (n) {
        *n = NULL;
    }

    /* If the list is empty, seed the first node.  By default when you
     * create a new cgul_rbtree_node object, it is black with a left
     * and right children counts of zero and has its parent, left,
     * right, older, and younger pointers set to NULL. */
    if (t->root == NULL) {
        tmp = (cgul_rbtree_node_t)cgul_cache__get(cex, t->cache);
        cgul_rbtree_node__set_key(cex, tmp, key);
        if (*cex) {
            goto out;
        }
        if (n) {
            *n = tmp;
        }
        t->root = tmp;
        t->front = tmp;
        t->back = tmp;
        t->oldest = tmp;
        t->youngest = tmp;
        t->size = 1;
        rv = 1;
    } else {
        cgul_rbtree_node_t p = NULL;
        /* Find the leaf node where insertion needs to take place. */
        p = cgul_rbtree__find_insertion_leaf_node(cex, t, key, &compare);
        if (compare == 0) {
            /* The key already exists.  Thus, the insertion fails, but
             * we return the node with "key". */
            if (n) {
                *n = p;
            }
            rv = 0;
        } else {
            /* Instantiate a new node. */
            tmp = (cgul_rbtree_node_t)cgul_cache__get(cex, t->cache);
            cgul_rbtree_node__set_key(cex, tmp, key);
            if (*cex) {
                goto out;
            }
            if (n) {
                *n = tmp;
            }

            /* Keep "front" and "back" up to date.  We do this here
             * because inserting *n into t may cause *n to be
             * re-parented, and it makes a little more sense to access
             * p now before *n might be re-parented. */
            assert(t->front);
            assert(t->back);
            if ((p == t->front) && (compare < 0)) {
                t->front = tmp;
            } else if ((p == t->back) && (compare > 0)) {
                t->back = tmp;
            }

            /* Keep the "youngest" pointer for the tree up to date and
             * initialize the "older" pointer for the new node and the
             * "younger" pointer for the node that was the youngest.
             * (The "younger" pointer for the new node is already
             * initialized to NULL.)  The youngest node is always the
             * most recently inserted node. */
            tmp->older = t->youngest;
            t->youngest->younger = tmp;
            t->youngest = tmp;

            /* Insert the new node. */
            cgul_rbtree__insert_new_child_into_leaf_node(
                                                   cex, t, p, tmp, compare);

            /* Update the lcount and rcount for ancestors. */
            if (t->is_indexed) {
                cgul_rbtree__increment_counts_along_path(cex, tmp);
            }

            rv = 1;
        }
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__find(cgul_exception_t* cex,
                  cgul_rbtree_t t,
                  const void* key)
{
    int compare_rv = 0;
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->root;
    while (rv != NULL) {
        compare_rv = (*t->compare)(key, rv->key);
        if (compare_rv == 0) {
            break;
        } else if (compare_rv < 0) {
            rv = rv->left;
        } else {
            rv = rv->right;
        }
    }

 out:
    return rv;
}

/* Each node keeps tracks of how many left children it has (lcount)
 * and how many right children it has (rcount).  This lets you find
 * any node in logarithmic time by an index into its sorted
 * position. */
cgul_rbtree_node_t
cgul_rbtree__find_at(cgul_exception_t* cex,
                     cgul_rbtree_t t,
                     unsigned long int index)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = cgul_rbtree__find_node_at(cex, t, index, 0 /* is_for_removal */);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

int
cgul_rbtree__find_rank(cgul_exception_t* cex,
                       cgul_rbtree_t t,
                       const void* key,
                       unsigned long int* rank)
{
    int rv = 0;
    int compare_rv = 0;
    unsigned long int lesser_count = 0;
    cgul_rbtree_node_with_indexing_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Clear the user's return parameter. */
    if (rank) {
        *rank = 0;
    }

    /* Make sure the tree is using nodes from the
     * cgul_rbtree_node_with_indexing class instead of from the base
     * cgul_rbtree_node class. */
    if (!t->is_indexed) {
        *cex = cgul_exception__new(
                  "cgul_rbtree__find_rank: "
                  "Indexing not supported for this tree.\n"
                  "cgul_rbtree object not created with "
                  "cgul_rbtree__new_with_indexing().");
        goto out;
    }

    /* Sanity check. */
    if (!key) {
        goto out;
    }

    /*
     * See comments in cgul_rbtree__find_node_at() for how
     * "lesser_count" works.
     */

    curr = (cgul_rbtree_node_with_indexing_t)t->root;
    lesser_count = curr->lcount;
    while (curr != NULL) {
        compare_rv = (*t->compare)(key, curr->base.key);
        if (compare_rv == 0) {
            break;
        } else if (compare_rv < 0) {
            curr = (cgul_rbtree_node_with_indexing_t)curr->base.left;
            if (curr) {
                lesser_count -= curr->rcount + 1;
            }
        } else {
            curr = (cgul_rbtree_node_with_indexing_t)curr->base.right;
            if (curr) {
                lesser_count += curr->lcount + 1;
            }
        }
    }

    if (curr) {
        /* Success. */
        if (rank) {
            *rank = lesser_count;
        }
        rv = 1;
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__find_floor(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        const void* search_key)
{
    int compare_rv = 0;
    cgul_rbtree_node_t rv = NULL;
    cgul_rbtree_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    curr = t->root;
    while (curr != NULL) {
        compare_rv = (*t->compare)(search_key, curr->key);
        if (compare_rv == 0) {
            /* We found an exact match, return it. */
            rv = curr;
            break;
        } else if (compare_rv < 0) {
            /* When we go left, it means the current node's key is
             * greater than "search_key".  Thus, the current node
             * cannot possibly be the floor. */
            curr = curr->left;
        } else {
            /* When we go right, it means the current node's key is
             * less than "search_key".  Thus, if we do not go right
             * again, the current node will be the floor. */
            rv = curr;
            curr = curr->right;
        }
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__find_ceiling(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          const void* search_key)
{
    int compare_rv = 0;
    cgul_rbtree_node_t rv = NULL;
    cgul_rbtree_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    curr = t->root;
    while (curr != NULL) {
        compare_rv = (*t->compare)(search_key, curr->key);
        if (compare_rv == 0) {
            /* We found an exact match, return it. */
            rv = curr;
            break;
        } else if (compare_rv < 0) {
            /* When we go left, it means the current node's key is
             * greater than "search_key".  Thus, if we do not go left
             * again, the current node will hold the ceiling. */
            rv = curr;
            curr = curr->left;
        } else {
            /* When we go right, it means the current node's key is
             * less than "search_key".  Thus, the current node cannot
             * possibly be the ceiling. */
            curr = curr->right;
        }
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__find_special(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree__compare_t compare,
                          const void* data)
{
    int compare_rv = 0;
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->root;
    while (rv != NULL) {
        /* NOTE: This is (*compare)(), NOT (*t->compare)() !!! */
        compare_rv = (*compare)(rv, data);
        if (compare_rv == 0) {
            break;
        } else if (compare_rv < 0) {
            rv = rv->left;
        } else {
            rv = rv->right;
        }
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__get_front(cgul_exception_t* cex,
                       cgul_rbtree_t t)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->front;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__get_back(cgul_exception_t* cex,
                      cgul_rbtree_t t)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->back;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree__get_oldest(cgul_exception_t* cex,
                        cgul_rbtree_t t)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->oldest;

 out:
    return rv;
}

void
cgul_rbtree__set_oldest(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(t->oldest);
    assert(n);

    /* There is nothing to do if n is already the oldest. */
    if (n == t->oldest) {
        goto out;
    }

    /* If n is currently the youngest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new youngest node. */
    if (n == t->youngest) {
        t->youngest = n->older;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the oldest node.  We assert t->oldest above because
     * any non-empty tree has an oldest node, and because n is a node
     * in t, t is non-empty. */
    t->oldest->older = n;
    n->younger = t->oldest;
    n->older = NULL;

    /* Remember that n is now the oldest node. */
    t->oldest = n;

 out:
    return;
}

cgul_rbtree_node_t
cgul_rbtree__get_youngest(cgul_exception_t* cex,
                          cgul_rbtree_t t)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->youngest;

 out:
    return rv;
}

void
cgul_rbtree__set_youngest(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(t->youngest);
    assert(n);

    /* There is nothing to do if n is already the youngest. */
    if (n == t->youngest) {
        goto out;
    }

    /* If n is currently the oldest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new oldest node. */
    if (n == t->oldest) {
        t->oldest = n->younger;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the youngest node.  We assert t->youngest above because
     * any non-empty tree has an youngest node, and because n is a node
     * in t, t is non-empty. */
    t->youngest->younger = n;
    n->older = t->youngest;
    n->younger = NULL;

    /* Remember that n is now the youngest node. */
    t->youngest = n;

 out:
    return;
}

/* To understand how to keep the tree balanced when you remove a
 * node, you have to understand how the tree is kept balanced when
 * inserting a node.  See the comments directly above the
 * definition of cgul_rbtree__insert() above then return here.
 *
 * I bring all this up because it is important to understand that, as
 * you start deleting nodes from one subtree, you can introduce
 * imbalances.  So long as you do not delete the last key/value pair
 * from _any_ node (unless this the last key/value pair in the tree),
 * you will keep the tree balanced.  You can do this if, as you
 * descend the tree, you make sure that no node has just one key/value
 * pair.  Then, when you get to the node you want to delete, it will
 * always share the 2-3-4 node with another key/value pair.
 *
 * You have two techniques at your disposal for making sure all the
 * nodes you pass through have more than one key/value pair.  The
 * preferred technique is to find an adjacent sibling in the
 * corresponding 2-3-4 tree and rotate a value up out of the sibling
 * into the parent in such a way that the parent is forced to rotate a
 * value down to the node you are currently visiting.  Because you
 * only rotate when the current node has just one key/value pair, you
 * know there is room in the node for the value passed down from the
 * parent.  Thus, the effects of the rotation are not felt futher down
 * the tree.
 *
 * Sometimes, however, you cannot rotate.  The first example of this
 * is when you are visiting the root node.  Because the root has no
 * parents or siblings, it does not make sense to try to rotate a
 * value out of a non-existing sibling in order to force a
 * non-existing parent to rotate a value down to the root node.  The
 * second example of when you cannot rotate is when both adjacent
 * siblings from the corresponding 2-3-4 tree sibling also have just
 * one key/value pair.  At these times, you must "unsplit".
 *
 * The "unsplit" operation is the second technique at your disposal to
 * allow you to guarantee that all nodes through which you pass
 * durning a remove operation have more than one key/value pair.  The
 * unsplit operation is simply the reverse of the split operation.
 * The unsplit operation works for the exact case for which the rotate
 * operation is impossible.  When the adjacent siblings from the
 * corresponding 2-3-4 tree also have only one key/value pair, it is
 * possible to combine one key/value pair from the parent node with
 * the current node and one of the adjacent siblings in order to form
 * a 3 key/value pair 2-3-4 node.  You cannot take the last key/value
 * pair from a parent, but as we are descending we are guaranteeing
 * that each node through which we pass has more than one key/value
 * pair.  Thus, the unsplit operation is guaranteed to work.
 *
 * When you find the node you need to remove, it might be located
 * somewhere in the middle of the tree (with two children).  You then
 * have to resort to the usual binary tree node removal technique in
 * which you search for the next node (in search order).  You can then
 * move the next node and have it take the place of the node you are
 * removing.  This obviously maintains the search order. */
int
cgul_rbtree__remove(cgul_exception_t* cex,
                    cgul_rbtree_t t,
                    const void* key_in,
                    void** key_out,
                    void** value_out)
{
    int rv = 0;

    /* Node to remove. */
    cgul_rbtree_node_t n = NULL;

    /* Node to replace n with if necessary. */
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    if (t->root) {
        /* Walk down the tree rotating and unsplitting as necessary.
         * After this function call, if the node to remove was found,
         * n will point to it; otherwise, n will be NULL.  If n has
         * two children, this function call will continue walking down
         * the tree rotating and unsplitting as necessary until it
         * finds the node that follows n in sort order.  It will
         * return this node as "next". */
        cgul_rbtree__find_node_to_remove(cex, t, key_in, &n, &next);

        if (n) {

            /* While "n" is still valid, fetch its "key" and "value"
             * pointers. */
            if (key_out) {
                *key_out = (void*)n->key;
            }
            if (value_out) {
                *value_out = (void*)n->value;
            }

            /* Update the lcount and rcount while we can still
             * dereference next and n. */
            if (t->is_indexed) {
                cgul_rbtree__decrement_counts_along_path(cex,
                                                         (next ? next : n));
            }

            /* Remove the node. */
            cgul_rbtree__remove_node_after_proper_descent(cex, t, n, next);

            rv = 1;
        }
    }

 out:
    return rv;
}

void
cgul_rbtree__remove_node(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree_node_t n,
                         void** key_out,
                         void** value_out)
{
    const void* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* Remove the node. */
    key = cgul_rbtree_node__get_key(cex, n);
    cgul_rbtree__remove(cex, t, key, key_out, value_out);

 out:
    return;
}

int
cgul_rbtree__remove_at(cgul_exception_t* cex,
                       cgul_rbtree_t t,
                       unsigned long int index,
                       void** key_out,
                       void** value_out)
{
    int rv = 0;
    cgul_rbtree_node_t n = NULL;
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    /* index is the C-style index. */
    if (index < t->size) {

        assert(t->root);

        /* Walk down the tree rotating and unsplitting as necessary.
         * After this function call, if the node to remove was found,
         * n will point to it; otherwise, n will be NULL. */
        n = cgul_rbtree__find_node_at(cex, t, index, 1 /* is_for_removal */);
        if (*cex) {
            goto out;
        }

        if (n) {

            /* If n has two children, this function call will continue
             * walking down the tree rotating and unsplitting as
             * necessary until it finds the node that follows n in
             * sort order.  It will return this node as "next". */
            if (!cgul_rbtree_node__is_234_leaf_node(cex, n)) {
                next = cgul_rbtree__find_next_with_intent_to_remove(cex, t, n);
            }

            /* While "n" is still valid, fetch its "value" pointer. */
            if (key_out) {
                *key_out = (void*)n->key;
            }
            if (value_out) {
                *value_out = (void*)n->value;
            }

            /* Update the lcount and rcount while we can still
             * dereference next and n. */
            cgul_rbtree__decrement_counts_along_path(cex, (next ? next : n));

            /* Remove the node. */
            cgul_rbtree__remove_node_after_proper_descent(cex, t, n, next);

            rv = 1;
        }
    }

 out:
   return rv;
}

int
cgul_rbtree__remove_front(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          void** key_out,
                          void** value_out)
{
    int rv = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    if (t->front) {

        /* We MUST descend the tree in order to insure that it stays
         * balanced after we remove the node. */
        n = t->root;
        while (n != NULL) {

            /* Rotate or unsplit the node if necessary.  You only need
             * to rotate or unsplit when you first enter the 2-3-4
             * node, i.e., when you arrive at a black node. */
            if (!n->red) {
                cgul_rbtree__insure_node_has_more_than_one_key(cex, t, n);
            }

            n = n->left;
        }

        /* Return the "value" pointer while t->front is still valid. */
        if (key_out) {
            *key_out = (void*)t->front->key;
        }
        if (value_out) {
            *value_out = (void*)t->front->value;
        }

        /* The front node is always in a 2-3-4 leaf node.  So we do not
         * need to pass in a replacement node. */
        cgul_rbtree__remove_node_after_proper_descent(cex, t, t->front, NULL);
        rv = 1;
    }

 out:
    return rv;
}

int
cgul_rbtree__remove_back(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         void** key_out,
                         void** value_out)
{
    int rv = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Clear the user's output parameters. */
    if (key_out) {
        *key_out = NULL;
    }
    if (value_out) {
        *value_out = NULL;
    }

    if (t->back) {

        /* We MUST descend the tree in order to insure that it stays
         * balanced after we remove the node. */
        n = t->root;
        while (n != NULL) {

            /* Rotate or unsplit the node if necessary.  You only need
             * to rotate or unsplit when you first enter the 2-3-4
             * node, i.e., when you arrive at a black node. */
            if (!n->red) {
                cgul_rbtree__insure_node_has_more_than_one_key(cex, t, n);
            }

            n = n->right;
        }

        /* Return the "value" pointer while t->back is still valid. */
        if (key_out) {
            *key_out = (void*)t->back->key;
        }
        if (value_out) {
            *value_out = (void*)t->back->value;
        }

        /* The back node is always in a 2-3-4 leaf node.  So we do not
         * need to pass in a replacement node. */
        cgul_rbtree__remove_node_after_proper_descent(cex, t, t->back, NULL);
        rv = 1;
    }

 out:
    return rv;
}

extern int
cgul_rbtree__remove_range(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree_node_t first,
                          cgul_rbtree_node_t last,
                          cgul_cache_t keys_cache,
                          cgul_cache_t values_cache)
{
    int rv = 0;
    struct cgul_rbtree__clear_cb_data data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    data.keys_cache = keys_cache;
    data.values_cache = values_cache;
    data.rv = 0;

    cgul_rbtree__traverse_range(cex,
                                t,
                                first,
                                last,
                                &cgul_rbtree__clear_cb,
                                &data);

    rv = data.rv;

 out:
    return rv;
}

void
cgul_rbtree__clear(cgul_exception_t* cex,
                   cgul_rbtree_t t,
                   cgul_cache_t keys_cache,
                   cgul_cache_t values_cache)
{
    struct cgul_rbtree__clear_cb_data data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    data.keys_cache = keys_cache;
    data.values_cache = values_cache;
    data.rv = 0;

    cgul_rbtree__traverse(cex,
                          t,
                          &cgul_rbtree__clear_cb,
                          &data);

 out:
    return;
}

unsigned long int
cgul_rbtree__get_size(cgul_exception_t* cex,
                      cgul_rbtree_t t)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->size;

 out:
    return rv;
}

int
cgul_rbtree__is_ok(cgul_exception_t* cex,
                   cgul_rbtree_t t)
{
    int rv = 1;
    int max_black_depth = -1;
    int min_black_depth = -1;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* Search for NULL children in non-leaf nodes. */
    for (n = t->root ; n != NULL ; n = cgul_rbtree_node__get_next(cex, n)) {

        /* Because of the insertion and deletion algorithm NULL
         * children are only allowed in leaf nodes.  Even then if one
         * child is NULL the other must be NULL too unless this is a
         * 3-child leaf node. */

        /* Verify that there are never two consecutive red nodes. */
        if (n->red) {
            if (n->left && n->left->red) {
                rv = 0;
                goto out;
            }
            if (n->right && n->right->red) {
                rv = 0;
                goto out;
            }
        }

        /* Your leaf nodes should all be at the same black depth or
         * one level up.  As we walk the tree collect black depth
         * information. */
        if ((n->left == NULL) || (n->right == NULL)) {
            int real_black_depth = cgul_rbtree_node__get_black_depth(cex, n);
            if (min_black_depth == -1) {
                min_black_depth = real_black_depth;
            } else if (real_black_depth < min_black_depth) {
                min_black_depth = real_black_depth;
            } else if (real_black_depth > max_black_depth) {
                max_black_depth = real_black_depth;
            }
        }

    }

    if (max_black_depth - min_black_depth > 1) {
        rv = 0;
        goto out;
    }

 out:
    return rv;
}

void
cgul_rbtree__swap(cgul_exception_t* cex,
                  cgul_rbtree_t t1,
                  cgul_rbtree_t t2)
{
    struct cgul_rbtree tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t1);
    assert(t2);

    /* Nothing to swap if they point to the same objects. */
    if (t1 == t2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *t1;
    *t1 = *t2;
    *t2 = tmp;

 out:
    return;
}

void
cgul_rbtree__foldl_keys(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree__fold_key_t f,
                        void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(f);

    /* Iterate over each node in the tree. */
    for (curr = t->front ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_rbtree_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_rbtree__foldr_keys(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree__fold_key_t f,
                        void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(f);

    /* Iterate over each node in the tree. */
    for (curr = t->back; curr; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = cgul_rbtree_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_rbtree__foldl_values(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree__fold_value_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(f);

    /* Iterate over each node in the tree. */
    for (curr = t->front ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_rbtree_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_rbtree__foldr_values(cgul_exception_t* cex,
                          cgul_rbtree_t t,
                          cgul_rbtree__fold_value_t f,
                          void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(f);

    /* Iterate over each node in the tree. */
    for (curr = t->back; curr; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = cgul_rbtree_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_rbtree__foldl_pairs(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree__fold_pair_t f,
                         void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(f);

    /* Iterate over each node in the tree. */
    for (curr = t->front ; curr ; curr = next) {
        /* Get the next node now in case f deletes curr. */
        next = cgul_rbtree_node__get_next(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_rbtree__foldr_pairs(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree__fold_pair_t f,
                         void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(f);

    /* Iterate over each node in the tree. */
    for (curr = t->back; curr; curr = prev) {
        /* Get the prev node now in case f deletes curr. */
        prev = cgul_rbtree_node__get_prev(cex, curr);
        /* Call the combining function. */
        f_rv = (*f)(cex, curr->key, (void*)curr->value, data);
        if (*cex || !f_rv) {
            goto out;
        }
    }

 out:
    return;
}

void
cgul_rbtree__traverse(cgul_exception_t* cex,
                      cgul_rbtree_t t,
                      cgul_rbtree__traverse_t f,
                      void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_rbtree__traverse_range(cex, t, NULL, NULL, f, data);

 out:
    return;
}

void
cgul_rbtree__traverse_range(cgul_exception_t* cex,
                            cgul_rbtree_t t,
                            cgul_rbtree_node_t first,
                            cgul_rbtree_node_t last,
                            cgul_rbtree__traverse_t f,
                            void* data)
{
    int f_rv = 0;
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert(first) nor assert(last) because NULL values
     * indicate that we should begin or end with the first or last
     * node of the entire tree respectively. */
    assert(t);
    assert(f);

    if (first == NULL) {
        first = t->front;
    }

    if (last == NULL) {
        last = t->back;
    }

    /* In order to provide "f" with a context in which it is safe for
     * "f" to delete curr, you have to be careful to dereference
     * curr before calling "f". */
    for (curr = first ; curr ; curr = next) {
        next = cgul_rbtree_node__get_next(cex, curr);
        f_rv = (*f)(cex, t, curr, data);
        if (*cex) {
            goto out;
        }
        /* We are done if f returned false or if we just processed the
         * last node. */
        if (!f_rv || (curr == last)) {
            break;
        }
    }

 out:
    return;
}
