#include "cgul_rbtree_node_p.h"
#include "cgul_libc.h"

/*
 * This method is used as the "restructor" for the "cgul_cache" of
 * nodes in the "cgul_rbtree".  "restructor" methods are defined in
 * "cgul_cache" to not throw exceptions.
 *
 * While the signature for this method does not exactly match the
 * signature defined for "restructor" methods, it can be safely
 * typecast to match.
 */
void
cgul_rbtree_node__initialize(cgul_rbtree_node_t n)
{
    assert(n);
    memset(n, 0, sizeof(*n));
}

cgul_rbtree_node_t
cgul_rbtree_node__new(cgul_exception_t* cex)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_rbtree_node_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_rbtree_node__initialize(rv);

 out:

    /* Delete partially constructed objects. */
    if (*cex && rv) {
        cgul_rbtree_node__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_rbtree_node__delete(cgul_rbtree_node_t n)
{
    /* If you alter this, you will also likely need to alter
     * cgul_rbtree_node_with_indexing__delete() which inherits from
     * this class. */
    if (n) {
        free(n);
    }
}

void*
cgul_rbtree_node__get_key(cgul_exception_t* cex,
                          cgul_rbtree_node_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = (void*)n->key;

 out:
    return rv;
}

void
cgul_rbtree_node__set_key(cgul_exception_t* cex,
                          cgul_rbtree_node_t n,
                          const void* key)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->key = key;

 out:
    return;
}

void*
cgul_rbtree_node__get_value(cgul_exception_t* cex,
                            cgul_rbtree_node_t n)
{
    void * rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = (void*)n->value;

 out:
    return rv;
}

void
cgul_rbtree_node__set_value(cgul_exception_t* cex,
                            cgul_rbtree_node_t n,
                            const void* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->value = value;

 out:
    return;
}

int
cgul_rbtree_node__is_red(cgul_exception_t* cex,
                         cgul_rbtree_node_t n)
{
    int rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->red;

 out:
    return rv;
}

int
cgul_rbtree_node__is_black(cgul_exception_t* cex,
                           cgul_rbtree_node_t n)
{
    int rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = !(n->red);

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_prev(cgul_exception_t* cex,
                           cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /* If n has a left child, the next node is the largest value in
     * the left child's subtree. */
    if (n->left) {
        rv = cgul_rbtree_node__get_largest_node_in_subtree(cex, n->left);
    } else {
        /* If n does not have a left child, you have to walk up the
         * tree.  The prev node is going to be the first ancestor of n
         * that also has an ancestor of n attached as its right child.
         *
         * As a point of reference, if n is its parent's right child
         * and n does not have a left child (as specified by the IF
         * statement above), then the n's parent is the prev node.
         * The tricky case is what happens if n is its parent's left
         * child and n does not have a left child of its own.  n's
         * parent cannot be the prev node because the parent is
         * greater than n.
         */
        rv = n->parent;
        while (rv) {
            /* If rv's right child is n, then rv is the prev node. */
            if (n == rv->right) {
                break;
            }
            n = rv;
            rv = rv->parent;
        }
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_next(cgul_exception_t* cex,
                           cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /* If n has a right child, the next node is the smallest value in
     * the right child's subtree. */
    if (n->right) {
        rv = cgul_rbtree_node__get_smallest_node_in_subtree(cex, n->right);
    } else {
        /* If n does not have a right child, you have to walk up the
         * tree.  The next node is going to be the first ancestor of n
         * that also has an ancestor of n attached as its left
         * child.
         *
         * As a point of reference, if n is its parent's left child
         * and n does not have a right child (as specified by the IF
         * statement above), then the n's parent is the next node.
         * The tricky case is what happens if n is its parent's right
         * child and n does not have a right child of its own.  n's
         * parent cannot be the next node because the parent is
         * less than n.
         */
        rv = n->parent;
        while (rv) {
            /* If rv's left child is n, then rv is the next node. */
            if (n == rv->left) {
                break;
            }
            n = rv;
            rv = rv->parent;
        }
    }

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_older(cgul_exception_t* cex,
                            cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->older;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_younger(cgul_exception_t* cex,
                              cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->younger;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_smallest_node_in_subtree(cgul_exception_t* cex,
                                               cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    while (n->left) {
        n = n->left;
    }

    rv = n;

out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_largest_node_in_subtree(cgul_exception_t* cex,
                                              cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    while (n->right) {
        n = n->right;
    }

    rv = n;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_left_child(cgul_exception_t* cex,
                                 cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->left;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_right_child(cgul_exception_t* cex,
                                  cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->right;

 out:
    return rv;
}

cgul_rbtree_node_t
cgul_rbtree_node__get_sibling(cgul_exception_t* cex,
                              cgul_rbtree_node_t n)
{
    cgul_rbtree_node_t rv = NULL;
    cgul_rbtree_node_t p = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    p = n->parent;
    if (p) {
        if (n == p->left) {
            rv = p->right;
        } else {
            assert(n == p->right);
            rv = p->left;
        }
    }

 out:
    return rv;
}

int
cgul_rbtree_node__is_234_leaf_node(cgul_exception_t* cex,
                                   cgul_rbtree_node_t n)
{
    int rv = 0;
    cgul_rbtree_node_t left = NULL;
    cgul_rbtree_node_t right = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    if (n->red) {
        n = n->parent;
        assert(n);
    }

    assert(!n->red);

    left = n->left;
    right = n->right;

    /*
     * There are 4 configurations for leaf nodes.
     */

    /* Black node with no children. */
    if ((left == NULL) && (right == NULL)) {
        rv = 1;
        goto out;
    }

    /* Black node with only a left red child that has no children. */
    if (right == NULL) {
        assert(left->red);
        if ((left->left == NULL) && (left->right == NULL)) {
            rv = 1;
            goto out;
        }
    }

    /* Black node with only a right red child that has no children. */
    if (left == NULL) {
        assert(right->red);
        if ((right->left == NULL) && (right->right == NULL)) {
            rv = 1;
            goto out;
        }
    }

    /* Black node with two red children that have no children. */
    if (   ( left->left == NULL) && ( left->right == NULL)
        && (right->left == NULL) && (right->right == NULL))
    {
        rv = 1;
        goto out;
    }

 out:
    return rv;
}

int
cgul_rbtree_node__get_black_depth(cgul_exception_t* cex,
                                  cgul_rbtree_node_t n)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    for ( ; n != NULL ; n = n->parent) {
        if (!n->red) {
            ++rv;
            assert(rv > 0);
        }
    }

 out:
    return rv;
}

int
cgul_rbtree_node__get_raw_depth(cgul_exception_t* cex,
                                cgul_rbtree_node_t n)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    for ( ; n != NULL ; n = n->parent) {
        ++rv;
        assert(rv > 0);
    }

 out:
    return rv;
}
