#include "cgul_rolling_hash.h"
#include "cgul_sbox.h"

cgul_uint32_t
cgul_rolling_hash__hash(cgul_exception_t* cex,
                        const char* block,
                        size_t block_length)
{
    size_t i = 0;
    cgul_uint32_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!block) {
        goto out;
    }

    /*
     * This is what the "Rolling Hash" page on Wikipedia calls a
     * "cyclic polynomial" hash.  This is just a rotation followed by
     * a table lookup based directly off of the byte being hashed.
     * Originally, I tried the Rabin-Karp hash function but the
     * multiple modulo operations per byte killed performance.
     */

    /* Iterate over each character in the block. */
    for (i = 0 ; i < block_length ; ++i) {
        rv = ((rv << 1) | (rv >> 31))
            ^ cgul_sbox[(unsigned char)block[i]];
    }

 out:
    return rv;
}

cgul_uint32_t
cgul_rolling_hash__roll(cgul_exception_t* cex,
                        cgul_uint32_t hv,
                        const char c_old,
                        const char c_new,
                        size_t block_length)
{
    size_t rots = 0;
    cgul_uint32_t c = 0;
    cgul_uint32_t rv = hv;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (block_length <= 0) {
        goto out;
    }

    /* Determine the effective number of rotations.  We check above to
     * make sure block_length is at least 1.  The effective rotation
     * count can be determined by simply masking off the high-order
     * bits. */
    rots = (block_length & 0x1f) - 1;

    /* Remove the old byte. */
    c = cgul_sbox[(unsigned char)c_old];
    rv ^= ((c << rots) | (c >> (32 - rots)));

    /* Rotate and add the new byte. */
    rv = ((rv << 1) | (rv >> 31)) ^ cgul_sbox[(unsigned char)c_new];

 out:
    return rv;
}
