/*
 * This program reads base64 data from stdin, converts it to binary
 * data, and writes the resulting conversion to stdout.  It is a good
 * idea to redirect stdout to a file when running the program.
 */

#include "cgul.h"

#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#endif

int main(int argc, char* argv[])
{
    int rv = 0;
    size_t rcount = (size_t)-1;
    size_t wcount = (size_t)-1;
    cgul_exception_t cex = NULL;
    cgul_base64_t b64 = NULL;
    char* wbuf = NULL;
    char rbuf[4096];

#ifdef WIN32
    /* Convert stdin and stdout to binary mode. */
    if (_setmode(_fileno(stdin), _O_BINARY) < 0) {
        cex = cgul_exception__new("_setmode(stdin): %m");
        goto out;
    }
    if (_setmode(_fileno(stdout), _O_BINARY) < 0) {
        cex = cgul_exception__new("_setmode(stdout): %m");
        goto out;
    }
#endif

    /* Get the base64 object ready. */
    b64 = cgul_base64__new(&cex);
    cgul_base64__set_strict_decode(&cex, b64, 0);
    if (cex) {
        goto out;
    }

    while (cex == NULL) {

        /* Read a block from stdin. */
        rcount = cgul_libc__fread(&cex, rbuf, 1, sizeof(rbuf), stdin);
        if (cex) {
            goto out;
        }

        /* EOF */
        if (rcount == 0) {
            break;
        }

        /* Base64 decode the block. */
        cgul_base64__decode(&cex, b64, rbuf, rcount);
        if (cex) {
            goto out;
        }

        /* Write the decoding as it exists so far. */
        wbuf = cgul_base64__get_decode_conversion(&cex, b64, &wcount);
        cgul_libc__fwrite(&cex, wbuf, wcount, 1, stdout);
        if (cex) {
            goto out;
        }
    }

    /* Flush and print any remaining bytes. */
    cgul_base64__decode_finish(&cex, b64);
    wbuf = cgul_base64__get_decode_conversion(&cex, b64, &wcount);
    cgul_libc__fwrite(&cex, wbuf, wcount, 1, stdout);
    if (cex) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (cex) {
        const char* msg = cgul_exception__get_message(cex);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(cex);
        cex = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (b64) {
        cgul_base64__delete(b64);
    }

    return rv;
}
