#include "cgul.h"

/**
 * Insert the key <code>key</code> into the trie <code>trie</code>.
 * The caller is responsible for calling <code>free()</code> on the
 * key inserted into the trie.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie trie
 * @param[in] key key
 */
static void
main__insert_key(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 const char* key)
{
    int insert_rv = 0;
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    /* Sanity check. */
    if (!key) {
        goto out;
    }

    /* Duplicate the key. */
    tmp = cgul_libc__strdup(cex, key);
    if (*cex) {
        goto out;
    }

    /* Insert the key into the trie. */
    insert_rv = cgul_trie__insert(cex, trie, tmp, NULL);
    if (*cex) {
        goto out;
    }
    if (insert_rv) {
        /* Transfer ownership of the key to the trie. */
        tmp = NULL;
    }

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}

/**
 * Print the keys in the trie <code>trie</code> in sort order.
 *
 * @param[in] cex c-style exception
 * @param[in] trie trie
 */
static void
main__print_keys(cgul_exception_t* cex,
                 cgul_trie_t trie)
{
    const char* key = NULL;
    cgul_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    curr = cgul_trie__get_front(cex, trie);
    for ( ; curr ; curr = cgul_trie_node__get_next(cex, curr)) {
        key = cgul_trie_node__get_key(cex, curr);
        printf("%s\n", key);
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_trie_t trie = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    trie = cgul_trie__new(&local);
    if (local) {
        goto out;
    }

    /* Insert keys. */
    main__insert_key(&local, trie, "foo");
    main__insert_key(&local, trie, "bar");
    main__insert_key(&local, trie, "baz");
    if (local) {
        goto out;
    }

    /* Print the keys in sort order. */
    printf("===== Before call to cgul_trie__clear(). \n");
    main__print_keys(&local, trie);

    /* Clear the keys. */
    cgul_trie__clear(&local, trie, cgul_cache__get_freer(), NULL);

    /* Print the keys in sort order. */
    printf("===== After call to cgul_trie__clear(). \n");
    main__print_keys(&local, trie);

    /* Make sure cgul_cache__delete() ignores the request to delete
     * the *statically* allocated "freer" instance. */
    cgul_cache__delete(cgul_cache__get_freer());

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        /* The should have been freed above. */
        cgul_trie__delete(trie);
    }

    return rv;
}
