#include "cgul.h"

/**
 * In-memory CSV "file."
 */
static const char* memory = "\
header1,header2,header3,header4,header5,header6\n\
Now is the,time for all,good men to,come to the,aid of their,country.\n\
\"Hello, World!\n\
Goodbye, Cruel World!\",\"He said, \"\"Are you talking to me?\"\"\",,,,\n\
\"\"\"Hello, World!\"\"\",,,,,IN-MEMORY!!!\n\
";

/**
 * Print the rows and columns in the CSV file <code>csv</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] csv CSV file
 * @param[in] fout output file
 */
static void
print_csv(cgul_exception_t* cex,
          cgul_csv_t csv,
          FILE* fout)
{
    unsigned long int i = 0;
    unsigned long int n = 0;
    const char* field = NULL;
    const char* header = NULL;
    cgul_rbtree_t headers = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the headers for the cgul_csv instance. */
    headers = cgul_csv__get_headers(cex, csv);

    /* Iterate over each row in the csv file. */
    while (cgul_csv__load_next(cex, csv)) {
        if (*cex) {
            goto out;
        }

        /* Iterate over each field. */
        n = cgul_csv__get_number_of_fields(cex, csv);
        for (i = 0 ; i < n ; ++i) {
            field = cgul_csv__lookup(cex, csv, i);
            cgul_libc__fprintf(cex, fout, "field[%lu] = %s\n", i, field);
            if (*cex) {
                goto out;
            }
        }

        /* Iterate again but use the headers to do the lookup. */
        node = cgul_rbtree__get_front(cex, headers);
        for ( ; node ; node = cgul_rbtree_node__get_next(cex, node)) {
            header = cgul_rbtree_node__get_key(cex, node);
            field = cgul_csv__lookup_by_header(cex, csv, header);
            if (field) {
                cgul_libc__fprintf(
                    cex, fout, "field[\"%s\"] = %s\n", header, field);
            }
        }

        /* Print separator. */
        cgul_libc__fprintf(cex, fout, "=====\n");
        if (*cex) {
            goto out;
        }

    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_csv_t csv = NULL;
    cgul_crlf_file_t f = NULL;
    cgul_exception_t local = NULL;

    /* Create the interface that will be used to read lines of text
     * from a data source. */
    f = cgul_crlf_file__new_from_fname(&local,
                                       "main_csv_input.csv",
                                       1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Create a new cgul_csv instance and load a CSV file. */
    csv = cgul_csv__new_from_crlf_file(&local, f);
    if (local) {
        goto out;
    }

    /* Print the CSV file. */
    print_csv(&local, csv, stdout);
    if (local) {
        goto out;
    }

    printf("\n\n");

    /* Open memory. */
    cgul_crlf_file__open_memory(&local, f, memory, strlen(memory));
    cgul_csv__open_crlf_file(&local, csv, f);
    if (local) {
        goto out;
    }

    /* Print the CSV "file" resident in memory. */
    print_csv(&local, csv, stdout);
    if (local) {
        goto out;
    }

    printf("\n\n");

    /* Open the file again. */
    cgul_crlf_file__open_fname(&local,
                               f,
                               "main_csv_input.csv",
                               1 /* is_block_buffered */);
    cgul_csv__open_crlf_file(&local, csv, f);
    if (local) {
        goto out;
    }

    /* Print the CSV file. */
    print_csv(&local, csv, stdout);
    if (local) {
        goto out;
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (csv) {
        cgul_csv__delete(csv);
    }
    if (f) {
        cgul_crlf_file__delete(f);
    }

    return rv;
}
