/**
 * This program reads lines of text from <code>stdin</code>,
 * calculates the hash, and prints each to <code>stdout</code>.
 *
 * @file
 * @brief dump hash values for each line in a text file
 */

#include "cgul.h"

/**
 * This function calculates the hash value of the C-style string
 * <code>s</code>.  There are two flavors of this function depending
 * on whether you want to test <code>cgul_hash__hash_block()</code> or
 * a custom hash function.
 *
 * @param[in] s input string
 * @return hash value of <code>s</code>
 */
static inline unsigned int
calc_hash(const char* s)
{
#if 1
    /* Calculate the hash value. */
    return cgul_hash__hash_block((void*)s, (unsigned int)strlen(s));
#else
    /* Calculate the hash value. */
    unsigned c;
    unsigned int rv = 0;
    for (c = *s ; c ; c = *++s) {
        rv = rv * 31 + c;
    }
    return rv;
#endif
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned int hv = 0;
    const char* line = NULL;
    cgul_crlf_file_t crlf = NULL;
    cgul_exception_t local = NULL;

    /* Check the command line. */
    if (argc != 1) {
        fprintf(stderr, "Usage: main_hash_dump\n");
        rv = 1;
        goto out;
    }

    /* Allocate crlf. */
    crlf = cgul_crlf_file__new_from_file(&local,
                                         stdin,
                                         1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Calculate and print the hash of each line. */
    while ((line = cgul_crlf_file__get_line(&local, crlf)) != NULL) {
        if (local) {
            goto out;
        }
        hv = calc_hash(line);
        printf("%u\n", hv);
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (crlf) {
        cgul_crlf_file__delete(crlf);
    }

    return rv;
}
