/**
 * This program simulates feeding long, sequential blocks of zeros to
 * hash functions. <p>
 *
 * @file
 * @brief test hash against blocks of zeros
 */

#include "cgul.h"

/*
 * IMPORTANT:
 * =========
 *
 * The hash functions in this program are different from real hash
 * function in that they are not passed an entire block of data to
 * hash.  This would be inefficient because we are just passing the
 * hash function successively larger blocks that are all zero, and the
 * hash value up to the last, new byte does not change.  Instead,
 * <code>hv</code> is passed in with the previous hash value and
 * returns the new hash value calculated by "appending" another zero
 * byte to the imaginary block of zeros. <p>
 */

/*
 * Jenkins One-at-a-Time
 */
#if 1
#define INITIAL_HASH_BIAS 0xaa87231f
unsigned int
next_iteration(void* block,
               unsigned int block_size,
               unsigned int* prev)
{
    /* Pick up from where we left off. */
    unsigned int rv = *prev;

    /* This is the equivalent of the main loop. */
    rv += '\0';        /* Next byte is always zero. */
    rv += (rv << 10);
    rv ^= (rv >> 6);

    /* This is the value before post-processing which is what we need
     * to feed to the "main loop" next time. */
    *prev = rv;

    /* Post-processing. */
    rv += (rv << 3);
    rv ^= (rv >> 11);
    rv += (rv << 15);

    return rv;
}
#endif

/*
 * Multiplicative hash Function
 */
#if 0
#define INITIAL_HASH_BIAS 0xaa87231f
unsigned int
next_iteration(void* block,
               unsigned int block_size,
               unsigned int* prev)
{
    /* Pick up from where we left off. */
    unsigned int rv = *prev;

    /* This is the equivalent of the main loop. */
    rv += (rv << 5);
    rv += '\0';       /* Next byte is always zero. */

    /* Save "prev" before returning. */
    *prev = rv;

    return rv;
}
#endif

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int insert_rv = 0;
    unsigned int hv = 0;
    unsigned int prev = INITIAL_HASH_BIAS;
    unsigned long int i = 0;
    unsigned long int collisions = 0;
    char* key = NULL;
    cgul_hash_t h = NULL;
    cgul_string_t s = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    h = cgul_hash__new(&local);
    s = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Get the hash values for a large range of blocks all of which
     * have each byte set to zero. */
    for (i = 0 ; i <= 1000000 ; ++i) {

        /* Get the next hash value. */
        hv = next_iteration(NULL, i, &prev);

        /* For simplicity, convert the hash value to a string that we
         * can use as a key to insert into a working hash object.
         * (The hash object into which we are inserting is totally
         * unrelated to the hash function we are testing.  We are just
         * using it as a way to detect collisions.  A red-black tree
         * would also work.) */
        cgul_string__set_value(&local, s, "");
        cgul_string__append_sprintf(&local, s, "%#x", hv);
        if (local) {
            goto out;
        }

        /* Get the key to insert. */
        key = cgul_string__get_value(&local, s);
        //printf("%s\n", key);

        /* Try to insert the key into the working hash object. */
        insert_rv = cgul_hash__insert(&local, h, key, NULL);
        if (local) {
            goto out;
        }

        /* Check if the key could be inserted. */
        if (insert_rv) {
            /* The key was inserted which means there was no other key
             * in the hash with that value which means "hv" did not
             * collide with any other value.  The hash now owns the
             * string. */
            cgul_string__take_value(&local, s);
        } else {
            /* The key could not be inserted because there was a
             * collision. */
            ++collisions;
        }
    }

    /* Print final number of collisions. */
    printf("collisions = %lu\n", collisions);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (h) {
        /* Free each key. */
        while (cgul_hash__remove_back(&local, h, (void**)&key, NULL)) {
            free(key);
        }
        /* Free the hash itself. */
        cgul_hash__delete(h);
    }
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}
