#include "cgul.h"

/*
 * To change which sorting function is being profiled, set exactly one
 * of PROFILE_CGUL_DEFAULT, PROFILE_CGUL_ALT, or PROFILE_QSORT to 1.
 */

/**
 * Whether to profile the default cgul_merge_sort() implementation
 * (which is the recursive implementation which is 30% faster than the
 * iterative implementation but uses 50% more memory).
 */
#define PROFILE_CGUL_DEFAULT 1

/**
 * Whether to profile the alternative cgul_merge_sort() implementation
 * (which is the iterative implementation).
 */
#define PROFILE_CGUL_ALT 0

/**
 * Whether to profile qsort().
 */
#define PROFILE_QSORT 0

/**
 * Whether to seed the PRNG with the constant LCG_MAX_RANGE;
 * otherwise, seed with time().  Set this to 1 when you want to
 * compare the performance of cgul_merge_sort() with qsort().  Set
 * this to 0 when you want to test whether cgul_merge_sort() is
 * working properly by having it sort a new data sets of a different
 * size each time the value of time() changes.
 */
#define CONST_SEED 1

/**
 * Whether to print the unsorted and sorted vectors.
 */
#define ENABLE_PRINT 0

/**
 * This variable controls the range of the pseudo-random number
 * generator.  The numbers generated by the pseudo-random number
 * generator control how many numbers to sort as well as the actual
 * numbers that get sorted.  If CONST_SEED is defined to 1,
 * this value is also used to seed the PRNG instead of using time().
 */
static unsigned long int LCG_MAX_RANGE = 10000000;

/**
 * Structure used to test <code>cgul_vector__sort()</code> (which uses
 * <code>cgul_merge_sort_pointers()</code>).  The <code>x</code> value
 * is set to a random number and is the only part of the structure
 * checked by the comparison functions
 * <code>compare_points_cgul()</code> and
 * <code>compare_points_qsort()</code>.  The <code>y</code> value is
 * incremented.  This is used to make sure the sort is stable.
 */
typedef struct point {
    unsigned long int x;
    unsigned long int y;
} *point_t;

#if PROFILE_CGUL_DEFAULT || PROFILE_CGUL_ALT
static int
compare_points_cgul(point_t p1,
                    point_t p2)
{
    /* This function does not need the extra layer of indirection that
     * PROFILE_QSORT needs because this is the call back for
     * cgul_merge_sort_pointers() accessed indirectly through
     * cgul_vector__sort().  If we were testing cgul_merge_sort()
     * directly, then we need the same callback function as
     * PROFILE_QSORT uses. */
    int rv = 0;
    if (p1->x < p2->x) {
        rv = -1;
    } else if (p1->x > p2->x) {
        rv = 1;
    }
    return rv;
}
#endif

#if PROFILE_QSORT
static int
compare_points_qsort(const point_t* p1,
                     const point_t* p2)
{
    int rv = 0;
    if ((*p1)->x < (*p2)->x) {
        rv = -1;
    } else if ((*p1)->x > (*p2)->x) {
        rv = 1;
    }
    return rv;
}
#endif

#if ENABLE_PRINT
static void
print_points(cgul_exception_t* cex,
             cgul_vector_t v)
{
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    point_t* points = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the underlying array for the vector. */
    points = (point_t*)cgul_vector__get_array(cex, v);

    /* Iterate over each element in the array. */
    i_max = cgul_vector__get_size(cex, v);
    for (i = 0 ; i < i_max ; ++i) {
        printf("(%6lu, %6lu)\n", points[i]->x, points[i]->y);
    }

    puts("=====");

 out:
    return;
}
#endif

static void
insert_random_points(cgul_exception_t* cex,
                     cgul_vector_t v,
                     cgul_random_lcg32_t lcg,
                     cgul_uint32_t rmax)
{
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    point_t p = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next random number. */
    i_max = cgul_random_lcg32__next_in_range(cex, lcg, rmax);
    if (*cex) {
        goto out;
    }

    /* Create each point. */
    for (i = 0 ; i < i_max ; ++i) {

        /* Allocate the next point. */
        p = cgul_libc__calloc(cex, 1, sizeof(*p));
        if (*cex) {
            goto out;
        }

        /* Get the next random number. */
        p->x = cgul_random_lcg32__next_in_range(cex, lcg, rmax);
        if (*cex) {
            goto out;
        }

        /* Track the order in which the points are inserted so we can
         * check later to see if the sort is stable. */
        p->y = i;

        /* Insert the point in the vector. */
        cgul_vector__push_back(cex, v, p);
        if (*cex) {
            goto out;
        }

        /* The vector now owns the point. */
        p = NULL;
    }

 out:

    /* Clean up. */
    if (p) {
        free(p);
    }

    return;
}

static int
is_sorted_and_stable(cgul_exception_t* cex,
                     cgul_vector_t v)
{
    int rv = 0;
    unsigned long int i;
    unsigned long int i_max;
    unsigned long int prev_x;
    unsigned long int prev_y;
    point_t* points = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* If the vector is empty, it is sorted and stable. */
    if (cgul_vector__is_empty(cex, v)) {
        rv = 1;
        goto out;
    }

    /* Get the underlying array. */
    points = (point_t*)cgul_vector__get_array(cex, v);

    /* Get the first values of x and y. */
    prev_x = points[0]->x;
    prev_y = points[0]->y;

    /* Iterate over the remaining nodes. */
    i_max = cgul_vector__get_size(cex, v);
    for (i = 1 ; i < i_max ; ++i) {

        /* The x-value is a randomly generated number.  It is used to
         * test whether the vector is sorted.  If the current value of x
         * is less than the previous value of x, the vector is not
         * sorted. */
        if (points[i]->x < prev_x) {
            goto out;
        }

        /* The y-value is the order the nodes were inserted.  It is
         * used to test for stability.  Stability can only be tested
         * when prev_x and points[i]->x have the same value. */
        if ((prev_x == points[i]->x) && (points[i]->y < prev_y)) {
            goto out;
        }

        /* Remember the current point. */
        prev_x = points[i]->x;
        prev_y = points[i]->y;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_uint32_t rmax = LCG_MAX_RANGE;
    cgul_vector_t v = NULL;
    cgul_random_lcg32_t lcg = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    v = cgul_vector__new(&local);
#if CONST_SEED
    lcg = cgul_random_lcg32__new_from_seed(&local, rmax);
#else
    lcg = cgul_random_lcg32__new(&local);
#endif
    if (local) {
        goto out;
    }

    /* Insert random points. */
    insert_random_points(&local, v, lcg, rmax);

#if ENABLE_PRINT
    /* Print the points. */
    print_points(&local, v);
#endif

#if PROFILE_CGUL_DEFAULT
    cgul_vector__sort(&local, v, (cgul_vector__compare_t)compare_points_cgul);
#endif

#if PROFILE_CGUL_ALT
    cgul_merge_sort_pointers_iteratively(
        &local,
        cgul_vector__get_array(&local, v),
        cgul_vector__get_size(&local, v),
        (cgul_merge_sort__compare_t)compare_points_cgul);
#endif

#if PROFILE_QSORT
    qsort(cgul_vector__get_array(&local, v),
          cgul_vector__get_size(&local, v),
          sizeof(point_t),
          (int (*)(const void*, const void*))compare_points_qsort);
#endif

#if ENABLE_PRINT
    /* Print the points. */
    print_points(&local, v);
#endif

#if PROFILE_CGUL_DEFAULT || PROFILE_CGUL_ALT
    /* Test whether the vector is in fact sorted and stable. */
    if (!is_sorted_and_stable(&local, v)) {
        local = cgul_exception__new("Not sorted or not stable.");
        goto out;
    }
#endif

    printf("Total Points Sorted: %lu\n",
           (unsigned long)cgul_vector__get_size(&local, v));

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (lcg) {
        cgul_random_lcg32__delete(lcg);
    }
    if (v) {
        cgul_vector__free_values(v);
        cgul_vector__delete(v);
    }

    return rv;
}
