/*
 * This test program shows how to piece together multiple MicroXML
 * document fragments to simulate a MicroXML file that has multiple
 * root elements.  The approach shown here uses the ability of
 * cgul_stream_reader to read characters from a queue of block
 * readers.
 *
 * A common example of where this is needed is when logging in
 * MicroXML format where one would like to treat the entire log file
 * as one MicroXML document with multiple root elements.  The basic
 * idea below it to just wrap the document fragment that holds the log
 * entries in a real root element with cgul_microxml reading from
 * different data sources to get the complete MicroXML file.
 */

#include "cgul.h"

/**
 * Skip pass the wrapper start tag which is the start tag for the
 * actual root element when all the document fragments are pieced
 * together.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
main__skip_wrapper_start_tag(cgul_exception_t* cex,
                             cgul_microxml_t uxml)
{
    /* Skip our wrapper root element start tag. */
    if (!cgul_microxml__is_element_start(cex, uxml)) {
        if (*cex) {
            goto out;
        }
        *cex = cgul_exception__new("expected element start");
        goto out;
    } else {
        if (strcmp("log-file",
                   cgul_microxml__get_element_name(cex, uxml)) != 0)
        {
            *cex = cgul_exception__new("expected element start <log-file>");
            goto out;
        }
    }

    /* Read the next token. */
    cgul_microxml__read_next_token(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Process the document fragment in the MicroXML parser
 * <code>uxml</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
main__process_multiroot(cgul_exception_t* cex,
                        cgul_microxml_t uxml)
{
    const char* element_name = NULL;
    const char* priority = NULL;
    const char* logmsg = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (;;) {

        /* Expect <log-entry> element start. */
        if (!cgul_microxml__is_element_start(cex, uxml)) {
            break;
        }

        /* Get the element name. */
        element_name = cgul_microxml__get_element_name(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Get the log priority. */
        tmp = cgul_microxml__get_attribute_value(cex, uxml, "priority");
        priority = cgul_string__get_value(cex, tmp);
        if (!priority) {
            *cex = cgul_exception__new("expected priority attribute");
            goto out;
        }

        /* "element_name" and "priority" are only valid until the next
         * read so use it now. */
        printf("<%s priorty='%s'>", element_name, priority);
        element_name = NULL;
        priority = NULL;

        /* Read the next token. */
        cgul_microxml__read_next_token(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Expect the log message. */
        if (!cgul_microxml__is_text(cex, uxml)) {
            *cex = cgul_exception__new("expected log message");
            goto out;
        }

        /* Get the log message. */
        tmp = cgul_microxml__get_text(cex, uxml);
        logmsg = cgul_string__get_value(cex, tmp);
        if (*cex) {
            goto out;
        }

        /* "logmsg" is only valid until the next read so use it now. */
        printf("%s", logmsg);
        logmsg = NULL;

        /* Read the next token. */
        cgul_microxml__read_next_token(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Expect </log-entry> element end. */
        if (!cgul_microxml__is_element_end(cex, uxml)) {
            *cex = cgul_exception__new("multiroot element end not found");
            goto out;
        }

        /* Get the element name. */
        element_name = cgul_microxml__get_element_name(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* "element_name" is only valid until the next read so use it now. */
        printf("</%s>\n", element_name);
        element_name = NULL;

        /* Read the next token. */
        cgul_microxml__read_next_token(cex, uxml);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

/**
 * Skip pass the wrapper end tag which is the end tag for the actual
 * root element when all the document fragments are pieced together.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
main__skip_wrapper_end_tag(cgul_exception_t* cex,
                           cgul_microxml_t uxml)
{
    /* Skip our wrapper root element end tag. */
    if (!cgul_microxml__is_element_end(cex, uxml)) {
        if (*cex) {
            goto out;
        }
        *cex = cgul_exception__new("expected element end");
        goto out;
    } else {
        if (strcmp("log-file",
                   cgul_microxml__get_element_name(cex, uxml)) != 0)
        {
            *cex = cgul_exception__new("expected element end </log-file>");
            goto out;
        }
    }

    /* Read the next token. */
    cgul_microxml__read_next_token(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* wrapper_start_tag = "<log-file>";
    const char* wrapper_end_tag = "</log-file>";
    const char* multiroot_tags = "\
<log-entry priority='LOG_WARNING'>foo</log-entry>\n\
<log-entry priority='LOG_NOTICE'>bar</log-entry>\n\
<log-entry priority='LOG_INFO'>baz</log-entry>\n\
";
    /* You don't need all of these, but the improve testing. */
    cgul_block_reader_memory_t br_start = NULL;
    cgul_block_reader_memory_t br_multiroot = NULL;
    cgul_block_reader_memory_t br_end = NULL;
    cgul_stream_reader_t sr = NULL;
    cgul_microxml_t uxml = NULL;
    cgul_exception_t local = NULL;

    /*
     * Allocate.
     */

    /* wrapper_start_tag */
    br_start = cgul_block_reader_memory__new_from_memory(
       &local, (void*)wrapper_start_tag, strlen(wrapper_start_tag));

    /* multiroot_tags */
    br_multiroot = cgul_block_reader_memory__new_from_memory(
       &local, (void*)multiroot_tags, strlen(multiroot_tags));

    /* wrapper_end_tag */
    br_end = cgul_block_reader_memory__new_from_memory(
       &local, (void*)wrapper_end_tag, strlen(wrapper_end_tag));

    /* sr */
    sr = cgul_stream_reader__new(&local);

    /* uxml */
    uxml = cgul_microxml__new(&local);
    if (local) {
        goto out;
    }

    /*
     * Load wrapper_start_tag, multiroot_tags, and wrapper_end_tag into sr.
     */

    /* sr supports reading from a queue of block readers. */
    cgul_stream_reader__add_block_reader(
        &local,
        sr,
        cgul_block_reader_memory__as_block_reader(&local, br_start));
    cgul_stream_reader__add_block_reader(
        &local,
        sr,
        cgul_block_reader_memory__as_block_reader(&local, br_multiroot));
    cgul_stream_reader__add_block_reader(
        &local,
        sr,
        cgul_block_reader_memory__as_block_reader(&local, br_end));

    /* Have uxml use sr. */
    cgul_microxml__set_stream_reader(&local, uxml, sr);

    /*
     * Configure
     */

    /* Skip white space. */
    cgul_microxml__set_skip_white_space(&local, uxml, 1);

    /*
     * Process
     */

    /* Read the next token. */
    cgul_microxml__read_next_token(&local, uxml);
    if (local) {
        goto out;
    }

    /* Expect the document start token. */
    if (!cgul_microxml__is_document_start(&local, uxml)) {
        local = cgul_exception__new("expected document start");
        goto out;
    }

    /* Read the next token. */
    cgul_microxml__read_next_token(&local, uxml);
    if (local) {
        goto out;
    }

    /* Skip passed the wrapper start tag. */
    main__skip_wrapper_start_tag(&local, uxml);
    if (local) {
        goto out;
    }

    /* Process the multi-root part of the document. */
    main__process_multiroot(&local, uxml);
    if (local) {
        goto out;
    }

    /* Skip passed the wrapper end tag. */
    main__skip_wrapper_end_tag(&local, uxml);
    if (local) {
        goto out;
    }

    /* Expect the document end token. */
    if (!cgul_microxml__is_document_end(&local, uxml)) {
        local = cgul_exception__new("expected document end");
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (uxml) {
        cgul_microxml__delete(uxml);
    }
    if (sr) {
        cgul_stream_reader__delete(sr);
    }
    if (br_end) {
        cgul_block_reader_memory__delete(br_end);
    }
    if (br_multiroot) {
        cgul_block_reader_memory__delete(br_multiroot);
    }
    if (br_start) {
        cgul_block_reader_memory__delete(br_start);
    }

    return rv;
}
