#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * This function prints the value stored in the node <code>n</code>.
 *
 * @param[in] mm multimap
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_pair(cgul_multimap_cxx* mm,
           cgul_multimap_node_cxx* n,
           void* data)
{
    printf("key=%s  value=%d\n", (char*)n->get_key(), *(int*)n->get_value());
    return 1;
}

/**
 * This function calls <code>free()</code> on the value stored in the
 * node <code>n</code>.
 *
 * @param[in] mm multimap
 * @param[in] n node
 * @param[in] data client data
 */
static int
free_pair(cgul_multimap_cxx* mm,
          cgul_multimap_node_cxx* curr,
          void* data)
{
    void* key_curr = NULL;
    void* key_next = NULL;
    cgul_multimap_node_cxx* next;

    /* We always need to free the value. */
    free(curr->get_value());

    /* We only free this key if no more values are associated with it. */
    next = curr->get_next(0);
    if (next) {
        key_next = (void*)next->get_key();
    }
    key_curr = (void*)curr->get_key();
    if ((next == NULL) || (key_curr != key_next)) {
        free(key_curr);
    }

    return 1;
}

/**
 * Initialize the multimap <code>t</code> by loading key/value pairs.
 *
 * @param[in] mm multimap
 */
static void
init_multimap(cgul_multimap_cxx* mm)
{
    int i = 0;
    char* key1 = NULL;
    char* key2 = NULL;
    int* value1 = NULL;
    int* value2 = NULL;

    assert(mm);

    try {

        for (i = 0 ; i < 5 ; ++i) {

            /* Allocate. */
            key1 = (char*)calloc(1, sizeof(char) * 2);
            key2 = (char*)calloc(1, sizeof(char) * 2);
            value1 = (int*)calloc(1, sizeof(int));
            value2 = (int*)calloc(1, sizeof(int));
            if (!key1 || !key2 || !value1 || !value2) {
                throw runtime_error("out of memory");
            }

            /* Assign values. */
            *key1 = 'a' + i;
            *key2 = 'a' + i;
            *value1 = i * 2;
            *value2 = i * 2 + 1;

            /* Insert key1 and value1 into multimap. */
            mm->insert(key1, value1, NULL);

            /* The multimap now owns key1 and value1. */
            key1 = NULL;
            value1 = NULL;

            /* Insert value2 into multimap. */
            mm->insert(key2, value2, NULL);

            /* Free key2 because it is not owned by the multimap, but
             * do not free value2 because it is owned by the
             * multimap. */
            free(key2);
            key2 = NULL;
            value2 = NULL;
        }

    } catch (...) {
        /* Clean up. */
        if (key1) {
            free(key1);
        }
        if (key2) {
            free(key2);
        }
        if (value1) {
            free(value1);
        }
        if (value2) {
            free(value2);
        }
        /* Rethrow. */
        throw;
    }
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_multimap_cxx* mm = NULL;
    cgul_multimap_node_cxx* first = NULL;
    cgul_multimap_node_cxx* last = NULL;

    try {

        /* Allocate. */
        mm = new cgul_multimap_cxx((cgul_multimap_cxx::compare_t)strcmp);

        /* Initialize the multimap. */
        init_multimap(mm);

        /* Print the multimap. */
        mm->traverse(0, print_pair, NULL);

        cout << "=====" << endl;

        /* Print a range of the multimap. */
        first = mm->get_front(0);
        if (first) {
            first = first->get_next(0);
        }
        last = mm->get_back(0);
        if (last) {
            last = last->get_prev(0);
        }
        mm->traverse_range(first, last, 0, print_pair, NULL);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    /* Clean up. */
    if (mm) {
        mm->traverse(0, free_pair, NULL);
        delete mm;
    }

    return rv;
}
