/*
 * You can use this program to test the cgul_rbtree class.  The code
 * is set up to read a command string (like "i" for insert, "r" for
 * remove, or "p" for print) followed by an integer for the value you
 * would like to insert or remove.
 *
 * You can use "test" as the command.  This will cause this code to
 * randomly insert TEST_POOL_SIZE integers into the test tree.  It
 * will then randomly remove each integer.
 *
 * You need to type "q" for quit.
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include "cgul.h"
#include "cgul_rbtree_p.h"
#include "cgul_rbtree_node_p.h"
#include "cgul_rbtree_node_with_indexing_p.h"

#define TEST_POOL_SIZE (10240)

static int
myintcmp(const void* k1, const void* k2)
{
    return (long int)k1 - (long int)k2;
}

/**
 * Get the next word from the input file <code>f</code> and assign it
 * to the string <code>word</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f input file
 * @param[in,out] word next word
 */
static int
get_next_word(cgul_exception_t* cex,
              FILE* f,
              cgul_string_t word)
{
    int c;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (;;) {
        c = fgetc(f);
        if (c == EOF) {
            return 0;
        }
        if (!isspace(c)) {
            break;
        }
    }
    cgul_string__set_value(cex, word, "");
    cgul_string__append_char(cex, word, (char)c);
    for (;;) {
        c = fgetc(f);
        if (c == EOF) {
            return 0;
        }
        if (isspace(c)) {
            break;
        }
        cgul_string__append_char(cex, word, (char)c);
    }

 out:
    return 1;
}

static int
get_next_unsigned_long(cgul_exception_t* cex,
                       FILE* f,
                       cgul_string_t w,
                       unsigned long* n)
{
    int rv = 0;
    unsigned long tmp = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_word(cex, stdin, w)) {
        tmp = cgul_string__to_unsigned_long(cex, w, 0 /* base */);
        if (*cex) {
            goto out;
        }
        *n = tmp;
        rv = 1;
    }

 out:
    return rv;
}

static long int
get_next_int(cgul_exception_t* cex, FILE* f, cgul_string_t w, long int* n)
{
    int rv = 0;
    int tmp = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_word(cex, stdin, w)) {
        tmp = cgul_string__to_long(cex, w, 0 /* base */);
        if (*cex) {
            goto out;
        }
        *n = tmp;
        rv = 1;
    }

 out:
    return rv;
}

static void
print_rbtree_node_with_ints(cgul_exception_t* cex, cgul_rbtree_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    printf("%-6ld", (long int)n->key);
    printf(" | %-5s",  n->red ? "red" : "black");
    printf(" | lc = %-3lu", ((cgul_rbtree_node_with_indexing_t)n)->lcount);
    printf(" | rc = %-3lu", ((cgul_rbtree_node_with_indexing_t)n)->rcount);
    if (n->parent) {
        printf(" | p = %-6ld", (long int)n->parent->key);
    } else {
        printf(" | p = %-6s", "NULL");
    }
    if (n->left) {
        printf(" | l = %-6ld", (long int)n->left->key);
    } else {
        printf(" | l = %-6s", "NULL");
    }
    if (n->right) {
        printf(" | r = %-6ld", (long int)n->right->key);
    } else {
        printf(" | r = %-6s", "NULL");
    }
    puts("");

 out:
    return;
}

static void
print_rbtree_with_ints_for_keys(cgul_exception_t* cex, cgul_rbtree_t t)
{
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    n = cgul_rbtree__get_front(cex, t);

    for ( ; n != NULL ; n = cgul_rbtree_node__get_next(cex, n)) {
        print_rbtree_node_with_ints(cex, n);
    }

    if (t->front) {
        printf("front = %ld\n", (long int)t->front->key);
    }
    if (t->back) {
        printf("back = %ld\n", (long int)t->back->key);
    }

 out:
    return;
}

static void
print_chronologically_rbtree_with_ints_for_keys(cgul_exception_t* cex,
                                                cgul_rbtree_t t)
{
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    n = cgul_rbtree__get_oldest(cex, t);

    for ( ; n != NULL ; n = cgul_rbtree_node__get_younger(cex, n)) {
        print_rbtree_node_with_ints(cex, n);
    }

    if (t->oldest) {
        printf("oldest = %ld\n", (long int)t->oldest->key);
    }

    if (t->youngest) {
        printf("youngest = %ld\n", (long int)t->youngest->key);
    }

 out:
    return;
}

static void
print_node_at(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    unsigned long int index = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_unsigned_long(cex, stdin, s, &index)) {
        if ((n = cgul_rbtree__find_at(cex, t, index)) == NULL) {
            fprintf(stderr, "Warning: Unable to find node t[%lu]\n", index);
        } else {
            printf("t[%lu] = %ld\n", index, (long int)n->key);
        }
    }

 out:
    return;
}

static void
find_int(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    long int i = 0;
    long int key = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_int(cex, stdin, s, &i)) {
        n = cgul_rbtree__find(cex, t, (void*)i);
        if (n == NULL) {
            fprintf(stderr, "Warning: Unable to find: %ld\n", i);
            goto out;
        }
        key = (long int)cgul_rbtree_node__get_key(cex, n);
        cgul_libc__printf(cex, "found: %ld\n", key);
    }

 out:
    return;
}

static void
find_floor_int(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    long int i = 0;
    long int key = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_int(cex, stdin, s, &i)) {
        n = cgul_rbtree__find_floor(cex, t, (void*)i);
        if (n == NULL) {
            fprintf(stderr, "Warning: Unable to find floor: %ld\n", i);
            goto out;
        }
        key = (long int)cgul_rbtree_node__get_key(cex, n);
        cgul_libc__printf(cex, "found floor: %ld\n", key);
    }

 out:
    return;
}

static void
find_ceiling_int(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    long int i = 0;
    long int key = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_int(cex, stdin, s, &i)) {
        n = cgul_rbtree__find_ceiling(cex, t, (void*)i);
        if (n == NULL) {
            fprintf(stderr, "Warning: Unable to find ceiling: %ld\n", i);
            goto out;
        }
        key = (long int)cgul_rbtree_node__get_key(cex, n);
        cgul_libc__printf(cex, "found ceiling: %ld\n", key);
    }

 out:
    return;
}

static void
insert_int(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    long int key = 0;
    int insert_rv = 0;
    long int i = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_int(cex, stdin, s, &i)) {
        insert_rv = cgul_rbtree__insert(cex, t, (void*)i, &n);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            fprintf(stderr, "Warning: Key already exists: %ld\n", i);
            goto out;
        }
        key = (long int)cgul_rbtree_node__get_key(cex, n);
        cgul_libc__printf(cex, "inserted: %ld\n", key);
    }

 out:
    return;
}

static void
remove_int(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_int(cex, stdin, s, &i)) {
        if (!cgul_rbtree__remove(cex, t, (void*)i, NULL, NULL)) {
            fprintf(stderr, "Warning: Unable to remove missing key: %ld\n", i);
        }
    }

 out:
    return;
}

static void
remove_int_at(cgul_exception_t* cex, cgul_rbtree_t t, cgul_string_t s)
{
    unsigned long int index = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (get_next_unsigned_long(cex, stdin, s, &index)) {
        if (!cgul_rbtree__remove_at(cex, t, index, NULL, NULL)) {
            fprintf(stderr, "Warning: Index out of bounds: %lu\n", index);
        }
    }

 out:
    return;
}

static void
reload_pool(cgul_exception_t* cex, cgul_rbtree_t pool)
{
    int insert_rv = 0;
    long int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    for (i = 0 ; i < TEST_POOL_SIZE ; ++i) {
        insert_rv = cgul_rbtree__insert(cex, pool, (void*)i, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            cgul_libc__fprintf(cex, stderr, "***\n");
            cgul_libc__fprintf(cex, stderr,
                 "*** cgul_rbtree__insert() failed while initializing pool\n");
            cgul_libc__fprintf(cex, stderr, "***\n");
            exit(1);
        }
    }

 out:
    return;
}

static void
verify_tree_is_full(cgul_exception_t* cex, cgul_rbtree_t t)
{
    int i = 0;
    cgul_rbtree_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    n = cgul_rbtree__get_front(cex, t);
    for ( i = 0 ; n != NULL ; n = cgul_rbtree_node__get_next(cex, n), ++i) {
        if ((long int)n->key != i) {
            cgul_libc__fprintf(cex, stderr,
                           "*** Error: tree did not verify as being full.\n");
            exit(1);
        }
    }

 out:
    return;
}

static void
verify_tree_is_sorted(cgul_exception_t* cex, cgul_rbtree_t t)
{
    cgul_rbtree_node_t curr = NULL;
    cgul_rbtree_node_t prev = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    prev = cgul_rbtree__get_front(cex, t);
    if (prev) {
        curr = cgul_rbtree_node__get_next(cex, prev);
        while (curr != NULL) {
            if (myintcmp(prev->key, curr->key) >= 0) {
                cgul_libc__printf(cex,
                                   "Error: tree is not sorted: %ld <--> %ld\n",
                                   (long int)prev->key, (long int)curr->key);
                cgul_libc__printf(cex, "Current size: %lu\n",
                                  (unsigned long)cgul_rbtree__get_size(cex, t));
                exit(1);
            }
            prev = curr;
            curr = cgul_rbtree_node__get_next(cex, curr);
        }
    }

 out:
    return;
}

/* This function randomly removes a node from pool and inserts the
 * same key into t. */
static void
randomly_fill_test_tree(cgul_exception_t* cex,
                        cgul_rbtree_t t,
                        cgul_rbtree_t pool,
                        cgul_random_cmwc32_t prng)
{
    int insert_rv = 0;
    cgul_uint32_t i = 0;
    cgul_uint32_t next = 0;
    long int value = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (cgul_rbtree__get_size(cex, pool) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: pool is not full.\n");
        exit(1);
    }

    if (cgul_rbtree__get_size(cex, t) != 0) {
        fprintf(stderr, "*** Error: t is not empty.\n");
        exit(1);
    }

    for (i = TEST_POOL_SIZE ; i > 0 ; --i) {

        /* The idea is to have prng generate a number between 0 and
         * pool->size - 1.  We then use remove_at() (not remove()) to
         * remove the node at the index returned by prng and insert
         * the same key into t.  We then have prng generate the next
         * random number taking into account that we just removed a
         * node. */
        next = cgul_random_cmwc32__next_in_unit(cex, prng) * i;

#if 0
        printf("insert %d\n", next);
#endif

        /* Remove the value at the "next" index from the pool */
        if (!cgul_rbtree__remove_at(cex, pool, next, (void**)&value, NULL)) {
            fprintf(stderr, "*** Error: Unable to remove pool[%u]\n", next);
            exit(1);
        }

        /* Insert that same value into the tree. */
        insert_rv = cgul_rbtree__insert(cex, t, (void*)value, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            fprintf(stderr, "*** Error: Key already exists in t: %ld\n", value);
            exit(1);
        }
    }


    if (cgul_rbtree__get_size(cex, pool) != 0) {
        fprintf(stderr, "*** Error: pool is not empty.\n");
        exit(1);
    }

    if (cgul_rbtree__get_size(cex, t) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: t is not full.\n");
        exit(1);
    }

 out:
    return;
}

/* This function randomly removes a node from pool and inserts the
 * same key into t. */
static void
randomly_empty_test_tree(cgul_exception_t* cex,
                         cgul_rbtree_t t,
                         cgul_rbtree_t pool,
                         cgul_random_cmwc32_t prng)
{
    cgul_uint32_t i = 0;
    cgul_uint32_t next = 0;
    long int value = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (cgul_rbtree__get_size(cex, pool) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: pool is not full.\n");
        exit(1);
    }

    if (cgul_rbtree__get_size(cex, t) != TEST_POOL_SIZE) {
        fprintf(stderr, "*** Error: t is not full.\n");
        exit(1);
    }

    for (i = TEST_POOL_SIZE ; i > 0 ; --i) {

        /* The idea is to have prng generate a number between 0 and
         * pool->size - 1.  We then use remove_at() (not remove()) to
         * remove the node at the index returned by prng and insert
         * the same key into t.  We then have prng generate the next
         * random number taking into account that we just removed a
         * node. */
        next = cgul_random_cmwc32__next_in_unit(cex, prng) * i;

#if 0
        printf("remove %d\n", next);
#endif

        /* Remove the value at the "next" index from the pool */
        if (!cgul_rbtree__remove_at(cex, pool, next, (void**)&value, NULL)) {
            fprintf(stderr,
                    "*** Error: Unable to find pool[%u].  Exiting!\n",
                    next);
            exit(1);
        }
        verify_tree_is_sorted(cex, pool);

        /* Remove that same value from the tree. */
        if (!cgul_rbtree__remove(cex, t, (void*)value, NULL, NULL)) {
            fprintf(stderr,
                    "*** Error: Unable to remove key from t: %ld\n",
                    value);
            exit(1);
        }
        verify_tree_is_sorted(cex, t);
    }

    if (cgul_rbtree__get_size(cex, pool) != 0) {
        fprintf(stderr, "*** Error: pool is not empty.\n");
        exit(1);
    }

    if (cgul_rbtree__get_size(cex, t) != 0) {
        fprintf(stderr, "*** Error: t is not empty.\n");
        exit(1);
    }

 out:
    return;
}

static void
test_rbtree(cgul_exception_t* cex,
            cgul_random_cmwc32_t prng)
{
    unsigned long count = 0;
    cgul_rbtree_t pool = NULL;
    cgul_rbtree_t t = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    pool = cgul_rbtree__new_with_indexing(cex, &myintcmp);
    t = cgul_rbtree__new_with_indexing(cex, &myintcmp);

    /*
     * Run the main loop.
     */
    for (count = 1; count <= 100 ; ++count) {
        printf("Starting loop %lu\n", count);
        printf("    reload_pool()\n");
        reload_pool(cex, pool);
        printf("    verify_tree_is_full(pool)\n");
        verify_tree_is_full(cex, pool);
        printf("    randomly_fill_test_tree()\n");
        randomly_fill_test_tree(cex, t, pool, prng);
        printf("    verify_tree_is_full(t)\n");
        verify_tree_is_full(cex, t);
        printf("    reload_pool()\n");
        reload_pool(cex, pool);
        printf("    verify_tree_is_full(pool)\n");
        verify_tree_is_full(cex, pool);
        printf("    randomly_empty_test_tree()\n");
        printf("        (slow because of extra verification steps) \n");
        randomly_empty_test_tree(cex, t, pool, prng);
        printf("    Finished loop %lu\n", count);
    }

 out:

    if (t) {
        cgul_rbtree__delete(t);
    }
    if (pool) {
        cgul_rbtree__delete(pool);
    }

    return;
}

int main(int argc, char* argv[])
{
    int rv = 0;
    char* cmd = NULL;
    cgul_rbtree_t t = NULL;
    cgul_string_t s = NULL;
    cgul_random_cmwc32_t prng = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    t = cgul_rbtree__new_with_indexing(&local, &myintcmp);
    s = cgul_string__new(&local);
    prng = cgul_random_cmwc32__new(&local);
    if (local) {
        goto out;
    }

    while (get_next_word(&local, stdin, s)) {

        /* Display any exceptions from last command. */
        if (local) {
            const char* msg = cgul_exception__get_message(local);
            fprintf(stderr, "*** Error: %s\n", msg);
            cgul_exception__delete(local);
            local = NULL;
            continue;
        }

        cmd = cgul_string__get_value(&local, s);
        if (cgul_libc__strcmp(&local, "at", cmd) == 0) {
            print_node_at(&local, t, s);
            continue;
        }
        if (cgul_libc__strcmp(&local, "i", cmd) == 0) {
            insert_int(&local, t, s);
            /* print_rbtree_with_ints_for_keys(t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "f", cmd) == 0) {
            find_int(&local, t, s);
            /* print_rbtree_with_ints_for_keys(t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "fc", cmd) == 0) {
            find_ceiling_int(&local, t, s);
            /* print_rbtree_with_ints_for_keys(t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "ff", cmd) == 0) {
            find_floor_int(&local, t, s);
            /* print_rbtree_with_ints_for_keys(t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "p", cmd) == 0) {
            print_rbtree_with_ints_for_keys(&local, t);
            continue;
        }
        if (cgul_libc__strcmp(&local, "pc", cmd) == 0) {
            print_chronologically_rbtree_with_ints_for_keys(&local, t);
            continue;
        }
        if (cgul_libc__strcmp(&local, "r", cmd) == 0) {
            remove_int(&local, t, s);
            /* print_rbtree_with_ints_for_keys(&local, t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "rat", cmd) == 0) {
            remove_int_at(&local, t, s);
            /* print_rbtree_with_ints_for_keys(&local, t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "rb", cmd) == 0) {
            cgul_rbtree__remove_back(&local, t, NULL, NULL);
            /* print_rbtree_with_ints_for_keys(&local, t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "rf", cmd) == 0) {
            cgul_rbtree__remove_front(&local, t, NULL, NULL);
            /* print_rbtree_with_ints_for_keys(&local, t); */
            continue;
        }
        if (cgul_libc__strcmp(&local, "test", cmd) == 0) {
            test_rbtree(&local, prng);
            continue;
        }
        if (   (cgul_libc__strcmp(&local, "q", cmd) == 0)
            || (cgul_libc__strcmp(&local, "quit", cmd) == 0))
        {
            break;
        }

        cgul_libc__fprintf(&local,
                           stderr,
                           "Warning: Unknown command: %s\n",
                           cmd);
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    if (prng) {
        cgul_random_cmwc32__delete(prng);
    }
    if (s) {
        cgul_string__delete(s);
    }
    if (t) {
        cgul_rbtree__delete(t);
    }

    return rv;
}
