/*
 * Test the following methods using randomly generated keys:
 *
 *     cgul_rbtree__find_at()
 *     cgul_rbtree__find_rank()
 *     cgul_rbtree__remove_at()
 */

#include "cgul.h"

/**
 * Minimum number of keys to generate.
 */
#define KEY_COUNT_MIN (0)

/**
 * Maximum number of keys to generate.
 */
#define KEY_COUNT_MAX (1000000)

/**
 * Return less than zero, zero, or greater than zero if
 * <code>*n1</code> is less than, equal to, or greater than
 * <code>*n2</code>.
 *
 * @param[in] n1 first number
 * @param[in] n2 second number
 * @return ordering of the first number relative to the second number
 */
static int
main__compare_uint32(cgul_uint32_t* n1,
                     cgul_uint32_t* n2)
{
    int rv = 0;

    if (*n1 < *n2) {
        rv = -1;
    } else if (*n1 > *n2) {
        rv = 1;
    }

    return rv;
}

/**
 * Insert between <code>key_count_min</code> (inclusive) and
 * <code>key_count_max</code> (inclusive) randomly generated keys into
 * the red-black tree <code>t</code> and push a copy of each key in
 * the order generated onto the <code>keys</code> vector (which can be
 * used later for verification purposes).  The key are randomly
 * generated by the PRNG <code>prng</code>.  The user is responsible
 * for calling <code>free()</code> on the keys inserted into the tree.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] prng PRNG
 * @param[in] keys keys
 * @param[in] key_count_min minimum key count
 * @param[in] key_count_max maximum key count
 */
static void
main__insert_keys(cgul_exception_t* cex,
                  cgul_rbtree_t t,
                  cgul_vector_t keys,
                  cgul_random_lfg32_t prng,
                  cgul_uint32_t key_count_min,
                  cgul_uint32_t key_count_max)
{
    int insert_rv = 0;
    cgul_uint32_t i = 0;
    cgul_uint32_t key_count = 0;
    cgul_uint32_t* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Determine the number of keys to generate. */
    key_count = cgul_random_lfg32__next_in_range(
        cex, prng, key_count_max - key_count_min + 1);
    if (*cex) {
        goto out;
    }
    key_count += key_count_min;

#if 1
    printf("key_count = %lu\n", (unsigned long)key_count);
#endif

    /* Generate and insert the keys. */
    for (i = 0 ; i < key_count ; ++i) {

        /* Allocate space for a new key. */
        key = (cgul_uint32_t*)cgul_libc__calloc(cex, 1, sizeof(*key));
        if (*cex) {
            goto out;
        }

        /* Randomly generate the next key. */
        *key = cgul_random_lfg32__next(cex, prng);
        if (*cex) {
            goto out;
        }

        /* Try to insert the key which could fail if a key repeats. */
        insert_rv = cgul_rbtree__insert(cex, t, key, NULL);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            free(key);
            key = NULL;
            continue;
        }

        /* Push the key onto the verification vector. */
        cgul_vector__push_back(cex, keys, key);
        if (*cex) {
            goto out;
        }

        /* The red-black tree now owns the key. */
        key = NULL;

    }

 out:

    /* Clean up. */
    if (key) {
        free(key);
    }

    return;
}

/**
 * Test <code>cgul_rbtree__find_at()</code> by verifying the keys in
 * the <code>keys</code> vector exist in the red-black tree
 * <code>t</code> in the same order.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_at(cgul_exception_t* cex,
                   cgul_rbtree_t t,
                   cgul_vector_t keys)
{
    unsigned long int i = 0;
    cgul_uint32_t* key = NULL;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Verify that each key in "keys" exists in "t" and in the same order. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the node by index. */
        node = cgul_rbtree__find_at(cex, t, i);
        if (!node) {
            *cex = cgul_exception__new(
                "main__test_find_at: Missing key: %lu", i);
            goto out;
        }

        /* Verify the keys match. */
        key = (cgul_uint32_t*)cgul_rbtree_node__get_key(cex, node);
        if (*key != *arr[i]) {
            *cex = cgul_exception__new(
                "main__test_find_at: "
                "Invalid key at index %lu: (expecting, actual) = (%lu, %lu)",
                i, (unsigned long int)*arr[i], (unsigned long int)*key);
            goto out;
        }
    }

    /* Verify lookup misses return NULL. */
    node = cgul_rbtree__find_at(cex, t, arr_size);
    if (node) {
        key = (cgul_uint32_t*)cgul_rbtree_node__get_key(cex, node);
        *cex = cgul_exception__new(
            "main__test_find_at: "
            "Unexpectedly found key for out-of-bounds index: "
            "(index, key) = (%lu, %lu)",
            arr_size, (unsigned long int)*key);
        goto out;
    }

 out:
    return;
}

/**
 * Test <code>cgul_rbtree__find_rank()</code> by verifying the index
 * of each key in the <code>keys</code> vector equals the rank of that
 * same key in the red-black tree <code>t</code>.  If an error occurs,
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_rbtree</code> instance
 * @param[in] keys keys
 */
static void
main__test_find_rank(cgul_exception_t* cex,
                     cgul_rbtree_t t,
                     cgul_vector_t keys)
{
    int rank_rv = 0;
    unsigned long int i = 0;
    unsigned long int rank = 0;
    cgul_uint32_t** arr = NULL;
    unsigned long int arr_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(keys);

    /* Unpack the keys vector. */
    arr = (cgul_uint32_t**)cgul_vector__get_array(cex, keys);
    arr_size = cgul_vector__get_size(cex, keys);

    /* Verify that each key in "keys" exists in "t" and in the same order. */
    for (i = 0 ; i < arr_size ; ++i) {

        /* Find the rank of the key. */
        rank_rv = cgul_rbtree__find_rank(cex, t, arr[i], &rank);
        if (!rank_rv) {
            *cex = cgul_exception__new(
                "main__test_find_rank: Missing key: %lu", i);
            goto out;
        }

        /* Verify the index of the key match the rank. */
        if (rank != i) {
            *cex = cgul_exception__new(
                "main__test_find_at: Invalid rank for key: "
                "(key, expecting, actual) = (%lu, %lu, %lu)",
                (unsigned long int)*arr[i], i, rank);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Randomly remove all keys from the rbtree <code>rbtree</code> using
 * <code>cgul_rbtree__remove_at()</code>.  The order in which keys are
 * removed is determined by the PRNG <code>prng</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] rbtree <code>cgul_rbtree</code> instance
 * @param[in] PRNG prng
 */
static void
main__test_remove_at(cgul_exception_t* cex,
                     cgul_rbtree_t t,
                     cgul_random_lfg32_t prng)
{
    int remove_rv = 0;
    cgul_uint32_t index = 0;
    cgul_uint32_t* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    while (!cgul_rbtree__is_empty(cex, t)) {

        /* Generate the index of the next key to remove. */
        index = cgul_random_lfg32__next_in_range(
            cex, prng, cgul_rbtree__get_size(cex, t));

        /* Remove this key from the rbtree. */
        remove_rv = cgul_rbtree__remove_at(cex, t, index, (void**)&key, NULL);
        if (!remove_rv) {
            *cex = cgul_exception__new(
                "main__test_remove_at: "
                "Unable to remove key at index: %lu",
                (unsigned long int)index);
            goto out;
        }

        /* Free the key. */
        free(key);
        key = NULL;

    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_uint32_t seed = 0;
    cgul_rbtree_t t = NULL;
    cgul_vector_t keys = NULL;
    cgul_random_lfg32_t prng = NULL;
    cgul_exception_t local = NULL;

#if 1
    /* Use the time as the seed. */
    seed = (cgul_uint32_t)time(NULL);
#else
    /* Use specific seed value. */
    seed = 1425126242;
    printf("===== REUSING SEED!!!\n");
#endif
    printf("===== seed: %lu\n", (unsigned long)seed);

    /* Allocate. */
    keys = cgul_vector__new(&local);
    prng = cgul_random_lfg32__new_from_seed(&local, seed);
    t = cgul_rbtree__new_with_indexing(
            &local,
            (cgul_rbtree__compare_t)main__compare_uint32);
    if (local) {
        goto out;
    }

    /* Insert randomly generated numeric keys. */
    main__insert_keys(&local, t, keys, prng, KEY_COUNT_MIN, KEY_COUNT_MAX);
    if (local) {
        goto out;
    }

    /* Sort the keys vector. */
    cgul_vector__sort(
        &local, keys, (cgul_vector__compare_t)main__compare_uint32);

    /* Test cgul_rbtree__find_at(). */
    main__test_find_at(&local, t, keys);
    if (local) {
        goto out;
    }

    /* Test cgul_rbtree__find_rank(). */
    main__test_find_rank(&local, t, keys);
    if (local) {
        goto out;
    }

    /* Test cgul_rbtree__remove_at().  (This removes all the keys from
     * the rbtree in random order.) */
    main__test_remove_at(&local, t, prng);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        fprintf(stderr, "***  seed: %lu\n", (unsigned long)seed);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (t) {
        /* The keys should be freed above. */
        cgul_rbtree__delete(t);
    }
    if (prng) {
        cgul_random_lfg32__delete(prng);
    }
    if (keys) {
        cgul_vector__delete(keys);
    }

    return rv;
}
