/*
 * Copy stdin to stdout.
 */

#include "cgul_cxx.h"
#include <iostream>
#include <stdexcept>

#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#endif

using namespace std;

/**
 * Force the standard input <code>stdin</code> and standard output
 * <code>stdout</code> to be in binary mode (instead of text mode).
 * If an error occurs, an exception is thrown.
 */
static void
main__force_binary_mode()
{
#ifdef WIN32
    /* Convert stdin and stdout to binary mode. */
    if (_setmode(_fileno(stdin), _O_BINARY) < 0) {
        throw cgul_exception_cxx("_setmode(stdin): %m");
    }
    if (_setmode(_fileno(stdout), _O_BINARY) < 0) {
        throw cgul_exception_cxx("_setmode(stdout): %m");
    }
#endif
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    char utf8[CGUL_WCHAR__MB_LEN_MAX + 1] = {0};
    cgul_wchar_t utf32 = CGUL_WCHAR__NUL;

    try {

        /* Force stdin and stdout to binary mode. */
        main__force_binary_mode();

        /* Create the stream reader. */
        cgul_block_reader_file_cxx brfile(stdin);
        cgul_stream_reader_cxx sr(brfile);

        /* Configure the stream reader. */
        sr.set_normalize_eol(0);
        sr.set_line_and_column_tracking(0);

        /* Iterate. */
        for (;;) {

            /* Read the next character. */
            if (!sr.read(&utf32)) {
                /* EOF */
                break;
            }

            /* Convert the next character to UTF-8. */
            cgul_unicode_cxx::wctomb(utf32, utf8);

#if 0
            /* Print the UTF-8 character. */
            cout << utf8;
#else
            /* Print the UTF-8 character one byte at a time (which is
             * _much_ faster than using "cout << utf8"). */
            for (int i = 0 ; utf8[i] ; ++i) {
                putc(utf8[i], stdout);
            }
#endif
        }

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    return rv;
}
