/*
 * This program reads a dictionary file with one word or phrase per
 * line.  The dictionary file is a text file where each line of text
 * corresponds to one key in the trie.  The program then reads
 * prefixes from standard input and prints the dictionary entries that
 * have the same prefix.  The program can be run as follows:
 *
 *     ./main_trie_auto_complete_cxx <dictionary>
 *
 * To quit, type the following:
 *
 *     quit
 */

#include "cgul_cxx.h"
#include <iostream>
#include <stdexcept>
#include <string>

using namespace std;

/**
 * Print this program's usage message to the output stream
 * <code>ostrm</code>.  The base name of <code>progname</code> is
 * printed as part of the usage message.
 *
 * @param[in] progname program name
 * @param[in] ostrm output stream
 */
static void
main__print_usage(const char* progname,
                  ostream& ostrm)
{
    char* bname = NULL;

    assert(progname);

    try {

        /* Get the base name for the program. */
        bname = cgul_libc_cxx::basename(progname);

        /* Print the usage message. */
        ostrm << endl;
        ostrm << "Usage: " << bname << "<dictionary>" << endl;
        ostrm << endl;

    } catch (...) {
        /* Clean up. */
        if (bname) {
            free(bname);
        }
        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (bname) {
        free(bname);
    }
}

#if 1
/**
 * Insert the keys in the <code>keys</code> array into the trie
 * <code>trie</code> freeing the duplicate keys from the
 * <code>keys</code> array and setting their place in the
 * <code>keys</code> array to <code>NULL</code>.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] keys array of keys to insert into the trie
 * @param[in] key_count number of keys
 */
static void
main__insert_keys_freeing_duplicates(cgul_trie_cxx& trie,
                                     char** keys,
                                     unsigned long int key_count)
{
    unsigned long int i = 0;
    cgul_trie_node_cxx* node = NULL;

    /* Sanity check. */
    if (!keys || (key_count == 0)) {
        return;
    }

    /* Log. */
    cout << "Inserting keys ... " << flush;

    /* Insert the keys into the trie.  This transfers ownership of
     * all the keys that get inserted to the trie; however, for
     * duplicate keys, ownership stays with the "lines" array. */
    trie.insert_from_arrays(
        (const char**)keys,   /* keys array */
        NULL,                 /* values array */
        key_count,            /* number of key/value pairs */
        1);                   /* maintain_relative_ages */

    /* Log. */
    cout << "Freeing duplicate keys ... " << flush;

    /* Free duplicate keys from the "keys" array that are not
     * stored in the trie.  This occurs because the trie only
     * stores the first copy of the key when the user tries to
     * other copies. */
    for (i = 0 ; i < key_count ; ++i) {
        node = trie.find(keys[i]);
        if (!node || (keys[i] != node->get_key())) {
            free(keys[i]);
        }
    }
}

#else

/**
 * Insert the keys in the <code>keys</code> array into the trie
 * <code>trie</code> freeing the duplicate keys from the
 * <code>keys</code> array and setting their place in the
 * <code>keys</code> array to <code>NULL</code>.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] keys array of keys to insert into the trie
 * @param[in] key_count number of keys
 */
static void
main__insert_keys_freeing_duplicates(cgul_trie_cxx& trie,
                                     char** keys,
                                     unsigned long int key_count)
{
    int insert_rv = 0;
    unsigned long int i = 0;

    /* Sanity check. */
    if (!keys || (key_count == 0)) {
        return;
    }

    /* Log. */
    cout << "Inserting keys (and freeing duplicates) ... ") << flush;

    /* Insert the keys one at a time freeing duplicates as they are
     * detected as part of the insertion process.*/
    for (i = 0 ; i < key_count ; ++i) {

        /* Try to insert the key and free any duplicates. */
        if (!trie.insert(keys[i], NULL)) {
            free(keys[i]);
            keys[i] = NULL;
        }

    }

    /* Log. */
    cout << "balancing ... " << flush;

    /* Balance the trie to avoid linear searches if the keys are
     * inserted in sort order. */
    trie.balance(1 /* maintain_relative_ages */);

 out:
    return;
}
#endif

/**
 * Load the dictionary in the file <code>fname</code> into the trie
 * <code>trie</code>.  The dictionary file is a text file where each
 * line in the text corresponds to one key in the trie.  If an error
 * occurs, an exception is thrown. <p>
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] fname file name for the dictionary
 */
static void
main__load_dictionary(cgul_trie_cxx& trie,
                      const char* fname)
{
    char** lines = NULL;
    unsigned long int line_count = 0;

    try {

        /* Log. */
        cout << "Loading dictionary ... " << flush;

        /* Slurp the file. */
        cgul_crlf_file_cxx crlf(fname, true /* is_block_buffered */);
        crlf.get_lines(&lines, &line_count);

        /* Insert the lines as keys into a temporary trie to avoid
         * disturbing "trie" in case an error occurs.  Any duplicate
         * lines in the "lines" array will be freed.  This way only
         * the trie holds lines that still need to be freed. */
        cgul_trie_cxx tmp;
        main__insert_keys_freeing_duplicates(tmp, lines, line_count);

        /* Success. */
        trie.swap(tmp);

        /* Log. */
        cout << "Done." << endl;

    } catch (...) {
        /* Clean up. */
        if (lines) {
            free(lines);
        }
        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (lines) {
        free(lines);
    }
}

/**
 * Print to standard output all the keys in the trie <code>trie</code>
 * having prefix <code>prefix</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 * @param[in] prefix prefix
 */
static void
main__handle_prefix_request(cgul_trie_cxx& trie,
                            const char* prefix)
{
    cgul_trie_node_cxx* start = NULL;
    cgul_trie_node_cxx* stop = NULL;
    cgul_trie_node_cxx* curr = NULL;

    /* Sanity check. */
    if (!prefix) {
        return;
    }

    /* Get iterators over the result set for this prefix. */
    trie.find_half_opened_range_by_prefix(prefix, &start, &stop);

    /* Iterate over the result set. */
    for (curr = start; curr != stop; curr = curr->get_next()) {
        /* Print the key to standard output. */
        cout << curr->get_key() << endl;
    }
}

/**
 * Loop requesting the user to enter a prefix which is used to query
 * the trie <code>trie</code> to get a list of dictionary words that
 * match the prefix.  The list is then printed to standard output.
 *
 * @param[in] trie <code>cgul_trie_cxx</code> instance
 */
static void
main__loop(cgul_trie_cxx& trie)
{
    string prefix;

    for (;;) {

        /* Prompt the user to enter a prefix. */
        cout << "=====" << endl;
        cout << "Enter prefix: " << flush;

        /* Get the user's response. */
        getline(cin, prefix);

        /* EOF */
        if (!cin && (prefix.length() == 0)) {
            break;
        }

        /* quit */
        if (prefix == "quit") {
            break;
        }

        /* Handle prefix request. */
        main__handle_prefix_request(trie, prefix.c_str());

    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_trie_cxx trie;

    try {

        /* Validate command-line arguments. */
        if (argc != 2) {
            main__print_usage(argv[0], cerr);
            goto out;
        }
        
        /* Load the dictionary into the trie. */
        main__load_dictionary(trie, argv[1]);
        
        /* Enter the main loop. */
        main__loop(trie);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

 out:

    /* Clean up. */
    trie.free_keys();

    return rv;
}
