#include <assert.h>
#include "cgul.h"

/**
 * This function performs the lookup that converts
 * <code>varname</code> into its associated value.  This function then
 * validates the printf()-style conversion specification <code>cs</code>.
 * If <code>cs</code> is valid, it uses it to call
 * <code>cgul_string__append_sprintf()</code> to append the value
 * associated with <code>varname</code> to the destination string
 * <code>dst</code>.  If <code>cs</code> is not valid, it simply
 * appends the value associated with <code>varname<code> to the
 * destination string <code>dst</code> without using a format string.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] varname variable name
 * @param[in] cs conversion specification or <code>NULL</code>
 * @param[in,out] dst destination string
 * @param[in] data client data
 * @return whether the lookup was successful
 */
static int
do_lookup(cgul_exception_t* cex,
          const char* varname,
          cgul_convspec_t cs,
          cgul_string_t dst,
          cgul_stanza_t stanza)
{
    int rv = 0;
    int use_format = 0;
    const char* value = NULL;
    cgul_string_t format = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Do not assert "cs" because it can be NULL. */
    assert(varname);
    assert(dst);
    assert(stanza);

    /* Allocate the string that will hold the validated format string. */
    format = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Perform the lookup.
     */
    value = cgul_stanza__lookup(cex, stanza, varname);
    if (value == NULL) {
        /* Lookup failed. */
        goto out;
    }

    /*
     * Validate the conversion specification.
     */

    /* If no conversion specification or cgul_varsub_formatted()
     * detected a problem with it, skip the validation. */
    if (cs == NULL) {
        goto append;
    }

    /* All of our values are strings so a "%s" format string is required. */
    if (cgul_convspec__get_cspec(cex, cs) != 's') {
        goto append;
    }

    /* "cs" appears to be ok.  Get the format string. */
    cgul_convspec__serialize(cex, cs, format);
    if (*cex) {
        /* Ignore the exception. */
        cgul_exception__delete(*cex);
        *cex = NULL;
        goto append;        
    }

    /* Use the format_validated string. */
    use_format = 1;

 append:

    /* Perform the substitution by appending "value" to "dst". */
    if (use_format) {
        /* Append using the format string. */
        cgul_string__append_sprintf(cex,
                                    dst,
                                    cgul_string__get_value(cex, format),
                                    value);
    } else {
        /* Something was wrong with the format string.  Append without it. */
        cgul_string__append(cex, dst, value);
    }
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = 1;

 out:

    /* Clean up. */
    if (format) {
        cgul_string__delete(format);
    }

    return rv;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* src = NULL;
    cgul_string_t dst = NULL;
    cgul_crlf_file_t crlf = NULL;
    cgul_stanza_t stanza = NULL;
    cgul_exception_t local = NULL;

    /* Allocate dst. */
    dst = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Read the mapping of variable names to values. */
    stanza = cgul_stanza__new_from_fname(&local,
                                         "main_varsub_mapping.txt",
                                         1 /* is_block_buffered */);
    cgul_stanza__all_one_stanza(&local, stanza, 1);
    cgul_stanza__load_next(&local, stanza);
    if (local) {
        goto out;
    }

    /* Open the text file. */
    crlf = cgul_crlf_file__new_from_fname(&local,
                                          "main_varsub_formatted_text.txt",
                                          1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Iterate over each line in the text file. */
    while ((src = cgul_crlf_file__get_line(&local, crlf)) != NULL) {

        /* Clear dst. */
        cgul_string__set_value(&local, dst, "");

        /* Perform formatted variable substitution. */
        cgul_varsub_formatted(&local,
                              src,
                              dst,
                              (cgul_varsub_formatted__lookup_t)do_lookup,
                              stanza);

        /* Print results. */
        cgul_libc__puts(&local, cgul_string__get_value(&local, dst));

        /* Check for exceptions. */
        if (local) {
            goto out;
        }
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (crlf) {
        cgul_crlf_file__delete(crlf);
    }
    if (stanza) {
        cgul_stanza__delete(stanza);
    }
    if (dst) {
        cgul_string__delete(dst);
    }

    return rv;
}
