#include "cgul.h"

/**
 * Append the <code>unsigned long</code> values on the half-open range
 * <code>[start, stop)</code> to the vector <code>v</code>.  The
 * client is responsible for calling <code>free()</code> on all the
 * pointers added to <code>v</code>.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v vector
 * @param[in] start inclusive starting value
 * @param[in] stop exclusive stopping value
 */
static void
main__append_values(cgul_exception_t* cex,
                    cgul_vector_t v,
                    unsigned long int start,
                    unsigned long int stop)
{
    unsigned long int i = 0;
    unsigned long int* n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Iterate over the values */
    for (i = start ; i < stop ; ++i) {

        /* Allocate space for the value. */ 
        n = cgul_libc__calloc(cex, 1, sizeof(*n));
        if (*cex) {
            goto out;
        }

        /* Set the value. */
        *n = i;

        /* Append the value to the vector. */
        cgul_vector__push_back(cex, v, n);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

/**
 * Print the vector <code>v</code> to the output file
 * <code>fout</code> where each entry is output with the prefix
 * <code>prefix[<index>]</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] v vector
 * @param[in] prefix prefix
 * @param[in] fout output file
 */
static void
main__print_vector(cgul_exception_t* cex,
                   cgul_vector_t v,
                   const char* prefix,
                   FILE* fout)
{
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    unsigned long int* n = NULL;

    /* Print the vector. */
    i_max = cgul_vector__get_size(cex, v);
    for (i = 0 ; i < i_max ; ++i) {
        n = (unsigned long*)cgul_vector__at(cex, v, i);
        printf("%s[%lu] = %lu\n", prefix, i, *n);
    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_vector_t v1 = NULL;
    cgul_vector_t v2 = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    v1 = cgul_vector__new(&local);
    v2 = cgul_vector__new(&local);
    if (local) {
        goto out;
    }

    /* Add [0 - 5) to the v1 and [5 - 10) to v2. */
    main__append_values(&local, v1, 0, 5);
    main__append_values(&local, v2, 5, 10);
    if (local) {
        goto out;
    }

    /* Print v1 and v2. */
    main__print_vector(&local, v1, "v1", stdout);
    main__print_vector(&local, v2, "v2", stdout);

    printf("-----\n");

    /* Swap the vectors. */
    cgul_vector__swap(&local, v1, v2);

    /* Print v1 and v2. */
    main__print_vector(&local, v1, "v1", stdout);
    main__print_vector(&local, v2, "v2", stdout);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (v2) {
        cgul_vector__free_values(v2);
        cgul_vector__delete(v2);
    }
    if (v1) {
        cgul_vector__free_values(v1);
        cgul_vector__delete(v1);
    }

    return rv;
}
